Type: Package
Package: promises
Title: Abstractions for Promise-Based Asynchronous Programming
Version: 1.5.0
Authors@R: c(
    person("Joe", "Cheng", , "joe@posit.co", role = "aut"),
    person("Barret", "Schloerke", , "barret@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-9986-114X")),
    person("Winston", "Chang", , "winston@posit.co", role = "aut",
           comment = c(ORCID = "0000-0002-1576-2126")),
    person("Charlie", "Gao", , "charlie.gao@posit.co", role = "aut",
           comment = c(ORCID = "0000-0002-0750-061X")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Provides fundamental abstractions for doing asynchronous
    programming in R using promises. Asynchronous programming is useful
    for allowing a single R process to orchestrate multiple tasks in the
    background while also attending to something else. Semantics are
    similar to 'JavaScript' promises, but with a syntax that is idiomatic
    R.
License: MIT + file LICENSE
URL: https://rstudio.github.io/promises/,
        https://github.com/rstudio/promises
BugReports: https://github.com/rstudio/promises/issues
Depends: R (>= 4.1.0)
Imports: fastmap (>= 1.1.0), later, lifecycle, magrittr (>= 1.5), otel
        (>= 0.2.0), R6, rlang
Suggests: future (>= 1.21.0), knitr, mirai, otelsdk (>= 0.2.0), purrr,
        Rcpp, rmarkdown, spelling, testthat (>= 3.0.0), vembedr
VignetteBuilder: knitr
Config/Needs/website: rsconnect, tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-05-27
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2025-10-31 20:26:43 UTC; barret
Author: Joe Cheng [aut],
  Barret Schloerke [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9986-114X>),
  Winston Chang [aut] (ORCID: <https://orcid.org/0000-0002-1576-2126>),
  Charlie Gao [aut] (ORCID: <https://orcid.org/0000-0002-0750-061X>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Barret Schloerke <barret@posit.co>
Repository: CRAN
Date/Publication: 2025-11-01 06:30:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(as.promise,Future)
S3method(as.promise,default)
S3method(as.promise,promise)
S3method(format,promise)
S3method(is.promising,Future)
S3method(is.promising,default)
S3method(is.promising,promise)
S3method(print,promise)
export("%...!%")
export("%...>%")
export("%...T!%")
export("%...T>%")
export("%>%")
export("%T>%")
export(as.promise)
export(catch)
export(finally)
export(future_promise)
export(future_promise_queue)
export(hybrid_then)
export(is.promise)
export(is.promising)
export(local_ospan_promise_domain)
export(local_otel_promise_domain)
export(new_promise_domain)
export(promise)
export(promise_all)
export(promise_map)
export(promise_race)
export(promise_reduce)
export(promise_reject)
export(promise_resolve)
export(then)
export(with_ospan_async)
export(with_ospan_promise_domain)
export(with_otel_promise_domain)
export(with_otel_span)
export(with_promise_domain)
importFrom(R6,R6Class)
importFrom(later,later)
importFrom(lifecycle,deprecated)
importFrom(magrittr,"%>%")
importFrom(magrittr,"%T>%")
importFrom(otel,end_span)
importFrom(otel,get_active_span)
importFrom(otel,is_tracing_enabled)
importFrom(otel,start_span)
importFrom(otel,with_active_span)

# promises <a href="https://rstudio.github.io/promises/"><img src="man/figures/logo.svg" align="right" height="139" alt="promises website" /></a>

<!-- badges: start -->
[![R-CMD-check](https://github.com/rstudio/promises/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rstudio/promises/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

A promise library for R. https://rstudio.github.io/promises/

## Installation

```r
install.packages("promises")
```

To use promises with Shiny, make sure you have Shiny v1.1.0 or later.

## License

MIT

# NAME

Promises - An implementation of Promises in Perl

# VERSION

version 1.00

# SYNOPSIS

```perl
use AnyEvent::HTTP;
use JSON::XS qw[ decode_json ];
use Promises qw[ collect deferred ];

sub fetch_it {
    my ($uri) = @_;
    my $d = deferred;
    http_get $uri => sub {
        my ($body, $headers) = @_;
        $headers->{Status} == 200
            ? $d->resolve( decode_json( $body ) )
            : $d->reject( $body )
    };
    $d->promise;
}

my $cv = AnyEvent->condvar;

collect(
    fetch_it('http://rest.api.example.com/-/product/12345'),
    fetch_it('http://rest.api.example.com/-/product/suggestions?for_sku=12345'),
    fetch_it('http://rest.api.example.com/-/product/reviews?for_sku=12345'),
)->then(
    sub {
        my ($product, $suggestions, $reviews) = @_;
        $cv->send({
            product     => $product,
            suggestions => $suggestions,
            reviews     => $reviews,
        })
    },
    sub { $cv->croak( 'ERROR' ) }
);

my $all_product_info = $cv->recv;
```

# DESCRIPTION

This module is an implementation of the "Promise/A+" pattern for
asynchronous programming. Promises are meant to be a way to
better deal with the resulting callback spaghetti that can often
result in asynchronous programs.

# FUTURE BACKWARDS COMPATIBILITY WARNING

The version of this module is being bumped up to 0.90 as the first
step towards 1.0 in which the goal is to have full Promises/A+ spec
compatibility. This is a departure to the previous goal of being
compatible with the Promises/A spec, this means that behavior may
change in subtle ways (we will attempt to document this completely
and clearly whenever possible).

It is **HIGHLY** recommended that you test things very thoroughly
before upgrading to this version.

# BACKWARDS COMPATIBILITY WARNING

In version up to and including 0.08 there was a bug in how
rejected promises were handled. According to the spec, a
rejected callback can:

- Rethrow the exception, in which case the next rejected handler
in the chain would be called, or
- Handle the exception (by not `die`ing), in which case the next
**resolved** handler in the chain would be called.

In previous versions of [Promises](https://metacpan.org/pod/Promises), this last step was handled incorrectly:
a rejected handler had no way of handling the exception.  Once a promise
was rejected, only rejected handlers in the chain would be called.

## Relation to the various Perl event loops

This module is actually Event Loop agnostic, the SYNOPSIS above
uses [AnyEvent::HTTP](https://metacpan.org/pod/AnyEvent::HTTP), but that is just an example, it can work
with any of the existing event loops out on CPAN. Over the next
few releases I will try to add in documentation illustrating each
of the different event loops and how best to use Promises with
them.

## Relation to the Promise/A spec

We are, with some differences, following the API spec called
"Promise/A" (and the clarification that is called "Promise/A+")
which was created by the Node.JS community. This is, for the most
part, the same API that is implemented in the latest jQuery and
in the YUI Deferred plug-in (though some purists argue that they
both go it wrong, google it if you care). We differ in some
respects to this spec, mostly because Perl idioms and best
practices are not the same as Javascript idioms and best
practices. However, the one important difference that should be
noted is that "Promise/A+" strongly suggests that the callbacks
given to `then` should be run asynchronously (meaning in the
next turn of the event loop). We do not do this by default,
because doing so would bind us to a given event loop
implementation, which we very much want to avoid. However we
now allow you to specify an event loop "backend" when using
Promises, and assuming a Deferred backend has been written
it will provide this feature accordingly.

## Using a Deferred backend

As mentioned above, the default Promises::Deferred class calls the
success or error `then()` callback synchronously, because it isn't
tied to a particular event loop.  However, it is recommended that you
use the appropriate Deferred backend for whichever event loop you are
running.

Typically an application uses a single event loop, so all Promises
should use the same event-loop. Module implementers should just use the
Promises class directly:

```perl
package MyClass;
use Promises qw(deferred collect);
```

End users should specify which Deferred backend they wish to use. For
instance if you are using AnyEvent, you can do:

```perl
use Promises backend => ['AnyEvent'];
use MyClass;
```

The Promises returned by MyClass will automatically use whichever
event loop AnyEvent is using.

See:

- [Promises::Deferred::AE](https://metacpan.org/pod/Promises::Deferred::AE)
- [Promises::Deferred::AnyEvent](https://metacpan.org/pod/Promises::Deferred::AnyEvent)
- [Promises::Deferred::EV](https://metacpan.org/pod/Promises::Deferred::EV)
- [Promises::Deferred::Mojo](https://metacpan.org/pod/Promises::Deferred::Mojo)
- [Promises::Deferred::IO::Async](https://metacpan.org/pod/Promises::Deferred::IO::Async)

## Relation to Promises/Futures in Scala

Scala has a notion of Promises and an associated idea of Futures
as well. The differences and similarities between this module
and the Promises found in Scalar are highlighted in depth in a
cookbook entry below.

## Cookbook

- [Promises::Cookbook::GentleIntro](https://metacpan.org/pod/Promises::Cookbook::GentleIntro)

    Read this first! This cookbook provides a step-by-step explanation
    of how Promises work and how to use them.

- [Promises::Cookbook::SynopsisBreakdown](https://metacpan.org/pod/Promises::Cookbook::SynopsisBreakdown)

    This breaks down the example in the SYNOPSIS and walks through
    much of the details of Promises and how they work.

- [Promises::Cookbook::TIMTOWTDI](https://metacpan.org/pod/Promises::Cookbook::TIMTOWTDI)

    Promise are just one of many ways to do async programming, this
    entry takes the Promises SYNOPSIS again and illustrates some
    counter examples with various modules.

- [Promises::Cookbook::ChainingAndPipelining](https://metacpan.org/pod/Promises::Cookbook::ChainingAndPipelining)

    One of the key benefits of Promises is that it retains much of
    the flow of a synchronous program, this entry illustrates that
    and compares it with a synchronous (or blocking) version.

- [Promises::Cookbook::Recursion](https://metacpan.org/pod/Promises::Cookbook::Recursion)

    This entry explains how to keep the stack under control when
    using Promises recursively.

- [Promises::Cookbook::ScalaFuturesComparison](https://metacpan.org/pod/Promises::Cookbook::ScalaFuturesComparison)

    This entry takes some examples of Futures in the Scala language
    and translates them into Promises. This entry also showcases
    using Promises with [Mojo::UserAgent](https://metacpan.org/pod/Mojo::UserAgent).

# EXPORTS

- `deferred`

    This just creates an instance of the [Promises::Deferred](https://metacpan.org/pod/Promises::Deferred) class
    it is purely for convenience.

    Can take a coderef, which will be dealt with as a `then` argument.

    ```perl
    my $promise = deferred sub {
        ... do stuff ...

        return $something;
    };

    # equivalent to

    my $dummy = deferred;

    my $promise = $dummy->then(sub {
        ... do stuff ...

        return $something;
    });

    $dummy->resolve;
    ```

- `resolved( @values )`

    Creates an instance of [Promises::Deferred](https://metacpan.org/pod/Promises::Deferred) resolved with
    the provided `@values`. Purely a shortcut for

    ```perl
    my $promise = deferred;
    $promise->resolve(@values);
    ```

- `rejected( @values )`

    Creates an instance of [Promises::Deferred](https://metacpan.org/pod/Promises::Deferred) rejected with
    the provided `@values`. Purely a shortcut for

    ```perl
    my $promise = deferred;
    $promise->reject(@values);
    ```

- `collect( @promises )`

    Accepts a list of [Promises::Promise](https://metacpan.org/pod/Promises::Promise) objects and then
    returns a [Promises::Promise](https://metacpan.org/pod/Promises::Promise) object which will be called
    once all the `@promises` have completed (either as an error
    or as a success).

    The eventual result of the returned promise
    object will be an array of all the results of each
    of the `@promises` in the order in which they where passed
    to `collect` originally, wrapped in arrayrefs, or the first error if
    at least one of the promises fail.

    If `collect` is passed a value that is not a promise, it'll be wrapped
    in an arrayref and passed through.

    ```perl
    my $p1 = deferred;
    my $p2 = deferred;
    $p1->resolve(1);
    $p2->resolve(2,3);

    collect(
        $p1,
        'not a promise',
        $p2,
    )->then(sub{
        print join ' : ', map { join ', ', @$_ } @_; # => "1 : not a promise : 2, 3"
    })
    ```

- `collect_hash( @promises )`

    Like `collect`, but flatten its returned arrayref into a single
    hash-friendly list.

    `collect_hash` can be useful to a structured hash instead
    of a long list of promise values.

    For example,

    ```perl
    my $id = 12345;

    collect(
        fetch_it("http://rest.api.example.com/-/product/$id"),
        fetch_it("http://rest.api.example.com/-/product/suggestions?for_sku=$id"),
        fetch_it("http://rest.api.example.com/-/product/reviews?for_sku=$id"),
    )->then(
        sub {
            my ($product, $suggestions, $reviews) = @_;
            $cv->send({
                product     => $product,
                suggestions => $suggestions,
                reviews     => $reviews,
                id          => $id
            })
        },
        sub { $cv->croak( 'ERROR' ) }
    );
    ```

    could be rewritten as

    ```perl
    my $id = 12345;

    collect_hash(
        id          => $id,
        product     => fetch_it("http://rest.api.example.com/-/product/$id"),
        suggestions => fetch_it("http://rest.api.example.com/-/product/suggestions?for_sku=$id"),
        reviews     => fetch_it("http://rest.api.example.com/-/product/reviews?for_sku=$id"),
    )->then(
        sub {
            my %results = @_;
            $cv->send(\%results);
        },
        sub { $cv->croak( 'ERROR' ) }
    );
    ```

    Note that all promise values of the key/value pairs passed to `collect_hash`
    must return a scalar or nothing, as returning more than one value would
    mess up the returned hash format. If a promise does return more than
    one value, `collect_hash` will consider it as having failed.

    If you know that a
    promise can return more than one value, you can do:

    ```perl
    my $collected = collect_hash(
        this => $promise_returning_scalar,
        that => $promise_returning_list->then(sub{ [ @_ ] } ),
    );
    ```

# SEE ALSO

## Promises in General

- [You're Missing the Point of Promises](http://domenic.me/2012/10/14/youre-missing-the-point-of-promises/)
- [Systems Programming at Twitter](http://monkey.org/~marius/talks/twittersystems/)
- [SIP-14 - Futures and Promises](http://docs.scala-lang.org/sips/pending/futures-promises.html)
- [Promises/A+ spec](http://promises-aplus.github.io/promises-spec/)
- [Promises/A spec](http://wiki.commonjs.org/wiki/Promises/A)

## Perl Alternatives

- [Future](https://metacpan.org/pod/Future)
- [Mojo::Promise](https://metacpan.org/pod/Mojo::Promise)

    Part of the [Mojolicious](https://metacpan.org/pod/Mojolicious) package.

- [Promise::ES6](https://metacpan.org/pod/Promise::ES6)
- [Promise::Tiny](https://metacpan.org/pod/Promise::Tiny)

# AUTHOR

Stevan Little <stevan.little@iinteractive.com>

# COPYRIGHT AND LICENSE

This software is copyright (c) 2019, 2017, 2014, 2012 by Infinity Interactive, Inc..

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

This archive contains the distribution Promises,
version 0.93:

  An implementation of Promises in Perl

This software is copyright (c) 2014 by Infinity Interactive, Inc..

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.


This README file was generated by Dist::Zilla::Plugin::Readme v5.012.

# Determine whether an expression is of the type that needs a first argument.
#
# @param a non-evaluated expression.
# @return logical - TRUE if expr is of "first-argument" type, FALSE otherwise.
is_first <- function(expr)
{
  !any(vapply(expr[-1L], identical, logical(1L), quote(.)))
}

# Prepare a magrittr rhs of "first-argument" type.
#
# @param a an expression which passes `is_first`
# @return an expression prepared for functional sequence construction.
prepare_first <- function(expr)
{
  as.call(c(expr[[1L]], quote(.), as.list(expr[-1L])))
}

library(later)

# Create a promise that can be EXTernally resolved/rejected/inspected
ext_promise <- function() {
  res <- NULL
  p <- promise(function(resolve, reject) {
    res <<- list(resolve = resolve, reject = reject)
  })

  list(
    promise = p,
    resolve = res$resolve,
    reject = res$reject,
    status = attr(p, "promise_impl", exact = TRUE)$status
  )
}

# Block until all pending later tasks have executed
wait_for_it <- function(timeout = 30) {
  start <- Sys.time()
  while (!loop_empty()) {
    if (difftime(Sys.time(), start, units = "secs") > timeout) {
      stop("Waited too long")
    }
    run_now()
    Sys.sleep(0.01)
  }
}

# Block until the promise is resolved/rejected. If resolved, return the value.
# If rejected, throw (yes throw, not return) the error.
extract <- function(promise) {
  promise_value <- NULL
  error <- NULL
  promise %...>%
    (function(value) promise_value <<- value) %...!%
    (function(reason) error <<- reason)

  wait_for_it()
  if (!is.null(error))
    stop(error)
  else
    promise_value
}

resolve_later <- function(value, delaySecs) {
  force(value)
  promise(~later::later(~resolve(value), delaySecs))
}

# Prevent "Unhandled promise error" warning that happens if you don't handle the
# rejection of a promise
squelch_unhandled_promise_error <- function(promise) {
  promise %...!% {}
}

.GlobalEnv$.Last <- function() {
  # Detect unexpected "Unhandled promise error" warnings.
  wait_for_it()
}

create_counting_domain <- function(trackFinally = FALSE) {
  counts <- list2env(parent = emptyenv(), list(
    onFulfilledBound = 0L,
    onFulfilledCalled = 0L,
    onFulfilledActive = 0L,
    onRejectedBound = 0L,
    onRejectedCalled = 0L,
    onRejectedActive = 0L
  ))

  incr <- function(field) {
    field <- as.character(substitute(field))
    counts[[field]] <- counts[[field]] + 1L
  }

  decr <- function(field) {
    field <- as.character(substitute(field))
    counts[[field]] <- counts[[field]] - 1L
  }

  pd <- new_promise_domain(
    wrapOnFulfilled = function(onFulfilled) {
      incr(onFulfilledBound)
      function(...) {
        incr(onFulfilledCalled)
        incr(onFulfilledActive)
        on.exit(decr(onFulfilledActive))

        onFulfilled(...)
      }
    },
    wrapOnRejected = function(onRejected) {
      incr(onRejectedBound)
      function(...) {
        incr(onRejectedCalled)
        incr(onRejectedActive)
        on.exit(decr(onRejectedActive))

        onRejected(...)
      }
    },
    counts = counts
  )

  if (trackFinally) {
    counts$onFinallyBound <- 0L
    counts$onFinallyCalled <- 0L
    counts$onFinallyActive <- 0L

    pd$wrapOnFinally <- function(onFinally) {
      incr(onFinallyBound)
      function() {
        incr(onFinallyCalled)
        incr(onFinallyActive)
        on.exit(incr(onFinallyActive))

        onFinally()
      }
    }
  }

  pd
}

library(testthat)

source("common.R")

# 2.3. The Promise Resolution Procedure

describe("2.3.1. If promise and x refer to the same object, reject promise with a TypeError as the reason.", {
  it(" ", {
    p <- ext_promise()
    p$resolve(p$promise)
    expect_error(extract(p$promise), "^Chaining cycle detected for promise$")
  })
})

describe("2.3.2. If x is a promise, adopt its state [3.4]:", {
  it("2.3.2.1. If x is pending, promise must remain pending until x is fulfilled or rejected.", {
    p <- ext_promise()
    x <- ext_promise()
    p$resolve(x$promise)
    wait_for_it()
    expect_identical(x$status(), "pending")
    expect_identical(p$status(), "pending")
  })
  it("2.3.2.2. If/when x is fulfilled, fulfill promise with the same value.", {
    p <- ext_promise()
    x <- ext_promise()
    p$resolve(x$promise)
    wait_for_it()
    expect_identical(x$status(), "pending")
    expect_identical(p$status(), "pending")
    x$resolve(100)
    expect_identical(extract(x$promise), 100)
    expect_identical(extract(p$promise), 100)
  })
  it("2.3.2.3. If/when x is rejected, reject promise with the same reason.", {
    p <- ext_promise()
    x <- ext_promise()
    p$resolve(x$promise)
    wait_for_it()
    expect_identical(x$status(), "pending")
    expect_identical(p$status(), "pending")
    squelch_unhandled_promise_error(p$promise)
    x$reject("bad")
    expect_error(extract(x$promise), "^bad$")
    expect_error(extract(p$promise), "^bad$")
  })
})
# 2.3.3. Otherwise, if x is an object or function,
# 2.3.3.1. Let then be x.then. [3.5]
# 2.3.3.2. If retrieving the property x.then results in a thrown exception e, reject promise with e as the reason.
# 2.3.3.3. If then is a function, call it with x as this, first argument resolvePromise, and second argument rejectPromise, where:
# 2.3.3.3.1. If/when resolvePromise is called with a value y, run [[Resolve]](promise, y).
# 2.3.3.3.2. If/when rejectPromise is called with a reason r, reject promise with r.
# 2.3.3.3.3. If both resolvePromise and rejectPromise are called, or multiple calls to the same argument are made, the first call takes precedence, and any further calls are ignored.
# 2.3.3.3.4. If calling then throws an exception e,
# 2.3.3.3.4.1. If resolvePromise or rejectPromise have been called, ignore it.
# 2.3.3.3.4.2. Otherwise, reject promise with e as the reason.
# 2.3.3.4. If then is not a function, fulfill promise with x.
describe("2.3.4. If x is not an object or function, fulfill promise with x.", {
  it(" ", {
    p <- ext_promise()
    p$resolve(10)
    expect_identical(extract(p$promise), 10)
  })
})

async_true <- function() {
  promise_resolve(TRUE)
}

describe("Promise domains", {
  it("are reentered during handlers", {
    cd <- create_counting_domain(trackFinally = TRUE)
    p <- with_promise_domain(cd, {
      async_true() %...>%
        {
          expect_identical(cd$counts$onFulfilledCalled, 1L)
          expect_identical(cd$counts$onFulfilledActive, 1L)
          10 # sync result
        } %...>%
        {
          expect_identical(cd$counts$onFulfilledCalled, 2L)
          expect_identical(cd$counts$onFulfilledActive, 1L)
          promise_resolve(20) # async result
        }
    })

    expect_identical(cd$counts$onFulfilledBound, 2L)

    p <- p %...>%
      {
        expect_identical(cd$counts$onFulfilledCalled, 2L)
        expect_identical(cd$counts$onFulfilledActive, 0L)
      }

    expect_identical(cd$counts$onFulfilledBound, 2L)
    p %>% wait_for_it()

    with_promise_domain(cd, {
      p <- async_true() %>%
        finally(
          ~ {
            expect_identical(cd$counts$onFinallyCalled, 1L)
            expect_identical(cd$counts$onFinallyActive, 1L)
          }
        )
      expect_identical(cd$counts$onFinallyBound, 1L)

      expect_identical(cd$counts$onFulfilledBound, 2L)
      expect_identical(cd$counts$onRejectedBound, 0L)

      p %>% wait_for_it()
    })

    expect_identical(cd$counts$onFulfilledBound, 2L)

    with_promise_domain(cd, {
      p <- async_true() %...>%
        {
          expect_identical(., TRUE)
          expect_identical(cd$counts$onFulfilledCalled, 3L)
          ten <- 10
          # This tests if promise domain membership infects subscriptions made
          # in handlers.
          promise_resolve(invisible(ten)) %...>%
            (function(value, .visible) {
              expect_identical(value, 10)
              expect_false(.visible)
              expect_true(!is.null(current_promise_domain()))
              expect_identical(cd$counts$onFulfilledCalled, 4L)
            })
        }
    })

    expect_true(is.null(current_promise_domain()))
    expect_identical(cd$counts$onFulfilledCalled, 2L)

    p %>% wait_for_it()
  })

  it("pass finally binding to fulfill/reject by default", {
    cd1 <- create_counting_domain(trackFinally = FALSE)

    with_promise_domain(cd1, {
      p1 <- async_true() %>%
        finally(
          ~ {
            expect_identical(cd1$counts$onFulfilledActive, 1L)
            expect_identical(cd1$counts$onRejectedActive, 0L)
          }
        )
      expect_identical(cd1$counts$onFulfilledBound, 1L)
      expect_identical(cd1$counts$onRejectedBound, 1L)
      p1 %>% wait_for_it()
      expect_identical(cd1$counts$onFulfilledCalled, 1L)
      expect_identical(cd1$counts$onRejectedCalled, 0L)
    })

    cd2 <- create_counting_domain(trackFinally = FALSE)

    p2 <- with_promise_domain(cd2, {
      promise_reject("a problem") %>%
        finally(
          ~ {
            expect_identical(cd2$counts$onFulfilledActive, 0L)
            expect_identical(cd2$counts$onRejectedActive, 1L)
          }
        )
    }) %>%
      squelch_unhandled_promise_error()

    expect_identical(cd2$counts$onFulfilledBound, 1L)
    expect_identical(cd2$counts$onRejectedBound, 1L)
    p2 %>% wait_for_it()
    expect_identical(cd2$counts$onFulfilledCalled, 0L)
    expect_identical(cd2$counts$onRejectedCalled, 1L)
  })

  it("doesn't intercept fulfill/reject on finally, if finally is explicitly intercepted", {
    cd1 <- create_counting_domain(trackFinally = TRUE)

    with_promise_domain(cd1, {
      p1 <- async_true() %>%
        finally(
          ~ {
            expect_identical(cd1$counts$onFinallyActive, 1L)
            expect_identical(cd1$counts$onFulfilledActive, 0L)
            expect_identical(cd1$counts$onRejectedActive, 0L)
          }
        )
      expect_identical(cd1$counts$onFinallyBound, 1L)
      expect_identical(cd1$counts$onFulfilledBound, 0L)
      expect_identical(cd1$counts$onRejectedBound, 0L)
      p1 %>% wait_for_it()
      expect_identical(cd1$counts$onFinallyCalled, 1L)
      expect_identical(cd1$counts$onFulfilledCalled, 0L)
      expect_identical(cd1$counts$onRejectedCalled, 0L)
    })

    cd2 <- create_counting_domain(trackFinally = TRUE)

    p2 <- with_promise_domain(cd2, {
      promise_reject(TRUE) %>%
        finally(
          ~ {
            expect_identical(cd2$counts$onFinallyActive, 1L)
            expect_identical(cd2$counts$onFulfilledActive, 0L)
            expect_identical(cd2$counts$onRejectedActive, 0L)
          }
        )
    }) %>%
      squelch_unhandled_promise_error()

    expect_identical(cd2$counts$onFinallyBound, 1L)
    expect_identical(cd2$counts$onFulfilledBound, 0L)
    expect_identical(cd2$counts$onRejectedBound, 0L)
    p2 %>% wait_for_it()
    expect_identical(cd2$counts$onFinallyCalled, 1L)
    expect_identical(cd2$counts$onFulfilledCalled, 0L)
    expect_identical(cd2$counts$onRejectedCalled, 0L)
  })

  it("handles weird edge case relating to symbols", {
    # This test resulted from a bug in wrap_callback_reenter() where do.call()'s
    # default behavior of quote=FALSE would cause a resolved value that happens
    # to be a symbol, to be evaluated. This would only happen when a promise
    # domain was in effect, and the symbol was passed to an onFulfilled. Fixed
    # by using rlang::exec() instead of do.call().
    cd <- create_counting_domain()
    with_promise_domain(cd, {
      promise_resolve(as.symbol("foo")) %...>%
        {
          expect_identical(., as.symbol("foo"))
        } %>%
        wait_for_it()
    })
  })

  it("executes wrap_callback_reenter handlers in the right lexical environment", {
    # No reason this shouldn't work, I haven't seen it fail, just making sure.
    cd <- create_counting_domain()
    x <- NULL
    with_promise_domain(cd, {
      promise_resolve(NULL) %...>%
        {
          x <<- 1L
        } %...>%
        {
          x <<- x + 1L
        } %>%
        wait_for_it()
    })
    expect_identical(x, 2L)
  })

  it("doesn't grow the call stack", {
    # See https://github.com/rstudio/promises/issues/114
    # and also https://github.com/jcheng5/shinychat/issues/16

    recursive_promise <- function(n, .last_callstack_depth = NULL) {
      if (n == 0) {
        return(0)
      } else {
        promise_resolve(TRUE) %...>%
          {
            current_callstack_depth <- length(sys.calls())
            if (!is.null(.last_callstack_depth)) {
              expect_identical(current_callstack_depth, .last_callstack_depth)
            }

            recursive_promise(
              n - 1,
              .last_callstack_depth = current_callstack_depth
            )
          }
      }
    }

    cd <- create_counting_domain()
    with_promise_domain(cd, {
      recursive_promise(5) %...>%
        {
          # 5 (from inside recursive_promise) + 1 (for the current handler)
          expect_identical(cd$counts$onFulfilledCalled, 6L)
        } %>%
        wait_for_it()
    })

    cd2 <- create_counting_domain()
    p <- recursive_promise(5)
    with_promise_domain(cd2, {
      p %...>%
        {
          # This time, none of the ones inside recursive_promise count, since
          # they were bound outside of the influence of cd2 (even though they
          # are resolved within the influence of cd2, thanks to wait_for_it()).
          expect_identical(cd2$counts$onFulfilledCalled, 1L)
        } %>%
        wait_for_it()
    })
  })
})

test_that("complex promise chains with mixed success/failure", {
  counter <- 0
  promise_resolve(1) |>
    then(\(x) {
      counter <<- counter + 1
      stop("error1")
    }) |>
    catch(\(err) {
      counter <<- counter + 1
      "recovered"
    }) |>
    then(\(x) {
      counter <<- counter + 1
      x
    }) |>
    catch(\(err) {
      counter <<- counter + 1
      "should not reach"
    }) |>
    then(\(x) {
      counter <<- counter + 1
      expect_equal(x, "recovered")
    }) |>
    wait_for_it()
  expect_equal(counter, 4)
})

test_that("then() handles NULL callbacks correctly", {
  # NULL onFulfilled should pass through value
  expect_assertions(n = 1, {
    promise_resolve(42) |>
      then(NULL, \(err) "error") |>
      extract() |>
      expect_equal(42)
  })
})

test_that("catch() handles NULL callbacks correctly", {
  # NULL onRejected should pass through error
  expect_assertions(n = 1, {
    promise_resolve(stop("boom")) |>
      then(\(x) x, NULL) |>
      catch(\(err) expect_match(err$message, "boom")) |>
      wait_for_it()
  })
})

test_that("then() validates arguments properly", {
  expect_error(
    then(promise_resolve(1), onFulfilled = "not a function"),
    "function"
  )

  expect_error(
    catch(promise_resolve(1), "not a function"),
    "function"
  )
})

test_that("tee types are handled correctly", {
  expect_silent({
    p <- promise_resolve(1) |> then(tee = TRUE)
    p2 <- promise_resolve(1) |> catch(\(err) {}, tee = TRUE)
  })

  expect_snapshot(
    then(promise_resolve(1), tee = 4),
    error = TRUE
  )

  expect_snapshot(
    promise_resolve(1) |> catch(\(err) {}, tee = "4"),
    error = TRUE
  )
})

test_that("then(tee = FALSE) functions are normalized", {
  counter <- 0
  promise_resolve(1) |>
    then(
      \() {
        counter <<- counter + 1
        2
      }
    ) |>
    then(
      \(value) {
        expect_equal(value, 2)
        counter <<- counter + 1
        invisible(3)
      }
    ) |>
    then(
      \(value, .visible) {
        expect_equal(value, 3)
        expect_false(.visible)
        counter <<- counter + 1
        stop("boom")
      }
    ) |>
    then(
      onRejected = \() {
        counter <<- counter + 1
        stop("blast")
      }
    ) |>
    then(
      onRejected = \(err) {
        expect_true(grepl("blast", err$message))
        counter <<- counter + 1
      }
    )
  wait_for_it()
  expect_equal(counter, 5)
})
test_that("catch(tee = FALSE) functions are normalized", {
  counter <- 0
  promise_resolve(stop("boom")) |>
    catch(
      \() {
        counter <<- counter + 1
        stop("blast")
      }
    ) |>
    catch(
      \(err) {
        expect_true(grepl("blast", err$message))
        counter <<- counter + 1
      }
    ) |>
    catch(\(err) {})
  wait_for_it()
  expect_equal(counter, 2)
})

test_that("then(tee=TRUE) functions are normalized", {
  then_counter <- 0
  error_counter <- 0
  promise_resolve(1) |>
    then(
      tee = TRUE,
      \() {
        then_counter <<- then_counter + 1
      }
    ) |>
    then(
      tee = TRUE,
      \(value) {
        expect_equal(value, 1)
        then_counter <<- then_counter + 1
        3
      }
    ) |>
    then(
      tee = TRUE,
      \(value, .visible) {
        expect_equal(value, 1)
        expect_true(.visible)
        then_counter <<- then_counter + 1

        stop("boom")
      }
    ) |>
    then(
      tee = TRUE,
      onRejected = \() {
        error_counter <<- error_counter + 1
      }
    ) |>
    then(
      tee = TRUE,
      onRejected = \(err) {
        expect_true(grepl("boom", err$message))
        error_counter <<- error_counter + 1
      }
    ) |>
    then(onRejected = \(err) {
      expect_true(grepl("boom", err$message))
      error_counter <<- error_counter + 1
    }) |>
    wait_for_it()

  expect_equal(then_counter, 3)
  expect_equal(error_counter, 3)
})

test_that("catch(tee=TRUE) functions are normalized", {
  error_counter <- 0
  promise_resolve(stop("boom")) |>
    catch(
      tee = TRUE,
      \() {
        error_counter <<- error_counter + 1
      }
    ) |>
    catch(
      tee = TRUE,
      \(err) {
        expect_true(grepl("boom", err$message))
        error_counter <<- error_counter + 1
      }
    ) |>
    catch(\() {
      error_counter <<- error_counter + 1
    }) |>
    wait_for_it()
  expect_equal(error_counter, 3)
})

test_that("promise_map handles NULL values", {
  prom2_map <- function(first_res, second_res) {
    f <- function(index) {
      switch(index, first_res, second_res)
    }
    promise_map(1:2, f)
  }

  counter <- 0

  prom2_map("anything", "anything_else") %...>%
    {
      counter <<- counter + 1
      expect_equal(., list("anything", "anything_else"))
    } %>%
    wait_for_it()

  prom2_map(NULL, "anything_else") %...>%
    {
      counter <<- counter + 1
      expect_equal(., list(NULL, "anything_else"))
    } %>%
    wait_for_it()

  prom2_map("anything", NULL) %...>%
    {
      counter <<- counter + 1
      expect_equal(., list("anything", NULL))
    } %>%
    wait_for_it()

  prom2_map(NULL, NULL) %...>%
    {
      counter <<- counter + 1
      expect_equal(., list(NULL, NULL))
    } %>%
    wait_for_it()

  expect_equal(counter, 4)
})

library(testthat)

source("common.R")

describe("2.1. Promise States", {
  describe("2.1.1. When pending, a promise:", {
    it("2.1.1.1. may transition to either the fulfilled or rejected state.", {
      a <- ext_promise()
      expect_identical(a$status(), "pending")
      a$resolve(0)
      expect_identical(a$status(), "fulfilled")

      b <- ext_promise()
      expect_identical(b$status(), "pending")
      squelch_unhandled_promise_error(b$promise)
      b$reject("err")
      expect_identical(b$status(), "rejected")
    })
  })

  describe("2.1.2. When fulfilled, a promise:", {
    it("2.1.2.1. must not transition to any other state.", {
      a <- ext_promise()
      a$resolve(TRUE)
      expect_identical(a$status(), "fulfilled")
      a$reject("err")
      expect_identical(a$status(), "fulfilled")
    })
    it("2.1.2.2. must have a value, which must not change.", {
      a <- ext_promise()
      a$resolve(TRUE)
      expect_identical(a$status(), "fulfilled")
      a$resolve(FALSE)
      expect_identical(extract(a$promise), TRUE)
    })
  })

  describe("2.1.3. When rejected, a promise:", {
    it("2.1.3.1. must not transition to any other state.", {
      a <- ext_promise()
      squelch_unhandled_promise_error(a$promise)
      a$reject("err")
      expect_identical(a$status(), "rejected")
      a$resolve(TRUE)
      expect_identical(a$status(), "rejected")
    })
    it("2.1.3.2. must have a reason, which must not change.", {
      a <- ext_promise()
      a$reject("err1")
      expect_identical(a$status(), "rejected")
      a$reject("err2")
      expect_error(extract(a$promise), "err1")
    })
  })
})

async_true <- function() {
  promise_resolve(TRUE)
}

describe("Promise domains", {
  it("are reentered during handlers", {
    cd <- create_counting_domain(trackFinally = TRUE)
    p <- with_promise_domain(cd, {
      async_true() |>
        then(\(x) {
          expect_identical(cd$counts$onFulfilledCalled, 1L)
          expect_identical(cd$counts$onFulfilledActive, 1L)
          10 # sync result
        }) |>
        then(\(x) {
          expect_identical(x, 10)
          expect_identical(cd$counts$onFulfilledCalled, 2L)
          expect_identical(cd$counts$onFulfilledActive, 1L)
          promise_resolve(20) # async result
        })
    })

    expect_identical(cd$counts$onFulfilledBound, 2L)

    p <- p |>
      then(\(x) {
        expect_identical(cd$counts$onFulfilledCalled, 2L)
        expect_identical(cd$counts$onFulfilledActive, 0L)
      })

    expect_identical(cd$counts$onFulfilledBound, 2L)
    p |> wait_for_it()

    with_promise_domain(cd, {
      p <- async_true() |>
        finally(\() {
          expect_identical(cd$counts$onFinallyCalled, 1L)
          expect_identical(cd$counts$onFinallyActive, 1L)
        })
      expect_identical(cd$counts$onFinallyBound, 1L)

      expect_identical(cd$counts$onFulfilledBound, 2L)
      expect_identical(cd$counts$onRejectedBound, 0L)

      p |> wait_for_it()
    })

    expect_identical(cd$counts$onFulfilledBound, 2L)

    with_promise_domain(cd, {
      p <- async_true() |>
        then(\(x) {
          expect_identical(x, TRUE)
          expect_identical(cd$counts$onFulfilledCalled, 3L)
          ten <- 10
          # This tests if promise domain membership infects subscriptions made
          # in handlers.
          promise_resolve(invisible(ten)) |>
            then(\(value, .visible) {
              expect_identical(value, 10)
              expect_false(.visible)
              expect_true(!is.null(current_promise_domain()))
              expect_identical(cd$counts$onFulfilledCalled, 4L)
            })
        })
    })
    expect_true(is.null(current_promise_domain()))
    expect_identical(cd$counts$onFulfilledCalled, 2L)

    p |> wait_for_it()
  })

  it("pass finally binding to fulfill/reject by default", {
    cd1 <- create_counting_domain(trackFinally = FALSE)

    with_promise_domain(cd1, {
      p1 <- async_true() |>
        finally(\() {
          expect_identical(cd1$counts$onFulfilledActive, 1L)
          expect_identical(cd1$counts$onRejectedActive, 0L)
        })

      expect_identical(cd1$counts$onFulfilledBound, 1L)
      expect_identical(cd1$counts$onRejectedBound, 1L)
      p1 |> wait_for_it()
      expect_identical(cd1$counts$onFulfilledCalled, 1L)
      expect_identical(cd1$counts$onRejectedCalled, 0L)
    })

    cd2 <- create_counting_domain(trackFinally = FALSE)

    p2 <- with_promise_domain(cd2, {
      promise_reject("a problem") |>
        finally(\() {
          expect_identical(cd2$counts$onFulfilledActive, 0L)
          expect_identical(cd2$counts$onRejectedActive, 1L)
        })
    }) |>
      squelch_unhandled_promise_error()

    expect_identical(cd2$counts$onFulfilledBound, 1L)
    expect_identical(cd2$counts$onRejectedBound, 1L)
    p2 |> wait_for_it()
    expect_identical(cd2$counts$onFulfilledCalled, 0L)
    expect_identical(cd2$counts$onRejectedCalled, 1L)
  })

  it("doesn't intercept fulfill/reject on finally, if finally is explicitly intercepted", {
    cd1 <- create_counting_domain(trackFinally = TRUE)

    with_promise_domain(cd1, {
      p1 <- async_true() |>
        finally(\() {
          expect_identical(cd1$counts$onFinallyActive, 1L)
          expect_identical(cd1$counts$onFulfilledActive, 0L)
          expect_identical(cd1$counts$onRejectedActive, 0L)
        })

      expect_identical(cd1$counts$onFinallyBound, 1L)
      expect_identical(cd1$counts$onFulfilledBound, 0L)
      expect_identical(cd1$counts$onRejectedBound, 0L)
      p1 |> wait_for_it()
      expect_identical(cd1$counts$onFinallyCalled, 1L)
      expect_identical(cd1$counts$onFulfilledCalled, 0L)
      expect_identical(cd1$counts$onRejectedCalled, 0L)
    })

    cd2 <- create_counting_domain(trackFinally = TRUE)

    p2 <- with_promise_domain(cd2, {
      promise_reject(TRUE) |>
        finally(\() {
          expect_identical(cd2$counts$onFinallyActive, 1L)
          expect_identical(cd2$counts$onFulfilledActive, 0L)
          expect_identical(cd2$counts$onRejectedActive, 0L)
        })
    }) |>
      squelch_unhandled_promise_error()

    expect_identical(cd2$counts$onFinallyBound, 1L)
    expect_identical(cd2$counts$onFulfilledBound, 0L)
    expect_identical(cd2$counts$onRejectedBound, 0L)
    p2 |> wait_for_it()
    expect_identical(cd2$counts$onFinallyCalled, 1L)
    expect_identical(cd2$counts$onFulfilledCalled, 0L)
    expect_identical(cd2$counts$onRejectedCalled, 0L)
  })

  it("handles weird edge case relating to symbols", {
    # This test resulted from a bug in wrap_callback_reenter() where do.call()'s
    # default behavior of quote=FALSE would cause a resolved value that happens
    # to be a symbol, to be evaluated. This would only happen when a promise
    # domain was in effect, and the symbol was passed to an onFulfilled. Fixed
    # by using rlang::exec() instead of do.call().
    cd <- create_counting_domain()
    with_promise_domain(cd, {
      promise_resolve(as.symbol("foo")) |>
        then(\(x) {
          expect_identical(x, as.symbol("foo"))
        }) |>
        wait_for_it()
    })
  })

  it("executes wrap_callback_reenter handlers in the right lexical environment", {
    # No reason this shouldn't work, I haven't seen it fail, just making sure.
    cd <- create_counting_domain()
    x <- NULL
    with_promise_domain(cd, {
      promise_resolve(NULL) |>
        then(\() {
          x <<- 1L
        }) |>
        then(\() {
          x <<- x + 1L
        }) |>
        wait_for_it()
    })
    expect_identical(x, 2L)
  })

  it("doesn't grow the call stack", {
    # See https://github.com/rstudio/promises/issues/114
    # and also https://github.com/jcheng5/shinychat/issues/16

    recursive_promise <- function(n, .last_callstack_depth = NULL) {
      if (n == 0) {
        return(0)
      } else {
        promise_resolve(TRUE) |>
          then(\() {
            current_callstack_depth <- length(sys.calls())
            if (!is.null(.last_callstack_depth)) {
              expect_identical(current_callstack_depth, .last_callstack_depth)
            }

            recursive_promise(
              n - 1,
              .last_callstack_depth = current_callstack_depth
            )
          })
      }
    }

    cd <- create_counting_domain()
    with_promise_domain(cd, {
      recursive_promise(5) |>
        then(\() {
          # 5 (from inside recursive_promise) + 1 (for the current handler)
          expect_identical(cd$counts$onFulfilledCalled, 6L)
        }) |>
        wait_for_it()
    })

    cd2 <- create_counting_domain()
    p <- recursive_promise(5)
    with_promise_domain(cd2, {
      p |>
        then(\() {
          # This time, none of the ones inside recursive_promise count, since
          # they were bound outside of the influence of cd2 (even though they
          # are resolved within the influence of cd2, thanks to wait_for_it()).
          expect_identical(cd2$counts$onFulfilledCalled, 1L)
        }) |>
        wait_for_it()
    })
  })

  it("apply nested domains in reverse order (innermost domain closest to callback)", {
    # Test that when nesting promise domains, the latest domain to be added
    # (innermost) should be the one that wraps the onFulfilled/onRejected most closely.

    # Create a test environment to track domain activation order
    test_env <- new.env()
    test_env$x <- "initial"
    test_env$activations <- character(0)

    createVarPromiseDomain <- function(env, name, value) {
      force(env)
      force(name)
      force(value)

      new_promise_domain(
        wrapOnFulfilled = function(onFulfilled) {
          function(...) {
            orig <- env[[name]]
            env[[name]] <- value
            env$activations <- c(env$activations, paste0(name, "=", value))
            on.exit({
              env[[name]] <- orig
            })

            onFulfilled(...)
          }
        },
        wrapSync = function(expr) {
          orig <- env[[name]]
          env[[name]] <- value
          on.exit(env[[name]] <- orig)

          force(expr)
        }
      )
    }

    domain1 <- createVarPromiseDomain(test_env, "x", 1)
    domain2 <- createVarPromiseDomain(test_env, "x", 2)

    # Test the nested domain behavior
    result <- NULL

    with_promise_domain(domain1, {
      # Outer domain should set x=1
      expect_identical(test_env$x, 1)

      with_promise_domain(domain2, {
        # Inner domain should set x=2
        expect_identical(test_env$x, 2)

        # Create and resolve a promise
        promise_resolve(TRUE) |>
          then(function(value) {
            # The inner domain (domain2) should be closest to this callback,
            # so x should be 2 when this executes
            result <<- test_env$x
            test_env$x
          }) |>
          wait_for_it()
      })
    })

    # Verify that the innermost domain took precedence
    expect_identical(result, 2)

    # Verify the activation order: domain1 (outer) activates first,
    # then domain2 (inner) activates and takes precedence
    expect_length(test_env$activations, 2)
    expect_identical(test_env$activations[1], "x=1")
    expect_identical(test_env$activations[2], "x=2")
  })
})

library(testthat)

describe("promise_all", {
  it("preserves element ordering", {
    a <- resolve_later(1, 0.5)
    b <- resolve_later(2, 0.3)
    c <- resolve_later(3, 0.1)

    x <- promise_all(.list = list(a = a, b = b, c = c))
    expect_identical(extract(x), list(a = 1, b = 2, c = 3))
  })

  it("Handles NULLs correctly", {
    x <- promise_all(
      promise_resolve(NULL),
      promise_resolve(NULL),
      promise_resolve(NULL)
    )
    expect_identical(extract(x), list(NULL, NULL, NULL))

    x <- promise_all(
      a = promise_resolve(NULL),
      b = promise_resolve(NULL),
      c = promise_resolve(NULL)
    )
    expect_identical(extract(x), list(a = NULL, b = NULL, c = NULL))
  })

  it("handles duplicate names correctly", {
    x <- resolve_later(1, 0.5)
    y <- resolve_later(2, 0.3)
    z <- resolve_later(3, 0.1)

    x <- promise_all(.list = list(a = x, a = y, a = z))
    expect_identical(extract(x), list(a = 1, a = 2, a = 3))
  })
})

describe("C++ interface", {
  env <- new.env()
  Rcpp::sourceCpp(
    system.file("promise_task.cpp", package = "promises"),
    env = env
  )

  it("basically works", {
    promise(function(resolve, reject) {
      env$asyncFib(resolve, reject, 3)
    }) |>
      then(\(x) {
        expect_identical(x, 2)
      }) |>
      wait_for_it()
  })

  it("works with domains", {
    cd <- create_counting_domain()

    expect_true(is.null(current_promise_domain()))
    with_promise_domain(cd, {
      promise(function(resolve, reject) {
        env$asyncFib(resolve, reject, 3)
      }) |>
        then(\(x) {
          expect_identical(x, 2)
          expect_identical(cd$counts$onFulfilledCalled, 1L)
          promise_resolve(TRUE) |>
            then(\(y) {
              expect_true(!is.null(current_promise_domain()))
              expect_identical(cd$counts$onFulfilledCalled, 2L)
            })
        }) |>
        wait_for_it()
    })
  })
})

tryCatch <- function(expr, ..., finally) {
  base::tryCatch(
    withCallingHandlers(
      expr,
      error = function(e) {
        promiseDomain$onError(e)
      }
    ),
    ...,
    finally = finally
  )
}

spliceOnFinally <- function(onFinally) {
  list(
    onFulfilled = finallyToFulfilled(onFinally),
    onRejected = finallyToRejected(onFinally)
  )
}

finallyToFulfilled <- function(onFinally) {
  force(onFinally)
  function(value, .visible) {
    onFinally()
    if (.visible)
      value
    else
      invisible(value)
  }
}

finallyToRejected <- function(onFinally) {
  force(onFinally)
  function(reason) {
    onFinally()
    stop(reason)
  }
}

promiseDomain <- list(
  onThen = function(onFulfilled, onRejected, onFinally) {
    force(onFulfilled)
    force(onRejected)
    force(onFinally)

    # Verify that if onFinally is non-NULL, onFulfilled and onRejected are NULL
    if (!is.null(onFinally) && (!is.null(onFulfilled) || !is.null(onRejected))) {
      stop("A single `then` call cannot combine `onFinally` with `onFulfilled`/`onRejected`")
    }

    # TODO: All wrapped functions should also be rewritten to reenter the domain
    # jcheng 2019-07-26: Actually, this seems not to be necessary--the domain
    # is getting reentered during callbacks. But I can't figure out now how it's
    # happening.

    domain <- current_promise_domain()

    shouldWrapFinally <- !is.null(onFinally) && !is.null(domain) && !is.null(domain$wrapOnFinally)

    newOnFinally <- if (shouldWrapFinally) {
      domain$wrapOnFinally(onFinally)
    } else {
      onFinally
    }

    if (!is.null(newOnFinally)) {
      spliced <- spliceOnFinally(newOnFinally)
      onFulfilled <- spliced$onFulfilled
      onRejected <- spliced$onRejected
    }

    shouldWrapFulfilled <- !is.null(onFulfilled) && !is.null(domain) && !shouldWrapFinally
    shouldWrapRejected <- !is.null(onRejected) && !is.null(domain) && !shouldWrapFinally

    results <- list(
      onFulfilled = if (shouldWrapFulfilled) domain$wrapOnFulfilled(onFulfilled) else onFulfilled,
      onRejected = if (shouldWrapRejected) domain$wrapOnRejected(onRejected) else onRejected
    )
    results[!vapply(results, is.null, logical(1))]
  },
  onError = function(error) {
    domain <- current_promise_domain()
    if (is.null(domain))
      return()
    domain$onError(error)
  }
)

globals <- new.env(parent = emptyenv())

current_promise_domain <- function() {
  globals$domain
}

#' Promise domains
#'
#' Promise domains are used to temporarily set up custom environments that
#' intercept and influence the registration of callbacks. Create new promise
#' domain objects using `new_promise_domain`, and temporarily activate a promise
#' domain object (for the duration of evaluating a given expression) using
#' `with_promise_domain`.
#'
#' While `with_promise_domain` is on the call stack, any calls to [then()] (or
#' higher level functions or operators, like [catch()] or the various [pipes])
#' will belong to the promise domain. In addition, when a `then` callback that
#' belongs to a promise domain is invoked, then any new calls to `then` will
#' also belong to that promise domain. In other words, a promise domain
#' "infects" not only the immediate calls to `then`, but also to "nested" calls
#' to `then`.
#'
#' For more background, read the
#' [original design doc](https://gist.github.com/jcheng5/b1c87bb416f6153643cd0470ac756231).
#'
#' For examples, see the source code of the Shiny package, which uses promise
#' domains extensively to manage graphics devices and reactivity.
#'
#' @param domain A promise domain object to install while `expr` is evaluated.
#' @param expr Any R expression, to be evaluated under the influence of
#'   `domain`.
#' @param replace If `FALSE`, then the effect of the `domain` will be added
#'   to the effect of any currently active promise domain(s). If `TRUE`, then
#'   the current promise domain(s) will be ignored for the duration of the
#'   `with_promise_domain` call.
#'
#' @export
with_promise_domain <- function(domain, expr, replace = FALSE) {
  oldval <- current_promise_domain()
  if (replace)
    globals$domain <- domain
  else
    globals$domain <- compose_domains(oldval, domain)
  on.exit(globals$domain <- oldval)

  if (!is.null(domain))
    domain$wrapSync(expr)
  else
    force(expr)
}

# Like with_promise_domain, but doesn't include the wrapSync call.
reenter_promise_domain <- function(domain, expr, replace = FALSE) {
  oldval <- current_promise_domain()
  if (replace)
    globals$domain <- domain
  else
    globals$domain <- compose_domains(oldval, domain)
  on.exit(globals$domain <- oldval)

  force(expr)
}

#' @param wrapOnFulfilled A function that takes a single argument: a function
#'   that was passed as an `onFulfilled` argument to [then()]. The
#'   `wrapOnFulfilled` function should return a function that is suitable for
#'   `onFulfilled` duty.
#' @param wrapOnRejected A function that takes a single argument: a function
#'   that was passed as an `onRejected` argument to [then()]. The
#'   `wrapOnRejected` function should return a function that is suitable for
#'   `onRejected` duty.
#' @param wrapSync A function that takes a single argument: a (lazily evaluated)
#'   expression that the function should [force()]. This expression represents
#'   the `expr` argument passed to [with_promise_domain()]; `wrapSync` allows
#'   the domain to manipulate the environment before/after `expr` is evaluated.
#' @param onError A function that takes a single argument: an error. `onError`
#'   will be called whenever an exception occurs in a domain (that isn't caught
#'   by a `tryCatch`). Providing an `onError` callback doesn't cause errors to
#'   be caught, necessarily; instead, `onError` callbacks behave like calling
#'   handlers.
#' @param ... Arbitrary named values that will become elements of the promise
#'   domain object, and can be accessed as items in an environment (i.e. using
#'   `[[` or `$`).
#' @param wrapOnFinally A function that takes a single argument: a function
#'   that was passed as an `onFinally` argument to [then()]. The
#'   `wrapOnFinally` function should return a function that is suitable for
#'   `onFinally` duty. If `wrapOnFinally` is `NULL` (the default), then the
#'   domain will use both `wrapOnFulfilled` and `wrapOnRejected` to wrap the
#'   `onFinally`. If it's important to distinguish between normal
#'   fulfillment/rejection handlers and finally handlers, then be sure to
#'   provide `wrapOnFinally`, even if it's just [base::identity()].
#' @rdname with_promise_domain
#' @export
new_promise_domain <- function(
  wrapOnFulfilled = identity,
  wrapOnRejected = identity,
  wrapSync = force,
  onError = force,
  ...,
  wrapOnFinally = NULL
) {
  list2env(list(
    wrapOnFulfilled = wrapOnFulfilled,
    wrapOnRejected = wrapOnRejected,
    wrapOnFinally = wrapOnFinally,
    wrapSync = wrapSync,
    onError = onError,
    ...
  ), parent = emptyenv())
}


compose_domains <- function(base, new) {
  if (is.null(base)) {
    return(new)
  }

  list(
    wrapOnFulfilled = function(onFulfilled) {
      # Force eager evaluation of base$wrapOnFulfilled(onFulfilled)
      base <- base$wrapOnFulfilled(onFulfilled)
      new$wrapOnFulfilled(base)
    },
    wrapOnRejected = function(onRejected) {
      # Force eager evaluation of base$wrapOnRejected(onRejected)
      base <- base$wrapOnRejected(onRejected)
      new$wrapOnRejected(base)
    },
    # Only include the new wrapSync, assuming that we've already applied the
    # base domain's wrapSync. This assumption won't hold if we either export
    # compose_domains in the future, or if we use it in cases where the base
    # domain isn't currently active.
    wrapSync = new$wrapSync,
    onError = function(e) {
      base$onError(e)
      new$onError(e)
    }
  )
}

without_promise_domain <- function(expr) {
  with_promise_domain(NULL, expr, replace = TRUE)
}

#' @importFrom magrittr %>%
#' @export
magrittr::"%>%"

#' @importFrom magrittr %T>%
#' @export
magrittr::"%T>%"

#' Promise pipe operators
#'
#' Promise-aware pipe operators, in the style of [magrittr](https://CRAN.R-project.org/package=magrittr/vignettes/magrittr.html).
#' Like magrittr pipes, these operators can be used to chain together pipelines
#' of promise-transforming operations. Unlike magrittr pipes, these pipes wait
#' for promise resolution and pass the unwrapped value (or error) to the `rhs`
#' function call.
#'
#' The `>` variants are for handling successful resolution, the `!` variants are
#' for handling errors. The `T` variants of each return the lhs instead of the
#' rhs, which is useful for pipeline steps that are used for side effects
#' (printing, plotting, saving).
#'
#' 1. \code{promise \%...>\% func()} is equivalent to \code{promise \%>\% then(func)}.
#' 2. \code{promise \%...!\% func()} is equivalent to \code{promise \%>\% catch(func)}.
#' 3. \code{promise \%...T>\% func()} is equivalent to \code{promise \%T>\% then(func)}.
#' 4. \code{promise \%...T!\% func()} is equivalent to \code{promise \%T>\%
#' catch(func)} or \code{promise \%>\% catch(func, tee = TRUE)}.
#'
#' One situation where 3. and 4. above break down is when `func()` throws an
#' error, or returns a promise that ultimately fails. In that case, the failure
#' will be propagated by our pipe operators but not by the
#' magrittr-plus-function "equivalents".
#'
#' For simplicity of implementation, we do not support the magrittr feature of
#' using a `.` at the head of a pipeline to turn the entire pipeline into a
#' function instead of an expression.
#'
#' @param lhs A promise object.
#' @param rhs A function call using the magrittr semantics. It can return either
#'   a promise or non-promise value, or throw an error.
#'
#' @examples
#' \dontrun{
#' library(future)
#' plan(multisession)
#'
#' future_promise(cars) %...>%
#'   head(5) %...T>%
#'   print()
#'
#' # If the read.csv fails, resolve to NULL instead
#' future_promise(read.csv("http://example.com/data.csv")) %...!%
#'   { NULL }
#' }
#'
#' @return A new promise.
#'
#' @seealso https://rstudio.github.io/promises/articles/overview.html#using-pipes
#'
#' @name pipes
#' @export
`%...>%` <- function(lhs, rhs) {
  # the parent environment
  parent <- parent.frame()

  # the environment in which to evaluate pipeline
  env <- new.env(parent = parent)

  parts <- match.call()
  func <- pipeify_rhs(parts[[3L]], env)
  then(lhs, func)
}

#' @rdname pipes
#' @export
`%...T>%` <- function(lhs, rhs) {
  # the parent environment
  parent <- parent.frame()

  # the environment in which to evaluate pipeline
  env <- new.env(parent = parent)

  parts <- match.call()
  func <- pipeify_rhs(parts[[3L]], env)
  lhs %>%
    then(func) %>%
    then(function(value) lhs)
}

#' @rdname pipes
#' @export
`%...!%` <- function(lhs, rhs) {
  # the parent environment
  parent <- parent.frame()

  # the environment in which to evaluate pipeline
  env <- new.env(parent = parent)

  parts <- match.call()
  func <- pipeify_rhs(parts[[3L]], env)
  catch(lhs, func)
}

#' @rdname pipes
#' @export
`%...T!%` <- function(lhs, rhs) {
  # the parent environment
  parent <- parent.frame()

  # the environment in which to evaluate pipeline
  env <- new.env(parent = parent)

  parts <- match.call()
  func <- pipeify_rhs(parts[[3L]], env)
  catch(lhs, func, tee = TRUE)
}

has.visible <- function(func) {
  ".visible" %in% names(formals(func))
}

pipeify_rhs <- function(rhs, env) {
  if (is_parenthesized(rhs)) {
    rhs <- eval(rhs, env, env)
  }

  rhs <- if (is_funexpr(rhs)) {
    rhs
  } else if (is_lambda(rhs)) {
    # We can remove this conditional if we want this behavior to be supported.
    # The next conditional checks for is_lambda too, and does the right thing.
    # Keeping the error for now in deference to magrittr pipe behavior.
    stop("Anonymous functions must be parenthesized")
  } else if (is_function(rhs) || is_colexpr(rhs) || is_lambda(rhs)) {
    # This block diverges from magrittr because we have an optional .visible
    # argument that can be passed to the function. If the function takes a
    # parameter called .visible then we will pass it, otherwise no.

    real_rhs <- if (is.function(rhs)) {
      rhs
    } else {
      eval(rhs, env, env)
    }

    if (has.visible(real_rhs)) {
      as.call(list(rhs, quote(.), .visible = quote(.visible)))
    } else {
      as.call(list(rhs, quote(.)))
    }
  } else if (is_first(rhs)) {
    prepare_first(rhs)
  } else {
    rhs
  }

  eval(call("function", as.pairlist(alist(.=, .visible=)), rhs), env, env)
}

# Determine whether an expression counts as a function in a magrittr chain.
#
# @param a non-evaluated expression.
# @return logical - TRUE if expr represents a function, FALSE otherwise.
is_function <- function(expr)
{
  is.symbol(expr) || is.function(expr)
}

# Prepare a magrittr rhs of funtion type
#
# @param a an expression which passes `is_function`
# @return an expression prepared for functional sequence construction.
prepare_function <- function(f)
{
  as.call(list(f, quote(.)))
}

is_lambda <- function(expr) {
  is.call(expr) && identical(expr[[1L]], quote(`function`))
}

#' Combine multiple promise objects
#'
#' Use `promise_all` to wait for multiple promise objects to all be successfully
#' fulfilled. Use `promise_race` to wait for the first of multiple promise
#' objects to be either fulfilled or rejected.
#'
#' @param ... Promise objects. Either all arguments must be named, or all
#'   arguments must be unnamed. If `.list` is provided, then these arguments are
#'   ignored.
#' @param .list A list of promise objects--an alternative to `...`.
#'
#' @return A promise.
#'
#'   For `promise_all`, if all of the promises were successful, the returned
#'   promise will resolve to a list of the promises' values; if any promise
#'   fails, the first error to be encountered will be used to reject the
#'   returned promise.
#'
#'   For `promise_race`, the first of the promises to either fulfill or reject
#'   will be passed through to the returned promise.
#'
#' @examples
#' p1 <- promise(~later::later(~resolve(1), delay = 1))
#' p2 <- promise(~later::later(~resolve(2), delay = 2))
#'
#' # Resolves after 1 second, to the value: 1
#' promise_race(p1, p2) %...>% {
#'   cat("promise_race:\n")
#'   str(.)
#' }
#'
#' # Resolves after 2 seconds, to the value: list(1, 2)
#' promise_all(p1, p2) %...>% {
#'   cat("promise_all:\n")
#'   str(.)
#' }
#'
#' @export
promise_all <- function(..., .list = NULL) {
  if (missing(.list)) {
    .list <- list(...)
  }

  if (length(.list) == 0) {
    return(promise(~resolve(list())))
  }

  # Verify that .list members are either all named or all unnamed
  nameCount <- sum(nzchar(names(.list)))
  if (nameCount != 0 && nameCount != length(.list)) {
    stop("promise_all expects promise arguments (or list) to be either all named or all unnamed")
  }

  done <- list()
  results <- list()

  promise(function(resolve, reject) {
    keys <- if (is.null(names(.list))) {
      1:length(.list)
    } else {
      names(.list)
    }

    lapply(keys, function(key) {
      done[[key]] <<- FALSE
      # Forces correct/deterministic ordering of the result list's elements
      results[[key]] <<- NA

      then(.list[[key]],
        onFulfilled = function(value) {
          # Save the result so we can return it to the user.
          # This weird assignment is similar to `results[[key]] <- value`, except
          # that it handles NULL values correctly.
          results[key] <<- list(value)

          # Record the fact that the promise was completed.
          done[[key]] <<- TRUE
          # If all of the tasks are done, resolve.
          if (all(as.logical(done))) {
            resolve(results)
          }
        },
        onRejected = function(reason) {
          # TODO: Cancel promises that are still running
          reject(reason)
        }
      )
    })
  })
}

#' @rdname promise_all
#' @export
promise_race <- function(..., .list = NULL) {
  if (missing(.list)) {
    .list <- list(...)
  }

  promise(function(resolve, reject) {
    lapply(.list, function(promise) {
      then(promise, resolve, reject)
    })
  })
}

#' Promise-aware lapply/map
#'
#' Similar to [`base::lapply()`] or [`purrr::map`], but promise-aware: the `.f`
#' function is permitted to return promises, and while `lapply` returns a list,
#' `promise_map` returns a promise that resolves to a similar list (of resolved
#' values only, no promises).
#'
#' `promise_map` processes elements of `.x` serially; that is, if `.f(.x[[1]])`
#' returns a promise, then `.f(.x[[2]])` will not be invoked until that promise
#' is resolved. If any such promise rejects (errors), then the promise returned
#' by `promise_map` immediately rejects with that err.
#'
#' @param .x A vector (atomic or list) or an expression object (but not a
#'   promise). Other objects (including classed objects) will be coerced by
#'   base::as.list.
#' @param .f The function to be applied to each element of `.x`. The function is
#'   permitted, but not required, to return a promise.
#' @param ... Optional arguments to `.f`.
#' @return A promise that resolves to a list (of values, not promises).
#'
#' @examples
#' # Waits x seconds, then returns x*10
#' wait_this_long <- function(x) {
#'   promise(~later::later(~{
#'     resolve(x*10)
#'   }, delay = x))
#' }
#'
#' promise_map(list(A=1, B=2, C=3), wait_this_long) %...>%
#'   print()
#'
#' @export
promise_map <- function(.x, .f, ...) {
  .f <- match.fun(.f)
  if (!is.vector(.x) || is.object(.x))
    .x <- as.list(.x)
  x_names <- names(.x)
  results <- vector("list", length(.x))

  do_next <- function(pos) {
    if (pos > length(results)) {
      return(stats::setNames(results, x_names))
    } else {
      # The next line may throw, that's fine, it will be caught by resolve() and
      # reject the promise
      this_result <- .f(.x[[pos]], ...)
      promise_resolve(this_result) %...>%
        (function(this_value) {
          results[[pos]] <<- this_value
          do_next(pos + 1)
        })
    }
  }

  promise(function(resolve, reject) {
    resolve(do_next(1))
  })
}

#' Promise-aware version of Reduce
#'
#' Similar to [`purrr::reduce`] (left fold), but the function `.f` is permitted
#' to return a promise. `promise_reduce` will wait for any returned promise to
#' resolve before invoking `.f` with the next element; in other words, execution
#' is serial. `.f` can return a promise as output but should never encounter a
#' promise as input (unless `.x` itself is a list of promises to begin with, in
#' which case the second parameter would be a promise).
#'
#' @param .x A vector or list to reduce. (Not a promise.)
#' @param .f A function that takes two parameters. The first parameter will be
#'   the "result" (initially `.init`, and then set to the result of the most
#'   recent call to `func`), and the second parameter will be an element of `.x`.
#' @param ... Other arguments to pass to `.f`
#' @param .init The initial result value of the fold, passed into `.f` when it
#'   is first executed.
#'
#' @return A promise that will resolve to the result of calling `.f` on the last
#'   element (or `.init` if `.x` had no elements). If any invocation of `.f`
#'   results in an error or a rejected promise, then the overall
#'   `promise_reduce` promise will immediately reject with that error.
#'
#' @examples
#' # Returns a promise for the sum of e1 + e2, with a 0.5 sec delay
#' slowly_add <- function(e1, e2) {
#'   promise(~later::later(~resolve(e1 + e2), delay = 0.5))
#' }
#'
#' # Prints 55 after a little over 5 seconds
#' promise_reduce(1:10, slowly_add, .init = 0) %...>% print()
#'
#' @export
promise_reduce <- function(.x, .f, ..., .init) {
  p <- promise_resolve(.init)
  lapply(.x, function(item) {
    p <<- p %...>% .f(item, ...)
  })
  p
}

# Placeholder to make R cmd check.
# * Need purrr for docs, but don't need purrr for functionality
# * Get failure if we don't include it
# * Get failure if we do include it, but don't use it. So using it below.
function() {
  purrr::reduce
}

Promise <- function() {
  self <- new.env(parent = emptyenv())
  private <- new.env(parent = emptyenv())

  private$state <- "pending"
  private$value <- NULL
  private$visible <- TRUE
  private$publicResolveRejectCalled <- FALSE
  private$onFulfilled <- list()
  private$onRejected <- list()
  private$onFinally <- list()
  private$rejectionHandled <- FALSE

  private$doResolve <- function(value) {
    val <- withVisible(value)
    value <- val$value
    visible <- val$visible

    if (is.promising(value)) {
      value <- as.promise(value)
      if (identical(self, attr(value, "promise_impl", exact = TRUE))) {
        return(private$doReject(simpleError(
          "Chaining cycle detected for promise"
        )))
      }
      # This then() call doesn't need promise domains; semantically, it doesn't
      # really exist, as it's just a convenient way to implement the new promise
      # inhabiting the old promise's corpse.
      without_promise_domain({
        value$then(
          private$doResolve,
          private$doReject
        )
      })
    } else {
      private$doResolveFinalValue(value, visible)
    }
  }

  private$doReject <- function(reason) {
    if (is.promising(reason)) {
      reason <- as.promise(reason)
      # This then() call doesn't need promise domains; semantically, it doesn't
      # really exist, as it's just a convenient way to implement the new promise
      # inhabiting the old promise's corpse.
      without_promise_domain({
        reason$then(
          private$doResolve,
          private$doReject
        )
      })
    } else {
      private$doRejectFinalReason(reason)
    }
  }

  # These "final" versions of resolve/reject are for when we've
  # established that the value/reason is not itself a promise.
  private$doResolveFinalValue <- function(value, visible) {
    private$value <- value
    private$visible <- visible
    private$state <- "fulfilled"

    later(function() {
      lapply(private$onFulfilled, function(f) {
        f(private$value, private$visible)
      })
      private$onFulfilled <- list()
    })
  }

  private$doRejectFinalReason <- function(reason) {
    private$value <- reason
    private$state <- "rejected"

    later(function() {
      lapply(private$onRejected, function(f) {
        private$rejectionHandled <- TRUE
        f(private$value)
      })
      private$onRejected <- list()

      later(
        function() {
          if (!private$rejectionHandled) {
            # warning() was unreliable here
            cat(
              file = stderr(),
              "Unhandled promise error: ",
              reason$message,
              "\n",
              sep = ""
            )
          }
        }
      )
    })
  }

  self$status <- function() {
    private$state
  }

  self$resolve <- function(value) {
    # Only allow this to be called once, then no-op.
    if (private$publicResolveRejectCalled) {
      return(invisible())
    }
    private$publicResolveRejectCalled <- TRUE

    tryCatch(
      {
        # Important: Do not trigger evaluation of value before
        # passing to doResolve. doResolve calls withVisible() on
        # value, so evaluating it before that point will cause
        # the visibility to be lost.
        private$doResolve(value)
      },
      error = function(err) {
        private$doReject(err)
      }
    )

    invisible()
  }

  self$reject <- function(reason) {
    # Only allow this to be called once, then no-op.
    if (private$publicResolveRejectCalled) {
      return(invisible())
    }
    private$publicResolveRejectCalled <- TRUE

    tryCatch(
      {
        force(reason)
        if (is.character(reason)) {
          reason <- simpleError(reason)
        }
        private$doReject(reason)
      },
      error = function(err) {
        private$doReject(err)
      }
    )

    invisible()
  }

  self$then <- function(
    onFulfilled = NULL,
    onRejected = NULL,
    onFinally = NULL
  ) {
    onFulfilled <- normalizeOnFulfilled(onFulfilled)
    onRejected <- normalizeOnRejected(onRejected)
    if (!is.function(onFinally)) {
      onFinally <- NULL
    }

    promise2 <- promise(function(resolve, reject) {
      res <- promiseDomain$onThen(onFulfilled, onRejected, onFinally)

      if (!is.null(res)) {
        onFulfilled <- res$onFulfilled
        onRejected <- res$onRejected
      }

      handleFulfill <- function(value, visible) {
        if (is.function(onFulfilled)) {
          resolve(onFulfilled(value, visible))
        } else {
          resolve(if (visible) value else invisible(value))
        }
      }

      handleReject <- function(reason) {
        if (is.function(onRejected)) {
          # Yes, resolve, not reject.
          resolve(onRejected(reason))
        } else {
          # Yes, reject, not resolve.
          reject(reason)
        }
      }

      if (private$state == "pending") {
        private$onFulfilled <- c(
          private$onFulfilled,
          list(
            handleFulfill
          )
        )
        private$onRejected <- c(
          private$onRejected,
          list(
            handleReject
          )
        )
      } else if (private$state == "fulfilled") {
        later(function() {
          handleFulfill(private$value, private$visible)
        })
      } else if (private$state == "rejected") {
        later(function() {
          private$rejectionHandled <- TRUE
          handleReject(private$value)
        })
      } else {
        stop("Unexpected state ", private$state)
      }
    })

    invisible(promise2)
  }

  self$catch <- function(onRejected) {
    invisible(self$then(onRejected = onRejected))
  }

  self$finally <- function(onFinally) {
    invisible(self$then(
      onFinally = onFinally
    ))
  }

  self$format <- function() {
    if (private$state == "pending") {
      "<Promise [pending]>"
    } else {
      classname <- class(private$value)[[1]]
      if (length(classname) == 0) {
        classname <- ""
      }

      sprintf("<Promise [%s: %s]>", private$state, classname)
    }
  }

  self$.__enclos_env__ <- list(private = private)
  class(self) <- "Promise"
  self
}

normalizeOnFulfilled <- function(onFulfilled) {
  if (!is.function(onFulfilled)) {
    if (!is.null(onFulfilled)) {
      warning("`onFulfilled` must be a function or `NULL`")
    }
    return(NULL)
  }

  args <- formals(onFulfilled)
  arg_count <- length(args)

  if (arg_count >= 2 && names(args)[[2]] == ".visible") {
    onFulfilled
  } else if (arg_count > 0) {
    function(value, .visible) {
      if (isTRUE(.visible)) {
        onFulfilled(value)
      } else {
        onFulfilled(invisible(value))
      }
    }
  } else {
    function(value, .visible) {
      onFulfilled()
    }
  }
}

normalizeOnRejected <- function(onRejected) {
  if (!is.function(onRejected)) {
    if (!is.null(onRejected)) {
      warning("`onRejected` must be a function or `NULL`")
    }
    return(NULL)
  }

  args <- formals(onRejected)
  arg_count <- length(args)

  if (arg_count >= 1) {
    onRejected
  } else if (arg_count == 0) {
    function(reason) {
      onRejected()
    }
  }
}

#' Create a new promise object
#'
#' `promise()` creates a new promise. A promise is a placeholder object for the
#' eventual result (or error) of an asynchronous operation. This function is not
#' generally needed to carry out asynchronous programming tasks; instead, it is
#' intended to be used mostly by package authors who want to write asynchronous
#' functions that return promises.
#'
#' The `action` function should be a piece of code that returns quickly, but
#' initiates a potentially long-running, asynchronous task. If/when the task
#' successfully completes, call `resolve(value)` where `value` is the result of
#' the computation (like the return value). If the task fails, call
#' `reject(reason)`, where `reason` is either an error object, or a character
#' string.
#'
#' It's important that asynchronous tasks kicked off from `action` be coded very
#' carefully--in particular, all errors must be caught and passed to `reject()`.
#' Failure to do so will cause those errors to be lost, at best; and the caller
#' of the asynchronous task will never receive a response (the asynchronous
#' equivalent of a function call that never returns, i.e. hangs).
#'
#' The return value of `action` will be ignored.
#'
#' @section `action=` formulas:
#'
#' `r lifecycle::badge("superseded")`
#'
#' With `{promises}` depending on R >= 4.1, the shorthand of a formula, `~
#' fn(.)` for `action` is no longer recommended by the `{promises}` package or
#' tidyverse (for example,
#' [`{purrr}`](https://github.com/tidyverse/purrr/commit/670c3ed9920f15da0d4175068ecddc41f0f1f335#diff-c4dcc43795da5c7f6bf5a94d957b5507ce795fedd6d3eb092ccad03678c4f76dR15))
#' as we now have access to the function shorthand, `\(x) fn(x)`. Please update
#' your `action` code to use the new function shorthand syntax `\(resolve,
#' reject) resolve(arg1, arg2)` instead of `~ { resolve(arg1, arg2) }`. The
#' magically created `resolve`/`reject` functions can be confusing when chained
#' with other methods.
#'
#' @param action A function with signature `function(resolve, reject)`.
#'
#' @return A promise object (see \code{\link{then}}).
#'
#' @examples
#' # Create a promise that resolves to a random value after 2 secs
#' p1 <- promise(\(resolve, reject) {
#'   later::later(\() resolve(runif(1)), delay = 2)
#' })
#'
#' p1 |> then(print)
#'
#' # Create a promise that errors immediately
#' p2 <- promise(\(resolve, reject) {
#'   reject("An error has occurred")
#' })
#' then(p2,
#'   onFulfilled = \(value) message("Success"),
#'   onRejected = \(err) message("Failure")
#' )
#'
#' @export
promise <- function(action) {
  if (is.function(action)) {
    if (length(formals(action)) != 2) {
      stop("'action' function must have two arguments")
    }
  } else if (inherits(action, "formula")) {
    if (length(action) != 2) {
      stop("'action' must be a function or one-sided formula")
    }
  } else {
    stop("Invalid action argument--should be a function")
  }

  p <- Promise()

  tryCatch(
    {
      if (is.function(action)) {
        action(p$resolve, p$reject)
      } else if (inherits(action, "formula")) {
        eval(
          action[[2]],
          list(
            resolve = p$resolve,
            reject = p$reject,
            return = function(value) {
              warning("Can't return a value from a promise, use resolve/reject")
            }
          ),
          environment(action)
        )
      }
    },
    error = function(e) {
      if (p$status() == "pending") {
        p$reject(e)
      } else {
        # Too late to do anything useful. Just notify.
        warning(e)
      }
    }
  )
  structure(
    list(
      then = p$then,
      catch = p$catch,
      finally = p$finally
    ),
    class = "promise",
    promise_impl = p
  )
}

#' Create a resolved or rejected promise
#'
#' Helper functions to conveniently create a promise that is resolved to the
#' given value (or rejected with the given reason).
#'
#' @param value A value, or promise, that the new promise should be resolved to.
#'   This expression will be lazily evaluated, and if evaluating the expression
#'   raises an error, then the new promise will be rejected with that error as
#'   the reason.
#' @param reason An error message string, or error object.
#'
#' @examples
#' promise_resolve(mtcars) |>
#'   then(head) |>
#'   then(print)
#'
#' promise_reject("Something went wrong") |>
#'   catch(tee = TRUE, \(e) message(conditionMessage(e)))
#'
#' @export
promise_resolve <- function(value) {
  promise(function(resolve, reject) resolve(value))
}

#' @rdname promise_resolve
#' @export
promise_reject <- function(reason) {
  promise(function(resolve, reject) reject(reason))
}

#' Coerce to a promise
#'
#' Use `is.promise` to determine whether an R object is a promise. Use
#' `as.promise` (an S3 generic method) to attempt to coerce an R object to a
#' promise, and `is.promising` (another S3 generic method) to test whether
#' `as.promise` is supported. [mirai::mirai] objects have an `as.promise` method
#' defined in the mirai package, and this package provides one for converting
#' [future::Future] objects into promises.
#'
#' @param x An R object to test or coerce.
#'
#' @return `as.promise` returns a promise object, or throws an error if the
#'   object cannot be converted.
#'
#'   `is.promise` returns `TRUE` if the given value is a promise object, and
#'   `FALSE` otherwise.
#'
#'   `is.promising` returns `TRUE` if the given value is a promise object or
#'   if it can be converted to a promise object using `as.promise`, and `FALSE`
#'   otherwise.
#'
#' @export
is.promise <- function(x) {
  inherits(x, "promise")
}

#' @rdname is.promise
#' @export
is.promising <- function(x) {
  UseMethod("is.promising")
}

#' @export
is.promising.default <- function(x) {
  FALSE
}

#' @export
is.promising.promise <- function(x) {
  TRUE
}

#' @export
is.promising.Future <- function(x) {
  TRUE
}

#' @rdname is.promise
#' @export
as.promise <- function(x) {
  UseMethod("as.promise", x)
}

#' @export
as.promise.promise <- function(x) {
  x
}

#' @export
as.promise.Future <- function(x) {
  # We want to create a promise only once for each Future object, and cache it
  # as an attribute. This spares us from having multiple polling loops waiting
  # for the same Future.
  #
  ## Note: 2025-07-barret:
  ## Active bindings will not work as `{future}` does not proactively retrieve results. A poll must be used.
  #
  ## Note: 2025-08-barret:
  ## From conversation with Henrik, while it may be possible to use `immediateCondition` within future to _immediately_ send a value back from a worker to the main R session,
  ## `future::resolved()` is still required to pull the _immediate_ value into the R session (even when the future has not fully resolved yet).
  ## Therefore, the polling mechanism below is still required.
  ## For comparison, mirai has a background process that waits for results and will interrupt the main R session with the value.

  cached <- attr(x, "converted_promise", exact = TRUE)
  if (!is.null(cached)) {
    return(cached)
  }

  p <- promise(function(resolve, reject) {
    poll_interval <- 0.1
    check <- function() {
      # timeout = 0 is important, the default waits for 200ms
      is_resolved <- tryCatch(
        {
          future::resolved(x, timeout = 0)
        },
        FutureError = function(e) {
          reject(e)
          TRUE
        }
      )
      if (is_resolved) {
        tryCatch(
          {
            result <- future::value(x, signal = TRUE)
            resolve(result)
          },
          FutureError = function(e) {
            reject(e)
            TRUE
          },
          error = function(e) {
            reject(e)
          }
        )
      } else {
        later(check, poll_interval)
      }
    }
    check()
  })

  # Store the new promise for next time
  attr(x, "converted_promise") <- p
  p
}

#' @export
as.promise.default <- function(x) {
  # TODO: If x is an error or try-error, should this return a rejected promise?
  stop(
    "Don't know how to convert object of class ",
    class(x)[[1L]],
    " into a promise"
  )
}

#' Fulfill a promise
#'
#' Use these functions to satisfy a promise with either success (\code{resolve})
#' or failure (\code{reject}). These functions are not exported, but rather, are
#' passed as arguments to the \code{action} function you pass to a [promise]
#' constructor.
#'
#' @param value The result from a successful calculation.
#' @param reason An error or string that explains why the operation failed.
#'
#' @keywords internal
resolve <- function(value = NULL) {
  stop("resolve() must be called from within a promise constructor")
}

#' @rdname resolve
#' @keywords internal
reject <- function(reason) {
  stop("reject() must be called from within a promise constructor")
}

library(testthat)

source("common.R")

describe("2.2. The `then` Method", {
  # A promise must provide a `then` method to access its current or eventual
  # value or reason.
  #
  # A promise’s then method accepts two arguments:
  # promise.then(onFulfilled, onRejected)

  describe("2.2.1. Both onFulfilled and onRejected are optional arguments:", {
    it("2.2.1.1. If onFulfilled is not a function, it must be ignored.", {
      # NOTE: This behavior varies; we throw an error
      # p <- promise(~resolve(10)) %>% then(20)
      # expect_identical(extract(p), 10)
    })
    it("2.2.1.1. If onRejected is not a function, it must be ignored.", {
      # NOTE: This behavior varies; we throw an error
      # p <- promise(~reject("foo")) %>% then(onRejected = "bar")
      # expect_error(extract(p), "foo")
    })
  })

  describe("2.2.2. If onFulfilled is a function:", {
    it("2.2.2.1. it must be called after promise is fulfilled, with promise’s value as its first argument.", {
      x <- NULL
      p <- ext_promise()

      p$promise %>% then(function(value) { x <<- value })
      wait_for_it()
      expect_identical(x, NULL)

      p$resolve(10)
      wait_for_it()
      expect_identical(x, 10)
    })
    it("2.2.2.2. it must not be called before promise is fulfilled.", {

    })
    it("2.2.2.3. it must not be called more than once.", {

    })
  })
  describe("2.2.3. If onRejected is a function,", {
    it("2.2.3.1. it must be called after promise is rejected, with promise’s reason as its first argument.", {
      x <- NULL
      p <- ext_promise()

      p$promise %>% then(onRejected = function(reason) { x <<- reason })
      wait_for_it()
      expect_identical(x, NULL)

      p$reject(simpleError("boom"))
      wait_for_it()
      expect_identical(x, simpleError("boom"))
    })
  })
  describe("2.2.4. onFulfilled or onRejected must not be called until the execution context stack contains only platform code. [3.1].", {
    it(" ", {
      x <- NULL
      promise(~resolve(TRUE)) %>% then(function(value) {x <<- value})
      expect_identical(x, NULL)
      wait_for_it()
      expect_identical(x, TRUE)
    })
  })
  describe("2.2.5. onFulfilled and onRejected must be called as functions (i.e. with no this value). [3.2]", {
    # Not relevant for R.
  })
  describe("2.2.6. `then` may be called multiple times on the same promise.", {
    it("2.2.6.1. If/when promise is fulfilled, all respective onFulfilled callbacks must execute in the order of their originating calls to then.", {
      p <- ext_promise()
      callbacks_called <- 0L
      results <- new.env(parent = emptyenv())

      lapply(1:10, function(i) {
        results[[as.character(i)]] <- p$promise %>%
          then(function(value) {
            callbacks_called <<- callbacks_called + 1L
            expect_identical(callbacks_called, i)
            value
          })
      })

      p$resolve(cars)
      wait_for_it()

      lapply(as.list(results), function(x) {
        expect_identical(extract(x), cars)
      })
    })
  })

  describe("2.2.6.2. If/when promise is rejected, all respective onRejected callbacks must execute in the order of their originating calls to then.", {
    p <- ext_promise()
    callbacks_called <- 0L
    results <- new.env(parent = emptyenv())

    lapply(1:10, function(i) {
      results[[as.character(i)]] <- p$promise %>%
        catch(function(err) {
          callbacks_called <<- callbacks_called + 1L
          expect_identical(callbacks_called, i)
          err
        })
    })

    p$reject(simpleError("an error"))
    wait_for_it()

    lapply(as.list(results), function(x) {
      expect_identical(extract(x), simpleError("an error"))
    })
  })

  describe("2.2.7. `then` must return a promise [3.3].", {
    it(" ", {
      promise(~{}) %>% then() %>% is.promise() %>% expect_true()
    })

    it("2.2.7.1. If either onFulfilled or onRejected returns a value x, run the Promise Resolution Procedure [[Resolve]](promise2, x).", {
      p1 <- promise(~resolve(TRUE)) %>% then(~"foo")
      expect_identical(extract(p1), "foo")

      p2 <- promise(~reject("boom")) %>% catch(~"bar")
      expect_identical(extract(p2), "bar")
    })

    it("2.2.7.2. If either onFulfilled or onRejected throws an exception e, promise2 must be rejected with e as the reason.", {
      p1 <- promise(~resolve(TRUE)) %>% then(~stop("foo"))
      expect_error(extract(p1), "^foo$")

      p2 <- promise(~reject("boom")) %>% catch(~stop("bar"))
      expect_error(extract(p2), "^bar$")
    })

    it("2.2.7.3. If onFulfilled is not a function and promise1 is fulfilled, promise2 must be fulfilled with the same value as promise1.", {
      p <- promise(~resolve("baz")) %>% then()
      expect_identical(extract(p), "baz")
    })

    it("2.2.7.4. If onRejected is not a function and promise1 is rejected, promise2 must be rejected with the same reason as promise1.", {
      p <- promise(~reject("qux")) %>% then()
      expect_error(extract(p), "^qux$")
    })
  })
})

test_that("promise resolution order matches expectations", {
  # See discussion in #151 - this is just the current behaviour. There may be
  # grounds for change if we find a legitimate reason for doing so.
  y <- 0L
  p1 <- promise_resolve(1) |> then(\(x) y <<- y + 1) |> then(\(x) y <<- y / 2)
  p2 <- promise_resolve(2) |> then(\(x) y <<- y + 3) |> then(\(x) y <<- y * 2)
  later::run_now()
  expect_equal(y, 4)
  later::run_now()
  expect_equal(y, 4)
})

library(testthat)

describe("then()", {
  it("honors .visible argument", {
    result <- NULL
    p <- promise(\(resolve, reject) resolve(invisible(1))) |>
      then(function(value, .visible) {
        result <<- list(value = value, visible = .visible)
      })
    p |> wait_for_it()
    expect_identical(result$value, 1)
    expect_identical(result$visible, FALSE)

    p <- promise(\(resolve, reject) resolve(1)) |>
      then(function(value, .visible) {
        result <<- list(value = value, visible = .visible)
      })
    p |> wait_for_it()
    expect_identical(result$value, 1)
    expect_identical(result$visible, TRUE)

    # .visible is preserved even with an intermediate then() or catch()
    p <- promise(\(resolve, reject) resolve(invisible(1))) |>
      then() |>
      catch(\(err) "what error?") |>
      then(function(value, .visible) {
        result <<- list(value = value, visible = .visible)
      })
    p |> wait_for_it()
    expect_identical(result$value, 1)
    expect_identical(result$visible, FALSE)
  })
  it("method ignores non-functions or NULL...", {
    p1 <- promise(\(resolve, reject) resolve(1))
    expect_warning(
      {
        p1 <- p1$then(10)
      },
      "`onFulfilled` must be a function or `NULL`",
      fixed = TRUE
    )
    expect_warning(
      {
        p1 <- p1$then(NULL)
      },
      NA
    )
    expect_identical(extract(p1), 1)
  })
  it("...but function only ignores NULL, not non-functions", {
    expect_error(promise(\(resolve, reject) resolve(1)) |> then(10))
    expect_error(promise(\(resolve, reject) resolve(1)) |> then(NULL), NA)
  })

  it("honors visibility with no .visible argument", {
    result <- NULL
    p <- promise_resolve(invisible(1))$then(function(value) {
      result <<- withVisible(value)
    })
    p |> wait_for_it()
    expect_identical(result$value, 1)
    expect_identical(result$visible, FALSE)

    result <- NULL
    p <- promise_resolve(2)$then(function(value) {
      result <<- withVisible(value)
    })
    p |> wait_for_it()
    expect_identical(result$value, 2)
    expect_identical(result$visible, TRUE)
  })
})

describe("catch()", {
  it("catches", {
    p <- ext_promise()
    p2 <- p$promise |> catch(\(err) TRUE)
    p$reject("boom")
    expect_identical(extract(p2), TRUE)
  })
  it("can throw", {
    p <- promise(\(resolve, reject) stop("foo")) |> catch(\(err) stop("bar"))
    expect_error(extract(p), "^bar$")
  })
  it("method ignores non-functions or NULL...", {
    p1 <- promise(\(resolve, reject) resolve(1))
    expect_warning(
      {
        p1 <- p1$catch(10)
      },
      "`onRejected` must be a function or `NULL`",
      fixed = TRUE
    )
    expect_warning(
      {
        p1 <- p1$catch(NULL)
      },
      NA
    )
    expect_identical(extract(p1), 1)
  })
  it("...but function only ignores NULL, not non-functions", {
    expect_error(promise(\(resolve, reject) resolve(1)) |> catch(10))
    expect_error(promise(\(resolve, reject) resolve(1)) |> catch(NULL), NA)
  })
})

describe("finally()", {
  it("calls back when a promise is resolved", {
    called <- FALSE
    p <- promise(\(resolve, reject) resolve(10)) |>
      finally(\() {
        called <<- TRUE
      })
    p |> wait_for_it()
    expect_identical(called, TRUE)
    expect_identical(extract(p), 10)
  })
  it("calls back when a promise is rejected", {
    called <- FALSE
    p <- promise(\(resolve, reject) reject("foobar"))
    p |>
      finally(\() {
        called <<- TRUE
      }) |>
      squelch_unhandled_promise_error() |>
      wait_for_it()
    expect_identical(called, TRUE)
    expect_error(extract(p), "^foobar$")
  })
  it("does not affect the return value of the promise", {
    p1 <- promise(\(resolve, reject) resolve(1)) |>
      finally(\() {
        20
      })
    expect_identical(extract(p1), 1)

    p2 <- promise(\(resolve, reject) reject("err")) |>
      finally(\() {
        20
      })
    expect_error(extract(p2), "^err$")
  })
  it("errors replace the result of the promise", {
    p1 <- promise(\(resolve, reject) resolve(1)) |>
      finally(\() {
        stop("boom")
      })
    expect_error(extract(p1), "^boom$")

    p2 <- promise(\(resolve, reject) reject("foo")) |>
      finally(\() {
        stop("bar")
      })
    expect_error(extract(p2), "^bar$")
  })
  it("method ignores non-functions or NULL...", {
    p1 <- promise(\(resolve, reject) resolve(1))$finally(10)$finally(NULL)
    expect_identical(extract(p1), 1)
  })
  it("...but function only ignores NULL, not non-functions", {
    expect_error(promise(\(resolve, reject) resolve(1)) |> finally(10))
    expect_error(promise(\(resolve, reject) resolve(1)) |> finally(NULL), NA)
  })
})

describe("future", {
  it("is treated as promise when used as resolution", {
    p <- promise_resolve(future::future(1))
    expect_identical(extract(p), 1)

    p2 <- promise_resolve(future::future(stop("boom")))
    expect_error(extract(p2))
  })

  it("is treated as promise when used as resolution", {
    p <- promise_reject(future::future(1))
    expect_identical(extract(p), 1)

    p2 <- promise_reject(future::future(stop("boom")))
    expect_error(extract(p2))
  })
})

skip_on_cran()
skip_on_os("windows") # timing is not consistent on Windows GHA

skip_if_not_installed("future", "1.21.0")
skip_if_not_installed("fastmap", "1.1.0")

source(test_path("common.R"))

local({
  ## Setup ##

  # Set up a plan with 2 future workers
  with_test_workers <- function(code) {
    # (Can not use a variable for workers if in a local({}))
    old_plan <- future::plan(future::multisession(workers = 2))
    on.exit({future::plan(old_plan)}, add = TRUE)

    force(code)
  }


  start <- Sys.time()
  time_diffs <- c()

  reset_baselines <- function() {
    start <<- Sys.time()
    time_diffs <<- c()
  }

  time_diff <- function() {
    difftime(Sys.time(), start, units = "secs")
  }

  # This function will print every `delay` seconds
  # The thing to notes is that the function will not execute unless the main R session is free
  # We should expect to see `run_every()` statements interleaved with `future_promise()` calls
  # We should NOT expect to see `run_every()` statements while `future::future()` is blocking the main R session
  run_every <- function(i = 0, max = 2 / delay, delay = 0.1) {
    if (i > max) return()
    time_diffs <<- c(time_diffs, time_diff())
    # Do it again, later
    later::later(function() { run_every(i + 1, max = max, delay = delay) }, delay = delay)
  }

  worker_jobs <- 8
  worker_job_time <- 1
  expected_total_time <- worker_jobs * worker_job_time / 2 # 2 workers

  do_future_test <- function(
    prom_fn = future_promise,
    # Introduce extra future workers mid execution
    block_mid_session = FALSE,
    # expect that the average finish lag time is less than 2 * n_time
    expect_reasonable_exec_lag_time = TRUE,
    # expect the lapply to finish in less than 1s
    expect_immediate_lapply = TRUE,
    # expect `run_every()` delay to be < 1s (Expected 0.1s)
    expect_no_main_blocking = TRUE
  ) {

    with_test_workers({
      # prep future sessions
      f1 <- future::future({1})
      f2 <- future::future({2})
      c(future::value(future::resolve(f1)), future::value(future::resolve(f2)))

      expect_true(future_worker_is_free())
      expect_equal(future::nbrOfWorkers(), 2)
      expect_equal(future::nbrOfFreeWorkers(), 2)

      reset_baselines()
      run_every()

      future_exec_times <- c()
      if (block_mid_session) {
        # Have `future` block the main R session 1 second into execution
        lapply(1:8, function(i) {
          later::later(
            function() {
              future::future({
                Sys.sleep(1)
                time_diff()
              }) %...>% {
                future_exec_times <<- c(future_exec_times, .)
              }
            },
            delay = 1
          )
        })
      }


      exec_times <- NA
      lapply(seq_len(worker_jobs), function(i) {
        prom_fn({
          Sys.sleep(worker_job_time)
          time_diff()
        })
      }) %>%
        promise_all(.list = .) %...>% {
          exec_times <<- unlist(.)
        }
      post_lapply_time_diff <- time_diff()

      wait_for_it()

      # expect that the average time is less than the expected total time
      expect_equal(median(exec_times) < expected_total_time, !block_mid_session)

      # expect prom_fn to take a reasonable amount of time to finish
      exec_times_lag <- exec_times[-1] - exec_times[-length(exec_times)]
      expect_equal(all(exec_times_lag < (2 * worker_job_time)), expect_reasonable_exec_lag_time)

      # post_lapply_time_diff should be ~ 0s
      expect_equal(post_lapply_time_diff < worker_job_time, expect_immediate_lapply)

      # time_diffs should never grow by more than 1s; (Expected 0.1)
      time_diffs_lag <- time_diffs[-1] - time_diffs[-length(time_diffs)]
      expect_equal(all(time_diffs_lag < worker_job_time), expect_no_main_blocking)
    })

  }


  test_that("future_promise() allows the main thread to keep the main R process open", {

    do_future_test(
      prom_fn = future_promise,
      # expect that the average finish lag time is less than 2 * n_time
      expect_reasonable_exec_lag_time = TRUE,
      # expect the lapply to finish in less than 1s
      expect_immediate_lapply = TRUE,
      # expect `run_every()` delay to be < 1s (Expected 0.1s)
      expect_no_main_blocking = TRUE
    )
  })


  test_that("future::future() does not keep the main process open when all workers are busy", {
    do_future_test(
      prom_fn = future::future,
      # expect that the average finish lag time is less than 2 * n_time
      expect_reasonable_exec_lag_time = TRUE,
      # expect the lapply to finish after 1s
      expect_immediate_lapply = FALSE,
      # expect one `run_every()` delay to be >= 1s (Expected 0.1s)
      expect_no_main_blocking = FALSE
    )
  })


  test_that("future_promise() recovers from losing all future workers", {

    do_future_test(
      prom_fn = future_promise,
      block_mid_session = TRUE,
      # expect that the average finish lag time is less than 2 * n_time
      # b/c `future::future()` blocks the main R session, expect >=1 unreasonable lag time
      expect_reasonable_exec_lag_time = FALSE,
      # expect the lapply to finish in less than 1s
      expect_immediate_lapply = TRUE,
      # expect one `run_every()` delay to be >= 1s (Expected 0.1s)
      # b/c `future::future()` blocks the main R session, expect >=1 unreasonable lag time
      expect_no_main_blocking = FALSE
    )
  })

})

test_that("hybrid_then() preserves visible values from expr (no callback)", {
  result <- hybrid_then(42)
  expect_equal(result, 42)
  # Check visibility by capturing with withVisible
  vis <- withVisible(hybrid_then(42))
  expect_true(vis$visible)
})

test_that("hybrid_then() with invisible expr (no callback)", {
  result <- hybrid_then(invisible(42))
  expect_equal(result, 42)
  vis <- withVisible(hybrid_then(invisible(42)))
  expect_false(vis$visible)
})

test_that("hybrid_then() with visible expr and on_success returning visible", {
  result <- hybrid_then(
    42,
    on_success = \(x) x + 8
  )
  expect_equal(result, 50)
  vis <- withVisible(hybrid_then(42, on_success = \(x) x + 8))
  expect_true(vis$visible)
})

test_that("hybrid_then() with visible expr and on_success returning invisible", {
  result <- hybrid_then(
    42,
    on_success = \(x) invisible(x + 8)
  )
  expect_equal(result, 50)
  vis <- withVisible(hybrid_then(42, on_success = \(x) invisible(x + 8)))
  expect_false(vis$visible)
})

test_that("hybrid_then() with invisible expr and on_success returning visible", {
  result <- hybrid_then(
    invisible(42),
    on_success = \(x) x + 8
  )
  expect_equal(result, 50)
  vis <- withVisible(hybrid_then(invisible(42), on_success = \(x) x + 8))
  expect_true(vis$visible)
})

test_that("hybrid_then() with invisible expr and on_success returning invisible", {
  result <- hybrid_then(
    invisible(42),
    on_success = \(x) invisible(x + 8)
  )
  expect_equal(result, 50)
  vis <- withVisible(hybrid_then(invisible(42), on_success = \(x) {
    invisible(x + 8)
  }))
  expect_false(vis$visible)
})

test_that("hybrid_then() with tee=TRUE preserves expr visibility (visible expr)", {
  result <- hybrid_then(
    42,
    on_success = \(x) invisible(999),
    tee = TRUE
  )
  expect_equal(result, 42)
  vis <- withVisible(hybrid_then(
    42,
    on_success = \(x) invisible(999),
    tee = TRUE
  ))
  expect_true(vis$visible)
})

test_that("hybrid_then() with tee=TRUE preserves expr visibility (invisible expr)", {
  result <- hybrid_then(
    invisible(42),
    on_success = \(x) 999, # callback returns visible
    tee = TRUE
  )
  expect_equal(result, 42)
  vis <- withVisible(hybrid_then(
    invisible(42),
    on_success = \(x) 999,
    tee = TRUE
  ))
  expect_false(vis$visible)
})

test_that("hybrid_then() on_success receives .visible parameter correctly (visible)", {
  visible_captured <- NULL
  result <- hybrid_then(
    42,
    on_success = \(x, .visible) {
      visible_captured <<- .visible
      x * 2
    }
  )
  expect_equal(result, 84)
  expect_true(visible_captured)
})

test_that("hybrid_then() on_success receives .visible parameter (invisible)", {
  visible_captured <- NULL
  result <- hybrid_then(
    invisible(42),
    on_success = \(x, .visible) {
      visible_captured <<- .visible
      x * 2
    }
  )
  expect_equal(result, 84)
  expect_false(visible_captured)
})

test_that("hybrid_then() on_success with .visible conditionally returning", {
  make_conditional <- function(val) {
    hybrid_then(
      val,
      on_success = \(x, .visible) {
        if (.visible) {
          x * 2
        } else {
          invisible(x * 2)
        }
      }
    )
  }

  vis1 <- withVisible(make_conditional(42))
  expect_equal(vis1$value, 84)
  expect_true(vis1$visible)

  vis2 <- withVisible(make_conditional(invisible(42)))
  expect_equal(vis2$value, 84)
  expect_false(vis2$visible)
})

test_that("hybrid_then() async with visible promise and on_success returning visible", {
  expect_assertions(n = 2, {
    result <- hybrid_then(
      promise_resolve(42),
      on_success = \(x) x + 8
    )
    expect_true(is.promise(result))
    result |> extract() |> expect_equal(50)
  })
})

test_that("hybrid_then() async with visible promise and on_success returning invisible", {
  expect_assertions(n = 2, {
    result <- hybrid_then(
      promise_resolve(42),
      on_success = \(x) invisible(x + 8)
    )
    expect_true(is.promise(result))
    result |> extract() |> expect_equal(50)
  })
})

test_that("hybrid_then() async visibility is handled by underlying then()", {
  expect_assertions(n = 3, {
    # The visibility semantics for async are handled by then()
    # Just verify the values work correctly
    visible_captured <- NULL
    result <- hybrid_then(
      promise_resolve(42),
      on_success = \(x, .visible) {
        visible_captured <<- .visible
        x * 2
      }
    )
    expect_true(is.promise(result))
    result |> extract() |> expect_equal(84)
    expect_true(visible_captured)
  })
})

test_that("hybrid_then() with tee=TRUE and visible expr", {
  # With tee=TRUE, expr visibility should be preserved regardless of callback
  vis_test <- function(expr_vis, callback_vis) {
    withVisible(
      hybrid_then(
        if (expr_vis) 42 else invisible(43),
        on_success = if (callback_vis) {
          \(x) x * 2
        } else {
          \(x) invisible(x * 3)
        },
        tee = TRUE
      )
    )
  }

  # Visible expr preserves visibility
  expect_equal(vis_test(TRUE, TRUE)$value, 42)
  expect_true(vis_test(TRUE, TRUE)$visible)

  expect_equal(vis_test(TRUE, FALSE)$value, 42)
  expect_true(vis_test(TRUE, FALSE)$visible)

  expect_equal(vis_test(FALSE, TRUE)$value, 43)
  expect_false(vis_test(FALSE, TRUE)$visible)

  expect_equal(vis_test(FALSE, FALSE)$value, 43)
  expect_false(vis_test(FALSE, FALSE)$visible)
})

test_that("hybrid_then() no callback with visible expr preserves visibility", {
  # Visible expr
  vis1 <- withVisible(hybrid_then(42))
  expect_equal(vis1$value, 42)
  expect_true(vis1$visible)
})

test_that("hybrid_then() no callback with invisible expr", {
  # Invisible expr via invisible()
  vis2 <- withVisible(hybrid_then(invisible(42)))
  expect_equal(vis2$value, 42)
  expect_false(vis2$visible)

  # Expression that returns invisible
  vis3 <- withVisible(hybrid_then({
    x <- 42
    invisible(x)
  }))
  expect_equal(vis3$value, 42)
  expect_false(vis3$visible)
})

test_that("hybrid_then() on_failure with invisible error recovery", {
  result <- hybrid_then(
    stop("error"),
    on_failure = \(err) invisible(0)
  )
  expect_equal(result, 0)
  vis <- withVisible(hybrid_then(stop("error"), on_failure = \(err) {
    invisible(0)
  }))
  expect_false(vis$visible)
})

test_that("hybrid_then() on_failure with visible error recovery", {
  result <- hybrid_then(
    stop("error"),
    on_failure = \(err) 0
  )
  expect_equal(result, 0)
  vis <- withVisible(hybrid_then(stop("error"), on_failure = \(err) 0))
  expect_true(vis$visible)
})

test_that("hybrid_then() handles synchronous values with on_success", {
  result <- hybrid_then(
    42,
    on_success = \(x) x + 8
  )
  expect_equal(result, 50)
})

test_that("hybrid_then() handles synchronous values without callbacks", {
  result <- hybrid_then(42)
  expect_equal(result, 42)
})

test_that("hybrid_then() handles synchronous errors with on_failure", {
  expect_assertions(n = 2, {
    result <- hybrid_then(
      stop("sync error"),
      on_failure = \(err) {
        expect_match(err$message, "sync error")
        "recovered"
      }
    )
    expect_equal(result, "recovered")
  })
})

test_that("hybrid_then() re-throws synchronous errors without on_failure", {
  expect_error(
    hybrid_then(stop("sync error")),
    "sync error"
  )
})

test_that("hybrid_then() handles asynchronous values with on_success", {
  expect_assertions(n = 2, {
    result <- hybrid_then(
      promise_resolve(42),
      on_success = \(x) x + 8
    )
    expect_true(is.promise(result))
    result |> extract() |> expect_equal(50)
  })
})

test_that("hybrid_then() handles asynchronous values without callbacks", {
  expect_assertions(n = 2, {
    result <- hybrid_then(promise_resolve(42))
    expect_true(is.promise(result))
    result |> extract() |> expect_equal(42)
  })
})

test_that("hybrid_then() handles asynchronous errors with on_failure", {
  expect_assertions(n = 3, {
    result <- hybrid_then(
      promise_resolve(stop("async error")),
      on_failure = \(err) {
        expect_match(err$message, "async error")
        "recovered"
      }
    )
    expect_true(is.promise(result))
    result |> extract() |> expect_equal("recovered")
  })
})

test_that("hybrid_then() propagates asynchronous errors without on_failure", {
  expect_assertions(n = 2, {
    result <- hybrid_then(promise_resolve(stop("async error")))
    expect_true(is.promise(result))
    result |>
      catch(\(err) {
        expect_match(err$message, "async error")
      }) |>
      wait_for_it()
  })
})

test_that("hybrid_then() with tee=TRUE preserves synchronous values", {
  counter <- 0
  result <- hybrid_then(
    42,
    on_success = \(x) {
      counter <<- counter + 1
      99
    },
    tee = TRUE
  )
  expect_equal(result, 42)
  expect_equal(counter, 1)
})

test_that("hybrid_then() with tee=TRUE preserves synchronous errors", {
  counter <- 0
  expect_error(
    hybrid_then(
      stop("sync error"),
      on_failure = \(err) {
        counter <<- counter + 1
        "recovered"
      },
      tee = TRUE
    ),
    "sync error"
  )
  expect_equal(counter, 1)
})


test_that("hybrid_then() with tee=TRUE preserves asynchronous values", {
  expect_assertions(n = 3, {
    counter <- 0
    result <- hybrid_then(
      promise_resolve(42),
      on_success = \(x) {
        counter <<- counter + 1
        99
      },
      tee = TRUE
    )
    expect_true(is.promise(result))
    result |> extract() |> expect_equal(42)
    expect_equal(counter, 1)
  })
})

test_that("hybrid_then() with tee=TRUE preserves asynchronous errors", {
  expect_assertions(n = 3, {
    counter <- 0
    result <- hybrid_then(
      promise_resolve(stop("async error")),
      on_failure = \(err) {
        counter <<- counter + 1
        "recovered"
      },
      tee = TRUE
    )
    expect_true(is.promise(result))
    result |>
      catch(\(err) {
        expect_match(err$message, "async error")
      }) |>
      wait_for_it()
    expect_equal(counter, 1)
  })
})

test_that("hybrid_then() handles both callbacks with synchronous success", {
  success_called <- FALSE
  failure_called <- FALSE

  result <- hybrid_then(
    42,
    on_success = \(x) {
      success_called <<- TRUE
      x * 2
    },
    on_failure = \(err) {
      failure_called <<- TRUE
      "error"
    }
  )

  expect_equal(result, 84)
  expect_true(success_called)
  expect_false(failure_called)
})

test_that("hybrid_then() handles both callbacks with synchronous failure", {
  success_called <- FALSE
  failure_called <- FALSE

  result <- hybrid_then(
    stop("error"),
    on_success = \(x) {
      success_called <<- TRUE
      x * 2
    },
    on_failure = \(err) {
      failure_called <<- TRUE
      "recovered"
    }
  )

  expect_equal(result, "recovered")
  expect_false(success_called)
  expect_true(failure_called)
})

test_that("hybrid_then() handles both callbacks with asynchronous success", {
  success_called <- FALSE
  failure_called <- FALSE

  result <- hybrid_then(
    promise_resolve(42),
    on_success = \(x) {
      success_called <<- TRUE
      x * 2
    },
    on_failure = \(err) {
      failure_called <<- TRUE
      "error"
    }
  )

  expect_true(is.promise(result))
  result |> extract() |> expect_equal(84)
  expect_true(success_called)
  expect_false(failure_called)
})

test_that("hybrid_then() handles both callbacks with asynchronous failure", {
  success_called <- FALSE
  failure_called <- FALSE

  result <- hybrid_then(
    promise_resolve(stop("error")),
    on_success = \(x) {
      success_called <<- TRUE
      x * 2
    },
    on_failure = \(err) {
      failure_called <<- TRUE
      "recovered"
    }
  )

  expect_true(is.promise(result))
  result |> extract() |> expect_equal("recovered")
  expect_false(success_called)
  expect_true(failure_called)
})

test_that("hybrid_then() can chain synchronous and asynchronous operations", {
  # Start with sync, end with async
  result1 <- hybrid_then(
    42,
    on_success = \(x) promise_resolve(x + 8)
  )
  expect_true(is.promise(result1))
  result1 |> extract() |> expect_equal(50)

  # Start with async, callback returns sync
  result2 <- hybrid_then(
    promise_resolve(42),
    on_success = \(x) x + 8
  )
  expect_true(is.promise(result2))
  result2 |> extract() |> expect_equal(50)
})

test_that("hybrid_then() validates tee parameter", {
  expect_snapshot(
    hybrid_then(42, tee = "invalid"),
    error = TRUE
  )

  expect_snapshot(
    hybrid_then(42, tee = 1),
    error = TRUE
  )
})

test_that("hybrid_then() validates `on_success` and `on_failure` parameters", {
  expect_snapshot(
    hybrid_then(42, on_success = "invalid"),
    error = TRUE
  )

  expect_snapshot(
    hybrid_then(42, on_failure = "invalid"),
    error = TRUE
  )
})

test_that("hybrid_then() example from documentation works", {
  add_to <- function(x, k) {
    hybrid_then(x, on_success = function(value) {
      value + k
    })
  }

  # Synchronous
  result_sync <- 42 |> add_to(100)
  expect_equal(result_sync, 142)

  # Asynchronous
  async_called <- FALSE
  result_async <- promise_resolve(42) |> add_to(8)
  expect_true(is.promise(result_async))
  result_async |>
    then(\(x) {
      async_called <<- TRUE
      expect_equal(x, 50)
    }) |>
    wait_for_it()
  expect_true(async_called)
})

test_that("hybrid_then() on_success callback can throw errors synchronously", {
  expect_error(
    hybrid_then(
      42,
      on_success = \(x) stop("callback error")
    ),
    "callback error"
  )
})

test_that("hybrid_then() on_success callback can throw errors asynchronously", {
  expect_assertions(n = 2, {
    result <- hybrid_then(
      promise_resolve(42),
      on_success = \(x) stop("callback error")
    )
    expect_true(is.promise(result))
    result |>
      catch(\(err) {
        expect_match(err$message, "callback error")
      }) |>
      wait_for_it()
  })
})

test_that("hybrid_then() on_failure callback can throw errors synchronously", {
  expect_error(
    hybrid_then(
      stop("original error"),
      on_failure = \(err) stop("callback error")
    ),
    "callback error"
  )
})

test_that("hybrid_then() on_failure callback can throw errors asynchronously", {
  expect_assertions(n = 2, {
    result <- hybrid_then(
      promise_resolve(stop("original error")),
      on_failure = \(err) stop("callback error")
    )
    expect_true(is.promise(result))
    result |>
      catch(\(err) {
        expect_match(err$message, "callback error")
      }) |>
      wait_for_it()
  })
})

test_that("hybrid_then() with tee=TRUE and on_success that throws (sync)", {
  counter <- 0
  expect_error(
    hybrid_then(
      42,
      on_success = \(x) {
        counter <<- counter + 1
        stop("callback error")
      },
      tee = TRUE
    ),
    "callback error"
  )
  expect_equal(counter, 1)
})

test_that("hybrid_then() with tee=TRUE and on_success that throws (async)", {
  expect_assertions(n = 3, {
    counter <- 0
    result <- hybrid_then(
      promise_resolve(42),
      on_success = \(x) {
        counter <<- counter + 1
        stop("callback error")
      },
      tee = TRUE
    )
    expect_true(is.promise(result))
    result |>
      catch(\(err) {
        expect_match(err$message, "callback error")
      }) |>
      wait_for_it()
    expect_equal(counter, 1)
  })
})

test_that("hybrid_then() can chain multiple hybrid_then calls synchronously", {
  result <- 1 |>
    hybrid_then(on_success = \(x) x + 1) |>
    hybrid_then(on_success = \(x) x * 2) |>
    hybrid_then(on_success = \(x) x - 1)
  expect_equal(result, 3)
})

test_that("hybrid_then() can chain multiple hybrid_then calls asynchronously", {
  expect_assertions(n = 2, {
    result <- promise_resolve(1) |>
      hybrid_then(on_success = \(x) x + 1) |>
      hybrid_then(on_success = \(x) x * 2) |>
      hybrid_then(on_success = \(x) x - 1)
    expect_true(is.promise(result))
    result |> extract() |> expect_equal(3)
  })
})

test_that("hybrid_then() can chain with mix of sync and async results", {
  expect_assertions(n = 2, {
    result <- 1 |>
      hybrid_then(on_success = \(x) promise_resolve(x + 1)) |>
      hybrid_then(on_success = \(x) x * 2) |>
      hybrid_then(on_success = \(x) promise_resolve(x - 1))
    expect_true(is.promise(result))
    result |> extract() |> expect_equal(3)
  })
})

test_that("hybrid_then() handles invisible values synchronously", {
  result <- hybrid_then(
    invisible(42),
    on_success = \(x) x + 8
  )
  expect_equal(result, 50)
})

test_that("hybrid_then() handles NULL return from on_success callback", {
  result <- hybrid_then(
    42,
    on_success = \(x) NULL
  )
  expect_null(result)
})

test_that("hybrid_then() handles NULL return from on_failure callback", {
  result <- hybrid_then(
    stop("error"),
    on_failure = \(err) NULL
  )
  expect_null(result)
})

test_that("hybrid_then() handles on_success returning a promise (sync to async)", {
  expect_assertions(n = 2, {
    result <- hybrid_then(
      42,
      on_success = \(x) promise_resolve(x * 2)
    )
    expect_true(is.promise(result))
    result |> extract() |> expect_equal(84)
  })
})

test_that("hybrid_then() handles on_failure returning a promise (sync to async)", {
  expect_assertions(n = 2, {
    result <- hybrid_then(
      stop("error"),
      on_failure = \(err) promise_resolve("recovered via promise")
    )
    expect_true(is.promise(result))
    result |> extract() |> expect_equal("recovered via promise")
  })
})

test_that("hybrid_then() with only on_failure and synchronous success", {
  result <- hybrid_then(
    42,
    on_failure = \(err) "should not be called"
  )
  expect_equal(result, 42)
})

test_that("hybrid_then() with only on_failure and asynchronous success", {
  expect_assertions(n = 2, {
    result <- hybrid_then(
      promise_resolve(42),
      on_failure = \(err) "should not be called"
    )
    expect_true(is.promise(result))
    result |> extract() |> expect_equal(42)
  })
})

test_that("hybrid_then() preserves error class information", {
  expect_assertions(n = 2, {
    custom_error <- structure(
      list(message = "custom error"),
      class = c("custom_error", "error", "condition")
    )

    result <- hybrid_then(
      stop(custom_error),
      on_failure = \(err) {
        expect_s3_class(err, "custom_error")
        "recovered"
      }
    )
    expect_equal(result, "recovered")
  })
})

test_that("hybrid_then() works with complex expressions", {
  result <- hybrid_then(
    {
      x <- 10
      y <- 20
      x + y
    },
    on_success = \(val) val * 2
  )
  expect_equal(result, 60)
})

#' @export
format.promise <- function(x, ...) {
  p <- attr(x, "promise_impl", exact = TRUE)
  p$format()
}

#' @export
print.promise <- function(x, ...) {
  cat(paste(format(x), collapse = "\n"), "\n", sep = "")
}

debug_msg_can_print <- FALSE
debug_msg <- function(...) {
  if (debug_msg_can_print) {
    message(...)
  }
}

assert_work_queue_pkgs <- local({
  val <- NULL
  function() {
    if (!is.null(val)) return()
    for (pkg in list(
      list(name = "future", version = "1.21.0"),
      list(name = "fastmap", version = "1.1.0")
    )) {
      if (!is_available(pkg$name, pkg$version)) {
        stop("Package `", pkg$name, "` (", pkg$version, ") needs to be installed")
      }
    }
    val <<- TRUE
    return()
  }
})
future_worker_is_free <- function() {
  future::nbrOfFreeWorkers() > 0
}


Delay <- R6::R6Class("Delay",
  private = list(
    delay_count = 0
  ),
  public = list(
    is_reset = function() {
      private$delay_count == 0
    },
    reset = function() {
      private$delay_count <- 0
    },

    increase = function() {
      private$delay_count <- private$delay_count + 1
    },

    delay = function() {
      stop("$delay() not implemented")
    }
  ), active = list(
    count = function() {
      private$delay_count
    }
  )
)

ExpoDelay <- R6::R6Class("ExpoDelay",
  inherit = Delay,
  private = list(
    base = 1 / 100,
    min_seconds = 0.01,
    max_seconds = 2
  ),
  public = list(
    initialize = function(
      base = 1 / 100,
      min_seconds = 0.01,
      max_seconds = 2
    ) {
      stopifnot(length(base) == 1 && is.numeric(base) && base >= 0)
      stopifnot(length(min_seconds) == 1 && is.numeric(min_seconds) && min_seconds >= 0)
      stopifnot(length(max_seconds) == 1 && is.numeric(max_seconds) && max_seconds >= 0)

      private$base <- base
      private$max_seconds <- max_seconds
      private$min_seconds <- min_seconds

      self
    },

    # return number of milliseconds until next attempt
    # will randomly backoff to avoid extra work on long poll times
    delay = function() {
      # calculate expo backoff value
      expo_val <- private$base * ((2 ^ private$delay_count) - 1)
      # find random value
      random_val <- runif(n = 1, max = min(private$max_seconds, expo_val))
      # perform `min()` on second step to avoid `runif(1, min = 5, max = 4)` which produces `NaN`
      max(private$min_seconds, random_val)
    }
  )
)


# Situations
# √ No future workers are busy. All future calls are `future_promises()`
#  * Can be accomplished using future job followup promise which calls `$attempt_work()`
# √ All future workers are busy with other tasks (but will become available).
#  * Require using delay
# √ While processing the first batch, existing future workers are taken over
#  * Require delay

# FIFO queue of workers
#' Future promise work queue
#'
#' #' `r lifecycle::badge('experimental')`
#'
#' An \pkg{R6} class to help with scheduling work to be completed. `WorkQueue` will only execute work if the `can_proceed()` returns `TRUE`. For the use case of `future`, `can_proceed()` defaults to `future::nbrOfFreeWorkers() > 0` which will not allow for work to be executed if a \pkg{future} worker is not available.
#'
#' `WorkQueue` will constantly try to start new work once prior work item finishes.  However, if `can_proceed()` returns `FALSE` (no future workers are available) and there is more work to be done, then work is attempted later a random amount of time later using exponential backoff.  The exponential backoff will cap out at 10 seconds to prevent unnecessarily large wait times.
#'
#' Each time `WorkQueue` tries to start more work, it will repeat until `can_proceed()` returns `FALSE` or there is no more work in the `queue`.
#'
#' @section Global event loop:
#'
#' The global loop is used by default as the internal `WorkQueue` "delayed check" uses a single delay check for the whole queue, rather than having each item in the queue attempt to process.
#' This behavior might change in the future, but we are not exactly sure how at this point.
#'
#' If a private `later` loop wants to become synchronous by running until all jobs are completed but is waiting on a `future_promise()`, the private loop will not complete unless the global loop is allowed to move forward.
#'
#' However, it is possible to use a private loop inside a user-defined `WorkQueue` may work which can be provided directly to `future_promise(queue=custom_queue)`. Having a concrete example (or need) will help us understand the problem better. If you have an example, please reach out .
#'
#' @seealso [future_promise_queue()] which returns a `WorkQueue` which is cached per R session.
#' @keywords internal
WorkQueue <- R6::R6Class("WorkQueue",

  # TODO - private loop proposal:
  # The queued data would actually be a list of queues whose _key_ matches
  # the loop ID. This would require that `schedule_work()` take in `loop` and have each loop have its own queue.
  # The scheduled work in each queue would contain the `work` function and `submission_time`.
  # Once `can_proceed_fn()` returns TRUE, the queue with the earliest `submission_time` should be processed.
  # This concept is similar to a merge sort when trying to merge two pre-sorted lists.
  # Check time: O(k), k = number of later loops ever registered. (This could become big!)
  # Maybe, if `ID != 0`, the queue is removed if the number of elements goes to 0.
  # Check time: O(kk), kk <= k, kk = number of _active_ later loops.
  #
  # Thought process, let's say chromote used the global WorkQueue
  # and wanted to have its local loop synchronize (.. while(later::loop_empty(local_loop)) later::run_now(local_loop))  ..).
  # [ ] In `do_work()`, Get loop from `later::current_loop()`
  #   * If is global loop, get item with earliest submission time
  #   * If is private loop (ID != 0), get first item in private queue
  # [ ] Validate that an work item finishing in loop X, has a promise created using loop X
  # [ ] Would each private queue need its own delay check?
  #
  #
  # Implementation question:
  # * Would it be better if we have a larger WorkQueue class that managed many WorkQueues for each loop?
  #    * This would allow for each loop to have its own delay check, delay counter, and loop
  private = list(
    queue = "fastmap::fastqueue()",
    can_proceed_fn = "future_worker_is_free()",
    # only _really_ used for delay checking
    loop = "later::global_loop()",
    delay = "ExpoDelay$new()",

    # Used as a semaphore to make sure only 1 call of `attempt_work()` is delayed
    cancel_delayed_attempt_work = NULL,

    # Increment delay. Used with ExpoDelay
    increase_delay = function() {
      debug_msg("increase_delay()...", private$delay$count)
      # Increment delay and try again later
      private$delay$increase()
    },

    # Reset delay and cancel any delayed `attempt_work()` calls
    reset_delay = function() {
      debug_msg("reset_delay()")
      private$delay$reset()

      # disable any delayed executions of `attempt_work`
      if (is.function(private$cancel_delayed_attempt_work)) {
        private$cancel_delayed_attempt_work()
      }
      private$cancel_delayed_attempt_work <- NULL
    },

    # Returns a logical which let's work begin
    can_proceed = function() {
      isTRUE(private$can_proceed_fn())
    },

    # Function to attempt as much work as possible
    # If no workers are available and a queue has elements,
    #   If a delayed check has already been registered, Return
    #   Else, check again after some delay
    attempt_work = function(can_delay = FALSE) {
      debug_msg('attempt_work()')
      # If nothing to start, return early
      if (private$queue$size() == 0) return()

      # If we are not waiting on someone else, we can do work now
      while ((private$queue$size() > 0) && private$can_proceed()) {
        # Do work right away
        private$reset_delay()
        private$do_work()
      }

      # If there are still items to be processed, but we can not proceed...
      if (private$queue$size() > 0 && ! private$can_proceed()) {
        # If we are allowed to delay (default FALSE), or nothing is currently delaying
        if (can_delay || is.null(private$cancel_delayed_attempt_work)) {

          # Try again later
          private$increase_delay()
          private$cancel_delayed_attempt_work <-
            later::later(
              loop = private$loop,
              delay = private$delay$delay(),
              function() {
                private$attempt_work(can_delay = TRUE)
              }
            )
        }
      }
    },

    # Actually process an item in the queue
    do_work = function() {
      debug_msg("do_work()")

      # Get first item in queue
      work_fn <- private$queue$remove()

      # Safety check...
      # If nothing is returned, no work to be done. Return early
      if (!is.function(work_fn)) return()

      # Do scheduled work
      debug_msg("execute work")
      future_job <- work_fn()

      # Try to attempt work immediately after the future job has finished
      finally(future_job, function() {
        debug_msg("finished work. queue size: ", private$queue$size())
        private$attempt_work()
      })

      return()
    }
  ),
  public = list(
    #' @description Create a new `WorkQueue`
    #' @param can_proceed Function that should return a logical value. If `TRUE` is returned, then the next scheduled work will be executed. By default, this function checks if \code{\link[future:nbrOfWorkers]{future::nbrOfFreeWorkers()} > 0}
    #' @param queue Queue object to use to store the scheduled work. By default, this is a "First In, First Out" queue using [fastmap::fastqueue()]. If using your own queue, it should have the methods `$add(x)`, `$remove()`, `$size()`.
    #' @param loop \pkg{later} loop to use for calculating the next delayed check. Defaults to [later::global_loop()].
    initialize = function(
      # defaults to a future::plan agnostic function
      can_proceed = future_worker_is_free,
      queue = fastmap::fastqueue(), # FIFO
      loop = later::global_loop()
    ) {
      stopifnot(is.function(can_proceed))
      stopifnot(
        is.function(queue$add) &&
        is.function(queue$remove) &&
        is.function(queue$size)
      )
      stopifnot(inherits(loop, "event_loop"))
      delay <- ExpoDelay$new()
      stopifnot(inherits(delay, "Delay"))

      private$can_proceed_fn <- can_proceed
      private$queue <- queue
      private$loop <- loop
      private$delay <- delay

      # make sure delay is reset
      private$reset_delay()

      self
    },

    # add to schedule only
    #' Schedule work
    #' @param fn function to execute when `can_proceed()` returns `TRUE`.
    schedule_work = function(fn) {
      debug_msg("schedule_work()")
      stopifnot(is.function(fn))
      private$queue$add(fn)

      private$attempt_work()

      invisible(self)
    }
  )
)


#' @describeIn future_promise Default `future_promise()` work queue to use. This function returns a [WorkQueue] that is cached per R session.
#' @seealso [`WorkQueue`]
#' @export
future_promise_queue <- local({
  future_promise_queue_ <- NULL
  function() {
    if (is.null(future_promise_queue_)) {
      assert_work_queue_pkgs()
      future_promise_queue_ <<- WorkQueue$new()
    }
    future_promise_queue_
  }
})


#' \pkg{future} promise
#'
#' `r lifecycle::badge('experimental')`
#'
#' When submitting \pkg{future} work, \pkg{future} (by design) will block the main R session until a worker becomes available.
#' This occurs when there is more submitted \pkg{future} work than there are available \pkg{future} workers.
#' To counter this situation, we can create a promise to execute work using future (using `future_promise()`) and only begin the work if a \pkg{future} worker is available.
#'
#' Using `future_promise()` is recommended when ever a continuous runtime is used, such as with \pkg{plumber} or \pkg{shiny}.
#'
#' For more details and examples, please see the [`vignette("future_promise", "promises")`](https://rstudio.github.io/promises/articles/future_promise.html) vignette().
#' @describeIn future_promise Creates a [promise()] that will execute the `expr` using [future::future()].
#' @inheritParams future::future
#' @param expr An R expression. While the `expr` is eventually sent to [`future::future()`], please use the same precautions that you would use with regular `promises::promise()` expressions. `future_promise()` may have to hold the `expr` in a [promise()] while waiting for a \pkg{future} worker to become available.
#' @param ... extra parameters provided to [`future::future()`]
#' @param queue A queue that is used to schedule work to be done using [future::future()].  This queue defaults to [future_promise_queue()] and requires that method `queue$schedule_work(fn)` exist.  This method should take in a function that will execute the promised \pkg{future} work.
#' @return Unlike [future::future()`], `future_promise()` returns a [promise()] object that will eventually resolve the \pkg{future} `expr`.
#' @examples
#' \donttest{# Relative start time
#' start <- Sys.time()
#' # Helper to force two `future` workers
#' with_two_workers <- function(expr) {
#'   if (!require("future")) {
#'     message("`future` not installed")
#'     return()
#'   }
#'   old_plan <- future::plan(future::multisession(workers = 2))
#'   on.exit({future::plan(old_plan)}, add = TRUE)
#'   start <<- Sys.time()
#'   force(expr)
#'   while(!later::loop_empty()) {Sys.sleep(0.1); later::run_now()}
#'   invisible()
#' }
#' # Print a status message. Ex: `"PID: XXX; 2.5s promise done"`
#' print_msg <- function(pid, msg) {
#'   message(
#'     "PID: ", pid, "; ",
#'     round(difftime(Sys.time(), start, units = "secs"), digits = 1), "s " ,
#'     msg
#'   )
#' }
#'
#' # `"promise done"` will appear after four workers are done and the main R session is not blocked
#' # The important thing to note is the first four times will be roughly the same
#' with_two_workers({
#'   promise_resolve(Sys.getpid()) %...>% print_msg("promise done")
#'   for (i in 1:6) future::future({Sys.sleep(1); Sys.getpid()}) %...>% print_msg("future done")
#' })
#' {
#' #> PID: XXX; 2.5s promise done
#' #> PID: YYY; 2.6s future done
#' #> PID: ZZZ; 2.6s future done
#' #> PID: YYY; 2.6s future done
#' #> PID: ZZZ; 2.6s future done
#' #> PID: YYY; 3.4s future done
#' #> PID: ZZZ; 3.6s future done
#' }
#'
#' # `"promise done"` will almost immediately, before any workers have completed
#' # The first two `"future done"` comments appear earlier the example above
#' with_two_workers({
#'   promise_resolve(Sys.getpid()) %...>% print_msg("promise")
#'   for (i in 1:6) future_promise({Sys.sleep(1); Sys.getpid()}) %...>% print_msg("future done")
#' })
#' {
#' #> PID: XXX; 0.2s promise done
#' #> PID: YYY; 1.3s future done
#' #> PID: ZZZ; 1.4s future done
#' #> PID: YYY; 2.5s future done
#' #> PID: ZZZ; 2.6s future done
#' #> PID: YYY; 3.4s future done
#' #> PID: ZZZ; 3.6s future done
#' }}
#' @export
future_promise <- function(
  expr = NULL,
  envir = parent.frame(),
  substitute = TRUE,
  globals = TRUE,
  packages = NULL,
  ...,
  queue = future_promise_queue()
) {

  # make sure queue is the right structure
  stopifnot(is.function(queue$schedule_work) && length(formals(queue$schedule_work)) >= 1)

  if (substitute) expr <- substitute(expr)

  # Force all variables to curb values changing before execution
  # Does NOT fix R environment values changing
  force(envir)
  force(substitute)
  force(globals)
  force(packages)
  force(list(...))

  ## Record globals
  gp <- future::getGlobalsAndPackages(expr, envir = envir, globals = globals)
  force(gp)

  promise(function(resolve, reject) {
    # add to queue
    queue$schedule_work(function() {
      ### Should the worker function be taken at creation time or submission time?
      ## The current implementation has `$can_proceed()` method of `WorkQueue` be plan agnostic.
      ## Therefore, it will always ask the current plan if a worker is available.
      ## If so, then the _current_ plan should be used. Not a plan that existed at initialization time.
      exec_future <- future::plan()

      # execute the future and return a promise so the schedule knows exactly when it is done
      future_job <- exec_future(
        gp$expr,
        envir = envir,
        substitute = FALSE,
        globals = gp$globals,
        packages = unique(c(packages, gp$packages)),
        ...
      )

      # Resolve the outer promising job value
      resolve(future_job)
      # Return a promising object that can be chained by the `queue` after executing this _work_
      future_job
    })
  })
}



if (FALSE) {

  # ConstDelay <- R6::R6Class("ConstDelay",
  #   inherit = Delay,
  #   private = list(
  #     const = 0.1,
  #     random = TRUE
  #   ),
  #   public = list(
  #     initialize = function(const = 0.1, random = TRUE) {
  #       stopifnot(length(const) == 1 && is.numeric(const) && const >= 0)
  #       private$const <- const
  #       private$random <- isTRUE(random)

  #       self
  #     },
  #     delay = function() {
  #       if (private$random) {
  #         runif(n = 1, max = private$const)
  #       } else {
  #         private$const
  #       }
  #     }
  #   )
  # )
  # LinearDelay <- R6::R6Class("LinearDelay",
  #   inherit = Delay,
  #   private = list(
  #     delta = 0.05,
  #     random = TRUE
  #   ),
  #   public = list(
  #     initialize = function(delta = 0.03, random = TRUE) {
  #       stopifnot(length(delta) == 1 && is.numeric(delta) && delta >= 0)
  #       private$delta <- delta
  #       private$random <- isTRUE(random)

  #       self
  #     },

  #     delay = function() {
  #       delta_delay <- private$delay_count * private$delta
  #       if (private$random) {
  #         runif(n = 1, max = delta_delay)
  #       } else {
  #         delta_delay
  #       }
  #     }
  #   )
  # )


  # dev_load <- pkgload::load_all

  # ## test
  # dev_load(); print_i(); start <- Sys.time(); promise_all(.list = lapply(1:10, function(x) { future_promise({ Sys.sleep(1); print(paste0(x)) })})) %...>% { print(Sys.time() - start) };

  # ## block workers mid job
  # dev_load(); print_i(); start <- Sys.time(); promise_all(.list = lapply(1:10, function(x) { future_promise({ Sys.sleep(1); print(paste0(x)) })})) %...>% { print(Sys.time() - start) }; lapply(1:2, function(i) { later::later(function() { message("*************** adding blockage", i); fj <- future::future({ Sys.sleep(4); message("*************** blockage done", i); i}); then(fj, function(x) { print(paste0("block - ", i))}); }, delay = 0.5 + i) }) -> ignore;

  # ## block main worker mid job
  # dev_load(); print_i(); start <- Sys.time(); promise_all(.list = lapply(1:10, function(x) { future_promise({ Sys.sleep(1); print(paste0(x)) })})) %...>% { print(Sys.time() - start) }; lapply(1:4, function(i) { later::later(function() { message("*************** adding blockage", i); fj <- future::future({ Sys.sleep(4); message("*************** blockage done", i); i}); then(fj, function(x) { print(paste0("block - ", i))}); }, delay = 0.5 + i/4) }) -> ignore;


  # ## block workers pre job
  # dev_load(); print_i(); lapply(1:2, function(i) { message("*************** adding blockage", i); future::future({ Sys.sleep(4); message("*************** blockage done", i); i}) }) -> future_jobs; lapply(future_jobs, function(fj) { as.promise(fj) %...>% { print(.) } }); start <- Sys.time(); promise_all(.list = lapply(1:10, function(x) { future_promise({ Sys.sleep(1); print(paste0(x)) })})) %...>% { print(Sys.time() - start) };

  # ## block main worker workers pre job
  # dev_load(); print_i(); start <- Sys.time(); promise_all(.list = lapply(1:10, function(x) { future_promise({ Sys.sleep(1); print(paste0(x)) })})) %...>% { print(Sys.time() - start) }; lapply(1:4, function(i) { later::later(function() { message("*************** adding blockage", i); fj <- future::future({ Sys.sleep(4); message("*************** blockage done", i); i}); then(fj, function(x) { print(paste0("block - ", i))}); }, delay = 0.5 + i/4) }) -> ignore;

  future::plan(future::multisession(workers = 2))

  debug_msg_can_print <- TRUE

  print_i <- function(i = 0) { if (i <= 50) { print(i); later::later(function() { print_i(i + 1) }, delay = 0.1) } }

  slow_calc <- function(n) {
    Sys.sleep(n)
    "slow!"
  }
  n <- 2
  prom <- future_promise
  # prom <- future::future
  a1 <- prom({
    print(paste0("start 1 - ", Sys.time()))
    print(slow_calc(n))
  })
  a2 <- prom({
    print(paste0("start 2 - ", Sys.time()))
    print(slow_calc(n))
  })
  a3 <- prom({
    print(paste0("start 3 - ", Sys.time()))
    print(slow_calc(n))
  })
  a4 <- prom({
    print(paste0("start 4 - ", Sys.time()))
    print(slow_calc(n))
  })

  print("done assignement!")

  a1 %...>% { message("end 1 - ", format(Sys.time())) }
  a2 %...>% { message("end 2 - ", format(Sys.time())) }
  a3 %...>% { message("end 3 - ", format(Sys.time())) }
  a4 %...>% { message("end 4 - ", format(Sys.time())) }

}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

asyncFib <- function(resolve, reject, x) {
    invisible(.Call('_promises_asyncFib', PACKAGE = 'promises', resolve, reject, x))
}

#' Access the results of a promise
#'
#' Use the `then` function to access the eventual result of a promise (or, if the operation fails, the reason for that failure). Regardless of the state of the promise, the call to `then` is non-blocking, that is, it returns immediately; so what it does *not* do is immediately return the result value of the promise. Instead, you pass logic you want to execute to `then`, in the form of function callbacks (or formulas, see Details). If you provide an `onFulfilled` callback, it will be called upon the promise's successful resolution, with a single argument `value`: the result value. If you provide an `onRejected` callback, it will be called if the operation fails, with a single argument `reason`: the error that caused the failure.
#'
#' @section Formulas:
#'
#' For convenience, the `then()`, `catch()`, and `finally()` functions use
#' [rlang::as_function()] to convert `onFulfilled`, `onRejected`, and
#' `onFinally` arguments to functions. This means that you can use formulas to
#' create very compact anonymous functions, using `.` to access the value (in
#' the case of `onFulfilled`) or error (in the case of `onRejected`).
#'
#' @section Chaining promises:
#'
#' The first parameter of `then` is a promise; given the stated purpose of the
#' function, this should be no surprise. However, what may be surprising is that
#' the return value of `then` is also a (newly created) promise. This new
#' promise waits for the original promise to be fulfilled or rejected, and for
#' `onFulfilled` or `onRejected` to be called. The result of (or error raised
#' by) calling `onFulfilled`/`onRejected` will be used to fulfill (reject) the
#' new promise.
#'
#' ```
#' promise_a <- get_data_frame_async()
#' promise_b <- then(promise_a, onFulfilled = head)
#' ```
#'
#' In this example, assuming `get_data_frame_async` returns a promise that
#' eventually resolves to a data frame, `promise_b` will eventually resolve to
#' the first 10 or fewer rows of that data frame.
#'
#' Note that the new promise is considered fulfilled or rejected based on
#' whether `onFulfilled`/`onRejected` returns a value or throws an error, not on
#' whether the original promise was fulfilled or rejected. In other words, it's
#' possible to turn failure to success and success to failure. Consider this
#' example, where we expect `some_async_operation` to fail, and want to consider
#' it an error if it doesn't:
#'
#' ```
#' promise_c <- some_async_operation()
#' promise_d <- then(promise_c,
#'   onFulfilled = function(value) {
#'     stop("That's strange, the operation didn't fail!")
#'   },
#'   onRejected = function(reason) {
#'     # Great, the operation failed as expected
#'     NULL
#'   }
#' )
#' ```
#'
#' Now, `promise_d` will be rejected if `promise_c` is fulfilled, and vice
#' versa.
#'
#' **Warning:** Be very careful not to accidentally turn failure into success,
#' if your error handling code is not the last item in a chain!
#'
#' ```
#' some_async_operation() %>%
#'   catch(function(reason) {
#'     warning("An error occurred: ", reason)
#'   }) %>%
#'   then(function() {
#'     message("I guess we succeeded...?")  # No!
#'   })
#' ```
#'
#' In this example, the `catch` callback does not itself throw an error, so the
#' subsequent `then` call will consider its promise fulfilled!
#'
#' @section Convenience functions:
#'
#' For readability and convenience, we provide `catch` and `finally` functions.
#'
#' The `catch` function is equivalent to `then`, but without the `onFulfilled`
#' argument. It is typically used at the end of a promise chain to perform error
#' handling/logging.
#'
#' The `finally` function is similar to `then`, but takes a single no-argument
#' function (or formula) that will be executed upon completion of the promise,
#' regardless of whether the result is success or failure. It is typically used
#' at the end of a promise chain to perform cleanup tasks, like closing file
#' handles or database connections. Unlike `then` and `catch`, the return value
#' of `finally` is ignored; however, if an error is thrown in `finally`, that
#' error will be propagated forward into the returned promise.
#'
#' @section Visibility:
#'
#' `onFulfilled` functions can optionally have a second parameter `visible`,
#' which will be `FALSE` if the result value is [invisible][base::invisible()].
#'
#' @param promise A promise object. The object can be in any state.
#'
#' @param onFulfilled A function (or a formula--see Details) that will be
#'   invoked if the promise value successfully resolves. When invoked, the
#'   function will be called with a single argument: the resolved value.
#'   Optionally, the function can take a second parameter `.visible` if you care
#'   whether the promise was resolved with a visible or invisible value. The
#'   function can return a value or a promise object, or can throw an error;
#'   these will affect the resolution of the promise object that is returned
#'   by `then()`.
#'
#' @param onRejected A function taking the argument `error` (or a formula--see
#'   Details). The function can return a value or a promise object, or can throw
#'   an error. If `onRejected` is provided and doesn't throw an error (or return
#'   a promise that fails) then this is the async equivalent of catching an
#'   error.
#'
#' @export
then <- function(promise, onFulfilled = NULL, onRejected = NULL) {
  promise <- as.promise(promise)

  if (!is.null(onFulfilled))
    onFulfilled <- rlang::as_function(onFulfilled)
  if (!is.null(onRejected))
    onRejected <- rlang::as_function(onRejected)
  invisible(promise$then(onFulfilled = onFulfilled, onRejected = onRejected))
}

#' @param tee If `TRUE`, ignore the return value of the callback, and use the
#'   original value instead. This is useful for performing operations with
#'   side-effects, particularly logging to the console or a file. If the
#'   callback itself throws an error, and `tee` is `TRUE`, that error will still
#'   be used to fulfill the the returned promise (in other words, `tee` only has
#'   an effect if the callback does not throw).
#' @rdname then
#' @export
catch <- function(promise, onRejected, tee = FALSE) {
  promise <- as.promise(promise)

  if (!is.null(onRejected))
    onRejected <- rlang::as_function(onRejected)

  if (!tee) {
    return(promise$catch(onRejected))
  } else {
    promise$catch(function(reason) {
      onRejected(reason)
      stop(reason)
    })
  }
}

#' @rdname then
#'
#' @param onFinally A function with no arguments, to be called when the async
#'   operation either succeeds or fails. Usually used for freeing resources that
#'   were used during async operations.
#'
#' @export
finally <- function(promise, onFinally) {
  promise <- as.promise(promise)

  if (!is.null(onFinally))
    onFinally <- rlang::as_function(onFinally)
  promise$finally(onFinally)
}

library(testthat)

describe("visibility", {
  single_fn <- function(value) {
    info <- withVisible(value)
    if (info$visible) {
      info$value
    } else {
      invisible(info$value)
    }
  }
  double_fn <- function(value, .visible) {
    if (.visible) value else invisible(value)
  }

  # display in block to avoid indent of doom
  for (add_catch in c("false", "single", "double", "expr")) {
    for (add_finally in c("false", "expr")) {
      for (add_then in c("false", "single", "double", "expr")) {
        it(
          paste0(
            "survives ",
            paste0(
              c(
                if (add_then != "false") paste0("then-", add_then),
                if (add_catch != "false") paste0("catch-", add_catch),
                if (add_finally != "false") paste0("finally-", add_finally),
                "then"
              ),
              collapse = ", "
            )
          ),
          {
            p <- promise_resolve(invisible(1))

            p <-
              switch(
                add_then,
                "false" = p,
                "single" = p |> then(single_fn),
                "double" = p |> then(double_fn),
                "expr" = p |>
                  then(
                    \(value) {
                      info <- withVisible(value)
                      if (info$visible) {
                        info$value
                      } else {
                        invisible(info$value)
                      }
                    }
                  )
              )
            p <-
              switch(
                add_catch,
                "false" = p,
                "single" = p |> catch(single_fn),
                "double" = p |> catch(double_fn),
                "expr" = p |>
                  catch(
                    \(err) {}
                  )
              )

            finally_val <- NULL
            p <-
              switch(
                add_finally,
                "false" = p,
                "expr" = p |>
                  finally(\() {
                    finally_val <<- TRUE
                  })
              )

            extended_val <-
              p |>
              then(function(value, .visible) {
                list(value = value, visible = .visible)
              }) |>
              extract()

            regular_val <-
              p |>
              then(function(value) {
                withVisible(value)
              }) |>
              extract()

            if (add_finally != "false") {
              expect_true(finally_val)
            }

            expect_identical(extended_val$value, 1)
            expect_identical(extended_val$visible, FALSE)

            expect_identical(regular_val$value, 1)
            expect_identical(regular_val$visible, FALSE)
          }
        )
      }
    }
  }
})

#' `r lifecycle::badge("deprecated")`: Deprecated functions in promises
#'
#' These functions have been deprecated and will be removed in a future version.
#' Please use the suggested alternatives instead.
#'
#' @name promises-deprecated
#' @keywords internal
NULL

#' @describeIn promises-deprecated Use `with_otel_span()` instead
#' @export
with_ospan_async <- function(
  name,
  expr,
  ...,
  tracer,
  attributes = NULL
) {
  lifecycle::deprecate_warn(
    "1.5.0",
    "with_ospan_async()",
    "with_otel_span()"
  )
  with_otel_span(
    name = name,
    expr = expr,
    ...,
    tracer = tracer,
    attributes = attributes
  )
}

#' @describeIn promises-deprecated Use `with_otel_promise_domain()` instead
#' @export
with_ospan_promise_domain <- function(expr) {
  lifecycle::deprecate_warn(
    "1.5.0",
    "with_ospan_promise_domain()",
    "with_otel_promise_domain()"
  )
  with_otel_promise_domain(expr)
}

#' @describeIn promises-deprecated Use `local_otel_promise_domain()` instead
#' @export
local_ospan_promise_domain <- function(envir = parent.frame()) {
  lifecycle::deprecate_warn(
    "1.5.0",
    "local_ospan_promise_domain()",
    "local_otel_promise_domain()"
  )
  local_otel_promise_domain(envir = envir)
}

check_hybrid_callback <- function(callback, name) {
  if (!is.null(callback) && !is.function(callback)) {
    stop(sprintf("`%s=` must be a function or `NULL`", name))
  }
}


#' Asynchronous or synchronous `then()`
#'
#' This is a helper function that behaves like `then`, however if
#' `is.promising()` returns `FALSE` then the handlers will be executed
#' immediately.
#'
#' Execution paths:
#' * If `expr` evaluates to a promise (`p`), it will call `p |> then(on_success,
#'   on_failure)`.
#' * If `expr` evaluates to a non-promise value (`x`), it will call
#'   `on_success(x)`.
#' * If `expr` throws an error (`e`) during calculation, it will call
#'   `on_failure(e)`.
#'
#' In all cases, the `on_success` and `on_failure` callbacks are executed (when
#' provided).
#'
#' @section Utility:
#'
#' This function is useful for writing functions that need to execute followup
#' behavior _now_ or within a promise. This is different behavior than `then()`
#' where *everything* is made into a promise.
#'
#' `hybrid_then()` allows authors to keep synchronous execution on the same
#' _tick_ without requiring the use of a followup promise. This is particularly
#' appealing for situations where the author does not control the execution flow
#' for items that may be either synchronous or asynchronous, such as within
#' `{plumber2}`.
#'
#' @section Error Handling:
#'
#' If no `on_failure` callback is provided and an error occurs, the error is
#' re-thrown immediately (for synchronous errors) or propagated through the
#' returned promise (for asynchronous errors).
#'
#' If an `on_failure` callback is provided but it throws an error, that new
#' error replaces the original error. With `tee = TRUE`, even if `on_failure`
#' executes successfully, the original error is still re-thrown.
#'
#' @section Callback Return Values:
#'
#' Callbacks can return any value, including promises. If a callback returns a
#' promise, the entire `hybrid_then()` call will return a promise, even if the
#' input was synchronous. This allows seamless transitions between synchronous
#' and asynchronous execution.
#'
#' @param expr An expression that evaluates to either a promise or a non-promise
#'   value.
#' @param on_success A function to be called when no error occurs synchronously
#'   or asynchronously. When invoked, the function will be called with a single
#'   argument: the resolved value. Optionally, the function can take a second
#'   parameter `.visible` if you care whether the promise was resolved with a
#'   visible or invisible value. Can return a value or a promise.
#' @param on_failure A function to be called if an error occurs synchronously or
#'   asynchronously. Takes one argument: the error object. Can return a value or
#'   a promise to recover from the error, or throw a new error. If `on_failure`
#'   is provided and doesn't throw an error (or return a promise that fails)
#'   then this is the async equivalent of catching an error.
#' @param ... Reserved for future use. Currently must be empty.
#' @param tee If `TRUE`, ignore the return value of the callback, and use the
#'   original value of `expr` as the result. For `on_failure` with `tee = TRUE`,
#'   the callback executes but the original error is re-thrown afterward.
#' @return
#'   * If `expr` evaluates to a promise, a promise with a single followup
#'     promise to handle the `on_success` or `on_failure` callbacks.
#'   * If `expr` evaluates to a non-promise value, the result of the synchronous
#'     operation after being processed by `on_success` or `on_failure`.
#'   * If a callback returns a promise, the result is always a promise.
#' @seealso [then()], [is.promising()], [promise_resolve()]
#' @export
#' @examples
#' # Basic usage - works with both sync and async values
#' add_to <- function(x, k) {
#'   hybrid_then(
#'     x,
#'     on_success = function(value) {
#'       value + k
#'     },
#'     on_failure = function(err) {
#'       message("Error: ", err$message)
#'       NA_real_
#'     }
#'   )
#' }
#'
#' # Synchronous
#' 42 |> add_to(100)
#' #> [1] 142
#'
#' # Synchronous error
#' add_to({stop("Bad input!")}, 8)
#' #> Error: Bad input!
#' #> [1] NA
#'
#' \dontrun{
#' # Asynchronous
#' promise_resolve(42) |>
#'   add_to(8) |>
#'   then(print)
#' # When resolved...
#' #> [1] 50
#'
#' # Error handling - asynchronous
#' promise_resolve(stop("Bad async input!")) |>
#'   add_to(8) |>
#'   then(print)
#' # When resolved...
#' #> Error: Bad async input!
#' #> [1] NA
#'
#' # Chaining multiple operations
#' # (Move the `promise_resolve()` around to see sync vs async behavior)
#' 1 |>
#'   hybrid_then(on_success = \(x) x + 1) |>
#'   hybrid_then(on_success = \(x) promise_resolve(x * 2)) |>
#'   hybrid_then(on_success = \(x) x - 1) |>
#'   hybrid_then(print)
#' # When resolved...
#' #> [1] 3
#' }
hybrid_then <- function(
  expr,
  on_success = NULL,
  on_failure = NULL,
  ...,
  tee = FALSE
) {
  check_dots(...)
  check_tee(tee)

  check_hybrid_callback(on_success, "on_success")
  check_hybrid_callback(on_failure, "on_failure")

  # Add support for `.visible`
  on_success <- normalizeOnFulfilled(on_success)
  on_failure <- normalizeOnRejected(on_failure)

  result_is_sync_error <- FALSE

  result_visible <-
    withVisible(
      tryCatch(
        expr,
        error = function(e) {
          # If no `on_failure` callback, re-throw the error
          if (is.null(on_failure)) {
            stop(e)
          }

          # Perform synchronous `on_failure` callback now
          result_on_error <- withVisible(on_failure(e))
          if (tee) {
            # Re-throw the error
            stop(e)
          } else {
            result_is_sync_error <<- TRUE
            return_with_visible(result_on_error)
          }
        }
      )
    )

  if (result_is_sync_error) {
    # Return synchronous `tee=FALSE` `result_on_error` from above
    return(return_with_visible(result_visible))
  }

  if (is.promising(result_visible$value)) {
    # Return async result
    # Will handle async success, failure callbacks (and `tee`)
    then(return_with_visible(result_visible), on_success, on_failure, tee = tee)
  } else {
    # If no `on_success` callback, return sync result now
    if (is.null(on_success)) {
      return(return_with_visible(result_visible))
    }

    # Peform synchronous `on_success` callback now
    result_on_success_visible <-
      withVisible(
        on_success(
          result_visible$value,
          result_visible$visible
        )
      )

    # Return synchronous result
    if (tee) {
      return_with_visible(result_visible)
    } else {
      return_with_visible(result_on_success_visible)
    }
  }
}

return_with_visible <- function(visible_info) {
  if (visible_info$visible) {
    visible_info$value
  } else {
    invisible(visible_info$value)
  }
}

# ## Use when otel v0.3.0 is released: https://github.com/r-lib/otel/commit/43e59c45a7de50cfd8af95f73d45f9899f957b44
# # attributes = as_attributes(attributes)
# attributes = if (!is.null(attributes)) {
#   as_attributes(attributes)
# }
NULL

#' @importFrom otel
#'   start_span
#'   end_span
#'   get_active_span
#'   is_tracing_enabled
#'   with_active_span
NULL

# Fixed variable named required for `otel::default_tracer_name()`
# which is used in `otel::get_tracer()` within `promises_otel_tracer()`
otel_tracer_name <- "co.posit.r-package.promises"

# Using local scope avoids an environment object lookup on each call.
# Inspired by httr2:::get_tracer(). Taken from Shiny
# Benchmark: https://github.com/rstudio/shiny/pull/4269/files#diff-0cc4a76032b57fcc125d41dfa3fb0f0c39976bb00a1d84bb56a0b77c331ce2d1R42
testthat__is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

promises_otel_tracer <- local({
  tracer <- NULL
  function() {
    if (!is.null(tracer)) {
      return(tracer)
    }
    if (testthat__is_testing()) {
      # Don't cache the tracer in unit tests. It interferes with tracer provider
      # injection in otelsdk::with_otel_record().
      return(otel::get_tracer())
    }
    tracer <<- otel::get_tracer()
    tracer
  }
})

#' OpenTelemetry integration
#'
#' @description
#' \pkg{otel} provides tools for integrating with OpenTelemetry, a framework for
#' observability and tracing in distributed systems.
#'
#' These methods are intended to enhance the framework to be used with the
#' \pkg{promises} package, not as a generic replacement.
#'
#' Developer note - Barret 2025/09: This otel span handoff promise domain topic is
#' complex and has been discussed over many hours. Many advanced Shiny/R
#' developers are not even aware of promise domains (very reasonable!),
#' therefore this topic requires more in-depth documentation and examples.
#'
#' @section Definitions:
#'
#' * Promise domain: An environment in which has setup/teardown methods. These
#'   environments can be composed together to facilitate execution context for
#'   promises. In normal R execution, this can be achieved with scope / stack.
#'   But for complex situations, such as the currently open graphics device,
#'   async operations require promise domains to setup/teardown these contexts
#'   to function properly. Otherwise a multi-stage promise that adds to the
#'   graphics device at each stage will only ever print to the most recently
#'   created graphics device, not the associated graphics device. These promise
#'   domains are not automatically created, they must be manually added to the
#'   execution stack, for example `with_otel_promise_domain()` does this for
#'   OpenTelemetry spans.
#' * Promise domain restoration: When switching from one promise chain to
#'   another, the execution context is torn down and then re-established. This
#'   re-establishment is called "promise domain restoration". During this
#'   process, the promise domains are restored in their previously established
#'   combination order.
#' * Promise chain: A set of promise objects to execute over multiple async
#'   ticks.
#' * Async tick: the number of times an event loop must run to move computation
#'   forward. (Similar to a JavaScript event loop tick.)
#' * `then()` promise domain capture: When `then()` is called, it will capture
#'   the current promise domain. This promise domain is restored (only if
#'   needed) when evaluating the given `onFulfilled` and `onRejected` callbacks.
#'   This captured promise domain does not go into any downstream promise chain
#'   objects. The only way the promise domain is captured is exactly when the
#'   `then()` method is called.
#'
#' `with_otel_promise_domain()` creates a promise domain that restores the
#' currently active OpenTelemetry span from when a call to `promises::then()` is
#' executed. Given the special circumstance where only the current otel span is
#' needed to continue recording (not a full ancestry tree of otel spans), we can
#' capture _just_ the current otel span and reactivate that otel span during promise
#' domain restoration.
#'
#' @section When promise domains are captured:
#'
#' Asynchronous operation
#' * Creates `async_op` otel span
#' * Automatically ends the otel span (`async_op`) when the promise (p)
#'   resolves or rejects
#'
#' The code below illustrates an example of when the promise domain are
#' created/captured/restored and when otel span objects are
#' created/activated/reactivated/ended.
#'
#' ```r
#' # t0.0
#' p2 <- with_otel_promise_domain({
#'   # t0.1
#'   p <- with_otel_span("async_op", {
#'     # ... return a promise ...
#'     init_async_work() |> # t0.2
#'       then( # t0.3
#'         some_async_work # t1.0
#'       )
#'   }) # t0.4, t1.0, t2.0
#'   p |>
#'     then( # t0.5
#'       more_async_work # t3.0
#'     )
#' }) # t0.6
#'
#' p_final <-
#'   p2 |> then( # t0.7
#'     final_work # t4.0
#'   )
#' ```
#'
#' An in-depth explanation of the execution timeline is below.
#' * At the first initial tick, `t0.*`:
#'   * `t0.0`: The code is wrapped in `with_otel_promise_domain()`
#'   * `t0.1`: The `async_op` otel span is created and activated
#'   * `t0.2`: Some async work is initiated
#'   * `t0.3`: `then()` is called, capturing the active `async_op` otel span (as it
#'     is called within `with_otel_promise_domain()`)
#'   * `t0.4`: The `with_otel_span()` call exits, but the `async_op` otel span is
#'     not ended as the promise is still pending. The returned promise has a
#'     `finally()` step added to it that will end the otel span `async_op` when `p`
#'     is resolved.
#'   * `t0.5`: Another `then()` is called, but there is no active otel span to
#'     capture
#'   * `t0.6`: The otel span promise domain call exits
#'   * `t0.7`: Another `then()` is called. No otel span will be captured as there is
#'     no active otel span / promise domain
#' * At the first followup tick, `t1.0`:
#'   * The active `async_op` otel span is reactivated during promise domain
#'     restoration for the duration of the `then` callback
#'   * The `some_async_work` function is called
#' * At tick, `t2.0`:
#'   * `some_async_work` has resolved
#'   * A hidden `finally()` step closes the otel span, `async_op`
#'   * `p` is now resolved
#' * At tick, `t3.0`:
#'   * There is no active otel span at `t0.5`, so no otel span is reactivated during
#'     promise domain restoration
#'   * The `more_async_work` function is executed
#' * At tick, `t4.0`:
#'   * `more_async_work` has resolved, therefore `p2` is now resolved
#'   * There was no otel span promise domain at `t0.7`, so no attempt is made to
#'     reactivate any otel span
#'   * The `final_work` function is executed
#' * At tick, `t5.0`:
#'   * `p_final` has resolved

#'
#' @section Complexity:
#'
#' When reactivating the `k`th step in a promise chain, the currently active
#' otel span (during the call to `then()`) will be reactivated during promise domain
#' restoration (`O(1)`). To restore a chain of promises, the active otel span will
#' be restored at each step (`O(n)`) due to the **`n`** calls to wrapping each
#' `onFulfilled` and `onRejected` callbacks inside `then()`.
#'
#' If we did NOT have a handoff promise domain for otel span restoration, a regular
#' promise domain approach would be needed at each step to restore the active
#' otel span. Each step would call `with_active_span()` `k` times (`O(k)`, where as
#' handoff domain computes in `O(1)`). Taking a step back, to restore each otel span
#' at for every step in a promise chain would then take `O(n^2)` time, not
#' `O(n)`. The standard, naive promise domain approach does not scale for
#' multiple similar promise domain restorations.
#'
#'
#' @section Execution model for `with_otel_promise_domain()`:
#'
#' 1. `with_otel_promise_domain(expr)` is called.
#'    * The following steps all occur within `expr`.
#' 2. Create an otel span object using `otel::start_span()`.
#'    * We need the otel span to be active during the a followup async operation.
#'      Therefore, `otel::start_local_active_span()` is not appropriate as the
#'      otel span would be ended when the function exits, not when the promise chain
#'      resolves.
#' 2. Be sure your otel span is activated before calling `promises::then()`.
#'    * Activate it using `with_otel_span(name, expr)` (which also
#'      creates/ends the otel span) or `otel::with_active_span(span, expr)`.
#' 3. Call `promises::then()`
#'   * When `promises::then()` is called, the two methods (`onFulfilled` and
#'     `onRejected`) capture the currently active spans. (Performed by the
#'     initial `with_otel_promise_domain()`)
#' 4. During reactivation of the promise chain step, the previously captured
#'    otel span is reactivated via `with_active_span()`. (Performed by the initial
#'    `with_otel_promise_domain()`)
#'
#' @section OpenTelemetry span compatibility:
#'
#' For otel span objects to exist over may async ticks, the otel span must be created
#' using `otel::start_span()` and later ended using `otel::end_span()`. Ending
#' the otel span must occur **after** any promise chain work has completed.
#'
#' If we were to instead use `otel::start_local_active_span()`, the otel span would
#' be ended when the function exits, not when the promise chain completes. Even
#' though the local otel span is created, activated, and eventually ended, the otel span
#' will not exist during reactivation of the otel span promise domain.
#'
#' `with_otel_span()` is a convenience method that creates, activates, and
#' ends the otel span only after the returned promise (if any) resolves. It also
#' properly handles both synchronous (ending the otel span within `on.exit()`) and
#' asynchronous operations (ending the otel span within `promises::finally()`).
#'
#'
#'
#'
#' @describeIn otel Creates an OpenTelemetry span, executes the given expression within it, and
#' ends the span.
#'
#' This method **requires** the use of `with_otel_promise_domain()`
#' to be within the execution stack.
#'
#' This function is designed to handle both synchronous and asynchronous
#' (promise-based) operations. For promises, the span is automatically ended
#' when the promise resolves or rejects.
#'
#' Returns the result of evaluating `expr`. If `expr` returns a promise, the
#' span will be automatically ended when the promise completes.
#'
#' This function differs from synchronous otel span operations in that it
#' installs a promise domain and properly handles asynchronous operations. In
#' addition, the internal span will be ended either when the function exits (for
#' synchronous operations) or when a returned promise completes (for
#' asynchronous operations).
#'
#' If OpenTelemetry is not enabled, the expression will be evaluated without any
#' tracing context.
#'
#' @param name Character string. The name of the otel span.
#' @param expr An expression to evaluate within the otel span context.
#' @param ... Additional arguments passed to [`otel::start_span()`].
#' @param tracer (Required) An `{otel}` tracer. It is required to provide your
#'   own tracer from your own package. See [`otel::get_tracer()`] for more
#'   details.
#' @param attributes Attributes passed through [`otel::as_attributes()`] (when
#'   not `NULL`)
#' @export
#' @seealso [`otel::start_span()`], [`otel::with_active_span()`],
#'   [`otel::end_span()`]
#' @examples
#' \dontrun{
#' # Common usage:
#' with_otel_promise_domain({
#'   # ... deep inside some code execution ...
#'
#'   # Many calls to `with_otel_span()` within `with_otel_promise_domain()`
#'   with_otel_span("my_operation", {
#'     # ... do some work ...
#'   })
#' })
#' }
#' \dontrun{
#' with_otel_promise_domain({
#'   # ... deep inside some code execution ...
#'
#'   # Synchronous operation
#'   # * Creates `my_operation` span
#'   result <- with_otel_span("my_operation", {
#'     # ... do some work ...
#'     print(otel::get_active_span()$name) # "my_operation"
#'
#'     # Nest (many) more spans
#'     prom_nested <- with_otel_span("my_nested_operation", {
#'       # ... do some more work ...
#'       promise_resolve(42) |>
#'         then(\(value) {
#'           print(otel::get_active_span()$name) # "my_nested_operation"
#'           print(value) # 42
#'         })
#'     })
#'
#'     # Since `then()` is called during the active `my_operation` span,
#'     # the `my_operation` span will be reactivated in the `then()` callback.
#'     prom_nested |> then(\(value) {
#'       print(otel::get_active_span()$name) # "my_operation"
#'       value
#'     })
#'   })
#'
#'   # Since `then()` is called where there is no active span,
#'   # there is no _active_ span in the `then()` callback.
#'   result |> then(\(value) {
#'     stopifnot(inherits(otel::get_active_span(), "otel_span_noop"))
#'     print(value) # 42
#'   })
#' })
#' }
with_otel_span <- function(
  name,
  expr,
  ...,
  tracer,
  attributes = NULL
) {
  # If tracing is not enabled, just run the expression
  #
  # ```
  # tracer <- get_tracer(); bench::mark(check = FALSE, !is_tracing_enabled(tracer), inherits(tracer, "otel_tracer_noop"), !.subset2(tracer, "is_enabled")())[,1:4]
  # #> A tibble: 3 × 4
  # #>   expression                                    min   median `itr/sec`
  # #>   <bch:expr>                               <bch:tm> <bch:tm>     <dbl>
  # #> 1 "!is_tracing_enabled(tracer)"              2.79µs    3.4µs   278206.
  # #> 2 "inherits(tracer, \"otel_tracer_noop\")" 205.01ns    287ns  3184760.
  # #> 3 "!.subset2(tracer, \"is_enabled\")()"     82.02ns    164ns  5716255.
  # ```
  #
  # Speed is important here as to not slow down non-tracing operations.
  # Therefore, we use the fastest method and require `tracer=` to be provided.
  # The method is safe to call as it just returns either `TRUE` or `FALSE` (if a
  # noop tracer).
  if (!.subset2(tracer, "is_enabled")()) {
    return(force(expr))
  }

  span <- start_span(name, ..., tracer = tracer, attributes = attributes)

  needs_cleanup <- TRUE
  cleanup <- function() {
    end_span(span)
  }
  on.exit(if (needs_cleanup) cleanup(), add = TRUE)

  result <- with_active_span(span, expr)

  if (is.promising(result)) {
    needs_cleanup <- FALSE

    result <- finally(result, cleanup)
  }

  result
}


#' @describeIn otel Adds an idempotent handoff Active OpenTelemetry span promise domain.
#'
#' Package authors are required to use this function to have otel span context
#' persist across asynchronous boundaries. This method is only needed once per
#' promise domain stack. If you are unsure, feel free to call
#' `with_otel_promise_domain()` as the underlying promise domain will only be
#' added if not found within the current promise domain stack. If your package
#' **only** works within Shiny apps, Shiny will have already added the domain so
#' no need to add it yourself. If your package works outside of Shiny and you
#' use `{promises}` (i.e. `{chromote}`), then you'll need to use this wrapper
#' method.
#'
#' This method adds a _handoff_ Active OpenTelemetry span promise domain to the
#' expression evaluation. This _handoff_ promise domain will only run once on
#' reactivation. This is critical if there are many layered
#' `with_otel_span()` calls, such as within Shiny reactivity. For example, if
#' we nested many `with_otel_span()` calls of which each call added a promise
#' domain that reactivated each otel span on restore, we'd reactivate `k` otel
#' span objects (`O(k)`) when we only need to activate the **last** span
#' (`O(1)`).
#'
#' Returns the result of evaluating `expr` within the otel span promise domain.
#'
#' @export
with_otel_promise_domain <- function(expr) {
  # with_otel_promise_domain <- function(expr) {
  # Do not add the otel span handoff promise domain twice
  if (has_otel_promise_domain()) {
    return(force(expr))
  }
  act_span_pd <- create_otel_promise_domain()
  with_promise_domain(act_span_pd, expr)
}


#' @describeIn otel Local OpenTelemetry span promise domain
#'
#' Adds an OpenTelemetry span promise domain to the local scope. This is useful
#' for `{coro}` operations where encapsulating the coro operations inside a
#' `with_*()` methods is not allowed.
#'
#' When not using `{coro}`, please prefer to use `with_otel_span()` or
#' `with_otel_promise_domain()`.
#' @export
#' @param envir The "local" environment in which to add the promise domain. When
#' the environment is exited, the promise domain is removed.
local_otel_promise_domain <- function(envir = parent.frame()) {
  # local_otel_promise_domain <- function(envir = parent.frame()) {
  if (has_otel_promise_domain()) {
    return(invisible())
  }

  local_promise_domain(
    create_otel_promise_domain(),
    envir = envir
  )
}

# Modifies the current promise domain to include `domain` for the local scope.
local_promise_domain <- function(
  domain,
  envir = parent.frame(),
  replace = FALSE
) {
  oldval <- current_promise_domain()
  if (replace) {
    globals$domain <- domain
  } else {
    globals$domain <- compose_domains(oldval, domain)
  }

  # Inspired by `withr::defer()`
  defer(
    {
      globals$domain <- oldval
    },
    envir = envir
  )

  invisible()
}

# # TODO: Set attributes on the current active span
# # 5. Set attributes on the current active span
# set_otel_span_attrs(status = 200L)

# -- Helpers --------------------------------------------------------------

has_otel_promise_domain <- function(domain = current_promise_domain()) {
  # Works for:
  # * `list()` `compose_domains()` result
  # * `NULL` - current_promise_domain() init value
  # * environment objects - `new_promise_domain()` result
  isTRUE(domain[[".otel_promise_domain"]])
}

#' Create a OpenTelemetry handoff promise domain
#'
#' Creates a handoff promise domain captures the currently active span. Upon
#' promise domain restoration, the previously captured span will be reactivated.
#'
#' @param span An OpenTelemetry span object.
#' @noRd
create_otel_promise_domain <- function() {
  new_promise_domain(
    .otel_promise_domain = TRUE, # set flag for later discovery
    wrapOnFulfilled = function(onFulfilled) {
      force(onFulfilled)

      ## Motivation on why we can't use
      ## `promises_otel_tracer()$get_active_span()` here to save 10 microseconds:
      ## https://github.com/r-lib/otel/issues/31#issuecomment-3250669840
      #> ❯❯ bench::mark(promises_otel_tracer()$get_active_span(), get_active_span())
      #> # A tibble: 2 × 13
      #>   expression            min  median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result     memory     time
      #>   <bch:expr>         <bch:> <bch:t>     <dbl> <bch:byt>    <dbl> <int> <dbl>   <bch:tm> <list>     <list>     <list>
      #> 1 promises_otel_tra…  4.8µs  6.52µs   144270.        0B     14.4  9999     1     69.3ms <otl_spn_> <Rprofmem> <bench_tm>
      #> 2 get_active_span()  14.3µs 15.13µs    64876.    7.02KB     38.9  9994     6      154ms <otl_spn_> <Rprofmem> <bench_tm>
      # ℹ 1 more variable: gc <list>
      ## tl/dr: Speed comes at the cost of safety. Just use this for now as it is only a single cost on restore.

      # Instead, use `get_active_span()` directly so that the recording status
      # can be changed over time for testing and within the terminal. The 10
      # microseconds is not worth the inconsistent behavior or the time to debug
      # the inconsistent behavior.

      # The `is_recording()` method is deemed safe to call as the underlying C++
      # method just checks an internal class instance property, is
      # non-allocating and has been declared noexcept.

      span <- get_active_span()
      if (!.subset2(span, "is_recording")()) {
        return(onFulfilled)
      }

      function(...) {
        with_active_span(span, {
          onFulfilled(...)
        })
      }
    },
    wrapOnRejected = function(onRejected) {
      force(onRejected)

      span <- get_active_span()
      if (!.subset2(span, "is_recording")()) {
        return(onRejected)
      }

      function(...) {
        with_active_span(span, {
          onRejected(...)
        })
      }
    }
  )
}

skip_on_cran()

skip_if_not_installed("otelsdk")

with_otel_promise_domain({
  describe("OpenTelemetry integration", {
    describe("otel::start_span()", {
      it("creates spans appropriately", {
        records <- otelsdk::with_otel_record({
          span <- otel::start_span("test_span0", tracer = NULL)
          expect_true(inherits(span, "otel_span"))
          otel::end_span(span)
        })
        expect_true(!is.null(records$traces[["test_span0"]]))

        skip_if(!Sys.getenv("OTEL_TRACES_EXPORTER") %in% c("", "none"))
        expect_s3_class(
          otel::start_span("test_span"),
          "otel_span_noop"
        )
      })
    })

    describe("Deprecated functions", {
      it("with_ospan_async() is deprecated", {
        expect_snapshot({
          with_ospan_async(
            "test_span_deprecated",
            {
              42
            },
            tracer = promises_otel_tracer()
          )
        })
      })

      it("with_ospan_promise_domain() is deprecated", {
        expect_snapshot({
          with_ospan_promise_domain({
            42
          })
        })
      })

      it("local_ospan_promise_domain() is deprecated", {
        expect_snapshot({
          local({
            local_ospan_promise_domain()
            42
          })
        })
      })
    })

    describe("with_otel_span()", {
      it("executes synchronous expressions without otel", {
        # When recording is off, should still execute expression
        result <- with_otel_span(
          "test_span1",
          {
            42
          },
          tracer = promises_otel_tracer()
        )
        expect_equal(result, 42)
      })
      it("executes synchronous expressions", {
        records <- otelsdk::with_otel_record({
          result <- with_otel_span(
            "test_span1",
            {
              42
            },
            tracer = promises_otel_tracer()
          )
          expect_equal(result, 42)
        })
        expect_true(!is.null(records$traces[["test_span1"]]))
      })
      it("executes asynchronous expressions without otel", {
        # When recording is off, should still execute expression
        result <- with_otel_span(
          "test_span1",
          {
            promise_resolve(42)
          },
          tracer = promises_otel_tracer()
        )
        expect_true(is.promising(result))
        expect_equal(extract(result), 42)
      })

      it("handles promise results", {
        records <- otelsdk::with_otel_record({
          result <- with_otel_span(
            "test_span2",
            {
              promise_resolve(42)
            },
            tracer = promises_otel_tracer()
          )

          expect_true(is.promising(result))
          expect_equal(extract(result), 42)
        })
        expect_true(!is.null(records$traces[["test_span2"]]))
      })
    })

    describe("promise domain integration", {
      it("integrates with existing promise domains", {
        # Test domain composition works even when otel is not available
        counting_domain <- create_counting_domain()

        p <- with_promise_domain(counting_domain, {
          with_otel_span("test_operation", tracer = promises_otel_tracer(), {
            promise_resolve(42) |>
              then(function(x) x * 2)
          })
        })

        result <- extract(p)
        expect_equal(result, 84)

        # Verify the counting domain was active
        expect_gte(counting_domain$counts$onFulfilledBound, 1L)
      })
    })

    describe("error handling", {
      it("handles promise rejections gracefully", {
        # Should handle rejected promises without errors
        records <- otelsdk::with_otel_record({
          p <- with_otel_span("test_span3", tracer = promises_otel_tracer(), {
            promise_reject("test error") |>
              catch(function(reason) {
                with_otel_span("catch_span3", tracer = promises_otel_tracer(), {
                  42
                })

                expect_true(inherits(reason, "error"))
                expect_match(as.character(reason), "test error")

                "caught"
              }) |>
              then(function(result) {
                expect_equal(result, "caught")
                with_otel_span("then_span3", tracer = promises_otel_tracer(), {
                  42
                })
              })
          })
          wait_for_it(p)
          p
        })
        expect_true(is.promising(p))

        # Should be able to catch the error

        test_id <- records$traces[["test_span3"]]$span_id
        expect_equal(records$traces[["catch_span3"]]$parent, test_id)
        expect_equal(records$traces[["then_span3"]]$parent, test_id)
      })

      it("propagates regular errors", {
        # Regular errors should still propagate
        expect_error(
          with_otel_span("test_span4", tracer = promises_otel_tracer(), {
            stop("regular error")
          }),
          "regular error"
        )
      })
    })

    it("maintains span context across parallel promise chains", {
      # Track execution order and span context
      execution_order <- character(0)

      # Compare execution order after making `recording`
      # Compare span parents using `recording`

      recording <- otelsdk::with_otel_record({
        # Create two parallel promise chains, each with their own span
        chain1 <- with_otel_span("chain_1", tracer = promises_otel_tracer(), {
          promise_resolve("init1") |>
            then(function(x) {
              spn <- otel::start_span("chain1_step1")
              on.exit(otel::end_span(spn))
              execution_order <<- c(execution_order, "chain1_step1")
              paste0(x, "_step11")
            }) |>
            then(function(x) {
              spn <- otel::start_span("chain1_step2")
              on.exit(otel::end_span(spn))
              execution_order <<- c(execution_order, "chain1_step2")
              paste0(x, "_step12")
            }) |>
            then(function(x) {
              spn <- otel::start_span("chain1_step3")
              on.exit(otel::end_span(spn))
              execution_order <<- c(execution_order, "chain1_step3")
              paste0(x, "_step13")
            }) |>
            then(function(x) {
              spn <- otel::start_span("chain1_step4")
              on.exit(otel::end_span(spn))
              execution_order <<- c(execution_order, "chain1_step4")
              paste0(x, "_final1")
            })
        })

        chain2 <- with_otel_span("chain_2", tracer = promises_otel_tracer(), {
          promise_resolve("init2") |>
            then(function(x) {
              spn <- otel::start_span("chain2_step1")
              on.exit(otel::end_span(spn))
              execution_order <<- c(execution_order, "chain2_step1")
              paste0(x, "_step12")
            }) |>
            then(function(x) {
              spn <- otel::start_span("chain2_step2")
              on.exit(otel::end_span(spn))
              execution_order <<- c(execution_order, "chain2_step2")
              paste0(x, "_step22")
            }) |>
            then(function(x) {
              spn <- otel::start_span("chain2_step3")
              on.exit(otel::end_span(spn))
              execution_order <<- c(execution_order, "chain2_step3")
              paste0(x, "_step23")
            }) |>
            then(function(x) {
              spn <- otel::start_span("chain2_step4")
              on.exit(otel::end_span(spn))
              execution_order <<- c(execution_order, "chain2_step4")
              paste0(x, "_final2")
            })
        })

        # Wait for both chains to complete
        result1 <- extract(chain1)
        result2 <- extract(chain2)

        # Verify final recording
        expect_equal(result1, "init1_step11_step12_step13_final1")
        expect_equal(result2, "init2_step12_step22_step23_final2")

        # Check that execution alternates between chains
        expected_alternating_pattern <- c(
          "chain1_step1",
          "chain2_step1",
          "chain1_step2",
          "chain2_step2",
          "chain1_step3",
          "chain2_step3",
          "chain1_step4",
          "chain2_step4"
        )

        expect_equal(
          execution_order,
          expected_alternating_pattern,
          info = "Execution should alternate between chains"
        )
      })

      # * Verify spans at each step have the same parent, even though the
      #   previous calculation may have been a different promise domain.
      # * All steps in chain1 should have same parent span, all in chain2 should
      #   have same parent span.
      chain1_id <- recording$traces[["chain_1"]]$span_id
      for (name in c(
        "chain1_step1",
        "chain1_step2",
        "chain1_step3",
        "chain1_step4"
      )) {
        expect_equal(recording$traces[[name]]$parent, chain1_id)
      }

      chain2_id <- recording$traces[["chain_2"]]$span_id
      for (name in c(
        "chain2_step1",
        "chain2_step2",
        "chain2_step3",
        "chain2_step4"
      )) {
        expect_equal(recording$traces[[name]]$parent, chain2_id)
      }
    })
  })
})

describe("local_otel_promise_domain()", {
  it("sets up otel span promise domain for local scope", {
    # Test that the domain is active within the local scope
    original_domain <- current_promise_domain()

    local({
      local_otel_promise_domain()
      current_domain <- current_promise_domain()

      # Should have an otel span domain set
      expect_false(identical(current_domain, original_domain))
      expect_true(has_otel_promise_domain(current_domain))
    })

    # Should be restored after exiting the local scope
    expect_identical(current_promise_domain(), original_domain)
  })

  it("restores previous domain when scope exits", {
    counting_domain <- create_counting_domain()

    expect_equal(current_promise_domain(), NULL)

    with_promise_domain(counting_domain, {
      domain_before <- current_promise_domain()

      local({
        local_otel_promise_domain()
        current_domain <- current_promise_domain()

        # Should be a composed domain with otel span support
        expect_false(identical(current_domain, domain_before))
        expect_true(has_otel_promise_domain(current_domain))
      })

      # Should restore the counting domain
      expect_identical(current_promise_domain(), domain_before)
    })

    expect_equal(current_promise_domain(), NULL)
  })

  # it("works with custom environment", {
  #   test_env <- new.env()
  #   original_domain <- current_promise_domain()

  #   # Set up domain in custom environment
  #   local_otel_promise_domain(envir = test_env)

  #   # Domain should still be active since test_env hasn't been cleaned
  #   current_domain <- current_promise_domain()
  #   expect_false(identical(current_domain, original_domain))
  #   expect_true(has_otel_promise_domain(current_domain))

  #   # Clean up the environment to trigger restoration
  #   rm(list = ls(test_env), envir = test_env)

  #   # Force garbage collection to ensure cleanup handlers run
  #   gc()

  #   # Domain should be restored (though this timing is implementation dependent)
  #   # So we mainly test that the function doesn't error with custom envir
  #   expect_true(TRUE)
  # })

  it("integrates with promise execution", {
    records <- otelsdk::with_otel_record({
      result <- local({
        local_otel_promise_domain()

        otel::start_local_active_span("outer_test_span")

        # Create a promise within the otel span domain
        promise_resolve(21) |>
          then(function(x) {
            # This should be executed within the otel span domain
            span <- otel::start_span("inner_test_span")
            on.exit(otel::end_span(span))
            x * 2
          })
      })

      wait_for_it(result)
      result
    })

    expect_true(is.promising(result))
    expect_equal(extract(result), 42)
    expect_true(!is.null(records$traces[["outer_test_span"]]))
    expect_true(!is.null(records$traces[["inner_test_span"]]))
    expect_equal(
      records$traces[["inner_test_span"]]$parent,
      records$traces[["outer_test_span"]]$span_id
    )
  })

  it("can be nested without issues", {
    # Track how many times promise domain setup occurs
    domain_creation_count <- 0
    original_copd <- create_otel_promise_domain

    # Mock the domain creation to count calls
    with_mocked_bindings(
      create_otel_promise_domain = function() {
        domain_creation_count <<- domain_creation_count + 1
        original_copd()
      },
      {
        original_domain <- current_promise_domain()

        local({
          local_otel_promise_domain()
          expect_equal(domain_creation_count, 1)
          domain1 <- current_promise_domain()

          local({
            # This should not create another domain due to idempotency
            local_otel_promise_domain()
            expect_equal(domain_creation_count, 1) # Still 1, not 2
            domain2 <- current_promise_domain()

            # Both should have otel span domains and should be identical
            expect_true(has_otel_promise_domain(domain1))
            expect_true(has_otel_promise_domain(domain2))
            expect_identical(domain1, domain2) # Should be the same due to idempotency
          })

          # Should still be domain1 after nested scope exits
          expect_identical(current_promise_domain(), domain1)
        })

        # Should restore to original after all scopes exit
        # Note: This may not be immediate due to deferred cleanup
        expect_equal(domain_creation_count, 1) # Only created once total
      }
    )
  })
})

describe("has_otel_promise_domain()", {
  it("returns FALSE for NULL domain", {
    expect_false(has_otel_promise_domain(NULL))
  })

  it("returns FALSE for empty list", {
    expect_false(has_otel_promise_domain(list()))
  })

  it("returns FALSE for regular promise domain", {
    regular_domain <- new_promise_domain(
      wrapOnFulfilled = function(onFulfilled) onFulfilled
    )
    expect_false(has_otel_promise_domain(regular_domain))
  })

  it("returns TRUE for otel span promise domain", {
    otel_span_domain <- create_otel_promise_domain()
    expect_true(has_otel_promise_domain(otel_span_domain))
  })

  it("returns TRUE for composed domain with otel span", {
    counting_domain <- create_counting_domain()
    otel_span_domain <- create_otel_promise_domain()
    composed_domain <- compose_domains(counting_domain, otel_span_domain)

    expect_true(has_otel_promise_domain(composed_domain))
  })

  it("returns FALSE for composed domain without otel span", {
    counting_domain1 <- create_counting_domain()
    counting_domain2 <- create_counting_domain()
    composed_domain <- compose_domains(counting_domain1, counting_domain2)

    expect_false(has_otel_promise_domain(composed_domain))
  })

  it("uses current_promise_domain() when no argument provided", {
    # Initially no domain
    expect_false(has_otel_promise_domain())

    # With otel span domain active
    with_otel_promise_domain({
      expect_true(has_otel_promise_domain())
    })

    # Back to no domain
    expect_false(has_otel_promise_domain())
  })

  it("works with various data types", {
    # # Test edge cases
    # expect_false(has_otel_promise_domain(character(0)))
    # expect_false(has_otel_promise_domain(numeric(0)))
    # expect_false(has_otel_promise_domain("not a domain"))
    # expect_false(has_otel_promise_domain(42))

    # Test list with wrong flag
    fake_domain <- list(.some_other_flag = TRUE)
    expect_false(has_otel_promise_domain(fake_domain))

    # Test list with correct flag
    correct_domain <- list(.otel_promise_domain = TRUE)
    expect_true(has_otel_promise_domain(correct_domain))

    # Test list with correct flag but FALSE value
    false_domain <- list(.otel_promise_domain = FALSE)
    expect_false(has_otel_promise_domain(false_domain))
  })
})

describe("with_otel_promise_domain() idempotency", {
  it("is idempotent when called multiple times", {
    # Track how many times promise domain setup occurs
    domain_creation_count <- 0
    reset_count <- function() {
      domain_creation_count <<- 0
    }
    original_copd <- create_otel_promise_domain

    expect_false(has_otel_promise_domain())

    # Mock the domain creation to count calls
    with_mocked_bindings(
      create_otel_promise_domain = function() {
        domain_creation_count <<- domain_creation_count + 1
        original_copd()
      },
      {
        # First call should create the domain
        result1 <- with_otel_promise_domain({
          42
        })
        expect_equal(result1, 42)
        expect_equal(domain_creation_count, 1)

        reset_count()

        result2 <-
          with_otel_promise_domain({
            expect_equal(domain_creation_count, 1)

            # Nested calls should not create additional domains
            with_otel_promise_domain({
              with_otel_promise_domain({
                expect_equal(domain_creation_count, 1)

                84
              })
            })
          })
        expect_equal(result2, 84)
        # Should still be 1, even with two prom domain calls
        expect_equal(domain_creation_count, 1)
      }
    )
  })

  it("allows nested calls without duplicate domain setup", {
    # Test that nested with_otel_promise_domain calls work correctly
    # and don't interfere with each other

    records <- otelsdk::with_otel_record({
      result <- with_otel_promise_domain({
        with_otel_span("outer_span", tracer = promises_otel_tracer(), {
          # This nested call should be idempotent
          with_otel_promise_domain({
            with_otel_span("inner_span", tracer = promises_otel_tracer(), {
              promise_resolve(42) |>
                then(function(x) {
                  # Another nested call
                  with_otel_promise_domain({
                    x * 2
                  })
                })
            })
          })
        })
      })

      wait_for_it(result)
      result
    })

    expect_true(is.promising(result))
    expect_equal(extract(result), 84)

    # Verify spans were created correctly
    expect_true(!is.null(records$traces[["outer_span"]]))
    expect_true(!is.null(records$traces[["inner_span"]]))

    # Verify parent-child relationship
    outer_id <- records$traces[["outer_span"]]$span_id
    expect_equal(records$traces[["inner_span"]]$parent, outer_id)
  })
})

#' Determine whether an non-evaluated call is parenthesized
#'
#' @param a non-evaluated expression
#' @return logical - TRUE if expression is parenthesized, FALSE otherwise.
#' @noRd
is_parenthesized <- function(expr) {
  is.call(expr) && identical(expr[[1L]], quote(`(`))
}


#' Check whether expression is enclosed in curly braces.
#'
#' @param  expr An expression to be tested.
#' @return logical - TRUE if expr is enclosed in `{`, FALSE otherwise.
#' @noRd
is_funexpr <- function(expr) {
  is.call(expr) && identical(expr[[1L]], quote(`{`))
}

#' Check whether expression has double or triple colons
#'
#' @param  expr An expression to be tested.
#' @return logical - TRUE if expr contains `::` or `:::`, FALSE otherwise.
#' @noRd
is_colexpr <- function(expr) {
  is.call(expr) &&
    (identical(expr[[1L]], quote(`::`)) || identical(expr[[1L]], quote(`:::`)))
}


#' Determine whether an expression counts as a function in a magrittr chain.
#'
#' @param a non-evaluated expression.
#' @return logical - TRUE if expr represents a function, FALSE otherwise.
#' @noRd
is_function <- function(expr) {
  is.symbol(expr) || is.function(expr)
}

is_lambda <- function(expr) {
  is.call(expr) && identical(expr[[1L]], quote(`function`))
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom later later
#' @importFrom lifecycle deprecated
#' @importFrom R6 R6Class
## usethis namespace: end
NULL

ignore_unused_imports <- function() {
  fastmap::fastqueue
}

release_bullets <- function() {
  c(
    "Update staticimports: `staticimports::import()`",
    "Revert `future_promise()` example to use `\\donttest{}`, not `\\dontrun{}` - https://github.com/rstudio/promises/pull/134",
    "Revert `skip_on_os(\"mac\")` in `future_promise()` test - https://github.com/rstudio/promises/pull/134"
  )
}

---
title: "Working with promises in R"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Working with promises in R}
  %\VignetteEncoding{UTF-8}
---

One persistent challenge with developing Shiny apps for live deployment is the R language runtime’s single-threaded nature. Because of this, a given Shiny app process can only do one thing at a time: if it is fitting a linear model for one client, it cannot simultaneously prepare a CSV download for another client, and vice versa.

For many Shiny apps, this isn’t a big problem; because no one processing step takes very long, no client has to wait an undue amount of time before they start seeing results. But for apps that perform long-running operations — either expensive computations that take a while to complete, or waiting on slow network operations like database or web API queries — your users’ experience can suffer dramatically as traffic ramps up.

The traditional approach to scaling web applications is to launch multiple processes and balance traffic between them, and indeed, Shiny Server Pro and RStudio Connect both implement a variant of this strategy. You do some load testing to determine how many concurrent users a single process can support, then [configure Shiny Server Pro](https://docs.posit.co/shiny-server/) to launch new processes as those limits are approached.

But there are some applications that perform truly expensive operations, like simulations, training neural networks, or complex per-row processing, that take *minutes* to complete. Again, while this is happening, any other users that are unfortunate enough to be assigned to the same process are completely blocked from proceeding in any way — even loading static JavaScript/CSS assets must wait until the blocking operation is complete.

Asynchronous (async) programming offers a way to offload certain classes of long-running operations from the main R thread, such that Shiny apps can remain responsive.

A warning before we dive in: async code is hard to write! It is hard in C++, it is hard in Java, it is hard in JavaScript, and sadly, R is no exception. We have attempted to make the API as simple and elegant as possible, but just as with reactive programming, it will likely take some effort to internalize the main concepts, and plenty of practice before expressing logic this way begins to feel natural.

## Async programming in R

Integrating async programming capabilities into R involves two types of tasks:

1.  **Invoking:** Getting expensive operations to happen either on a different thread, or (more likely) in a different process, leaving the main R thread/process free to continue doing other work. Generally, an expensive operation will either produce a result value (e.g. a data frame), or cause a side effect (e.g. a write to a database).
2.  **Handling:** When an operation completes or fails, notify the main R thread/process so that it may make use of the resulting value or error in further logic. Handling logic may choose to perform side effects like logging or persisting, or to transform the value/error for further downstream processing.

In our vision for R async programming, there should be several different ways of invoking expensive operations asynchronously, each with different tradeoffs, depending on the type of task you are trying to execute. We will go into more detail later, but just to give you an idea, here are just a few of the different strategies you could use to invoke code asynchronously:

* Run it in the current process, but on a different thread. (This strategy is impossible for R code, but you can run C/C++ code on a different thread, even in an R process.)
* Launch a separate R process and pass it the R code to evaluate.
* Fork the R process and run the code in the child process. (Doesn’t work on Windows.)
* Farm the code out to a pre-allocated cluster of R processes, either on the same machine or distributed across a network.

Regardless of which approach you choose, the API for handling the result is identical. It’s centered around an abstraction that you will come to know very well: the **promise**.

## Promises: the central abstraction of async programming

> **Terminology note:** Advanced R users (or users who have at least read [Advanced R](http://adv-r.had.co.nz/Computing-on-the-language.html)) may be familiar with the term “promises” already: in R, unevaluated function arguments are technically called promises. Those types of promises have nothing to do with asynchronous programming, and the things we call “promises” in this document have nothing to do with those, so try to forget they exist for the time being. Sorry for the confusion.

A promise is an object that represents the *eventual result* of a specific asynchronous operation.

Whenever you launch an async task, you get a promise object back. That promise is what lets you know:

* When the task completes (if ever)
* Whether the task completed successfully or failed
* If success, the result value
* If failure, the error

So if a regular, synchronous function call generally looks like this:

```r
value <- read.csv("http://example.com/data/data.csv")
```

An asynchronous function call (which uses [the mirai package](promises_04_mirai.html)) will look instead like:

```r
promise <- as.promise(mirai(read.csv("http://example.com/data/data.csv")))
```

While the regular function call returns a data frame, the async call returns a promise, which is most definitely not a data frame. You cannot ask the promise how many rows it has, or the names of its columns. You cannot run dplyr operations on it, or turn it into a data.table.

You might guess that you could call a function or method on a promise to extract the value, like `value(promise)` or `promise$value()`. But that isn't how promises work. Instead, everything is based on a function called `then`.

## Accessing results with `then`

The `promises::then` function is what ultimately makes promise objects useful. It is used to register success and failure handlers on a promise. Its signature looks like:

```r
then(promise, onFulfilled = NULL, onRejected = NULL, ..., tee = FALSE)
```

In promise terminology, “fulfilled” (and equivalently, “resolved”) means success and “rejected” means failure. You can pass functions with single arguments to `onFulfilled` and `onRejected` to be notified when a promise succeeds or fails. (If the promise has already been fulfilled or resolved by the time `then` is called, don't worry—the appropriate callback will be still be called. It's never too late to call `then` on a promise.)

The promise library guarantees that only one of `onFulfilled` or `onRejected` will be called, never both. And a callback will never be invoked more than once. It is possible, though, that neither callback will ever be called, i.e. the async operation never completes. (This is analogous to calling a regular function that never returns.)

For now, we will focus on fulfillment, and come back to rejection in the [Error Handling](#error-handling) section below.

The following example shows a simple example of printing out a success message and the value.

```r
then(promise,
  \(value) {
    cat("The operation completed!\n")
    print(value)
  })
```

In the example above, we used the anonymous function shorthand (`\(x) { }`) to write an anonymous function. `function(x) { }` would behave just the same.

To help reduce the amount of in/out reading, we can use the R pipe operator (`|>`):

```r
promise |>
  then(\(value) {
    cat("The operation completed!\n")
    print(value)
  })
```

Note that the call to `then()` **always returns immediately**, without invoking the callback function. The callback function will be invoked sometime in the future—it could be very soon, or it could be hours, depending mostly on how long it takes the async operation to complete.





## Promise chaining

The `then()` function has an important function beyond registering callbacks. It also returns a promise—not the promise it takes as an argument, but a new, distinct promise. This new promise gets fulfilled after the input promise has resolved and the callback registered by `then()` has run; the return value of the callback is used to fulfill the new promise.

For example:

```r
p <- promise_resolve(palmerpenguins::penguins)
p2 <- p |>
  then(nrow)
```

In this case, after `p` is fulfilled with a data frame, `p2` will be fulfilled with the number of rows of that data frame.

Because `then()` uses promises for both input and output, you can chain multiple `then()` calls together directly:

```r
p |>
  then(\(df) filter(df, year == 2008)) |>
  then(\(df) group_by(df, species)) |>
  then(\(df) summarise(df, mean_bill = mean(bill_length_mm, na.rm = TRUE))) |>
  then(\(df) arrange(df, desc(mean_bill))) |>
  then(print)
#> # A tibble: 3 × 2
#>   species   mean_bill
#>   <fct>         <dbl>
#> 1 Chinstrap      48.7
#> 2 Gentoo         46.9
#> 3 Adelie         38.6
```

Or, equivalently:

```r
p |>
  then(\(df) {
    df |>
      filter(year == 2008) |>
      group_by(state) |>
      summarise(pop = sum(population)) |>
      arrange(desc(pop))
  })
```

Evaluating this expression results in a promise that will eventually resolve to the filtered, summarized, and ordered data.

## Tee operator

When working with promise pipelines, it may sometimes be useful to have a stage that performs an action but does not modify the value presented to downstream stages. For example, you may want to log the number of rows in a data frame for diagnostic purposes:

```r
# Incorrect!
promise |>
  then(\(df) df |> filter(year == 2008)) |>
  then(\(df) print(nrow(df))) |>
  then(\(df) df |> group_by(state)) |>
  then(\(df) df |> summarise(pop = sum(population))) |>
  then(\(df) df |> arrange(desc(pop))) |>
  then(print)
```

This is not correct, as the `print(nrow(df))` stage will not only print the desired value, but pass the return value of `print(nrow(.))`, which is just `invisible(nrow(df))`, to the next stage.

For synchronous code, the [`{magrittr}` package offered the `%T>%` (pronounced “tee”) operator](https://magrittr.tidyverse.org/reference/tee.html), which operates like a regular `|>` except that, after executing its right-hand side, it returns its left-hand side value.

Similarly, for asynchronous code, you can use the `then(tee = TRUE)` method, which is like `then()` except that after execution it resolves using its input promise. The only difference in the corrected code below is the operator immediately preceding `print(nrow(df))` has changed from `then()` to `then(tee = TRUE)`.

```r
# Correct.
promise |>
  then(\(df) df |> filter(year == 2008)) |>
  then(tee = TRUE, \(df) print(nrow(df))) |>
  then(\(df) df |> group_by(state)) |>
  then(\(df) df |> summarise(pop = sum(population))) |>
  then(\(df) df |> arrange(desc(pop))) |>
  then(print)
```

## Error handling

Many scripts and Shiny apps that use promises will not contain any explicit error handling code at all, just like most scripts and Shiny apps don’t contain `tryCatch` or `try` calls to handle errors in synchronous code. But if you need to handle errors, promises have a robust and flexible mechanism for doing so.

### Catching errors with `onRejected`

The lowest level of error handling is built into the `then` function. To review, the `then` function takes an input promise, and up to two callbacks: `onFulfilled` and `onRejected`; and it returns a new promise as output. If the operation behind by the input promise succeeds, the `onFulfilled` callback (if provided) will be invoked. If the input promise’s operation fails, then `onRejected` (if provided) will be invoked with an error object.

```r
promise2 <- promise1 |>
  then(
    onFulfilled = \(value) {
      # Getting here means promise1 succeeded
    },
    onRejected = \(err) {
      # Getting here means promise1 failed
    }
  )
```

In the code above, you can see that the success or failure of `promise1` is what will determine which of the two callbacks is invoked.

But what about the output promise, `promise2`? We know what happens if `promise1` succeeds and the `onFulfilled` callback returns normally: `promise2` is resolved with the return value of `onFulfilled` (and if that return value is itself a promise, then `promise2` will do whatever that promise does). What happens if `promise1` is rejected; does that automatically mean `promise2` is rejected as well?

The answer is no, `promise2` is not automatically rejected if `promise1` is rejected. The rejection of `promise1` causes `onRejected` to be called, but from there on, `onFulfilled` and `onRejected` are treated identically. Whichever callback is invoked, if the invocation of the callback succeeds (returns either a regular value, or, a promise that ultimately resolves successfully) then the output promise will be resolved/succeed. But if the invocation of the callback fails (either throws an error, or returns a promise that ultimately rejects) then the output promise will be rejected/fail.

If you think about it, this behavior makes sense; just like `tryCatch`, once you’ve caught an error, it doesn’t continue to propagate, unless you go out of your way to do so by re-throwing it using `stop(err)`.

So the equivalent to this (synchronous) code:

```r
value <- tryCatch(
  somepkg::operation(),
  error = \(err) {
    warning("An error occurred: ", err)
    warning("Using default value of 0 instead")
    0
  }
)
```

would be this, when the operation is performed asynchronously:

```r
promise <- mirai(somepkg::operation()) |>
  then(onRejected = \(err) {
    warning("An error occurred: ", err)
    warning("Using default value of 0 instead")
    0
  })
```

In the synchronous case, an error in `operation()` will result in the error being logged as a warning, and `0` being assigned to `value`. In the asynchronous case, the same warning log messages will happen but then the value of `0` will be used to resolve `promise`. In both cases, the error is caught, dealt with, and turned into a non-error.

### Default onRejected behavior

In many of the examples above, we called `then` with an `onFulfilled` but no `onRejected`. What is the behavior of `then` if its input promise is rejected with an error, but the caller has not provided an explicit `onRejected` callback?

```r
promise2 <- promise1 |>
  then(head) |>
  then(print)
```

Well, `then` has its own default version of `onRejected`. It’s *not* an empty `onRejected = \(err) { }`, as you might think. Even though this function has no code in its body, it still returns normally, and thus would cause any errors to be caught and swallowed. That’s not the behavior we want; in the code above, we want a failure in `promise1` to cause `promise2` to be rejected so we know that something went wrong. So the default callback actually looks like: `onRejected = stop`, meaning, do nothing but raise the error, pushing the responsibility for error handling downstream.

(Incidentally, it’s valid to call `then` with `onRejected` and not `onFulfilled`, and the default version of `onFulfilled` is not an empty function either; instead, it’s `onFulfilled = identity`, so that the input promise’s return value can be passed through to the output promise.)

### Syntactic sugar for onRejected

The same syntactic sugar that is offered for non-error cases, is available for error handling code as well. You can use formulas in `onRejected()`:

```r
mirai(somepkg::operation()) |>
  then(onRejected = \(err) warning(err))
```

There’s a `catch()` function that is just a shorthand for `then(onRejected)`. It saves a little typing, but more importantly, is easier to read:

```r
mirai(somepkg::operation()) |>
  catch(warning)
```

### Error tee

Because it’s fairly common to want to do something with an error without stopping it from propagating (such as logging), there are a couple of additional shorthands for doing so without having to explicitly call `stop(err)`. For example:

```r
promise |> catch(print)
```

will print the error, but also eat it. To print the error without eating it, you’d have to do this:

```r
promise |>
  catch(\(err) {
    print(err)
    stop(err)
  })
```

That’s a fair amount of boilerplate. Instead, you can either add `tee = TRUE` to your `catch` call. These two lines are equivalent to the previous code chunk:

```r
promise |>
  catch(print, tee = TRUE)
```

## Cleaning up with `finally`

In synchronous programming, you use either`tryCatch(expr, finally = ...)` or `on.exit(...)` to perform tasks (usually relating to freeing resources or reverting temporary changes) regardless of whether the main logic succeeds or fails (throws an error). When programming with promises, you can use the `finally()` call to do the same. The `finally()` function is similar to `then()` but it only takes a single callback that executes on both success and failure, and its return value is ignored.

```r
file_path <- tempfile(fileext = ".png")
png_bytes_promise <-
  mirai(
    {
      png(file_path)
      plot(cars)
      dev.off()
      file_path
    },
    file_path = file_path
  ) |>
  then(\() brio::read_file_raw(file_path)) |>
  finally(\() unlink(file_path))
```

In this example, we need a temp file for the duration of the pipeline. Our `finally()` code expression makes sure the temp file is deleted when the operation is done, regardless of whether it succeeded or failed.

<div style="font-size: 20px; margin-top: 40px; text-align: right;">
Next: [Launching tasks](promises_04_mirai.html)
</div>

---
title: "Combining promises"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Combining promises}
  %\VignetteEncoding{UTF-8}
---

So far, all of our examples have involved chaining operations onto a single promise. In practice, you'll often find yourself needing to perform tasks that require the results of more than one promise. These are some patterns you may find useful:

* [**Gathering:**](#gathering) Combining multiple independent promises into a single computation
* [**Nesting:**](#nesting) Using the result of one promise to affect the input or execution of another async operation
* [**Racing:**](#racing) Using the fastest of multiple promises
* [**Mapping:**](#mapping) Applying an async function to each of a list's elements and collecting the results
* [**Reducing:**](#reducing) Applying an async function to each of a list's elements and reducing

## Gathering

The most common pattern for combining promises is gathering, where you have two or more promises in hand and you want to use all of their results in a computation. The `promise_all` function is designed for this. Its signature looks like:

```r
promise_all(..., .list = NULL)
```

`promise_all` takes any number of promises as named arguments, and returns a promise of a list containing named elements with the results of those promises.

Here's an example using `promise_all` to combine the results of two async `read.csv` operations:

```r
library(promises)
library(mirai)

a <- mirai(read.csv("a.csv"))
b <- mirai(read.csv("b.csv"))

result <- promise_all(a = a, b = b) |>
  then(\(values) {
    rbind(values$a, values$b)
  })
```

In this example, the value of `values` within the curly braces is a list whose elements `a` and `b` are both data frames. We use `rbind` to combine them.

The `values$` prefix is a bit inelegant, so we recommend the use of the base R function `with`, which lets you skip the prefix. Here's the same example, with `with`:

```r
library(promises)
library(mirai)

a <- mirai(read.csv("a.csv"))
b <- mirai(read.csv("b.csv"))

promise_all(a = a, b = b) |>
  then(\(values) {
    with(values, {
      rbind(a, b)
    })
  })
```

(Note that since the `promise_all` argument names are the same as the variable names (`a = a`, `b = b`), the original variables are masked: inside the `with` block, `a` now refers to the *result* of the promise `a`, not the promise object itself. If you find this confusing, you can just choose a different argument name, like `promise_all(a_result = a, …)`.)

The combination of `promise_all` and `with` is a concise and powerful way to gather the results of multiple promises.

`promise_all` also gives you two other options for passing input promises. First, if you would rather your result list be unnamed, you can pass in promises as unnamed arguments: `promise_all(a, b)` would yield `list(1, 2)`. Second, if you have a list of promises already in hand, you can pass the list as a single argument using `promise_all(.list = x)`(instead of, say, using `do.call(promise_all, x)`).

```r
proms <-
  lapply(c(a = "a.csv", b = "b.csv"), \(file) {
    mirai(read.csv(file))
  })
promise_all(.list = proms) |>
  then(\(values) {
    # values is a named list with elements a and b
    with(values, {
      rbind(a, b)
    })
  })
```

## Nesting

Gathering is easy and convenient, but sometimes not flexible enough. For example, if you use the result of promise `a` to decide whether to launch a second async task, whose result you then use in combination with the result of `a`.

```r
library(promises)
library(mirai)

a <- mirai(1)

a |>
  then(\(a) {
    b <- mirai(2)
    b |> then(\(b) a + b)
  })
```

(We use anonymous functions here to mask the names of the original promises--i.e. once inside the first anonymous function, the symbol `a` now refers to the result of the promise `a`.)

The nesting pattern is effective and flexible. The main downside is the physical nesting of the source code; if you use this pattern to a depth of more than a couple of promises, your code will be quite indented (in programming jargon this is referred to as the "pyramid of doom").

## Racing

```r
library(promises)
library(mirai)

a <- mirai({ Sys.sleep(1); 1 })
b <- mirai({ Sys.sleep(0.5); 2 })

first <- promise_race(a, b)
```

`promise_race` takes multiple promises and returns a new promise that will be fulfilled with the first promise that succeeds. In the example above, `first` is a promise that will be fulfilled with `2` after 0.5 seconds.

If one of the input promises rejects before any succeed, then the returned promise will be rejected.

Note that promises does not currently support cancellation. So losing promises will attempt to run to completion even after the race ends.

## Mapping

Use `promise_map` to run an async operation on each element of a list or vector, and collect the results in a list. It's very similar to `lapply` or `purrr::map`, except that the function to apply can return a promise, and the return value is also a promise.

In the example below, we iterate over a named vector of package names. For each package name, we launch an async task to download the package's description file from CRAN pick out the last published date.

```r
library(promises)
library(mirai)

get_pub_date <- \(pkg) {
  desc_url <- paste0("https://cran.r-project.org/web/packages/", pkg, "/DESCRIPTION")
  mirai(
    {
      r <- read.dcf(url(desc_url))[, "Date/Publication"]
      unname(r)
    },
    desc_url = desc_url
  )
}

packages <- setNames(, c("ggplot2", "dplyr", "knitr"))

pkg_dates <- promise_map(packages, get_pub_date)

pkg_dates |> then(print)
```

The resulting output looks like this:

```
$ggplot2
[1] "2025-04-09 09:40:10 UTC"

$dplyr
[1] "2023-11-17 16:50:02 UTC"

$knitr
[1] "2025-03-16 09:20:02 UTC"
```

`promise_map` works serially; each time it calls the given function on an element of the list/vector, it will wait for the returned promise to resolve before proceeding to the next element. Furthermore, any error or rejected promise will cause the entire `promise_map` operation to reject.

If you want behavior that's similar to `promise_map` but for all the async operations to occur in parallel, you can achieve that with a combination of a regular `purrr::map` and `promise_all`:

```r
pkg_dates <- purrr::map(packages, get_pub_date) |>
  promise_all(.list = _)

pkg_dates |> then(print)
```

## Reducing

Use `promise_reduce` when you have a list where you want to run an async operation on each of the elements, and to do so serially (i.e. only one async operation runs at a time). This can be helpful when you're searching through some elements using an async operation and want to terminate early when your search succeeds.

The signature of `promise_reduce` is as follows:

```r
promise_reduce(x, func, init = NULL)
```

If you've worked with `base::Reduce()` or `purr:::reduce()`, this should seem reasonably familiar: `x` is a vector or list; `func` is a function that takes two arguments, the accumulated value and the "next" value; and `init` is the default accumulated value.

The main difference between `promise_reduce` and `purrr:::reduce` is that with `promise_reduce`, your `func` can return a promise. If it does, `promise_reduce` will wait for it to resolve before updating the accumulated value and invoking `func` on the next element. The result returned from `promise_reduce` is a promise that resolves to the ultimate accumulated value.

The following example loops through a partial list of CRAN mirrors, returning the first one that passes whatever check `http::http_error` performs.

```r
library(promises)
library(mirai)

cran_mirrors <- c(
  "https://cloud.r-project.org",
  "https://cran.usthb.dz",
  "https://cran.csiro.au",
  "https://cran.wu.ac.at"
)

promise_reduce(
  cran_mirrors,
  \(result, mirror) {
    if (!is.null(result)) {
      result
    } else {
      mirai(
        {
          # Test the URL; return the URL on success, or NULL on failure
          if (!httr::http_error(mirror)) mirror
        },
        mirror = mirror
      )
    }
  },
  .init = NULL
) |>
  then(print)
```

---
title: "Launching tasks with mirai"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Launching tasks with mirai}
  %\VignetteEncoding{UTF-8}
---

```{css echo=FALSE}
.alert-secondary a, .alert-secondary a:visited {
  color: inherit;
  text-decoration: underline;
}
.alert code {
  color: inherit;
  background-color: inherit;
}
```

We've updated this guide to using the `mirai` package from `future` as we believe the following benefits are compelling in the context of Shiny apps:

1. Faster startup times and much less per-task overhead, meaning you can boost performance by making even shorter tasks async.
1. More linear scaling, meaning you get the same relative benefits whether running 2 or 200 cores.
1. Event-driven promises using `mirai` vs. promises using `future`, which time-poll every 100 ms. Lower latency and response times can help with the user experience.

The previous guide using `future` is available [here](promises_05a_futures.html). Using `future` continues to be supported within the Shiny ecosystem, and Henrik Bengtsson‘s excellent work on the futureverse deserves credit for pushing the boundaries of parallelism in R farther than many thought possible.

<hr>

The `mirai` package provides a lightweight way to launch R tasks that don't block the current R session.

The `promises` package provides the API for working with the results of async tasks, but it totally abdicates responsibility for actually launching/creating async tasks. The idea is that any number of different packages could be capable of launching async tasks, using whatever techniques they want, but all of them would either return promise objects or objects that can be converted to promise objects, as is the case for `mirai`.

This document will give an introduction to the parts of `mirai` that are most relevant to promises. For more information, please consult the documentation and vignettes that come with [`mirai`](https://mirai.r-lib.org).

## How mirai works

The main API that `mirai` provides couldn't be simpler. You call `mirai()` and pass it the code that you want executed asynchronously:

```R
m <- mirai({
  # expensive operations go here...
  df <- download_lots_of_data()
  fit_model(df)
})
```

The object that's returned is a mirai, which for all intents and purposes is a promise object[^1], which will eventually resolve to the return value of the code block (i.e. the last expression) or an error if the code does not complete executing successfully. The important thing is that no matter how long the expensive operation takes, these lines will execute almost instantly, while the operation continues in the background.

[^1]: (The `mirai` package provides several functions for working with mirai objects, but they are not relevant for our purposes.)

But we know that R is single-threaded, so how does `mirai` accomplish this? The answer: by utilizing another R process. `mirai` delegates the execution of the expensive operation to a totally different R process, so that the original R process can move on.

## Choosing a launch method

In mirai, the `daemons()` function is used to set and launch background R processes (*daemons*).

These background processes will be used/recycled for the life of the originating R process. If a mirai is launched while all the background R processes are busy executing, then the new mirai is queued until one of the background processes frees up.

To launch `n` processes locally, you just need to call `daemons(n)`, supplying the value of `n`.

You need to determine `n` yourself, and typically this should be at most one less than the number of processor cores on your machine, to leave one for the main R process. The reason we don't automatically detect this for you is that you may also be running other tasks on your machine, and you should take this into account when supplying a value for `n`.

`daemons()` has further arguments `url` and `remote` for setting and launching remote daemons over the network for distributed computing. To learn more, see the [`mirai::daemons()` reference docs](https://mirai.r-lib.org/reference/daemons.html) as well as the daemons sections of the [`mirai` reference vignette](https://mirai.r-lib.org/articles/mirai.html).

If you don't set `daemons()` in a session, then each `mirai()` call will launch a new local R process solely for the purpose of performing that evaluation. Whilst this may be desirable in certain circumstances, this is rarely going to be the case for Shiny. This is as we cannot limit the total number of processes spawned at any one time. If a Shiny app has many simultaneous users, then this could lead to an excessive number of processes being created, overwhelming the system.

## Caveats and limitations

The abstractions that `mirai` presents are simple and consistent, although it may take some time to get used to them. Please read this entire section carefully before proceeding.

### Globals: Providing input to mirai code chunks

Most mirai code chunks will need to reference data from the original process, e.g. data to be fitted, URLs to be requested, file paths to read from.

As evaluation happens in another process, these won't be available to the code chunk by default. These objects will need to be passed to the `...` argument of your `mirai()` call. These are then serialized and sent to the other process along with the code to be executed.

These objects include any functions which are defined in your session and not in a package.

For example:
```R
download_data <- \(url) {
  file <- tempfile()
  download.file(url, file, "libcurl")
  file
}

url <- "http://example.com/data.csv"

m <- mirai(
  {
    file <- download_data(url)
    read.csv(file)
  },
  download_data = download_data,
  url = url
)
```
If there are many variables to pass through, mirai does offer a convenience feature to pass an environment instead of individual `...` pairs. The above call would then look like this instead:

```R
m <- mirai(
  {
    file <- download_data(url)
    read.csv(file)
  },
  environment()
)
```
This passes the calling environment, which includes both the `download_data` function as well as `url`.

Care should be taken when using this feature as it will also pass anything else that happens to be in the same environment. It is safer to use when mirai is called inside of another function, then `environment()` will only consist of variables passed as arguments to that function, or created locally within it.

### Package loading

Besides variables, package functions need to be declared with the full namespace so that they can be found in the other process. For example, using `dplyr::mutate()` instead of just `mutate()`, even if the `dplyr` package is loaded in your main session, as the other process will not have any packages loaded by default.

Alternatively, make a call to load the package inside your mirai code chunk, for example by adding `library(dplyr)`. Sometimes this may be the most convenient option, especially for for infix operators. For example the `{magrittr}` pipe `%>%`, requires a `library(magrittr)` to load the package beforehand.

### Custom Data Types

Certain objects are implemented at a low level, not using one of R's native vector types, and represented in R by an external pointer. An example of this is an Arrow table. It is not possible to serialize these to and from R's native rds format. Instead they provide their own serialization and deserialization methods.

mirai offers a seamless solution for working with these data types, integrating those custom serialization and deserialization methods with R's native serialization so that you don't need to manually handle each instance of these objects when moving them across processes.

This does require a one-off configuration step when you set up daemons, and you may read more about this in the [`mirai` serialization vignette](https://mirai.r-lib.org/articles/mirai-serialization.html).

### Native resources

Mirai code blocks cannot use resources such as database connections and network sockets that were created in the parent process. Even if it seems to work with a simple test, you are asking for crashes or worse by sharing these kinds of resources across processes.

Instead, make sure you create, use, and destroy such resources entirely within the scope of the mirai code block.

### Mutation

Reference class objects (including R6 objects, S7 objects, and data.table objects) and environments are among the few "native" R object types that are mutable, that is, can be modified in-place. Unless they contain native resources (see previous section), there's nothing wrong with using mutable objects from within mirai code blocks, even objects created in the parent process. However, note that any changes you make to these objects will not be visible from the parent process; the mirai code is operating on a copy of the object, not the original.

### Returning values

Mirai code blocks return a value—they'd be a lot less useful if they couldn't! Like everywhere else in R, the return value is determined by the last expression in the code block, unless `return()` is explicitly called earlier.

The return value will always be copied back into the parent process. This matters for two reasons.

First, if the return value is very large, the copying process can take some time — and because the data must essentially be serialized to and deserialized from rds format, it can take a surprising amount of time. In the case of mirai blocks that execute fairly quickly but return huge amounts of data, you may be better off not using async techniques at all.

Second, objects that refer to native resources are unlikely to work in this direction either; just as you can't use the parent's database connections in the child process, you also cannot have the child process return a database connection for the parent to use.

<div style="font-size: 20px; margin-top: 40px; text-align: right;">
Next: [Using `promises` with Shiny](promises_06_shiny.html)
</div>

---
title: "Why use promises?"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Why use promises?}
  %\VignetteEncoding{UTF-8}
---

The `promises` library is designed to make it possible for R users to integrate *asynchronous programming* techniques into their code. It is intended to improve the scalability of certain classes of Shiny applications, and can also be used completely separately from Shiny.

Asynchronous programming means starting an operation without waiting around for its results. Contrast this with standard (synchronous) R programming, where the caller of a function must wait for the function to either return a result or throw an error.

Asynchronous programming is a particularly important capability for languages that are both 1) single threaded, and 2) need to serve multiple users concurrently. R is single threaded, but until recent years has mostly been restricted to desktop use or batch execution. The rise of Shiny and other web frameworks for R has meant much greater demand for serving multiple users concurrently, and thus a greater need for asynchronous programming.

## What are promises good for?

Promises work well when you have a Shiny application with a particular operation, especially a calculation, that takes a long time (measured in seconds or even minutes). Without promises, such operations not only block the current user's session from proceeding, but also block *all* other requests being made to the R process, no matter how trivial. Even loading a small CSS file, which should be nearly instantaneous, can be delayed by many seconds if another user's Shiny session is busy crunching through hundreds of gigabytes of data or querying a particularly slow backend database.

With promises, you can convert long-running operations to be asynchronous, which frees up the R process to do other work. For Shiny applications, this has the potential to greatly increase scalability for each R process (depending on how slow the operations are in the first place, and what resources they make use of).

**DO use promises if you have a Shiny app with long-running operations, and want to serve multiple users simultaneously.**

## What _aren't_ promises good for?

While promises can make a huge difference in the scalability of a Shiny app, they make relatively little difference in the latency of a single session. That is to say, if a Shiny application is slow when only a single user is hitting it, converting it to use promises is unlikely to make it perform any faster (and in fact may slightly slow it down). Promises will just help prevent *other* sessions from being slowed down by one session's computations.

(The exception to this is if the Shiny app has several distinct long computations that don't depend very much on each other—then you could use asynchronous programming to exploit a little parallelism. But I think this will be a less common use of async programming, as R already has other good tools designed specifically for data parallelism.)

**DON'T use promises to improve the performance of Shiny apps for a single user.**

<div style="font-size: 20px; margin-top: 40px; text-align: right;">
Next: [An informal intro to async programming](promises_02_intro.html)
</div>

---
title: "Using promises with Shiny"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Using promises with Shiny}
  %\VignetteEncoding{UTF-8}
---

```{css echo=FALSE}
.alert-secondary a, .alert-secondary a:visited {
  color: inherit;
  text-decoration: underline;
}
```

<div class="alert alert-secondary">
As of Shiny 1.8.1, there is a new feature called **Extended Tasks** that has several advantages over the approach in this article. We highly recommend that you read [the Extended Task documentation](https://shiny.posit.co/r/articles/improve/nonblocking/) first.
</div>

Taking advantage of async programming from Shiny is not as simple as turning on an option or flipping a switch. If you have already written a Shiny application and are looking to improve its scalability, expect the changes required for async operation to ripple through multiple layers of server code.

Async programming with Shiny boils down to following a few steps.

1. Identify slow operations (function calls or blocks of statements) in your app.

2. Convert the slow operation into a 'mirai' using `mirai()`. (If you haven't read the [article on mirai](promises_04_mirai.html), definitely do that before proceeding!)

3. Any code that relies on the result of that operation (if any), whether directly or indirectly, now must be converted to promise handlers that operate on the mirai object.

We'll get into details for all these steps, but first, an example. Consider the following synchronous server code:

```R
function(input, output, session) {
  output$plot <- renderPlot({
    result <- expensive_operation()
    result <- head(result, input$n)
    plot(result)
  })
}
```

We'd convert it to async like this:

```R
library(promises)
library(mirai)
daemons(6)

function(input, output, session) {
  output$plot <- renderPlot({
    mirai({ expensive_operation() }) |>
    then(\(x) head(x, input$n)) |>
    then(plot)
  })
}
```

## Adding prerequisites

The easiest part is adding `library(promises)`, `library(mirai)`, and `daemons(6)` to the top of the app.

The `promises` library is necessary for the `then()` method. You may also want to use promise utility functions like `promise_all` and `promise_race`.

The `mirai` library is needed because the `mirai()` function call is how you will launch asynchronous tasks.

`daemons(6)` is a directive to the `mirai` package, telling it where and how tasks should actually be executed. See the [article on mirai](promises_04_mirai.html) for more details.

## Identifying slow operations

To find areas of your code that are good candidates for the mirai/promise treatment, let's start with the obvious: identifying the code that is making your app slow. You may assume it's your plotting code that's slow, but it's actually your database queries; or vice versa. If there's one thing that veteran programmers can agree on, it's that human intuition is a surprisingly unreliable tool for spotting performance problems.

Our recommendation is that you use the [profvis](https://profvis.r-lib.org/) profiler, which we designed to work with Shiny (see Example 3 in the profvis documentation). You can use profvis to help you focus in on where the time is actually being spent in your app.

>  **Note:** As of this writing, profvis doesn't work particularly well for diagnosing performance problems in parts of your code that you've already made asynchronous. In particular, we haven't done any work to help it profile code that executes in a mirai, and the mechanism we use to hide "irrelevant" parts of the stack trace doesn't work well with promises. These are ripe areas for future development.

Async programming works well when you can identify just a few "hotspots" in your app where lots of time is being spent. It works much less well if your app is too slow because of a generalized, diffuse slowness through every aspect of your app, where no one operation takes too much time but it all adds up to a lot. The more mirai you need to introduce into your app, the more fixed communication overhead you incur. So for the most bang-for-the-buck, we want to launch a small number of mirai per session but move a lot of the waited-on code into each one.

## Converting a slow operation into a mirai

Now that we've found hotspots that we want to make asynchronous, let's talk about the actual work of converting them to mirai.

Conceptually, mirai works like this:

```R
mirai({
  # Expensive code goes here
}) |>
  then(\(result) {
    # Code to handle result of expensive code goes here
  })
```

which seems incredibly simple. What's actually happening is that the mirai runs in a totally separate child R process, and then the result is collected up and returned to the main R process:

```R
# Code here runs in process A
mirai({
  # Code here runs in (child) process B
}) |>
  then(\(result) {
    # Code here runs in process A
  })
```

The fact that the mirai code block executes in a separate process means we have to take special care to deal with a number of practical issues. All the objects required by the code block must be passed along as part of the `mirai()` function call; certain objects may require custom serialization to be set up, and other objects cannot be safely used across process boundaries. Again, see the [article on mirai](promises_04_mirai.html) for more details.

### Shiny-specific caveats and limitations

In addition to the constraints that mirai faces, there is an additional one for Shiny: reactive values and reactive expressions cannot be read from within a mirai. Whenever reactive values/expressions are read, side effects are carried out under the hood so that the currently executing observer or reactive expression can be notified when the reactive value/expression becomes invalidated. If a reactive value/expression is created in one process, but read in another process, there will be no way for readers to be notified about invalidation.

This code, for example, will not work:

```R
function(input, output, session) {
  r1 <- reactive({ ... })

  r2 <- reactive({
    mirai(
      {
        r1() # Will error--don't do this!
      },
      r1 = r1
    )
  })
}
```

Even though `r1()` is called from inside the `r2` reactive expression, the fact that it's also in a mirai means the call will fail. Instead, you must read any reactive values/expressions you need in advance of launching the mirai:

```R
function(input, output, session) {
  r1 <- reactive({ ... })

  r2 <- reactive({
    val <- r1()
    mirai(
      {
        val # No problem!
      },
      val = val
    )
  })
}
```

However, it's perfectly fine to read reactive values/expressions from inside a promise _handler_. Handlers run in the original process, not a child process, so reactive operations are allowed.

```R
function(input, output, session) {
  r1 <- reactive({ ... })

  r2 <- reactive({
    mirai({ ... }) |>
      then(\(x) rbind(x, r1())) # OK!
  })
}
```

## Integrating promises with Shiny

Generally, you'll be using promises with Shiny from within outputs, reactive expressions, and observers. We've tried to integrate promises into these constructs in as natural a way as possible.

### Outputs

Most outputs (`renderXXX({ ... })`) functions expect your code block to return a value; for example, `renderText()` expects a character vector and `renderTable()` expects a data frame. All such render functions that are included within the `shiny` package can now optionally be given a promise for such a value instead.

So this:

```R
output$table <- renderTable({
  read.csv(url) |>
    filter(date == input$date)
})
```

could become:

```R
output$table <- renderTable({
  mirai(read.csv(url), url = url) |>
    then(\(df) df |> filter(date == input$date))
})
```

or, trading elegance for efficiency:

```R
output$table <- renderTable({
  input_date <- input$date
  mirai(
    {
      read.csv(url) |>
        filter(date == input_date)
    },
    url = url,
    input_date = input_date
  )
})
```

The important thing to keep in mind is that the promise (or promise pipeline) must be the final expression in the code block. Shiny only knows about promises you actually return to it when you hand control back.

#### Render functions with side effects: `renderPrint` and `renderPlot`

The render functions `renderPrint()` and `renderPlot()` are slightly different than other render functions, in that they can be affected by side effects in the code block you provide. In `renderPrint` you can print to the console, and in `renderPlot` you can plot to the active R graphics device.

With promises, these render functions can work in a similar way, but with a caveat. As you hopefully understand by now, mirai executes its code in a separate R process, and printing/plotting in a separate process won't have any effect on the Shiny output in the original process. These examples, then, are incorrect:

```R
output$summary <- renderPrint({
  mirai(
    {
      read.csv(url) |>
        summary() |>
        print()
    },
    url = url
  )
})

output$plot <- renderPlot({
  mirai(
    {
      df <- read.csv(url)
      ggplot(df, aes(length, width)) + geom_point()
    },
    url = url
  )
})
```

Instead, do printing and plotting after control returns back to the original process, via a promise handler:

```R
output$summary <- renderPrint({
  mirai(read.csv(url), url = url) |>
    then(summary) |>
    then(print)
})

output$plot <- renderPlot({
  mirai(read.csv(url), url = url) |>
    then(\(df) {
      ggplot(df, aes(length, width)) + geom_point()
    })
})
```

Again, you do need to be careful to make sure that the last expression in your code block is the promise/pipeline; this is the only way the rendering logic can know whether and when your logic has completed, and if any errors occurred (so they can be displayed to the user).

### Observers

Observers are very similar to outputs: you must make sure that the last expression in your code block is the promise/pipeline. Like outputs, observers need to know whether and when they're done running, and if any errors occured (so they can log them and terminate the user session). The way to communicate this from your async user code is by returning the promise.

Here's a synchronous example that we'll convert to async. Clicking the `refresh_data` action button causes data to be downloaded, which is then saved to disk as `cached.rds` and also used to update the reactive value `data`.

```R
data <- reactiveVal(readRDS("cached.rds"))

function(input, output, session) {
  observeEvent(input$refresh_data, {
    df <- read.csv(url)
    saveRDS(df, "cached.rds")
    data(df)
  })
}
```

And the async version:

```R
data <- reactiveVal(readRDS("cached.rds"))

function(input, output, session) {
  observeEvent(input$refresh_data, {
    mirai(
      {
        df <- read.csv(url)
        saveRDS(df, "cached.rds")
        df
      },
      url = url
    ) |>
      # Assign the result to the reactive value
      then(\(df) data(df))
  })
}
```

Note that in this version, we cannot call `data(df)` inside the mirai, as this would cause the update to happen in the wrong process. Instead, we use the `then()` promise method to perform the assignment back in the main process once the mirai resolves.

### Reactive expressions

Recall that reactive expressions are used to calculate values, and are cached until they are automatically invalidated by one of their dependencies. Unlike outputs and observers, reactive expressions can be used from other reactive consumers.

Asynchronous reactive expressions are similar to regular (synchronous) reactive expressions: instead of a "normal" value, they return a promise that will yield the desired value; and a normal reactive will cache a normal value, while an async reactive will cache the promise.

The upshot is that when defining an async reactive expression, your code block should return a promise or promise pipeline, following the same rules as reactive outputs. And when calling an async reactive expression, call it like a function like you would a regular reactive expression, and treat the value that's returned like any other promise.

```R
function(input, output, session) {
  data <- eventReactive(input$refresh_data, {
    read.csv(url)
  })

  filteredData <- reactive({
    data() |> filter(date == input$date)
  })

  output$table <- renderTable({
    filteredData() |> head(5)
  })
}
```

And now in async:

```R
function(input, output, session) {
  data <- eventReactive(input$refresh_data, {
    mirai(read.csv(url), url = url)
  })

  filteredData <- reactive({
    data() |> then(\(df) df |> filter(date == input$date))
  })

  output$table <- renderTable({
    filteredData() |> then(\(df) head(df, 5))
  })
}
```

## The flush cycle

In the past, Shiny's reactive programming model has operated using a mostly traditional [event loop](https://en.wikipedia.org/wiki/Event_loop) model. Somewhere many levels beneath `shiny::runApp()` was a piece of code that looked a bit like this:

```R
while (TRUE) {
  # Do nothing until a browser sends some data
  input <- receiveInputFromBrowser()
  # Use the received data to update reactive inputs
  session$updateInput(input)
  # Execute all invalidated outputs/observers
  flushReact()
  # After ALL outputs execute, send the results back
  flushOutputs()
}
```

We call this Shiny's "flush cycle". There are two important properties to our flush cycle.

1. Only one of the four steps—receiving, updating, reacting, and sending—can be executing a time. (Remember, R is single threaded.) In particular, it's not possible for inputs to be updated while outputs/observers are running. This is important in order to avoid race conditions that would be all but impossible to defend against.
2. Many outputs may change as a result of a single input value received from the browser, but none of them are sent back to the client until all of the outputs are ready. The advantage of this is a smoother experience for the end-user in most cases. (Admittedly, there has been some controversy regarding this property of Shiny; some app authors would strongly prefer to show outputs as soon as they are ready, or at least to have manual control over this behavior.)

While adding async support to Shiny, we aimed to keep these two properties intact. Imagine now that `flushReact()`, the line that executes invalidated outputs/observers, returns a promise that combines all of the async outputs/observers (i.e. a promise that resolves only after all of the async outputs/observers have resolved). The new, async-aware event loop is conceptually more like this:

```R
doEventLoop <- function() {
  # Do nothing until a browser sends some data
  input <- receiveInputFromBrowser()
  # Use the received data to update reactive inputs
  session$updateInput(input)
  # Execute all invalidated outputs/observers
  flushReact() |>
    finally(\() {
      # After ALL outputs execute, send the results back
      flushOutputs()
      # Continue the event loop
      doEventLoop()
    })
}
```

The resulting behavior matches the synchronous version of the event loop, in that:

1. No inputs are received from the browser until all pending async outputs/observers have completed. Unlike the synchronous version, this separation is enforced at the session level: if Session A has some async observers that have not finished executing, that only prevents Session A from processing new input values, while new input values from Session B can be handled immediately because they belong to a different session. Again, the goal of keeping input updates separate from output/observer execution is to prevent race conditions, which are even more pernicious to debug and understand when async code is involved.
2. For a given session, no outputs are sent back to the client, until all outputs are ready. It doesn't matter whether the outputs in question are synchronous, asynchronous, or some combination; they all must complete execution before any can be sent.

---
title: "Launching tasks with future"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Launching tasks with future}
  %\VignetteEncoding{UTF-8}
---

```{css echo=FALSE}
.alert-secondary a, .alert-secondary a:visited {
  color: inherit;
  text-decoration: underline;
}
.alert code {
  color: inherit;
  background-color: inherit;
}
```

The `future` package provides a lightweight way to launch R tasks that don't block the current R session. It was created by Henrik Bengtsson long before the `promises` package existed—the first CRAN release of `future` predates development of `promises` by almost two years!

The `promises` package provides the API for working with the results of async tasks, but it totally abdicates responsibility for actually launching/creating async tasks. The idea is that any number of different packages could be capable of launching async tasks, using whatever techniques they want, but all of them would either return promise objects or objects that can be converted to promise objects, as is the case for `future`.

This document will give an introduction to the parts of `future` that are most relevant to promises. For more information, please consult the vignettes that come with `future`, especially the [Comprehensive Overview](https://CRAN.R-project.org/package=future/vignettes/future-1-overview.html).

## How future works

The main API that `future` provides couldn't be simpler. You call `future()` and pass it the code that you want executed asynchronously:

```R
f <- future({
  # expensive operations go here...
  df <- download_lots_of_data()
  fit_model(df)
})
```

The object that's returned is a future, which for all intents and purposes is a promise object[^1], which will eventually resolve to the return value of the code block (i.e. the last expression) or an error if the code does not complete executing successfully. The important thing is that no matter how long the expensive operation takes, these lines will execute almost instantly, while the operation continues in the background.

[^1]: (The `future` package provides several functions for working with future objects, but they are not relevant for our purposes.)

But we know that R is single-threaded, so how does `future` accomplish this? The answer: by utilizing another R process. `future` delegates the execution of the expensive operation to a totally different R process, so that the original R process can move on.

## Choosing a launch method

There are several different methods we could use for launching R processes or attaching to existing R processes, and each method has its own advantages, disadvantages, limitations, and requirements. Rather than prescribing a single method, the `future` package provides an extensible mechanism that lets you, the R user, decide what method to use. Call the `plan()` function with one of the following values (without quotes—these are function names, not strings):

* `multisession`: Launches up to *n* background R processes on the same machine (where *n* is the number of processor cores on the system, minus 1). These background processes will be used/recycled for the life of the originating R process. If a future is launched while all the background R processes are busy executing, then the new future will be queued until one of the background processes free up.
* `multicore`: Each new task executes in its own forked child process. Forking is generally much faster than launching a new process from scratch, and most of the state of the original process is available to the child process without having to go through any extra effort (see the section about Globals below). The biggest limitation of forking is that it doesn't work at all on Windows operating systems, which is what the majority of R users use. There are also some dangerous edge cases with this style of execution (Google "fork without exec" for more information), though popular frameworks like RServe and OpenCPU rely heavily on this and don't seem to suffer for it.

The `future` package also includes a `sequential` method, which executes synchronously and is therefore not relevant for our purposes. Unfortunately, `sequential` is the default, hence explicitly calling `plan()` with a different method is a must.

There is also a `cluster` method, as well as a separate `future.batchtools` package, for doing distributed execution; those may work with promises, but have not been tested by our team and are not described further in this document.

To learn more, see the [`future::plan()` reference docs](https://future.futureverse.org/reference/plan.html) as well as the [`future` overview](https://future.futureverse.org/articles/future-1-overview.html#controlling-how-futures-are-resolved).

## Caveats and limitations

The abstractions that `future` presents are simple, but [leaky](https://en.wikipedia.org/wiki/Leaky_abstraction). You can't make effective use of `future` without understanding its various strategies for running R tasks asynchronously. Please read this entire section carefully before proceeding.

### Globals: Providing input to future code chunks

Most future code chunks will need to reference data from the original process, e.g. data to be fitted, URLs to be requested, file paths to read from. The future package goes to some lengths to try to make this process seamless for you, by inspecting your code chunk and predicting which variables from the original process should be copied to the child process. In our testing this works fairly reliably with multicore, somewhat less reliably with multisession.

Multisession also has the distinct disadvantage that any identified variables must be physically (though automatically) copied between the main and child processes, which can be extremely time-consuming if the data is large. (The multicore strategy does not need to do this, because every forked process starts out with its memory in the same state as its parent at the time of the fork.)

In summary, it's possible for both false positives (data copied that doesn't need to be) and false negatives (data not available when it's needed) to occur. Therefore, for all but the simplest cases, we suggest suppressing future's automated variable copying and instead manually specifying the relevant variables, using the `future()` function's `globals` parameter. You can pass it a character vector (`globals = c("var_a", "var_b")`) or a named list (`globals = c(data = mtcars, iterations = n)`).

One final note about globals: as a safety measure, `future()` will error if the size of the data to be shuttled between the processes exceeds 500MB. This is true whether the variables to copy were identified by automatic detection, or explicitly via the `globals` parameter; and it's even true if you're using the multicore strategy, where no copies are actually made. If your data is potentially large, you'll want to increase the limit by setting the `future.globals.maxSize` option to a suitably high number of bytes, e.g. `options(future.globals.maxSize=1e9)` for a billion bytes.

### Package loading

Besides variables, `future()` also tries to automatically infer what R packages need to be loaded in the child process. If the automatic detection is not sufficient, you can use the `future()` function's `packages` parameter to pass in a character vector of package names, e.g. `packages = c("dplyr", "ggplot2")`.

Again, this is especially important for multisession, because multicore will inherit all of the attached packages of the parent process.

### Native resources

Future code blocks cannot use resources such as database connections and network sockets that were created in the parent process. This is true regardless of what future implementation you use! Even if it seems to work with a simple test, you are asking for crashes or worse by sharing these kinds of resources across processes.

Instead, make sure you create, use, and destroy such resources entirely within the scope of the future code block.

### Mutation

Reference class objects (including R6 objects and data.table objects) and environments are among the few "native" R object types that are mutable, that is, can be modified in-place. Unless they contain native resources (see previous section), there's nothing wrong with using mutable objects from within future code blocks, even objects created in the parent process. However, note that any changes you make to these objects will not be visible from the parent process; the future code is operating on a copy of the object, not the original.

### Returning values

Future code blocks can return a value—they'd be a lot less useful if they couldn't! Like everywhere else in R, the return value is determined by the last expression in the code block, unless `return()` is explicitly called earlier.

Regardless of future method, the return value will be copied back into the parent process. This matters for two reasons.

First, if the return value is very large, the copying process can take some time—and because the data must essentially be serialized to and deserialized from rds format, it can take a surprising amount of time. In the case of future blocks that execute fairly quickly but return huge amounts of data, you may be better off not using future/async techniques at all.

Second, objects that refer to native resources are unlikely to work in this direction either; just as you can't use the parent's database connections in the child process, you also cannot have the child process return a database connection for the parent to use.

<div style="font-size: 20px; margin-top: 40px; text-align: right;">

Next:

* [Advanced `future` and `promises` usage](promises_05b_future_promise.html)
* [Using `promises` with Shiny](promises_06_shiny.html)

</div>

---
title: "An informal introduction to async programming"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{An informal introduction to async programming}
  %\VignetteEncoding{UTF-8}
---

Hello, R and/or Shiny user! Let’s talk about async programming!

**Async programming? Sounds complicated.**

It is, very! You may want to grab some coffee.

**Ugh. Tell me why I even need to know this?**

Async programming is a major new addition to Shiny that can make certain classes
of apps dramatically more responsive under load.

Because R is single threaded (i.e. it can only do one thing at a time), a given
Shiny app process can also only do one thing at a time: if it is fitting a
linear model for one client, it can’t simultaneously serve up a CSV download for
another client.

For many Shiny apps, this isn’t a big problem; if no one processing step takes
very long, then no client has to wait an undue amount of time before they start
seeing results. But for apps that perform long-running operations — either
expensive computations that take a while to complete, or waiting on slow network
operations like database or web API queries — your users’ experience can suffer
dramatically as traffic ramps up. Operations that normally are lightning quick,
like downloading a small JavaScript file, can get stuck in traffic behind
something slow.

**Oh, OK—more responsiveness is always good. But you said this’ll only help for
certain classes of Shiny apps?**

It’s mostly helpful for apps that have a few specific operations that take a
long time, rather than lots of little operations that are all a bit slow on
their own and add up to one big slow mess. We’re looking for watermelons, not
blueberries.

**Watermelons… sure. So then, how does this all work?**

It all starts with *async functions*. An async function is one that performs an
operation that takes a long time, yet returns control to you immediately.
Whereas a normal function like `read.csv` will not return until its work is done
and it has the value you requested, an asynchronous `read.csv.async` function
would kick off the CSV reading operation, but then return immediately, long
before the real work has actually completed.

```r
library(mirai)

read.csv.async <- \(file, header = TRUE, stringsAsFactors = FALSE) {
  mirai(
    read.csv(file, header = header, stringsAsFactors = stringsAsFactors),
    .args = environment()
  )
}
```

(Don't worry about what this definition means for now. You'll learn more about defining async functions in [Launching tasks](promises_04_mirai.html).)

**So instead of “read this CSV file” it’s more like “begin reading this CSV
file”?**

Yes! That’s what async functions do: they start things, and give you back a
special object called a *promise*. If it doesn’t return a promise, it’s not an
async function.

**Oh, I’ve heard of promises in R! From [the NSE
chapter](http://adv-r.had.co.nz/Computing-on-the-language.html) in Hadley’s
Advanced R book!**

Ah... this is awkward, but no. I’m using the word “promise”, but I’m not referring
to *that* kind of promise. For the purposes of async programming, try to forget
that you’ve ever heard of that kind of promise, OK?

I know it seems needlessly confusing, but the promises we’re talking about here
are ~~shamelessly copied from~~ directly inspired by a central abstraction in modern
JavaScript, and the JS folks named them “promises”.

**Fine, whatever. So what are these promises?**

Conceptually, they’re a stand-in for the *eventual result* of the operation. For
example, in the case of our `read.csv.async` function, the
promise is a stand-in for a data frame. At some point, the operation is going to
finish, and a data frame is going to become available. The promise gives us a
way to get at that value.

**Let me guess: it’s an object that has `has_completed()` and
`get_value()` methods?**

Good guess, but no. Promises are *not* a way to directly inquire about the
status of an operation, nor to directly retrieve the result value. That is
probably the simplest and most obvious way to build an async framework, but in
practice it’s very difficult to build deeply async programs with an API like
that.

Instead, a promise lets you *chain together operations* that should be performed
whenever the operation completes. These operations might have side effects (like
plotting, or writing to disk, or printing to the console) or they might
transform the result values somehow.

**Chain together operations? Using the `|>` operator?**

A lot like that! You can’t use the `|>` operator itself as we need to wrap execution of the promise step within a function that may be called with the incoming promise resolves. So whereas
you might do this to a regular data frame:

```r
library(dplyr)
read.csv("https://rstudio.github.io/promises/data.csv") |>
  filter(state == "NY") |>
  View()
```

The async version would look like:

```r
library(dplyr)
read.csv.async("https://rstudio.github.io/promises/data.csv") |>
  then(\(df) df |> filter(state == "NY")) |>
  then(\(df) View(df))
```

The `then()` call here is the secret sauce. It creates a promise pipeline, chaining the provided callback by having it run on the result of the incoming promise value.

**Why can't I just use `|>` directly?**

Not all values are promises. So while the `|>` operator can be used to
chain together operations on an object, it does not wrap and unwrap promise objects.
Instead, the `then()` function is designed to work with promises, and it will
unwrap the promise for you, passing the result value to the function you provide.

`then()` *must* take a promise
as input; that is, the left-hand side of the `|>` operator must be an expression that
yields a promise. `then()` will _then_ pass the unwrapped, plain old, not-a-promise value
to the function—this is critically important. It means we can use promise
objects with non-promise-aware functions, with `then()` serving as the bridge
between asynchronous and synchronous code.

However, that work often can’t be done in the present, since the whole point of
a promise is that it represents work that hasn’t completed yet. So `then()` does
the work of extracting and piping not at the time that it’s called, but rather,
sometime in the future.

**You lost me.**

OK, let’s slow down and take this step by step. We’ll generate a promise by
calling an async function:

```r
df_promise <- read.csv.async("https://rstudio.github.io/promises/data.csv")
```

Even if `data.csv` is many gigabytes, `read.csv.async` returns immediately with
a new promise. We store it as `df_promise`. Eventually, when the CSV reading
operation successfully completes, the promise will contain a data frame, but for
now it’s just an empty placeholder.

One thing we definitely *can’t* do is treat `df_promise` as if it’s simply a
data frame:

```r
# Doesn't work!
dplyr::filter(df_promise, state == "NY")
```

Try this and you’ll get an error like `no applicable method for 'filter_'
applied to an object of class "promise"`. And the native pipe won’t help you either;
`df_promise |> filter(state == "NY")` will give you the same error.

**Right, that makes sense. `filter` is designed to work on data frames, and
`df_promise` isn’t a data frame.**

Exactly. Now let’s try something that actually works:

```r
df_promise |>
  then(\(df) df |> filter(state == "NY"))
```

At the moment it’s called, this code won’t appear to do much of anything,
really. But whenever the `df_promise` operation actually completes successfully,
then the result of that operation—the plain old data frame—will be passed to
`filter(df, state = "NY")`.

**OK, so that’s good. I see what you mean about `then()` letting you use
non-promise functions with promises. But the whole point of using the
`filter` function is to get a data frame back. If `filter` isn’t even
going to be called until some random time in the future, how do we get its value
back?**

I’ll tell you the answer, but it’s not going to be satisfying at first.

When you use a regular `|>`, the result you get back is the return value from
the right-hand side:

```r
df_filtered <- df |> filter(state == "NY")
```

When you use `then()`, the result you get back is a promise, whose *eventual*
result will be the return value from the right-hand side:

```r
df_filtered_promise <- df_promise |> then(\(df) df |> filter(state == "NY"))
```

**Wait, what? If I have a promise, I can do stuff to it using `then()`, but
then I just end up with another promise? Why not just have `then()` return a
regular value instead of a promise?**

Remember, the whole point of a promise is that we don’t know its value yet! So
to write a function that uses a promise as input and returns some non-promise
value as output, you’d need to either be a time traveler or an oracle.

To summarize, once you start working with a promise, any calculations and
actions that are “downstream” of that promise will need to become
promise-oriented. Generally, this means once you have a promise, you need to use
`then()` and keep using it until your pipeline terminates.

**I guess that makes sense. Still, if the only thing you can do with promises is
make more promises, that limits their usefulness, doesn’t it?**

It’s a different way of thinking about things, to be sure, but it turns out
there’s not much limit in usefulness—especially in the context of a Shiny app.

First, you can use promises with Shiny outputs. If you’re using an
async-compatible version of Shiny (version >=1.1), all of the
built-in `renderXXX` functions can deal with either regular values or promises.
An example of the latter:

```r
output$table <- renderTable({
  read.csv.async("https://rstudio.github.io/promises/data.csv") |>
    then(\(df) df |> filter(state == "NY"))
})
```

When `output$table` executes the `renderTable` code block, it will notice that
the result is a promise, and wait for it to complete before continuing with the
table rendering. While it’s waiting, the R process can move on to do other
things.

Second, you can use promises with reactive expressions. Reactive expressions
treat promises about the same as they treat other values, actually. But this
works perfectly fine:

```r
# A reactive expression that returns a promise
filtered_df <- reactive({
  read.csv.async("https://rstudio.github.io/promises/data.csv") |>
    then(\(df) df |> filter(state == "NY")) |>
    then(\(df) df |> arrange(median_income))
})

# A reactive expression that reads the previous
# (promise-returning) reactive, and returns a
# new promise
top_n_by_income <- reactive({
  filtered_df() |>
    then(\(df) df |> head(input$n))
})

output$table <- renderTable({
  top_n_by_income()
})
```

Third, you can use promises in reactive observers. Use them to perform
asynchronous tasks in response to reactivity.

```r
observeEvent(input$save, {
  filtered_df() |>
    then(\(df) write.csv(df, "ny_data.csv"))
})
```

**Alright, I think I see what you mean. You can’t escape from promise-land, but
there’s no need to, because Shiny knows what to do with them.**

Yes, that’s basically right. You just need to keep track of which functions and
reactive expressions return promises instead of regular values, and be sure to
interact with them using `then()` or other promise-aware functions.

**Wait, there are other promise-aware functions?**

Yes. The `then()` method is the one you’ll most commonly use, but there is a variant
`then(tee = TRUE)`, which creates a *promise tee* object (it’s analogous to the
magrittr `%T>%` operator). `then(tee = TRUE)` method acts like `then()`, but
instead of returning a promise for the result value, it returns the original
value instead. Meaning `p |> then(tee = TRUE, \(df) cat(df, "\n"))` won’t return a promise for the
return value of `cat()` (which is always `NULL`) but instead the value of `p`.
This is useful for logging, or other “side effecty” operations.

There’s also `catch()` and its tee version, `catch(tee = TRUE)`, which are used for error
handling. I won’t confuse you with more about that now, but you can read more
[here](promises_03_overview.html#error-handling).

The `promises` package is where all of these operators live, and it also comes
with some additional functions for working with promises.

So far, the only actual async function we’ve talked about has been
`read.csv.async`, which doesn’t actually exist. To learn where actual async
functions come from, read this guide to the [`mirai` package](promises_04_mirai.html).

There are the lower-level functions `then`, `catch`, and `finally`, which are
the non-pipe, non-operator equivalents of the promise operators we’ve been
discussing. See [reference](promises_03_overview.html#accessing-results-with-then).

And finally, there are `promise_all`, `promise_race`, and `promise_map`, used to combine
multiple promises into a single promise. Learn more about them [here](../reference/promise_all.html).

**OK, looks like I have a lot of stuff to read up on. And I’ll probably have to
reread this conversation a few times before it fully sinks in.**

Sorry. I told you it was complicated. If you make it through the rest of the guide, you’ll be 95% of the way there.

<div style="font-size: 20px; margin-top: 40px; text-align: right;">
Next: [Working with promises](promises_03_overview.html)
</div>

---
title: "Launching tasks with `future`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{4. Launching tasks with future}
  %\VignetteEncoding{UTF-8}
---

The `future` package provides a lightweight way to launch R tasks that don't block the current R session. It was created by Henrik Bengtsson long before the `promises` package existed—the first CRAN release of `future` predates development of `promises` by almost two years.

The `promises` package provides the API for working with the results of async tasks, but it totally abdicates responsibility for actually launching/creating async tasks. The idea is that any number of different packages could be capable of launching async tasks, using whatever techniques they want, but all of them would either return promise objects (or objects that can be converted to promise objects, as is the case for `future`). However, for now, `future` is likely to be the primary or even exclusive way that async tasks are created.

This document will give an introduction to the parts of `future` that are most relevant to promises. For more information, please consult the vignettes that come with `future`, especially the [Comprehensive Overview](https://CRAN.R-project.org/package=future/vignettes/future-1-overview.html).

## How future works

The main API that `future` provides couldn't be simpler. You call `future()` and pass it the code that you want executed asynchronously:

```R
f <- future({
  # expensive operations go here...
  df <- download_lots_of_data()
  fit_model(df)
})
```

The object that's returned is a future, which for all intents and purposes is a promise object[^1], which will eventually resolve to the return value of the code block (i.e. the last expression) or an error if the code does not complete executing successfully. The important thing is that no matter how long the expensive operation takes, these lines will execute almost instantly, while the operation continues in the background.

[^1]: (The `future` package provides several functions for working with future objects, but they are not relevant for our purposes.)

But we know that R is single-threaded, so how does `future` accomplish this? The answer: by utilizing another R process. `future` delegates the execution of the expensive operation to a totally different R process, so that the original R process can move on.

## Choosing a launch method

There are several different methods we could use for launching R processes or attaching to existing R processes, and each method has its own advantages, disadvantages, limitations, and requirements. Rather than prescribing a single method, the `future` package provides an extensible mechanism that lets you, the R user, decide what method to use. Call the `plan()` function with one of the following values (without quotes—these are function names, not strings):

* `multisession`: Launches up to *n* background R processes on the same machine (where *n* is the number of processor cores on the system, minus 1). These background processes will be used/recycled for the life of the originating R process. If a future is launched while all the background R processes are busy executing, then the new future will be queued until one of the background processes free up.
* `multicore`: Each new task executes in its own forked child process. Forking is generally much faster than launching a new process from scratch, and most of the state of the original process is available to the child process without having to go through any extra effort (see the section about Globals below). The biggest limitation of forking is that it doesn't work at all on Windows operating systems, which is what the majority of R users use. There are also some dangerous edge cases with this style of execution (Google "fork without exec" for more information), though popular frameworks like RServe and OpenCPU rely heavily on this and don't seem to suffer for it.
* `multiprocess`: This is a meta-method that uses `multicore` on platforms where it's available, and `multisession` on Windows. In theory this means that you get the best performance available for each platform. In practice, since the two implementations are so different, it's very important that you test your app on all of the platforms you intend your app to be used on (or, if your development machine is non-Windows, test with both `multicore` (or `multiprocess`) and `multisession`).

The `future` package includes several other methods. Two of them, `sequential` and `transparent`, execute synchronously and are therefore not relevant for our purposes. There are also `cluster` and `remote` methods, as well as a separate `future.batchtools` package, for doing distributed execution; those may work with promises, but have not been tested by our team and are not described further in this document.

## Caveats and limitations

The abstractions that `future` presents are simple, but [leaky](https://en.wikipedia.org/wiki/Leaky_abstraction). You can't make effective use of `future` without understanding its various strategies for running R tasks asynchronously. Please read this entire section carefully before proceeding.

### Globals: Providing input to future code chunks

Most future code chunks will need to reference data from the original process, e.g. data to be fitted, URLs to be requested, file paths to read from. The future package goes to some lengths to try to make this process seamless for you, by inspecting your code chunk and predicting which variables from the original process should be copied to the child process. In our testing this works fairly reliably with multicore, somewhat less reliably with multisession.

Multisession also has the distinct disadvantage that any identified variables must be physically (though automatically) copied between the main and child processes, which can be extremely time-consuming if the data is large. (The multicore strategy does not need to do this, because every forked process starts out with its memory in the same state as its parent at the time of the fork.)

In summary, it's possible for both false positives (data copied that doesn't need to be) and false negatives (data not available when it's needed) to occur. Therefore, for all but the simplest cases, we suggest suppressing future's automated variable copying and instead manually specifying the relevant variables, using the `future()` function's `globals` parameter. You can pass it a character vector (`globals = c("var_a", "var_b")`) or a named list (`globals = c(data = mtcars, iterations = n)`).

One final note about globals: as a safety measure, `future()` will error if the size of the data to be shuttled between the processes exceeds 500MB. This is true whether the variables to copy were identified by automatic detection, or explicitly via the `globals` parameter; and it's even true if you're using the multicore strategy, where no copies are actually made. If your data is potentially large, you'll want to increase the limit by setting the `future.globals.maxSize` option to a suitably high number of bytes, e.g. `options(future.globals.maxSize=1e9)` for a billion bytes.

### Package loading

Besides variables, `future()` also tries to automatically infer what R packages need to be loaded in the child process. If the automatic detection is not sufficient, you can use the `future()` function's `packages` parameter to pass in a character vector of package names, e.g. `packages = c("dplyr", "ggplot2")`.

Again, this is especially important for multisession, because multicore will inherit all of the attached packages of the parent process.

### Native resources

Future code blocks cannot use resources such as database connections and network sockets that were created in the parent process. This is true regardless of what future implementation you use! Even if it seems to work with a simple test, you are asking for crashes or worse by sharing these kinds of resources across processes.

Instead, make sure you create, use, and destroy such resources entirely within the scope of the future code block.

### Mutation

Reference class objects (including R6 objects and data.table objects) and environments are among the few "native" R object types that are mutable, that is, can be modified in-place. Unless they contain native resources (see previous section), there's nothing wrong with using mutable objects from within future code blocks, even objects created in the parent process. However, note that any changes you make to these objects will not be visible from the parent process; the future code is operating on a copy of the object, not the original.

### Returning values

Future code blocks can return a value—they'd be a lot less useful if they couldn't! Like everywhere else in R, the return value is determined by the last expression in the code block, unless `return()` is explicitly called earlier.

Regardless of future method, the return value will be copied back into the parent process. This matters for two reasons.

First, if the return value is very large, the copying process can take some time—and because the data must essentially be serialized to and deserialized from rds format, it can take a surprising amount of time. In the case of future blocks that execute fairly quickly but return huge amounts of data, you may be better off not using future/async techniques at all.

Second, objects that refer to native resources are unlikely to work in this direction either; just as you can't use the parent's database connections in the child process, you also cannot have the child process return a database connection for the parent to use.

<div style="display:flex;">
  <div style="font-size: 20px; margin-top: 40px; margin-left: auto;">

Next:

* [Advanced `future` and `promises` usage](future_promise.html)
* [Using `promises` with Shiny](shiny.html)

  </div>
</div>

---
title: "Combining `promises`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{7. Combining promises}
  %\VignetteEncoding{UTF-8}
---

So far, all of our examples have involved chaining operations onto a single promise. In practice, you'll often find yourself needing to perform tasks that require the results of more than one promise. These are some patterns you may find useful:

* [**Gathering:**](#gathering) Combining multiple independent promises into a single computation
* [**Nesting:**](#nesting) Using the result of one promise to affect the input or execution of another async operation
* [**Racing:**](#racing) Using the fastest of multiple promises
* [**Mapping:**](#mapping) Applying an async function to each of a list's elements and collecting the results
* [**Reducing:**](#reducing) Applying an async function to each of a list's elements and reducing

## Gathering

The most common pattern for combining promises is gathering, where you have two or more promises in hand and you want to use all of their results in a computation. The `promise_all` function is designed for this. Its signature looks like this:

```r
promise_all(..., .list = NULL)
```

`promise_all` takes any number of promises as named arguments, and returns a promise of a list containing named elements with the results of those promises.

Here's an example using `promise_all` to combine the results of two async `read.csv` operations:

```r
library(promises)
library(future)
plan(multisession)

a <- future_promise(read.csv("a.csv"))
b <- future_promise(read.csv("b.csv"))

result <- promise_all(a = a, b = b) %...>% {
  rbind(.$a, .$b)
}
```

In this example, the value of `.` within the curly braces is a list whose elements `a` and `b` are both data frames. We use `rbind` to combine them.

The `.$` prefix is a bit inelegant, so we recommend the use of the base R function `with`, which lets you skip the prefix. Here's the same example, with `with`:

```r
library(promises)
library(future)
plan(multisession)

a <- future_promise(read.csv("a.csv"))
b <- future_promise(read.csv("b.csv"))

promise_all(a = a, b = b) %...>%
  with({
    rbind(a, b)
  })
```

(Note that since the `promise_all` argument names are the same as the variable names (`a = a`, `b = b`), the original variables are masked: inside the `with` block, `a` now refers to the *result* of the promise `a`, not the promise object itself. If you find this confusing, you can just choose a different argument name, like `promise_all(a_result = a, …)`.)

The combination of `promise_all` and `with` is a concise and powerful way to gather the results of multiple promises.

`promise_all` also gives you two other options for passing input promises. First, if you would rather your result list be unnamed, you can pass in promises as unnamed arguments: `promise_all(a, b)` would yield `list(1, 2)`. Second, if you have a list of promises already in hand, you can pass the list as a single argument using `promise_all(.list = x)` (instead of, say, using `do.call(promise_all, x)`).

## Nesting

Gathering is easy and convenient, but sometimes not flexible enough. For example, if you use the result of promise `a` to decide whether to launch a second async task, whose result you then use in combination with the result of `a`.

```r
library(promises)
library(future)
plan(multisession)

a <- future_promise(1)

a %...>% (function(a) {
  b <- future_promise(2)
  b %...>% (function(b) {
    a + b
  })
})
```

(We use anonymous functions here to mask the names of the original promises--i.e. once inside the first anonymous function, the symbol `a` now refers to the result of the promise `a`.)

The nesting pattern is effective and flexible. The main downside is the physical nesting of the source code; if you use this pattern to a depth of more than a couple of promises, your code will be quite indented (in programming jargon this is referred to as the "pyramid of doom").

## Racing

```r
library(promises)
library(future)
plan(multisession)

a <- future_promise({ Sys.sleep(1); 1 })
b <- future_promise({ Sys.sleep(0.5); 2 })

first <- promise_race(a, b)
```

`promise_race` takes multiple promises and returns a new promise that will be fulfilled with the first promise that succeeds. In the example above, `first` is a promise that will be fulfilled with `2` after 0.5 seconds.

If one of the input promises rejects before any succeed, then the returned promise will be rejected.

Note that promises does not currently support cancellation. So losing promises will attempt to run to completion even after the race ends.

## Mapping

Use `promise_map` to run an async operation on each element of a list or vector, and collect the results in a list. It's very similar to `lapply` or `purrr::map`, except that the function to apply can return a promise, and the return value is also a promise.

In the example below, we iterate over a named vector of package names. For each package name, we launch an async task to download the package's description file from CRAN pick out the last published date.

```r
library(promises)
library(future)
plan(multisession)

get_pub_date <- function(pkg) {
  desc_url <- paste0("https://cran.r-project.org/web/packages/", pkg, "/DESCRIPTION")
  future_promise({
    read.dcf(url(desc_url))[, "Date/Publication"] %>% unname()
  })
}

packages <- setNames(, c("ggplot2", "dplyr", "knitr"))

pkg_dates <- promise_map(packages, get_pub_date)

pkg_dates %...>% print()
```

The resulting output looks like this:

```
$ggplot2
[1] "2016-12-30 22:45:17"

$dplyr
[1] "2017-09-28 20:43:29 UTC"

$knitr
[1] "2018-01-29 11:01:22 UTC"
```

`promise_map` works serially; each time it calls the given function on an element of the list/vector, it will wait for the returned promise to resolve before proceeding to the next element. Furthermore, any error or rejected promise will cause the entire `promise_map` operation to reject.

If you want behavior that's similar to `promise_map` but for all the async operations to occur in parallel, you can achieve that with a combination of a regular `purrr::map` and `promise_all`:

```r
pkg_dates <- purrr::map(packages, get_pub_date) %>%
  promise_all(.list = .)

pkg_dates %...>% print()
```

## Reducing

Use `promise_reduce` when you have a list where you want to run an async operation on each of the elements, and to do so serially (i.e. only one async operation runs at a time). This can be helpful when you're searching through some elements using an async operation and want to terminate early when your search succeeds.

The signature of `promise_reduce` is as follows:

```r
promise_reduce(x, func, init = NULL)
```

If you've worked with `base::Reduce()` or `purr:::reduce()`, this should seem reasonably familiar: `x` is a vector or list; `func` is a function that takes two arguments, the accumulated value and the "next" value; and `init` is the default accumulated value.

The main difference between `promise_reduce` and `purrr:::reduce` is that with `promise_reduce`, your `func` can return a promise. If it does, `promise_reduce` will wait for it to resolve before updating the accumulated value and invoking `func` on the next element. The result returned from `promise_reduce` is a promise that resolves to the ultimate accumulated value.

The following example loops through a partial list of CRAN mirrors, returning the first one that passes whatever check `http::http_error` performs.

```r
library(promises)
library(future)
plan(multisession)

cran_mirrors <- c(
  "https://cloud.r-project.org",
  "https://cran.usthb.dz",
  "https://cran.csiro.au",
  "https://cran.wu.ac.at"
)

promise_reduce(cran_mirrors, function(result, mirror) {
  if (!is.null(result)) {
    result
  } else {
    future_promise({
      # Test the URL; return the URL on success, or NULL on failure
      if (!httr::http_error(mirror)) mirror
    })
  }
}, .init = NULL) %...>% print()
```

---
title: "An informal introduction to async programming"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{2. An informal introduction to async programming}
  %\VignetteEncoding{UTF-8}
---

Hello, R and/or Shiny user! Let’s talk about async programming!

**Async programming? Sounds complicated.**

It is, very! You may want to grab some coffee.

**Ugh. Tell me why I even need to know this?**

Async programming is a major new addition to Shiny that can make certain classes
of apps dramatically more responsive under load.

Because R is single threaded (i.e. it can only do one thing at a time), a given
Shiny app process can also only do one thing at a time: if it is fitting a
linear model for one client, it can’t simultaneously serve up a CSV download for
another client.

For many Shiny apps, this isn’t a big problem; if no one processing step takes
very long, then no client has to wait an undue amount of time before they start
seeing results. But for apps that perform long-running operations — either
expensive computations that take a while to complete, or waiting on slow network
operations like database or web API queries — your users’ experience can suffer
dramatically as traffic ramps up. Operations that normally are lightning quick,
like downloading a small JavaScript file, can get stuck in traffic behind
something slow.

**Oh, OK—more responsiveness is always good. But you said this’ll only help for
certain classes of Shiny apps?**

It’s mostly helpful for apps that have a few specific operations that take a
long time, rather than lots of little operations that are all a bit slow on
their own and add up to one big slow mess. We’re looking for watermelons, not
blueberries.

**Watermelons… sure. So then, how does this all work?**

It all starts with *async functions*. An async function is one that performs an
operation that takes a long time, yet returns control to you immediately.
Whereas a normal function like `read.csv` will not return until its work is done
and it has the value you requested, an asynchronous `read.csv.async` function
would kick off the CSV reading operation, but then return immediately, long
before the real work has actually completed.

```r
library(future)
plan(multisession)

read.csv.async <- function(file, header = TRUE, stringsAsFactors = FALSE) {
  future_promise({
    read.csv(file, header = header, stringsAsFactors = stringsAsFactors)
  })
}
```

(Don't worry about what this definition means for now. You'll learn more about defining async functions in [Launching tasks](futures.html) and [Advanced `future` and `promises` usage](future_promise.html).)

**So instead of “read this CSV file” it’s more like “begin reading this CSV
file”?**

Yes! That’s what async functions do: they start things, and give you back a
special object called a *promise*. If it doesn’t return a promise, it’s not an
async function.

**Oh, I’ve heard of promises in R! From [the NSE
chapter](http://adv-r.had.co.nz/Computing-on-the-language.html) in Hadley’s
Advanced R book!**

Ah... this is awkward, but no. I’m using the word “promise”, but I’m not referring
to *that* kind of promise. For the purposes of async programming, try to forget
that you’ve ever heard of that kind of promise, OK?

I know it seems needlessly confusing, but the promises we’re talking about here
are ~~shamelessly copied from~~ directly inspired by a central abstraction in modern
JavaScript, and the JS folks named them “promises”.

**Fine, whatever. So what are these promises?**

Conceptually, they’re a stand-in for the *eventual result* of the operation. For
example, in the case of our `read.csv.async` function, the
promise is a stand-in for a data frame. At some point, the operation is going to
finish, and a data frame is going to become available. The promise gives us a
way to get at that value.

**Let me guess: it’s an object that has `has_completed()` and
`get_value()` methods?**

Good guess, but no. Promises are *not* a way to directly inquire about the
status of an operation, nor to directly retrieve the result value. That is
probably the simplest and most obvious way to build an async framework, but in
practice it’s very difficult to build deeply async programs with an API like
that.

Instead, a promise lets you *chain together operations* that should be performed
whenever the operation completes. These operations might have side effects (like
plotting, or writing to disk, or printing to the console) or they might
transform the result values somehow.

**Chain together operations? Using the `%>%` operator?**

A lot like that! You can’t use the `%>%` operator itself, but we provide a
promise-compatible version of it: `%...>%`. So whereas you might do this to a
regular data frame:

```r
library(dplyr)
read.csv("https://rstudio.github.io/promises/data.csv") %>%
  filter(state == "NY") %>%
  View()
```

The async version would look like:

```r
library(dplyr)
read.csv.async("https://rstudio.github.io/promises/data.csv") %...>%
  filter(state == "NY") %...>%
  View()
```

The `%...>%` operator here is the secret sauce. It’s called the *promise pipe*;
the `...` stands for promise, and `>` mimics the standard pipe operator.

**What a strange looking operator. Does it work just like a regular pipe?**

In many ways `%...>%` does work like a regular pipe: it rewrites each stage’s
function call to take the previous stage’s output as the first argument. (All
the [standard magrittr
tricks](https://CRAN.R-project.org/package=magrittr/vignettes/magrittr.html)
apply here: `.`, `{`, parenthesized lambdas, etc.) But the differences, while
subtle, are profound.

The first and most important difference is that `%...>%` *must* take a promise
as input; that is, the left-hand side of the operator must be an expression that
yields a promise. The `%...>%` will do the work of “extracting” the result value
from the promise, and passing that (unwrapped) result to the function call on
the right-hand side.

This last fact—that `%...>%` passes an unwrapped, plain old, not-a-promise value
to the right-hand side—is critically important. It means we can use promise
objects with non-promise-aware functions, with `%...>%` serving as the bridge
between asynchronous and synchronous code.

**So the left-hand side of `%...>%` needs to be one of these special promise
objects, but the right-hand side can be regular R base functions?**

Yes! R base functions, dplyr, ggplot2, or whatever.

However, that work often can’t be done in the present, since the whole point of
a promise is that it represents work that hasn’t completed yet. So `%...>%` does
the work of extracting and piping not at the time that it’s called, but rather,
sometime in the future.

**You lost me.**

OK, let’s slow down and take this step by step. We’ll generate a promise by
calling an async function:

```r
df_promise <- read.csv.async("https://rstudio.github.io/promises/data.csv")
```

Even if `data.csv` is many gigabytes, `read.csv.async` returns immediately with
a new promise. We store it as `df_promise`. Eventually, when the CSV reading
operation successfully completes, the promise will contain a data frame, but for
now it’s just an empty placeholder.

One thing we definitely *can’t* do is treat `df_promise` as if it’s simply a
data frame:

```r
# Doesn't work!
dplyr::filter(df_promise, state == "NY")
```

Try this and you’ll get an error like `no applicable method for 'filter_'
applied to an object of class "promise"`. And the pipe won’t help you either;
`df_promise %>% filter(state == "NY")` will give you the same error.

**Right, that makes sense. `filter` is designed to work on data frames, and
`df_promise` isn’t a data frame.**

Exactly. Now let’s try something that actually works:

```r
df_promise %...>% filter(state == "NY")
```

At the moment it’s called, this code won’t appear to do much of anything,
really. But whenever the `df_promise` operation actually completes successfully,
then the result of that operation—the plain old data frame—will be passed to
`filter(., state = "NY")`.

**OK, so that’s good. I see what you mean about `%...>%` letting you use
non-promise functions with promises. But the whole point of using the
`filter` function is to get a data frame back. If `filter` isn’t even
going to be called until some random time in the future, how do we get its value
back?**

I’ll tell you the answer, but it’s not going to be satisfying at first.

When you use a regular `%>%`, the result you get back is the return value from
the right-hand side:

```r
df_filtered <- df %>% filter(state == "NY")
```

When you use `%...>%`, the result you get back is a promise, whose *eventual*
result will be the return value from the right-hand side:

```r
df_filtered_promise <- df_promise %...>% filter(state == "NY")
```

**Wait, what? If I have a promise, I can do stuff to it using `%...>%`, but
then I just end up with another promise? Why not just have `%...>%` return a
regular value instead of a promise?**

Remember, the whole point of a promise is that we don’t know its value yet! So
to write a function that uses a promise as input and returns some non-promise
value as output, you’d need to either be a time traveler or an oracle.

To summarize, once you start working with a promise, any calculations and
actions that are “downstream” of that promise will need to become
promise-oriented. Generally, this means once you have a promise, you need to use
`%...>%` and keep using it until your pipeline terminates.

**I guess that makes sense. Still, if the only thing you can do with promises is
make more promises, that limits their usefulness, doesn’t it?**

It’s a different way of thinking about things, to be sure, but it turns out
there’s not much limit in usefulness—especially in the context of a Shiny app.

First, you can use promises with Shiny outputs. If you’re using an
async-compatible version of Shiny (version >=1.1), all of the
built-in `renderXXX` functions can deal with either regular values or promises.
An example of the latter:

```r
output$table <- renderTable({
  read.csv.async("https://rstudio.github.io/promises/data.csv") %...>%
    filter(state == "NY")
})
```

When `output$table` executes the `renderTable` code block, it will notice that
the result is a promise, and wait for it to complete before continuing with the
table rendering. While it’s waiting, the R process can move on to do other
things.

Second, you can use promises with reactive expressions. Reactive expressions
treat promises about the same as they treat other values, actually. But this
works perfectly fine:

```r
# A reactive expression that returns a promise
filtered_df <- reactive({
  read.csv.async("https://rstudio.github.io/promises/data.csv") %...>%
    filter(state == "NY") %...>%
    arrange(median_income)
})

# A reactive expression that reads the previous
# (promise-returning) reactive, and returns a
# new promise
top_n_by_income <- reactive({
  filtered_df() %...>%
    head(input$n)
})

output$table <- renderTable({
  top_n_by_income()
})
```

Third, you can use promises in reactive observers. Use them to perform
asynchronous tasks in response to reactivity.

```r
observeEvent(input$save, {
  filtered_df() %...>%
    write.csv("ny_data.csv")
})
```

**Alright, I think I see what you mean. You can’t escape from promise-land, but
there’s no need to, because Shiny knows what to do with them.**

Yes, that’s basically right. You just need to keep track of which functions and
reactive expressions return promises instead of regular values, and be sure to
interact with them using `%...>%` or other promise-aware operators and
functions.

**Wait, there are other promise-aware operators and functions?**

Yes. The `%...>%` is the one you’ll most commonly use, but there is a variant
`%...T>%`, which we call the *promise tee* operator (it’s analogous to the
magrittr `%T>%` operator). The `%...T>%` operator mostly acts like `%...>%`, but
instead of returning a promise for the result value, it returns the original
value instead. Meaning `p %...T>% cat("\n")` won’t return a promise for the
return value of `cat()` (which is always `NULL`) but instead the value of `p`.
This is useful for logging, or other “side effecty” operations.

There’s also `%...!%`, and its tee version, `%...T!%`, which are used for error
handling. I won’t confuse you with more about that now, but you can read more
[here](overview.html#error-handling).

The `promises` package is where all of these operators live, and it also comes
with some additional functions for working with promises.

So far, the only actual async function we’ve talked about has been
`read.csv.async`, which doesn’t actually exist. To learn where actual async
functions come from, read [this guide to the `future` package](futures.html).

There are the lower-level functions `then`, `catch`, and `finally`, which are
the non-pipe, non-operator equivalents of the promise operators we’ve been
discussing. See [reference](overview.html#accessing-results-with-then).

And finally, there are `promise_all`, `promise_race`, and `promise_lapply`, used to combine
multiple promises into a single promise. Learn more about them [here](../reference/promise_all.html).

**OK, looks like I have a lot of stuff to read up on. And I’ll probably have to
reread this conversation a few times before it fully sinks in.**

Sorry. I told you it was complicated. If you make it through the rest of the guide, you’ll be 95% of the way there.

<div style="font-size: 20px; margin-top: 40px; text-align: right;">
Next: [Working with promises](overview.html)
</div>

---
title: "Why use `promises`?"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{1. Why use promises?}
  %\VignetteEncoding{UTF-8}
---

The `promises` library is designed to make it possible for R users to integrate *asynchronous programming* techniques into their code. It is intended to improve the scalability of certain classes of Shiny applications, and can also be used completely separately from Shiny.

Asynchronous programming means starting an operation without waiting around for its results. Contrast this with standard (synchronous) R programming, where the caller of a function must wait for the function to either return a result or throw an error.

Asynchronous programming is a particularly important capability for languages that are both 1) single threaded, and 2) need to serve multiple users concurrently. R is single threaded, but until recent years has mostly been restricted to desktop use or batch execution. The rise of Shiny and other web frameworks for R has meant much greater demand for serving multiple users concurrently, and thus a greater need for asynchronous programming.

## What are promises good for?

Promises work well when you have a Shiny application with a particular operation, especially a calculation, that takes a long time (measured in seconds or even minutes). Without promises, such operations not only block the current user's session from proceeding, but also block *all* other requests being made to the R process, no matter how trivial. Even loading a small CSS file, which should be nearly instantaneous, can be delayed by many seconds if another user's Shiny session is busy crunching through hundreds of gigabytes of data or querying a particularly slow backend database.

With promises, you can convert long-running operations to be asynchronous, which frees up the R process to do other work. For Shiny applications, this has the potential to greatly increase scalability for each R process (depending on how slow the operations are in the first place, and what resources they make use of).

**DO use promises if you have a Shiny app with long-running operations, and want to serve multiple users simultaneously.**

## What _aren't_ promises good for?

While promises can make a huge difference in the scalability of a Shiny app, they make relatively little difference in the latency of a single session. That is to say, if a Shiny application is slow when only a single user is hitting it, converting it to use promises is unlikely to make it perform any faster (and in fact may slightly slow it down). Promises will just help prevent *other* sessions from being slowed down by one session's computations.

(The exception to this is if the Shiny app has several distinct long computations that don't depend very much on each other—then you could use asynchronous programming to exploit a little parallelism. But I think this will be a less common use of async programming, as R already has other good tools designed specifically for data parallelism.)

**DON'T use promises to improve the performance of Shiny apps for a single user.**

<div style="font-size: 20px; margin-top: 40px; text-align: right;">
Next: [An informal intro to async programming](intro.html)
</div>

---
title: "Using `promises` with Shiny"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{6. Using promises with Shiny}
  %\VignetteEncoding{UTF-8}
---

Taking advantage of async programming from Shiny is not as simple as turning on an option or flipping a switch. If you have already written a Shiny application and are looking to improve its scalability, expect the changes required for async operation to ripple through multiple layers of server code.

Async programming with Shiny boils down to following a few steps.

1. Identify slow operations (function calls or blocks of statements) in your app.

2. Convert the slow operation into a future using `future_promise()`. (If you haven't read the [article on futures](futures.html) and [`future_promise()`](future_promise.html), definitely do that before proceeding!)

3. Any code that relies on the result of that operation (if any), whether directly or indirectly, now must be converted to promise handlers that operate on the future object.

We'll get into details for all these steps, but first, an example. Consider the following synchronous server code:

```R
function(input, output, session) {
  output$plot <- renderPlot({
    result <- expensive_operation()
    result <- head(result, input$n)
    plot(result)
  })
}
```

We'd convert it to async like this:

```R
library(promises)
library(future)
plan(multisession)

function(input, output, session) {
  output$plot <- renderPlot({
    future_promise({ expensive_operation() }) %...>%
      head(input$n) %...>%
      plot()
  })
}
```

## Adding prerequisites

The easiest part is adding `library(promises)`, `library(future)`, and `plan(multisession)` to the top of the app.

The `promises` library is necessary for the `%...>%` operator. You may also want to use promise utility functions like `promise_all` and `promise_race`.

The `future` library is needed because the `future()` function call used inside `future_promise()` is how you will launch asynchronous tasks.

`plan(multisession)` is a directive to the `future` package, telling it how future tasks should actually be executed. See the [article on futures](futures.html) for more details.

## Identifying slow operations

To find areas of your code that are good candidates for the future/promise treatment, let's start with the obvious: identifying the code that is making your app slow. You may assume it's your plotting code that's slow, but it's actually your database queries; or vice versa. If there's one thing that veteran programmers can agree on, it's that human intuition is a surprisingly unreliable tool for spotting performance problems.

Our recommendation is that you use the [profvis](https://rstudio.github.io/profvis/) profiler, which we designed to work with Shiny (see Example 3 in the profvis documentation). You can use profvis to help you focus in on where the time is actually being spent in your app.

>  **Note:** As of this writing, profvis doesn't work particularly well for diagnosing performance problems in parts of your code that you've already made asynchronous. In particular, we haven't done any work to help it profile code that executes in a future, and the mechanism we use to hide "irrelevant" parts of the stack trace doesn't work well with promises. These are ripe areas for future development.

Async programming works well when you can identify just a few "hotspots" in your app where lots of time is being spent. It works much less well if your app is too slow because of a generalized, diffuse slowness through every aspect of your app, where no one operation takes too much time but it all adds up to a lot. The more futures you need to introduce into your app, the more fixed communication overhead you incur. So for the most bang-for-the-buck, we want to launch a small number of futures per session but move a lot of the waited-on code into each one.

## Converting a slow operation into a future

Now that we've found hotspots that we want to make asynchronous, let's talk about the actual work of converting them to futures.

Conceptually, futures work like this:

```R
future({
  # Expensive code goes here
}) %...>% (function(result) {
  # Code to handle result of expensive code goes here
})
```

which seems incredibly simple. What's actually happening is that the future runs in a totally separate child R process, and then the result is collected up and returned to the main R process:

```R
# Code here runs in process A
future({
  # Code here runs in (child) process B
}) %...>% (function(result) {
  # Code here runs in process A
})
```

The fact that the future code block executes in a separate process means we have to take special care to deal with a number of practical issues. There are extremely important constraints that futures impose on their code blocks; certain objects cannot be safely used across process boundaries, and some of the default behaviors of the future library may severely impact the performance of your app. Again, see the [article on futures](futures.html) for more details.

The remainder of this document will use `future_promise()` in place of `future()`. For more information on the differences, see the [article on the benefits of `future_promise()`](future_promise.html).

### Shiny-specific caveats and limitations

In addition to the constraints that all futures face, there is an additional one for Shiny: reactive values and reactive expressions cannot be read from within a future. Whenever reactive values/expressions are read, side effects are carried out under the hood so that the currently executing observer or reactive expression can be notified when the reactive value/expression becomes invalidated. If a reactive value/expression is created in one process, but read in another process, there will be no way for readers to be notified about invalidation.

This code, for example, will not work:

```R
function(input, output, session) {
  r1 <- reactive({ ... })

  r2 <- reactive({
    future_promise({
      r1() # Will error--don't do this!
    })
  })
}
```

Even though `r1()` is called from inside the `r2` reactive expression, the fact that it's also in a future means the call will fail. Instead, you must read any reactive values/expressions you need in advance of launching the future:

```R
function(input, output, session) {
  r1 <- reactive({ ... })

  r2 <- reactive({
    val <- r1()
    future_promise({
      val # No problem!
    })
  })
}
```

However, it's perfectly fine to read reactive values/expressions from inside a promise _handler_. Handlers run in the original process, not a child process, so reactive operations are allowed.

```R
function(input, output, session) {
  r1 <- reactive({ ... })

  r2 <- reactive({
    future_promise({ ... }) %...>%
      rbind(r1()) # OK!
  })
}
```

## Integrating promises with Shiny

Generally, you'll be using promises with Shiny from within outputs, reactive expressions, and observers. We've tried to integrate promises into these constructs in as natural a way as possible.

### Outputs

Most outputs (`renderXXX({ ... })`) functions expect your code block to return a value; for example, `renderText()` expects a character vector and `renderTable()` expects a data frame. All such render functions that are included within the `shiny` package can now optionally be given a promise for such a value instead.

So this:

```R
output$table <- renderTable({
  read.csv(url) %>%
    filter(date == input$date)
})
```

could become:

```R
output$table <- renderTable({
  future_promise({ read.csv(url) }) %...>%
    filter(date == input$date)
})
```

or, trading elegance for efficiency:

```R
output$table <- renderTable({
  input_date <- input$date
  future_promise({
    read.csv(url) %>% filter(date == input_date)
  })
})
```

The important thing to keep in mind is that the promise (or promise pipeline) must be the final expression in the code block. Shiny only knows about promises you actually return to it when you hand control back.

#### Render functions with side effects: `renderPrint` and `renderPlot`

The render functions `renderPrint()` and `renderPlot()` are slightly different than other render functions, in that they can be affected by side effects in the code block you provide. In `renderPrint` you can print to the console, and in `renderPlot` you can plot to the active R graphics device.

With promises, these render functions can work in a similar way, but with a caveat. As you hopefully understand by now, futures execute their code in a separate R process, and printing/plotting in a separate process won't have any effect on the Shiny output in the original process. These examples, then, are incorrect:

```R
output$summary <- renderPrint({
  future_promise({
    read.csv(url) %>%
      summary() %>%
      print()
  })
})

output$plot <- renderPlot({
  future_promise({
    df <- read.csv(url)
    ggplot(df, aes(length, width)) + geom_point()
  })
})
```

Instead, do printing and plotting after control returns back to the original process, via a promise handler:

```R
output$summary <- renderPrint({
  future_promise({ read.csv(url) }) %...>%
    summary() %...>%
    print()
})

output$plot <- renderPlot({
  future_promise({ read.csv(url) }) %...>%
  {
    ggplot(., aes(length, width)) + geom_point()
  }
})
```

Again, you do need to be careful to make sure that the last expression in your code block is the promise/pipeline; this is the only way the rendering logic can know whether and when your logic has completed, and if any errors occurred (so they can be displayed to the user).

### Observers

Observers are very similar to outputs: you must make sure that the last expression in your code block is the promise/pipeline. Like outputs, observers need to know whether and when they're done running, and if any errors occured (so they can log them and terminate the user session). The way to communicate this from your async user code is by returning the promise.

Here's a synchronous example that we'll convert to async. Clicking the `refresh_data` action button causes data to be downloaded, which is then saved to disk as `cached.rds` and also used to update the reactive value `data`.

```R
data <- reactiveVal(readRDS("cached.rds"))

function(input, output, session) {
  observeEvent(input$refresh_data, {
    df <- read.csv(url)
    saveRDS(df, "cached.rds")
    data(df)
  })
}
```

And the async version:

```R
data <- reactiveVal(readRDS("cached.rds"))

function(input, output, session) {
  observeEvent(input$refresh_data, {
    future_promise({
      df <- read.csv(url)
      saveRDS(df, "cached.rds")
      df
    }) %...>%
      data()
  })
}
```

Note that in this version, we cannot call `data(df)` inside the future, as this would cause the update to happen in the wrong process. Instead, we use the `%...>%` operator to perform the assignment back in the main process once the future resolves.

### Reactive expressions

Recall that reactive expressions are used to calculate values, and are cached until they are automatically invalidated by one of their dependencies. Unlike outputs and observers, reactive expressions can be used from other reactive consumers.

Asynchronous reactive expressions are similar to regular (synchronous) reactive expressions: instead of a "normal" value, they return a promise that will yield the desired value; and a normal reactive will cache a normal value, while an async reactive will cache the promise.

The upshot is that when defining an async reactive expression, your code block should return a promise or promise pipeline, following the same rules as reactive outputs. And when calling an async reactive expression, call it like a function like you would a regular reactive expression, and treat the value that's returned like any other promise.

```R
function(input, output, session) {
  data <- eventReactive(input$refresh_data, {
    read.csv(url)
  })

  filteredData <- reactive({
    data() %>% filter(date == input$date)
  })

  output$table <- renderTable({
    filteredData() %>% head(5)
  })
}
```

And now in async:

```R
function(input, output, session) {
  data <- eventReactive(input$refresh_data, {
    future_promise({ read.csv(url) })
  })

  filteredData <- reactive({
    data() %...>% filter(date == input$date)
  })

  output$table <- renderTable({
    filteredData() %...>% head(5)
  })
}
```

## The flush cycle

In the past, Shiny's reactive programming model has operated using a mostly traditional [event loop](https://en.wikipedia.org/wiki/Event_loop) model. Somewhere many levels beneath `shiny::runApp()` was a piece of code that looked a bit like this:

```R
while (TRUE) {
  # Do nothing until a browser sends some data
  input <- receiveInputFromBrowser()
  # Use the received data to update reactive inputs
  session$updateInput(input)
  # Execute all invalidated outputs/observers
  flushReact()
  # After ALL outputs execute, send the results back
  flushOutputs()
}
```

We call this Shiny's "flush cycle". There are two important properties to our flush cycle.

1. Only one of the four steps—receiving, updating, reacting, and sending—can be executing a time. (Remember, R is single threaded.) In particular, it's not possible for inputs to be updated while outputs/observers are running. This is important in order to avoid race conditions that would be all but impossible to defend against.
2. Many outputs may change as a result of a single input value received from the browser, but none of them are sent back to the client until all of the outputs are ready. The advantage of this is a smoother experience for the end-user in most cases. (Admittedly, there has been some controversy regarding this property of Shiny; some app authors would strongly prefer to show outputs as soon as they are ready, or at least to have manual control over this behavior.)

While adding async support to Shiny, we aimed to keep these two properties intact. Imagine now that `flushReact()`, the line that executes invalidated outputs/observers, returns a promise that combines all of the async outputs/observers (i.e. a promise that resolves only after all of the async outputs/observers have resolved). The new, async-aware event loop is conceptually more like this:

```R
doEventLoop <- function() {
  # Do nothing until a browser sends some data
  input <- receiveInputFromBrowser()
  # Use the received data to update reactive inputs
  session$updateInput(input)
  # Execute all invalidated outputs/observers
  flushReact() %...>% {
	# After ALL outputs execute, send the results back
    flushOutputs()
    # Continue the event loop
    doEventLoop()
  }
}
```

The resulting behavior matches the synchronous version of the event loop, in that:

1. No inputs are received from the browser until all pending async outputs/observers have completed. Unlike the synchronous version, this separation is enforced at the session level: if Session A has some async observers that have not finished executing, that only prevents Session A from processing new input values, while new input values from Session B can be handled immediately because they belong to a different session. Again, the goal of keeping input updates separate from output/observer execution is to prevent race conditions, which are even more pernicious to debug and understand when async code is involved.
2. For a given session, no outputs are sent back to the client, until all outputs are ready. It doesn't matter whether the outputs in question are synchronous, asynchronous, or some combination; they all must complete execution before any can be sent.

---
title: "Advanced `future` and `promises` usage"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{5. Advanced future and promises usage}
  %\VignetteEncoding{UTF-8}
resource_files:
  - future_promise/future.png
  - future_promise/blocked_future_promise.png
---

This article discusses the benefits of using `promises::future_promise()` over a combination of `future::future()` + `promises::promise()` to better take advantage of computing resources available to your main R session. To demonstrate these benefits, we'll walk-through a use-case with the `plumber` package. ([See here](https://www.rplumber.io/) to learn more about `plumber` and [the previous article](futures.html) to learn more about `future`.)

## The problem with `future()`+`promise()`

In an ideal situation, the number of available `future` _workers_ (`future::nbrOfFreeWorkers()`) is always **more than** the number of `future::future()` _jobs_. However, if a `future` job is attempted when the number of free workers is `0`, then `future` will block the current R session until one becomes available.

For a concrete example, let's imagine a scenario, where seven `plumber` requests are received at the same time with only two `future` workers available. Also, let's assume the `plumber` route(s) serving the first 6 requests use `future::future()` and take ~10s to compute `slow_calc()`:

```r
#* @get /slow/<k>
function() {
  future::future({
    slow_calc()
  })
}
```

Let's also assume the `plumber` route serving the last request does not use any form of `future` or `promises` and takes almost no time to compute.

```r
#* @get /fast/<k>
function() {
  fast_calc()
}
```


The figure below depicts the overall timeline of execution of these 7 requests under the conditions we've outlined above. Note that the y-axis is ordered from first request coming in (`/slow/1`) to the last request (`/fast/7`).

<img src = "future_promise/future.png" width = "75%" alt="Early workers take more time than expected. Main R session is blocked" />

Note how R has to wait 20s before processing the 7th request (shown in green). This is a big improvement over not using `future`+`promises` at all (in that case, R would have to wait 60s before processing). However, since there are only two `future` workers available R still has to wait longer than necessary to process that last request because the main R session must wait for a `future` worker to become available. The video below animates this behavior:


```{r, echo = FALSE}
library(vembedr)
embed_vimeo("505287449") %>%
  use_align("center")
```


## The solution: `future_promise()`

The advantage of using `future_promise()` over `future::future()` is that even if there aren't `future` workers available, the `future` is scheduled to be done when workers become available via `promises`. In other words, `future_promise()` ensures the main R thread isn't blocked when a `future` job is requested and can't immediately perform the work (i.e., the number of jobs exceeds the number of workers).

Continuing with the example above, we can swap out the calls to `future::future()` with `future_promise()`.

```r
#* @get /slow/<k>
function() {
  promises::future_promise({
    slow_calc()
  })
}
```


With this change to `future_promise()`, note how the `/fast/7` route now does not have to wait on `future` work to finish processing. Therefore, `plumber` can complete the last requests almost immediately:

<img src = "future_promise/blocked_future_promise.png" width="75%" alt="future_promise() keeps the main R session free" />

The vertical gray bars in the figure above represent timepoints where the main R session is actually busy. Outside of these gray areas, the R session is free to do other things, for example, executing other `promises` or, more generally, non-`future` work. The video below animates this behavior:

```{r, echo = FALSE}
library(vembedr)
embed_vimeo("505286442") %>%
  use_align("center")
```

<div style="font-size: 20px; margin-top: 40px; text-align: right;">
Next: [Using `promises` with Shiny](shiny.html)
</div>

---
title: "Case study: converting a Shiny app to async"
author: Joe Cheng (joe@rstudio.com)
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{8. Case study}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

In this case study, we'll work through an application of reasonable complexity, turning its slowest operations into futures/promises and modifying all the downstream reactive expressions and outputs to deal with promises.

## Motivation

> As a web service increases in popularity, so does the number of rogue scripts that abuse it for no apparent reason.
>
> _—Cheng's Law of Why We Can't Have Nice Things_

I first noticed this in 2011, when the then-new RStudio IDE was starting to gather steam. We had a dashboard that tracked how often RStudio was being downloaded, and the numbers were generally tracking smoothly upward. But once every few months, we'd have huge spikes in the download counts, ten times greater than normal—and invariably, we'd find that all of the unexpected increase could be tracked to one or two IP addresses.

For hours or days we'd be inundated with thousands of downloads per hour, then just as suddenly, they'd cease. I didn't know what was happening then, and I still don't know today. Was it the world's least competent denial-of-service attempt? Did someone write a download script with an accidental `while (TRUE)` around it?

Our application will let us examine downloads from CRAN for this kind of behavior. For any given day on CRAN, we'll see what the top downloaders are and how they're behaving.

## Our source data

RStudio maintains the popular `0-Cloud` CRAN mirror, and the log files it generates are freely available at http://cran-logs.rstudio.com/. Each day is a separate gzipped CSV file, and each row is a single package download. For privacy, IP addresses are anonymized by substituting each day's IP addresses with unique integer IDs.

Here are the first few lines of http://cran-logs.rstudio.com/2018/2018-05-26.csv.gz :

```
"date","time","size","r_version","r_arch","r_os","package","version","country","ip_id"
"2018-05-26","20:42:23",450377,"3.4.4","x86_64","linux-gnu","lubridate","1.7.4","NL",1
"2018-05-26","20:42:30",484348,NA,NA,NA,"homals","0.9-7","GB",2
"2018-05-26","20:42:21",98484,"3.3.1","x86_64","darwin13.4.0","miniUI","0.1.1.1","NL",1
"2018-05-26","20:42:27",518,"3.4.4","x86_64","linux-gnu","RCurl","1.95-4.10","US",3
```

Fortunately for our purposes, there's no need to analyze these logs at a high level to figure out which days are affected by badly behaved download scripts. These CRAN mirrors are popular enough that, according to Cheng's Law, there should be plenty of rogue scripts hitting it every day of the year.

## A tour of the app

The app I built to explore this data, **cranwhales**, let us examine the behavior of the top downloaders ("whales") for any given day, at varying levels of detail. You can view this app live at https://gallery.shinyapps.io/cranwhales/, or download and run the code yourself at https://github.com/rstudio/cranwhales.

When the app starts, the "All traffic" tab shows you the number of package downloads per hour for all users vs. whales. In this screenshot, you can see the proportion of files downloaded by the top six downloaders on May 28, 2018. It may not look like a huge fraction at first, but keep in mind, we are only talking about six downloaders out of 52,815 total!

![](case-study-tab1.png)

The "Biggest whales" tab simply shows the most prolific downloaders, with their number of downloads performed. Each anonymized IP address has been assigned an easier-to-remember name, and you can also see the country code of the original IP address.

![](case-study-tab2.png)

The "Whales by hour" tab shows the hourly download counts for each whale individually. In this screenshot, you can see that the Netherlands' `relieved_snake` downloaded at an extremely consistent rate during the whole day, while the American `curly_capabara` was active only during business hours in Eastern Standard Time. Still others, like `colossal_chicken` out of Hong Kong, was busy all day but at varying rates.

![](case-study-tab3.png)

The "Detail View" has perhaps the most illuminating information. It lets you view every download made by a given whale on the day in question. The x dimension is time and the y dimension is what package they downloaded, so you can see at a glance exactly how many packages were downloaded, and how their various package downloads relate to each other. In this case, `relieved_snake` downloaded 104 different packages, in the same order, continuously, for the entire day.

![](case-study-tab4.png)

Others behave very differently, like `freezing_tapir`, who downloaded `devtools`--and _only_ `devtools`--for the whole day, racking up 19,180 downloads totalling 7.9 gigabytes for that one package alone!

![](case-study-tab5.png)

Sadly, the app can't tell us any more than that--it can't explain _why_ these downloaders are behaving this way, nor can it tell us their street addresses so that we can send ninjas in black RStudio helicopters to make them stop.

## The implementation

Now that you've seen what the app does, let's talk about how it was implemented, then convert it from sync to async.

### User interface

The user interface is a pretty typical shinydashboard. It's important to note that the UI part of the app is entirely agnostic to whether the server is written in the sync or async style; when we port the app to async, we won't touch the UI at all.

There are two major pieces of input we need from users: what **date** to examine (this app only lets us look at one day at a time) and **how many** of the most prolific downloaders to look at. We'll put these two controls in the dashboard sidebar.

```r
dashboardSidebar(
  dateInput("date", "Date", value = Sys.Date() - 2),
  numericInput("count", "Show top N downloaders:", 6)
)
```

(We set `date` to two days ago by default, because there's some lag between when a day ends and when its logs are published.)

The rest of the UI code is just typical shinydashboard scaffolding, plus some `shinydashboard::valueBoxOutput`s and `plotOutputs`. These are so trivial that they're hardly worth talking about, but I'll include the code here for completeness. Finally, there's `detailViewUI`, a [Shiny module](https://shiny.rstudio.com/articles/modules.html) that just contains more of the same (value boxes and plots).

```r
  dashboardBody(
    fluidRow(
      tabBox(width = 12,
        tabPanel("All traffic",
          fluidRow(
            valueBoxOutput("total_size", width = 4),
            valueBoxOutput("total_count", width = 4),
            valueBoxOutput("total_downloaders", width = 4)
          ),
          plotOutput("all_hour")
        ),
        tabPanel("Biggest whales",
          plotOutput("downloaders", height = 500)
        ),
        tabPanel("Whales by hour",
          plotOutput("downloaders_hour", height = 500)
        ),
        tabPanel("Detail view",
          detailViewUI("details")
        )
      )
    )
  )
```

### Server logic

Based on these inputs and outputs, we'll write a variety of reactive expressions and output renderers to download, manipulate, and visualize the relevant log data.

The reactive expressions:

* `data` (`eventReactive`): Whenever `input$date` changes, the `data` reactive downloads the full log for that day from http://cran-logs.rstudio.com, and parses it.
* `whales` (`reactive`): Reads from `data()`, tallies the number of downloads performed by each unique IP, and returns a data frame of the top `input$count` most prolific downloaders, along with their download counts.
* `whale_downloads` (`reactive`): Joins the `data()` and `whales()` data frames, to return all of the details of the cetacean downloads.

The `whales` reactive expression depends on `data`, and `whale_downloads` depends on `data` and `whales`.

![](case-study-react.png)

The outputs in this app are mostly either `renderPlot`s that we populate with `ggplot2`, or `shinydashboard::renderValueBox`es. They all rely on one or more of the reactive expressions we just described. We won't catalog them all here, as they're not individually interesting, but we will look at some archetypes below.

## Improving performance and scalability

While this article is specifically about async, this is a good time to remind you that there are lots of ways to improve the performance of a Shiny app. Async is just one tool in the toolbox, and before reaching for that hammer, take a moment to consider your other options:

1. Have I used [profvis](https://rstudio.github.io/profvis/) to **profile my code** and determine what's actually taking so long? (Human intuition is a notoriously bad profiler!)
2. Can I perform any **calculations, summarizations, and aggregations offline**, when my Shiny app isn't even running, and save the results to .rds files to be read by the app?
3. Are there any opportunities to **cache**--that is, save the results of my calculations and use them if I get the same request later? (See [memoise](https://cran.r-project.org/package=memoise), or roll your own.)
4. Am I effectively leveraging [reactive programming](https://rstudio.com/resources/shiny-dev-con/reactivity-pt-1-joe-cheng/) to make sure my reactives are doing as little work as possible?
5. When deploying my app, am I load balancing across multiple R processes and/or multiple servers? ([Shiny Server Pro](https://docs.rstudio.com/shiny-server/#utilization-scheduler), [RStudio Connect](https://docs.rstudio.com/connect/admin/appendix/configuration/#Scheduler), [ShinyApps.io](https://shiny.rstudio.com/articles/scaling-and-tuning.html))

These options are more generally useful than using async techniques because they can dramatically speed up the performance of an app even if only a single user is using it. While it obviously depends on the particulars of the app itself, a few lines of precomputation or caching logic can often lead to 10X-100X better performance. Async, on the other hand, generally doesn't help make a single session faster. Instead, it helps a single Shiny process support more concurrent sessions without getting bogged down.

Async can be an essential tool when there is no way around performing expensive tasks (i.e. taking multiple seconds) while the user waits. For example, an app that analyzes any user-specified Twitter profile may get too many unique queries (assuming most people specify their own Twitter handle) for caching to be much help. And applications that invite users to upload their own datasets won't have an opportunity to do any offline summarizing in advance. If you need to run apps like that and support lots of concurrent users, async can be a huge help.

In that sense, the cranwhales app isn't a perfect example, because it has lots of opportunities for precomputation and caching that we'll willfully ignore today so that I can better illustrate the points I want to make about async. When you're working on your own app, though, please think carefully about _all_ of the different techniques you have for improving performance.

## Converting to async

To quote the article [*Using promises with Shiny*](https://rstudio.github.io/promises/articles/shiny.html), async programming with Shiny boils down to following a few steps:

1. Identify slow operations in your app.
2. Convert the slow operations into futures.
3. Any code that relies on the result of those operations (if any), whether directly or indirectly, now must be converted to promise handlers that operate on the future object.

In this case, the slow operations are easy to identify: the downloading and parsing that takes place in the `data` reactive expression can each take several long seconds.

Converting the download and parsing operations into futures turns out to be the most complicated part of the process, for reasons we'll get into later.

Assuming we do that successfully, the `data` reactive expression will no longer return a data frame, but a `promise` object (that resolves to a data frame). Since the `whales` and `whale_downloads` reactive expressions both rely on `data`, those will both also need to be converted to read and return `promise` objects. And therefore, because the outputs all rely on one or more reactive expressions, they will all need to know how to deal with `promise` objects.

Async code is infectious like that; once you turn the heart of your app into a promise, everything downstream must become promise-aware as well, all the way through to the observers and outputs.

With that overview out of the way, let's dive into the code.

In the sections below, we'll take a look at the code behind some outputs and reactive expressions. For each element, we'll look first at the sync version, then the async version.

In some cases, these code snippets may be slightly abridged. See the [GitHub repository](https://github.com/rstudio/cranwhales) for the full code.

Until you've received an introduction to the `%...>%` operator, the async code below will make no sense, so if you haven't read [*An informal intro to async programming*](https://rstudio.github.io/promises/articles/intro.html) and/or [*Working with promises in R*](https://rstudio.github.io/promises/articles/overview.html), I highly recommend doing so before continuing!

### Loading `promises` and `future`

The first thing we'll do is load the basic libraries of async programming.

```r
library(promises)
library(future)
plan(multisession)
```

I originally used `multiprocess` but file downloading inside a future seemed to fail on Mac. (I've found that it's usually not worth spending a lot of time trying to figure out why `multiprocess` doesn't work for some specific code; instead, just use `multisession`, since that's probably going to be the solution anyway.)

### The `data` reactive: future_promise() all the things

The next thing we'll do is convert the `data` event reactive to use `future` for the expensive bits. The original code looks lke this:

```r
# SYNCHRONOUS version

data <- eventReactive(input$date, {
  date <- input$date  # Example: 2018-05-28
  year <- lubridate::year(date)  # Example: "2018"

  url <- glue("http://cran-logs.rstudio.com/{year}/{date}.csv.gz")
  path <- file.path("data_cache", paste0(date, ".csv.gz"))

  withProgress(value = NULL, {

    if (!file.exists(path)) {
      setProgress(message = "Downloading data...")
      download.file(url, path)
    }

    setProgress(message = "Parsing data...")
    read_csv(path, col_types = "Dti---c-ci", progress = FALSE)

  })
})
```

(Earlier, I said we wouldn't take advantage of precomputation or caching. That wasn't entirely true; in the code above, we cache the log files we download in a `data_cache` directory. I couldn't bring myself to put my internet connection through that level of abuse, as I knew I'd be running this code thousands of times as I load tested it.)

For now, we'll lose the `withProgress`/`setProgress` reporting, since doing that correctly requires some more advanced techniques that we'll talk about later. We'll come back and fix this code later, but for now:

```r
# ASYNCHRONOUS version

data <- eventReactive(input$date, {
  date <- input$date
  year <- lubridate::year(date)

  url <- glue("http://cran-logs.rstudio.com/{year}/{date}.csv.gz")
  path <- file.path("data_cache", paste0(date, ".csv.gz"))

  future_promise({
    if (!file.exists(path)) {
      download.file(url, path)
    }
    read_csv(path, col_types = "Dti---c-ci", progress = FALSE)
  })
})
```

Pretty straightforward. This reactive now returns a future (which counts as a promise), not a data frame.

Remember that we **must** read any reactive values (including `input`) and reactive expressions [from **outside** the future](https://rstudio.github.io/promises/articles/shiny.html#shiny-specific-caveats-and-limitations). (You will get an error if you attempt to read one from inside the future.)

At this point, since there are no other long-running operations we want to make asynchronous, we're actually done interacting directly with the `future` package. The rest of the reactive expressions will deal with the future returned by `data` using general async functions and operators from `promises`.

### The `whales` reactive: simple pipelines are simple

The `whales` reactive takes the data frame from `data`, and uses dplyr to find the top `input$count` most prolific downloaders.

```r
# SYNCHRONOUS version

whales <- reactive({
  data() %>%
    count(ip_id) %>%
    arrange(desc(n)) %>%
    head(input$count)
})
```

Since `data()` now returns a promise, the whole function needs to be modified to deal with promises.

This is basically a best-case scenario for working with `promises`. The whole expression consists of a single magrittr pipeline. There's only one object  (`data()`) that's been converted to a promise. The promise object only appears once, at the head of the pipeline.

When the stars align like this, converting this code to async is literally as easy as replacing each `%>%` with `%...>%`:

```r
# ASYNCHRONOUS version

whales <- reactive({
  data() %...>%
    count(ip_id) %...>%
    arrange(desc(n)) %...>%
    head(input$count)
})
```

The input (`data()`) is a promise, the resulting output object is a promise, each stage of the pipeline returns a promise; but we can read and write this code almost as easily as the synchronous version!

An example this simple may seem reductive, but this best-case scenario happens surprisingly often, if your coding style is influenced by the tidyverse. In this example app, **59%** of the reactives, observers, and outputs were converted using nothing more than replacing `%>%` with `%...>%`.

One last thing before we move on. In the last section, I emphasized that reactive values cannot be read from inside a future. Here, we're using `head(input$count)` inside a promise-pipeline; since `data()` is written using a future, doesn't that mean… well… isn't this wrong?

Nope—this code is just fine. The prohibition is against reading reactive values/expressions from *inside* a future, because code inside a future is executed in a totally different R process. The steps in a promise-pipeline aren't futures, but promise handlers. These aren't executed in a different process; rather, they're executed back in the original R process after a promise is resolved. We're allowed and expected to access reactive values and expressions from these handlers.

### The `whale_downloads` reactive: reading from multiple promises

The `whale_downloads` reactive is a bit more complicated case.

```r
# SYNCHRONOUS version

whale_downloads <- reactive({
  data() %>%
    inner_join(whales(), "ip_id") %>%
    select(-n)
})
```

Looks simple, but we can't just do a simple replacement this time. Can you see why?

```r
# BAD VERSION DOESN'T WORK

whale_downloads <- reactive({
  data() %...>%
    inner_join(whales(), "ip_id") %...>%
    select(-n)
})
```

Remember, both `data()` and `whales()` now return a promise object, not a data frame. None of the dplyr verbs know how to deal with promises natively (and the same is true for almost every other R function, anywhere in the R universe).

We're able to use `%...>%` with promises on the left-hand side and regular dplyr calls on the right-hand side, only because the `%...>%` operator "unwraps" the promise object for us, yielding a regular object (data frame or whatever) to be passed to dplyr. But in this case, we're passing `whales()`, which a promise object, directly to `inner_join`, and `inner_join` has no idea what to do with it.

The fundamental thing to pattern-match on here, is that **we have a block of code that relies on more than one promise object**, and that means `%...>%` won't be enough. This is a pretty common situation as well, and occurs in **12%** of reactives and outputs in this example app.

Here's what the real solution looks like:

```r
# ASYNCHRONOUS version

whale_downloads <- reactive({
  promise_all(data_df = data(), whales_df = whales()) %...>% with({
    data_df %>%
      inner_join(whales_df, "ip_id") %>%
      select(-n)
  })
})
```

#### Promises: the Gathering

This solution uses the [promise gathering](https://rstudio.github.io/promises/articles/combining.html#gathering) pattern, which combines `promises_all`, `%...>%`, and `with`.

* The `promise_all` function gathers multiple promise objects together, and returns a single promise object. This new promise object doesn't resolve until all the input promise objects are resolved, and it yields a list of those results.

```r
> promise_all(a = future_promise("Hello"), b = future_promise("World")) %...>% print()
$a
[1] "Hello"

$b
[1] "World"
```

* The `%...>%`, as before, "unwraps" the promise object and passes the result to its right hand side.
* The `with` function (from base R) takes a named list, and makes it into a sort of virtual parent environment while evaluating a code block you pass it.

```r
> x + y
Error: object 'x' not found

> with(list(x = 1, y = 2), {
+   x + y
+ })
[1] 3
```

Let's once again combine the three, with the simplest possible example of the gathering pattern:

```r
> promise_all(x = future_promise("Hello"), y = future_promise("World")) %...>%
+   with({ paste(x, y) }) %...>%
+   print()
[1] "Hello World"
```

You can make use of this pattern without remembering exactly how these pieces combine. Just remember that the arguments to `promise_all` provide the promise objects (`future_promise(1)` and `future_promise(2)`), along with the names you want to use to refer to their yielded values (`x ` and `y`); and the code block you put in `with()` can refer to those names without worrying about the fact that they were ever promises to begin with.

### The `total_downloaders` value box: simple pipelines are for output, too

![](case-study-downloaders.png)

All of the value boxes in this app ended up looking a lot like this:

```r
# SYNCHRONOUS version

output$total_downloaders <- renderValueBox({
  data() %>%
    pull(ip_id) %>%
    unique() %>%
    length() %>%
    format(big.mark = ",") %>%
    valueBox("unique downloaders")
})
```

This is structurally no different than the `whales` best-case scenario reactive. One thing worth pointing out is that an async `renderValueBox` means you return a promise that returns a `valueBox`; you *don't* return a `valueBox` to whom you have passed a promise.

Meaning, you *don't* do this:

```r
# BAD VERSION DOESN'T WORK

output$total_downloaders <- renderValueBox({
  valueBox(
    data() %...>%
      pull(ip_id) %...>%
      unique() %...>%
      length() %...>%
      format(big.mark = ","),
    "unique downloaders"
  )
})
```

Instead, you do this:

```r
# ASYNCHRONOUS version

output$total_downloaders <- renderValueBox({
  data() %...>%
    pull(ip_id) %...>%
    unique() %...>%
    length() %...>%
    format(big.mark = ",") %...>%
    valueBox("unique downloaders")
})
```

The other trick worth nothing is the `pull` verb, which is used to retrieve a specific column of a data frame as a vector (similar to `$` or `[[`). In this case, `pull(data, ip_id)` is equivalent to `data[["ip_id"]]`. Note that `pull` is part of dplyr and isn't specific to promises.

### The `biggest_whales` plot: getting untidy

In a cruel twist of API design fate, one of the cornerstone packages of the tidyverse lacks a tidy API. I'm referring, of course, to `ggplot2`:

```r
# SYNCHRONOUS version

output$downloaders <- renderPlot({
  whales() %>%
    ggplot(aes(ip_name, n)) +
    geom_bar(stat = "identity") +
    ylab("Downloads on this day")
})
```

While `dplyr` and other tidyverse packages are designed to link calls together with `%>%`, the older `ggplot2` package uses the `+` operator. This is mostly a small aesthetic wart when synchronous code, but it's a real problem with async, because the `promises` package doesn't currently have a promise-aware replacement for `+` like it does for `%>%`.

Fortunately, there's a pretty good escape hatch for `%>%`, and `%...>%` inherited it too. Instead of a pipeline stage being a simple function call, you can put a `{` and `}` delimited code block, and inside of that code block, you can access the "it" value using a period (`.`).

```r
# ASYNCHRONOUS version

output$downloaders <- renderPlot({
  whales() %...>% {
    whale_df <- .
    ggplot(whale_df, aes(ip_name, n)) +
      geom_bar(stat = "identity") +
      ylab("Downloads on this day")
  }
})
```

**The importance of this pattern cannot be overstated!** Using `%...>%` and simple calls alone, you're restricted to doing pipeline-compatible operations. But `%...>%` together with a curly-brace code block means your handler code can be any shape or size. Once inside that code block, you have a regular, non-promise value in `.` (if you even want to use it—sometimes you don't, as we'll see later). You can have zero, one, or more statements. You can use the `.` multiple times, in nested expressions, whatever.

Tip: if you have extensive or complex code to put in a code block, start the block by creating a properly named variable to store the value of `.`. The reason for this is that `.` may acquire a different meaning than you intend as you add code to the code block. For example, if a magrittr pipeline starts with `.`, instead of evaluating the pipeline and returning a value, it creates a function that takes a single argument. So the following code wouldn't filter the resolved value of `whales()`, but instead, create an anonymous function that calls `filter(n > 1000)` on whatever you pass it.

```r
whales() %...>% {
  . %>% filter(n > 1000)
}
```

This fixes it:

```r
whales() %...>% {
  whales_df <- .
  whales_df %>% filter(n > 1000)
}
```

There are other ways to work around the above problem as well, but I like this fix because it doesn't require any thought or care. Just give the `.` value a new name, and forget the `.` exists.

For untidy code with a single promise object, just remember: pair a single `%...>%` with a code block and you should be able to do almost anything.

### Revisiting the `data` reactive: progress support

Now that we have discussed a few techniques for writing async code, let's come back to our original `data` event reactive, and this time do a more faithful async conversion that preserves the progress reporting functionality of the original.

Again, here's the original sync code:

```r
# SYNCHRONOUS version

data <- eventReactive(input$date, {
  date <- input$date  # Example: 2018-05-28
  year <- lubridate::year(date)  # Example: "2018"

  url <- glue("http://cran-logs.rstudio.com/{year}/{date}.csv.gz")
  path <- file.path("data_cache", paste0(date, ".csv.gz"))

  withProgress(value = NULL, {

    if (!file.exists(path)) {
      setProgress(message = "Downloading data...")
      download.file(url, path)
    }

    setProgress(message = "Parsing data...")
    read_csv(path, col_types = "Dti---c-ci", progress = FALSE)

  })
})
```

Progress reporting currently presents two challenges for future.

First, the `withProgress({...})` function cannot be used with async. `withProgress` is designed to wrap a slow synchronous action, and dismisses its progress dialog when the block of code it wraps is done executing. Since the call to `future_promise()` will return immediately even though the actual task is far from done, using `withProgress` won't work; the progress dialog would be dismissed before the download even got going.

It's conceivable that `withProgress` could gain promise compatibility someday, but it's not in Shiny v1.1.0. In the meantime, we can work around this by using the alternative, [object-oriented progress API](https://shiny.rstudio.com/reference/shiny/1.1.0/Progress.html) that Shiny offers. It's a bit more verbose and fiddly than `withProgress`/`setProgress`, but it is flexible enough to work with futures/promises.

Second, progress messages can't be sent from futures. This is simply because futures are executed in child processes, which don't have direct access to the browser like the main Shiny process does.

It's conceivable that `future` could gain the ability for child processes to communicate back to their parents, but no good solution exists at the time of this writing. In the meantime, we can work around this by taking the one future that does both downloading and parsing, and splitting it into two separate futures. After the download future has completed, we can send a progress message that parsing is beginning, and then start the parsing future.

The regrettably complicated solution is below.

```r
# ASYNCHRONOUS version

data <- eventReactive(input$date, {
  date <- input$date
  year <- lubridate::year(date)

  url <- glue("http://cran-logs.rstudio.com/{year}/{date}.csv.gz")
  path <- file.path("data_cache", paste0(date, ".csv.gz"))

  p <- Progress$new()
  p$set(value = NULL, message = "Downloading data...")
  future_promise({
    if (!file.exists(path)) {
      download.file(url, path)
    }
  }) %...>%
    { p$set(message = "Parsing data...") } %...>%
    { future_promise(read_csv(path, col_types = "Dti---c-ci", progress = FALSE)) } %>%
    finally(~p$close())
})
```

The single future we wrote earlier has now become a pipeline of promises:

1. future (download)
2. send progress message
3. future (parse)
4. dismiss progress dialog

Note that neither the R6 call `p$set(message = ...)` nor the second `future_promise()` call are tidy, so they use curly-brace blocks, as discussed in the above section about `biggest_whales`.

The final step of dismissing the progress dialog doesn't use `%...>%` at all; because we want the progress dialog to dismiss whether the download and parse operations succeed or fail, we use the regular pipe `%>%` and `finally()` function instead. See the relevant section in [*Working with promises in R*](https://rstudio.github.io/promises/articles/overview.html#cleaning-up-with-finally) to learn more.

With these changes in place, we've now covered all of the changes to the application. You can see the full changes side-by-side via [this GitHub diff](https://github.com/rstudio/cranwhales/compare/sync...async?diff=split).

## Measuring scalability

It was a fair amount of work to do the sync-to-async conversion. Now we'd like to know if the conversion to async had the desired effect: improved responsiveness (i.e. lower latency) when the number of simultaneous visitors increases.

### Load testing with Shiny (coming soon)

At the time of this writing, we are working on a suite of load testing tools for Shiny that is not publicly available yet, but was previewed by Sean Lopp during his [epic rstudio::conf 2018 talk](https://rstudio.com/resources/) about running a Shiny load test with 10,000 simulated concurrent users.

You use these tools to easily **record** yourself using your Shiny app, which creates a test script; then **play back** that test script, but multiplied by dozens/hundreds/thousands of simulated concurrent users; and finally, **analyze** the timing data generated during the playback step to see what kind of latency the simulated users experienced.

To examine the effects of my async refactor, I recorded a simple test script by loading up the app, waiting for the first tab to appear, then clicking through each of the other tabs, pausing for several seconds each time before moving on to the next. When using the app without any other visitors, the homepage fully loads in less than a second, and the initial loading of data and rendering of the plot on the default tab takes about 7 seconds. After that, each tab takes no more than a couple of seconds to load. Overall, the entire test script, including time where the user is thinking, takes about 40 seconds under ideal settings (i.e. only a single concurrent user).

I then used this test script to generate load against the Shiny app running in my local RStudio. With the settings I chose, the playback tool introduced one new "user" session every 5 seconds, until 50 sessions total had been launched; then it waited until all the sessions were complete. I ran this test on both the sync and async versions in turn, which generated the following results.

### Sync vs. async performance

![](case-study-gantt-async.png)

In this plot, each row represents a single session, and the x dimension represents time. Each of the rectangles represents a single "step" in the test script, be it downloading the HTML for the homepage, fetching one of the two dozen JavaScript/CSS files, or waiting for the server to update outputs. So the wider a rectangle is, the longer the user had to wait. (The empty gaps between rectangles represents time the app is waiting for the user to click an input; their widths are hard-coded into the test script.)

Of particular importance are the red and pink rectangles, as these represent the initial page load. While these are taking place, the user is staring at a blank page, probably wondering if the server is down. Long waits during this stage are not only undesirable, but surprising and incomprehensible to the user; whereas the same user is probably prepared to wait a little while for a complicated visualization to be rendered in response to an input change.

And as you can see from this plot, the behavior of the async app is much improved in the critical metric of homepage/JS/CSS loading time. The sync version of the app starts displaying unacceptably long red/pink loading times as early as session 15, and by session #44 the maximum page load time has exceeded one minute. The async version at that point is showing 25 second load times, which is far from great, but still a significant step in the right direction.

### Further optimizations

I was surprised that the async version's page load times weren't even faster, and even more surprised to see that the blue rectangles were just as wide as the sync version. Why isn't the async version way faster? The sync version does all of its work on a single thread, and I specifically designed this app to be a nightmare for scalability by having each session kick off by parsing hundreds of megabytes of CSV, an operation that is quite expensive. The async version gets to spread these jobs across several workers. Why aren't we seeing a greater time savings?

Mostly, it's because calling `future_promise(read_csv("big_file.csv"))` is almost a worst-case scenario for future and async. `read_csv` is generally fast, but because the CRAN log files are so big, `read_csv("big_file.csv")` is slow. The value it returns is a very large data frame, that has now been loaded not into the Shiny process, but a `future` worker process. In order to return that data frame to the Shiny process, that data must first be serialized (I believe `future` essentially uses `saveRDS` for this), transmitted to the Shiny process, and then deserialized; to make matters worse, the transmitting and deserialization steps happen on the main R thread that we're working so hard to try to keep idle. **The larger the data we send back and forth to the future, the more performance suffers,** and in this case we're sending back quite a lot of data.

We can make our code significantly faster by doing more summarizing, aggregation, and filtering _inside_ the future; not only does this make more of the work happen in parallel, but by returning the data in already-processed form, we can have much less data to transfer from the worker process back to the Shiny process. (For example, the data for May 31, 2018 weighs 75MB before optimization, and 8.8MB afterwards.)

Compare all three runs in the image below (the newly optimized version is labelled "async2"). The homepage load times have dropped further, and the calculation times are now dramatically faster than the sync code.

![](case-study-gantt-async2.png)

Looking at the "async2" graph, the leading (bottom-left) edge has the same shape as before, as that's simply the rate at which the load testing tool launches new sessions. But notice how much more closely the trailing (upper-right) edge matches the leading edge! It means that even as the number of active sessions ramped up, the amount of latency didn't get dramatically worse, unlike with the "sync" and "async" versions. And each of the individual blue rectangles in the "async2" are comparatively tiny, meaning that users never have to wait more than a dozen seconds at the most for plots to update.

This last plot shows the same data as above, but with the sessions aligned by start time. You can clearly see how the sessions are both shorter and less variable in "async2" compared to the others. I've added a yellow vertical line at the 10 second mark; if the page load (red/pink) has not completed at this point, it's likely that your visitor has left in disgust. While "async" does better than "sync", they both break through the 10 second mark early and often. In contrast, the "async2" version just barely peeks over the line three times.

![](case-study-gantt-aligned.png)

To get a visceral sense for what it feels like to use the app under load, here's a video that shows what it's like to browse the app while the load test is running at its peak. The left side of the screen shows "sync", the right shows "async2". In both cases, I navigated to the app when session #40 was started.

<p class="embed-responsive embed-responsive-16by9">
<iframe class="embed-responsive-item" src="https://www.youtube-nocookie.com/embed/HsjdEZMnb0w?rel=0&amp;showinfo=0&amp;ecver=1" frameborder="0" allow="autoplay; encrypted-media" allowfullscreen></iframe>
</p>

Take a look at the [code diff for async vs. async2](https://github.com/rstudio/cranwhales/compare/async...async2?diff=split). While the code has not changed very dramatically, it has lost a little elegance and maintainability: the code for each of the affected outputs now has one foot in the the render function and one foot in the future. If your app's total audience is a team of a hundred analysts and execs, you may choose to forgo the extra performance and stick with the original async (or even sync) code. But if you have serious scaling needs, the refactoring is probably a small price to pay.

Let's get real for a second, though. If this weren't an example app written for exposition purposes, but a real production app that was intended to scale to thousands of concurrent users across dozens of R processes, we wouldn't download and parse CSV files on the fly. Instead, we'd establish a proper [ETL procedure](https://solutions.rstudio.com/examples/apps/twitter-etl/) to run every night and put the results into a properly indexed database table, or RDS files with just the data we need. As I said [earlier](#improving-performance-and-scalability), a little precomputation and caching can make a huge difference!

Much of the remaining latency for the async2 branch is from ggplot2 plotting. [Sean's talk](https://rstudio.com/resources/) alluded to some upcoming plot caching features we're adding to Shiny, and I imagine they will have as dramatic an effect for this test as they did for Sean.

## Summing up

With async programming, expensive computations and tasks no longer need to be the scalability killers that they once were for Shiny. Armed with this and other common techniques like precomputation, caching, and load balancing, it's possible to write responsive and scalable Shiny applications that can be safely deployed to thousands of concurrent users.

# tee types are handled correctly

    Code
      then(promise_resolve(1), tee = 4)
    Condition
      Error in `then()`:
      ! `tee` must be `TRUE` or `FALSE`

---

    Code
      catch(promise_resolve(1), function(err) { }, tee = "4")
    Condition
      Error in `catch()`:
      ! `tee` must be `TRUE` or `FALSE`

# hybrid_then() validates tee parameter

    Code
      hybrid_then(42, tee = "invalid")
    Condition
      Error in `hybrid_then()`:
      ! `tee` must be `TRUE` or `FALSE`

---

    Code
      hybrid_then(42, tee = 1)
    Condition
      Error in `hybrid_then()`:
      ! `tee` must be `TRUE` or `FALSE`

# hybrid_then() validates `on_success` and `on_failure` parameters

    Code
      hybrid_then(42, on_success = "invalid")
    Condition
      Error in `check_hybrid_callback()`:
      ! `on_success=` must be a function or `NULL`

---

    Code
      hybrid_then(42, on_failure = "invalid")
    Condition
      Error in `check_hybrid_callback()`:
      ! `on_failure=` must be a function or `NULL`

# Deprecated functions: with_ospan_async() is deprecated

    Code
      with_ospan_async("test_span_deprecated", {
        42
      }, tracer = promises_otel_tracer())
    Condition
      Warning:
      `with_ospan_async()` was deprecated in promises 1.5.0.
      i Please use `with_otel_span()` instead.
    Output
      [1] 42

# Deprecated functions: with_ospan_promise_domain() is deprecated

    Code
      with_ospan_promise_domain({
        42
      })
    Condition
      Warning:
      `with_ospan_promise_domain()` was deprecated in promises 1.5.0.
      i Please use `with_otel_promise_domain()` instead.
    Output
      [1] 42

# Deprecated functions: local_ospan_promise_domain() is deprecated

    Code
      local({
        local_ospan_promise_domain()
        42
      })
    Condition
      Warning:
      `local_ospan_promise_domain()` was deprecated in promises 1.5.0.
      i Please use `local_otel_promise_domain()` instead.
    Output
      [1] 42

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/promise.R
\name{promise_resolve}
\alias{promise_resolve}
\alias{promise_reject}
\title{Create a resolved or rejected promise}
\usage{
promise_resolve(value)

promise_reject(reason)
}
\arguments{
\item{value}{A value, or promise, that the new promise should be resolved to.
This expression will be lazily evaluated, and if evaluating the expression
raises an error, then the new promise will be rejected with that error as
the reason.}

\item{reason}{An error message string, or error object.}
}
\description{
Helper functions to conveniently create a promise that is resolved to the
given value (or rejected with the given reason).
}
\examples{
promise_resolve(mtcars) \%...>\%
  head() \%...>\%
  print()

promise_reject("Something went wrong") \%...T!\%
  { message(conditionMessage(.)) }

}

package Promises::Cookbook::ChainingAndPipelining;

# ABSTRACT: Examples of chaining/pipelining of asynchronous operations

__END__

=pod

=head1 NAME

Promises::Cookbook::ChainingAndPipelining - Examples of chaining/pipelining of asynchronous operations

=head1 VERSION

version 0.95

=head1 SYNOPSIS

  my $cv = AnyEvent->condvar;

  fetch_it(
      'http://rest.api.example.com/-/user/search?access_level=admin'
  )->then(
      sub {
          my $admins = shift;
          collect(
              map {
                  fetch_it( 'http://rest.api.example.com/-/user/' . url_encode( $_->{user_id} ) )
              } @$admins
          );
      }
  )->then(
      sub { $cv->send( @_ ) },
      sub { $cv->croak( 'ERROR' ) }
  );

  my @all_admins = $cv->recv;

=head1 DESCRIPTION

So one of the real benefits of the Promise pattern is how it allows
you to write code that flows and reads more like synchronous code
by using the chaining nature of Promises. In example above we are
first fetching a list of users whose access level is 'admin', in
our fictional web-service we get back a list of JSON objects with
only minimal information, just a user_id and full_name for instance.
From here we can then loop through the results and fetch the full
user object for each one of these users, passing all of the promises
returned by C<fetch_it> into C<collect>, which itself returns a promise.

So despite being completely asynchronous, this code reads much like
a blocking synchronous version would read, from top to bottom.

  my @all_admins;
  try {
      my $admins = fetch_it( 'http://rest.api.example.com/-/user/search?access_level=admin' );
      @all_admins = map {
          fetch_it( 'http://rest.api.example.com/-/user/' . url_encode( $_->{user_id} ) )
      } @$admins;
  } catch {
      die $_;
  };
  # do something with @all_admins ...

The only difference really are the C<then> wrappers and the way in
which we handle errors, but even that is very similar since we are
not including an error callback in the first C<then> and allowing
the errors to bubble till the final C<then>, which maps very closely
to the C<try/catch> block. And of course the Promise version runs
asynchronously and reaps all the benefits that brings.

=head1 AUTHOR

Stevan Little <stevan.little@iinteractive.com>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2014 by Infinity Interactive, Inc..

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{promise_map}
\alias{promise_map}
\title{Promise-aware lapply/map}
\usage{
promise_map(.x, .f, ...)
}
\arguments{
\item{.x}{A vector (atomic or list) or an expression object (but not a
promise). Other objects (including classed objects) will be coerced by
base::as.list.}

\item{.f}{The function to be applied to each element of \code{.x}. The function is
permitted, but not required, to return a promise.}

\item{...}{Optional arguments to \code{.f}.}
}
\value{
A promise that resolves to a list (of values, not promises).
}
\description{
Similar to \code{\link[base:lapply]{base::lapply()}} or \code{\link[purrr:map]{purrr::map}}, but promise-aware: the \code{.f}
function is permitted to return promises, and while \code{lapply} returns a list,
\code{promise_map} returns a promise that resolves to a similar list (of resolved
values only, no promises).
}
\details{
\code{promise_map} processes elements of \code{.x} serially; that is, if \code{.f(.x[[1]])}
returns a promise, then \code{.f(.x[[2]])} will not be invoked until that promise
is resolved. If any such promise rejects (errors), then the promise returned
by \code{promise_map} immediately rejects with that err.
}
\examples{
# Waits x seconds, then returns x*10
wait_this_long <- function(x) {
  promise(~later::later(~{
    resolve(x*10)
  }, delay = x))
}

promise_map(list(A=1, B=2, C=3), wait_this_long) \%...>\%
  print()

}

package Promises::Cookbook::GentleIntro;

# ABSTRACT: All you need to know about Promises

__END__

=pod

=head1 NAME

Promises::Cookbook::GentleIntro - All you need to know about Promises

=head1 VERSION

version 0.93

=head1 All you need to know about Promises

=encoding utf8

If you have every done any async programming, you will be familiar with
"callback hell", where one callback calls another, calls another, calls
another... Promises give us back a top-to-bottom coding style, making async
code easier to manage and understand. It looks like synchronous code, but
execution is asynchronous.

The L<Promises> module is event loop agnostic - it can be used with any event
loop. Backends exist for L<AnyEvent> (and thus all the event loops supported
by AnyEvent) and L<Mojo::IOLoop>.  But more of this later in
L</Integration with event loops>.

There are two moving parts:

=over

=item Deferred objects

Deferred objects provide the interface to a specific async request.  They
execute some asynchronous action and return a promise.

=item Promise objects

A promise is like a placeholder for a future result.  The promise will either
be I<resolved> in case of success, or I<rejected> in case of failure. Promises
can be chained together, and each step in the chain is executed sequentially.

=back

The easiest way to understand how Deferred and Promise objects work is by
example.

=head2 Deferred objects

A deferred object is used to signal the success or failure of some async
action which can be implemented in the async library of your choice.  For
instance:

    use Promises qw(deferred);
    use AnyEvent::HTTP qw(http_get);
    use JSON qw(decode_json);

    sub fetch_it {
        my ($uri) = @_;
        my $deferred = deferred->new;
        http_get $uri => sub {
            my ($body, $headers) = @_;
            $headers->{Status} == 200
                ? $deferred->resolve( decode_json($body) )
                : $deferred->reject( $headers->{Reason} )
        };
        $deferred->promise;
    }

The above code makes an asynchronous C<http_get> request to the specified
C<$uri>. The result of the request at the time the subroutine returns is like
Schrödinger's cat: both dead and alive.  In the future it may succeed or it
may fail.

This sub creates a L<Promises::Deferred> object using C<deferred>, which is either:

=over

=item * resolved on success, in which case it returns the request C<body>, or

=item * rejected on failure, in which case it returns the reason for failure.

=back

As a final step, the deferred object returns a L<Promises::Promise> object
which represents the future result.

That's all there is to know about L<Promises::Deferred>.

=head2 Promise objects

Promises are a lot like C<try>/C<catch>/C<finally> blocks except that they can
be chained together. The most important part of a promise is the C<then()>
method:

    $promise->then(
        sub { success! },
        sub { failure }
    );

The C<then()> method takes two arguments: a success callback and a failure
callback.  But the important part is that it returns a B<new> promise, which
is the thing that allows promises to be chained together.

The simple genius of promises (and I can say that because I didn't invent them)
will not be immediately obvious, but bear with me. Promises are very simple,
as long as you understand the execution flow:

=head3 Resolving or rejecting a Promise

    use Promises qw(deferred);

    my $deferred = deferred;
    $deferred->promise->then(
        sub { say "OK! We received: ".shift(@_)},       # on resolve
        sub { say "Bah! We failed with: ". shift(@_)}   # on reject
    );

What this code does depends on what happens to the C<$deferred> object:

    $deferred->resolve('Yay!');
    # prints: "OK! We received: Yay!"

    $deferred->reject('Pooh!');
    # prints "Bah! We failed with: Pooh!"

A Deferred object can only be resolved or rejected once.  Once it is resolved
or rejected, it informs all its promises of the outcome.

=head3 Chaining resolve callbacks

As mentioned earlier, the C<then()> method returns a new promise which will be
resolved or rejected in turn. Each C<resolve> callback will receive the return
value of the previous C<resolve> callback:

    deferred
    ->resolve('red','green')
    ->promise

    ->then(sub {
        # @_ contains ('red','green')
        return ('foo','bar');
    })

    ->then(sub {
        # @_ contains ('foo,bar');
        return 10;
    })

    ->then( sub {
        # @_ contains (10)
    });

All of these example callbacks have just returned a simple value (or values),
so execution has moved from one callback to the next.

=head3 Chaining reject callbacks

Note that in the above example, in each call to C<then()> we specified only a
I<resolved> callback, not a I<rejected> callback.  If a promise is resolved or
rejected, the action gets passed down the chain until it finds a resolved or
rejected handler.  This means that errors can be handled in the appropriate
place in the chain:

    my $deferred = deferred;

    $deferred->promise
    ->then(
        sub {
            my $count = shift();
            say "Count: $count";
            return $count+1;
        }
    )
    ->then(
        sub {
            my $count = shift();
            say "Count: $count";
            return $count+1;
        }
    )->then(
        sub {
            my $count = shift();
            say "Final count: $count";
            return $count+1;
        },
        sub {
            my $reason = shift;
            warn "Failed to count: $reason"
        }
    );

If the C<$deferred> object is resolved, it will call each resolved callback in
turn:

    $deferred->resolve(5);
    # prints:
    #   Count: 5
    #   Count: 6
    #   Final count: 7

If the C<$deferred> object is rejected, however, it will skip all of the steps
in the chain until it hits the first rejected callback:

    $deferred->reject('Poor example');
    # warns:
    #    "Failed to count: Poor example"

B<Important>: Event loops do not like fatal exceptions! For this reason the
I<resolved> and I<rejected> callbacks are run in C<eval> blocks. Exceptions
thrown in either type of callback are passed down the chain to the next
I<rejected> handler.  If there are no more I<rejected> handlers, then the
error is silently swallowed.

=head3 Throwing and handling exceptions

While you can signal success or failure by calling C<resolve()> or C<reject()>
on the C<$deferred> object, you can also signal success or failure in each
step of the promises chain.

=over

=item *

I<Resolved> callbacks are like C<try> blocks: they can either execute some
code successfully or throw an exception.

=item *

I<Rejected> callbacks are like C<catch> blocks: they can either handle the
exception or rethrow it.

=back

    $deferred = deferred;

    $deferred->promise
    ->then(
        sub {
            my $count = shift;
            if ( $count > 100 ) {
                die "Count too high!"
            }
            return $count
        }
    )->then(
        sub {
            say "The count is OK. Continuing";
            return @_
        },
        sub {
            my $error = shift;
            warn "We have a problem: $error";
            die $error;
        }
    )
    )->then(
        undef,  # no resolved handler
        sub {
            return 1;
        }
    )-> then(
        sub {
            my $count = shift;
            say "Got count: $count";
        }
    )

There are a few ways this code can execute. We can resolve the C<$deferred>
object with a reasonable count:

    $deferred->resolve(5);
    # prints:
    #   The count is OK. Continuing
    #   Got count: 5

    $defer

If we reject the C<$deferred> object, the first I<rejected> handler is called.
It warns, then rethrows the exception with C<die> which calls the next
I<rejected> handler.  This handler resolves the exception (that is, it doesn't
call C<die>) and returns a value which gets passed to the next I<resolved>
handler:

    $deferred->reject('For example purposes')
    # warns:
    #    We have a problem: For example purposes
    # prints:
    #    Got count: 1

Finally, if we resolve the C<$deferred> object with a too large count, the
first I<resolved> handler throws an exception, which calls the next
I<rejected> handler:

    $deferred->resolve(1000);
    # warns:
    #    We have a problem: Count too high!
    # prints:
    #    Got count: 1

=head3 C<catch()>

In the above example, we called C<then()> with C<undef> instead of a
I<resolved> callback. This could be rewritten to look a bit cleaner using the
C<catch()> method, which takes just a I<rejected> callback.

    # these two lines are equivalent:
    $promise->then( undef, sub { rejected cb} )
    $promise->catch( sub { rejected cb } )

=head3 C<finally()>

Any C<try>/C<catch> implementation has a C<finally> block, which can be used
to clean up resources regardless of whether the code in the C<try> block
succeeded or failed. Promises offer this functionality too.

The C<finally()> method accepts a single callback which is called regardless
of whether the previous step was resolved or rejected. The return value
(or any exception thrown in the callback) are thrown away, and the chain
continues as if it were not there:

    $deferred = deferred;
    $deferred->promise
    ->then(
        sub {
            my $count = shift;
            if ($count > 10) { die "Count too high"}
            return $count
        }
    )->finally(
        sub { say "Finally got: ".shift(@_) }
    )->then(
        sub { say "OK: ". shift(@_)   },
        sub { say "Bah!: ". shift(@_) }
    );

If we resolve the C<$deferred> object with a good count, we see:

    $d->resolve(5);
    # prints:
    #   Finally got: 5
    #   OK: 5

With a high count we get:

    $d->resolve(20);
    # prints:
    #   Finally got: Count to high
    #   Bah: 20

=head3 Chaining async callbacks

This is where the magic starts: each I<resolved>/I<rejected> handler can not
only return a value (or values), it can also B<return a new Promise>. Remember
that a Promise represents a future value, which means that execution of the
chain will stop until the new Promise has been either resolved or rejected!

For instance, we could write the following code using the C<fetch_it()> sub
(see  L</"Deferred objects">) which returns a promise:

    fetch_it('http://domain.com/user/123')
    ->then(
        sub {
            my $user = shift;
            say "User name: ".$user->{name};
            say "Fetching total comments";
            return fetch_id($user->{total_comments_url});
        }
    )->then(
        sub {
            my $total = shift;
            say "User has left $total comments"
        }
    )
    ->catch(
        sub {
            warn @_
        }
    );

This code sends an asynchronous request to get the page for user C<123> and
returns a promise. Once the promise is resolved, it sends an asynchronous
request to get the total comments for that user and again returns a promise.
Once the second promise is resolved, it prints out the total number of
comments. If either promise were to be rejected, it would skip down the chain
looking for the first I<rejected> handler and execute that.

This is organised to look like synchronous code.  Each step is executed
sequentially, it is easy to read and easy to understand, but it works
asynchronously.  While we are waiting for a response from C<domain.com>
(while our promise remains unfulfilled), the event loop can happily continue
running code elsewhere in the application.

In fact, it's not just L<Promises::Promise> objects that can be returned, it
can be any object that is ``thenable'' (ie it has a C<then()> method). So
if you want to integrate your Promises code with a library which is using
L<Future> objects, you should be able to do it.

=head3 Running async requests in parallel

Sometimes order doesn't matter: perhaps we want to retrieve several web pages
at the same time.  For that we can use the C<collect> helper:

    use Promises qw(collect);

    collect(
        fetch_it('http://rest.api.example.com/-/product/12345'),
        fetch_it('http://rest.api.example.com/-/product/suggestions?for_sku=12345'),
        fetch_it('http://rest.api.example.com/-/product/reviews?for_sku=12345'),
    )->then(
        sub {
            my ($product, $suggestions, $reviews) = @_;
            # do something with these values
        },
        sub { warn @_ }
    );

C<collect()> accepts a list of promises and returns a new promise (which we'll
call C<$p> for clarification purposes.  When all of its promises have been
resolved, it resolves C<$p> with the values returned by every promise, in the
same order as they were passed in to C<collect()>.

B<Note:> Each promise can return multiple values, so C<$product>,
C<$suggestions> and C<$reviews> in the example above will all be array refs.

If any of the passed in promises is rejected, then C<$p> will also be rejected
with the reason for the failure.  C<$p> can only be rejected once, so we wil
only find out about the first failure.

=head2 Integration with event loops

In order to run asynchronous code, you need to run some event loop.  That can
be as simple as using L<AnyEvent/"CONDITION VARIABLES"> to run the event loop
just until a particular condition is met:

    use AnyEvent;

    my $cv = AnyEvent->condvar;
    collect(
        fetch_it('http://rest.api.example.com/-/product/12345'),
        fetch_it('http://rest.api.example.com/-/product/suggestions?for_sku=12345'),
        fetch_it('http://rest.api.example.com/-/product/reviews?for_sku=12345'),
    )->then(
        sub {
            my ($product, $suggestions, $reviews) = @_;
            $cv->send({
                product     => $product->[0],
                suggestions => $suggestions->[0],
                reviews     => $reviews->[0],
            })
        },
        sub { $cv->croak( 'ERROR' ) }
    );

    # wait for $cv->send or $cv->croak
    my $results = $cv->recv;

More usually though, a whole application is intended to be asynchronous, in
which case the event loop just runs continuously. Normally you would only need
to use C<$cv>'s or the equivalent at the point where your application uses a
specific async library, as explained in L</Deferred objects>. The rest of your
code can deal purely with Promises.

=head3 Event loop specific backends

The I<resolved> and I<rejected> callbacks should be run by the event loop,
rather than having one callback call the next, which calls the next etc.

In other words, if a promise is resolved, it doesn't call the I<resolved>
callback directly. Instead it adds it to the event loop's queue, then returns
immediately. The next time the event loop checks its queue, it'll find the
callback in the queue and will call it.

By default, L<Promises> is event loop agnostic, which means that it doesn't
know which event loop to use and so each callback ends up calling the next,
etc.  If you're writing L<Promises>-based modules for CPAN, then your code
should also be event loop agnostic, in which case you want to use Promises
like this:

    use Promises qw(deferred collect);

However, if you are an end user, then you should specify which event loop
you are using at the start of your application:

    use Promises backed => ['AnyEvent']; # or "EV" or "Mojo"

You only need to specify the backend once - any code in the application
which uses L<Promises> will automatically use the specified backend.

=head2 Recursing safely with with C<done()>

One of the cool things about working with promises is that the return value
gets passed down the chain as if we the code were synchronous. However that is
not always what we want.

Imagine that we want to process every line in a file, which could be millions
of lines. We don't care about the results from each line, all we care about is
whether the whole file was processed successfully, or whether something
failed.

In sync code we'd write something like this:

    sub process_file {
        my $fh = shift;
        while (my $line = <$fh>) {
            process_line($line)
                || die "Failed"
        }
    }

Now imagine that C<process_line()> runs asynchronously and returns a promise.
By the time it returns, it probably hasn't executed anything yet.  We can't go
ahead and read the next line of the file otherwise we could generate a billion
promises before any of them has had time to execute.

Instead, we need to wait for C<process_line()> to complete and only then move
on to reading the next line.  We could do this as follows:

    # WARNING: EXAMPLE OF INCORRECT CODE #

    use Promises qw(deferred);

    sub process_file {
        my $fh        = shift;
        my $deferred  = deferred;
        my $processor = sub {
            my $line = <$fh>;
            unless (defined $line) {
                # we're done
                return $deferred->resolve;
            }
            process_line($line)->then(

                # on success, call $processor again
                __SUB__,

                # on failure:
                sub {
                    return $deferred->reject("Failed")
                }
            )
        }

        # start the loop
        $processor->();

        return $deferred->promise
    }

This code has two stack problems. The first is that, every time we process a
line, we recurse into the current C<__SUB__> B<from> the current sub.  This
problem is solved by specifying an L</Event loop specific backend> somewhere
in our application, which we discussed above.

The second problem is that every time we recurse into the current
C<__SUB__> we're waiting for the return value. Other languages use the
L<Tail Call optimization|http://en.wikipedia.org/wiki/Tail_call> to
keep the return stack flat, but we don't have this option.

Instead, we have the C<done()> method which, like C<then()>, accepts a I<resolved> callback
and a I<rejected> callback. But it differs from C<then()> in two ways:

=over

=item *

It doesn't return a promise, which means that the chain ends with the C<done()> step.

=item *

Callbacks are not run in an C<eval> block, so calling C<die()> will throw a
fatal exception. (Most event loops, however will catch the exception, warn,
and continue running.)

=back

The code can be rewritten using C<done()> instead of C<then()> and an event
loop specific backend, and it will happily process millions of lines without
memory leaks or stack oveflows:

    use Promises backend => ['EV'], 'deferred';

    sub process_file {
        my $fh        = shift;
        my $deferred  = deferred;
        my $processor = sub {
            my $line = <$fh>;
            unless (defined $line) {
                # we're done
                return $deferred->resolve;
            }
            #### USE done() TO END THE CHAIN ####
            process_line($line)->done(

                # on success, call $processor again
                __SUB__,

                # on failure:
                sub {
                    return $deferred->reject("Failed")
                }
            )
        }

        # start the loop
        $processor->();

        return $deferred->promise
    }

=head1 AUTHOR

Stevan Little <stevan.little@iinteractive.com>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2014 by Infinity Interactive, Inc..

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut

package Promises::Cookbook::TIMTOWTDI;

# ABSTRACT: Counter examples to Promises

__END__

=pod

=head1 NAME

Promises::Cookbook::TIMTOWTDI - Counter examples to Promises

=head1 VERSION

version 1.00

=head1 DESCRIPTION

So, like I said before, Promises are a means by which you can more
effectively manage your async operations and avoid callback spaghetti.
But of course this is Perl and therefore there is always another way
to do it. In this section I am going to show a few examples of other
ways you could accomplish the same thing.

=head2 Caveat

Please note that I am specifically illustrating ways to do this which
I feel are inferior or less elegant then Promises. This is not meant
to be a slight on the API of other modules at all, I am simply using
these modules to try and illustrate other (perhaps more familiar)
idioms in hopes that it will help people understand Promises.

I am sure there are other ways to do some of these things and do
them more effectively, and I am fully willing to admit my ignorance
here. I welcome any patches which might illustrate said ignorance, as
I do not claim at all to be an expert in async programming.

=head1 AnyEvent::HTTP

So, enough caveating, please consider this (more traditional) version
of our the L<Promises> SYNOPSIS example using L<AnyEvent::HTTP>.

  my $cv = AnyEvent->condvar;

  http_get('http://rest.api.example.com/-/product/12345', sub {
      my ($product) = @_;
      http_get('http://rest.api.example.com/-/product/suggestions?for_sku=12345', sub {
          my ($suggestions) = @_;
          http_get('http://rest.api.example.com/-/product/reviews?for_sku=12345', sub {
              my ($reviews) = @_;
              $cv->send({
                  product     => $product,
                  suggestions => $suggestions,
                  reviews     => $reviews,
              })
          }),
      });
  });

  my $all_product_info = $cv->recv;

Not only do we have deeply nested callbacks, but we have an enforced
order of operations. If you wanted to try and avoid that order of
operations, you might end up writing something like this:

   my $product_cv    = AnyEvent->condvar;
   my $suggestion_cv = AnyEvent->condvar;
   my $review_cv     = AnyEvent->condvar;

   http_get('http://rest.api.example.com/-/product/12345', sub {
       my ($product) = @_;
       $product_cv->send( $product );
   });

   http_get('http://rest.api.example.com/-/product/suggestions?for_sku=12345', sub {
       my ($suggestions) = @_;
       $suggestion_cv->send( $suggestions );
   });

   http_get('http://rest.api.example.com/-/product/reviews?for_sku=12345', sub {
       my ($reviews) = @_;
       $reviews_cv->send( $reviews )
   }),

   my $all_product_info = {
       product     => $product_cv->recv,
       suggestions => $suggestions_cv->recv,
       reviews     => $reviews_cv->recv
   };

But actually, this doesn't work either, while we do gain something by
allowing the C<http_get> calls to be run in whatever order works best,
we still end up still enforcing some order in the way we call C<recv>
on our three C<condvars> (Oh yeah, and we had to create and manage three
C<condvars> as well).

The following example was submitted to me by James Wright (via RT #83992)
as an alternate approach which is non-nested, uses only one condvar, and
has no fixed-order.

  my $cv = AnyEvent->condvar;
  my ( $product, $suggestions, $reviews ) = ( [], [], [] );

  $cv->begin;
  http_get('http://rest.api.example.com/-/product/12345', sub {
      ($product) = @_;
      $cv->end;
  });

  $cv->begin;
  http_get('http://rest.api.example.com/-/product/suggestions?for_sku=12345', sub {
      ($suggestions) = @_;
      $cv->end;
  });

  $cv->begin;
  http_get('http://rest.api.example.com/-/product/reviews?for_sku=12345', sub {
      ($reviews) = @_;
      $cv->end;
  });

  $cv->cb(sub {
      $cv->send({
          product     => $product,
          suggestions => $suggestions,
          reviews     => $reviews,
      });
  });

  my $all_product_info = $cv->recv;

The only real issue I have with this approach is the semi-global variable usage
(C<$product>, C<$suggestions> and C<$reviews>), but otherwise it works fine.

NOTE: Again, if you can think of a better way to do this that I missed,
please let me know.

=head1 Mojo::UserAgent

    #!/usr/bin/env perl

    use Mojo::Base -strict;
    use Mojo::UserAgent;

    my $titles;

    my $ua = Mojo::UserAgent->new;
    Mojo::IOLoop->delay(
        sub {
            my $delay = shift;
            $ua->get('http://google.com/', $delay->begin);
            $ua->get('http://yahoo.com/', $delay->begin);
            $ua->get('http://perlmonks.org/', $delay->begin);
        },
        sub {
            my ($delay, $tx1, $tx2, $tx3) = @_;
            $titles = {
                google      => $tx1->res->dom->at('title')->text,
                yahoo       => $tx2->res->dom->at('title')->text,
                perlmonks   => $tx3->res->dom->at('title')->text,
            };
        },
    )->catch(
        sub {
            my ($delay, $err) = @_;
            warn "failed to download or parse title\n";
        }
    )->wait;

    say Mojo::Util::dumper($titles);

=head1 AUTHOR

Stevan Little <stevan.little@iinteractive.com>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2019, 2017, 2014, 2012 by Infinity Interactive, Inc..

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/promise.R
\name{resolve}
\alias{resolve}
\alias{reject}
\title{Fulfill a promise}
\usage{
resolve(value = NULL)

reject(reason)
}
\arguments{
\item{value}{The result from a successful calculation.}

\item{reason}{An error or string that explains why the operation failed.}
}
\description{
Use these functions to satisfy a promise with either success (\code{resolve})
or failure (\code{reject}). These functions are not exported, but rather, are
passed as arguments to the \code{action} function you pass to a \link{promise}
constructor.
}
\keyword{internal}

package Promises::Cookbook::Recursion;

# ABSTRACT: Examples of recursive asynchronous operations

__END__

=pod

=head1 NAME

Promises::Cookbook::Recursion - Examples of recursive asynchronous operations

=head1 VERSION

version 1.00

=head1 SYNOPSIS

    package MyClass;

    use Promises backend => ['AE'], 'deferred';

    sub new                 {...}
    sub process             {...}
    sub is_finished         {...}
    sub fetch_next_from_db  {...} # returns a promise

    sub fetch_all {
        my $self     = shift;
        my $deferred = deferred;

        $self->_fetch_loop($deferred);
        return $deferred->promise;
    }

    sub _fetch_loop {
        my ($self,$deferred) = @_;
        if ( $self->is_finished ) {
            $deferred->resolve;
            return
        }
        $self->fetch_next_from_db
             ->then( sub { $self->process(@_) })
             ->done(
                sub { $self->_fetch_loop($deferred) }
                sub { $deferred->reject(@_) }
             );
    }

    package main;

    my $cv  = AnyEvent->condvar;
    my $obj = MyClass->new(...);
    $obj->fetch_all->then(
        sub { $cv->send(@_)          },
        sub { $cv->croak('ERROR',@_) }
    );

    $cv->recv;

=head1 DESCRIPTION

While C<collect()> allows you to wait for multiple promises which
are executing in parallel, sometimes you need to execute each step
in order, by using promises recursively. For instance:

=over

=item 1

Fetch next page of results

=item 2

Process page of results

=item 3

If there are no more results, return success

=item 4

Otherwise, goto step 1

=back

However, recursion can result in very deep stacks and out of memory
conditions.  There are two important steps for dealing with recursion
effectively.

The first is to use one of the event-loop backends:

    use Promises backend => ['AE'], 'deferred';

While the default L<Promises::Deferred> implementation calls the
C<then()> callbacks synchronously, the event-loop backends call
the callbacks asynchronously in the context of the event loop.

However, each C<promise> passes its return value on to the next
C<promise> etc, so you still end up using a lot of memory with
recursion. We can avoid this by breaking the chain.

In our example, all we care about is whether all the steps in our
process completed successfully or not.  Each execution of steps 1 to
4 is independent. Step 1 does not need to receive the return value
from step 4.

We can break the chain by using C<done()> instead of C<then()>.
While C<then()> returns a new C<promise> to continue the chain,
C<done()> will execute either the success callback or the
error callback and return an empty list, breaking the chain and
rolling back the stack.

To work through the code in the L</SYNOPSIS>:

    sub fetch_all {
        my $self     = shift;
        my $deferred = deferred;

        $self->_fetch_loop($deferred);
        return $deferred->promise;
    }

The C<$deferred> variable (and the promise that we return to the caller)
will either be resolved once all results have been fetched and
processed by the C<_fetch_loop()>, or rejected if an error occurs at
any stage of execution.

    sub _fetch_loop {
        my ($self,$deferred) = @_;

        if ( $self->is_finished ) {
            $deferred->resolve;
            return;
        }

If C<is_finished> returns a true value (eg there are no more results to fetch),
then we can resolve our promise, indicating success, and exit the loop.

        $self->fetch_next_from_db
             ->then( sub { $self->process(@_) })
             ->done(
                sub { $self->_fetch_loop($deferred) }
                sub { $deferred->reject(@_) }
             );
    }

Otherwise we fetch the next page of results aynchronously from the DB and
process them. If either of these steps (fetching or processing) fails,
then we signal failure by rejecting our deferred promise and exiting the loop.
If there is no failure, we recurse back into our loop by calling
C<_fetch_loop()> again.

However,this recursion happens asynchronously. What this code actually does
is to schedule the call to C<_fetch_loop()> in the next tick of the event
loop. And because we used C<done()> instead of C<then()>, we don't
wait around for the return result but instead return immediately,
exiting the current execution, discarding the return results and
rolling back the stack.

=head1 AUTHOR

Stevan Little <stevan.little@iinteractive.com>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2019, 2017, 2014, 2012 by Infinity Interactive, Inc..

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut

package Promises::Cookbook::ScalaFuturesComparison;

# ABSTRACT: A comparison of Scala Futures with Promises

__END__

=pod

=head1 NAME

Promises::Cookbook::ScalaFuturesComparison - A comparison of Scala Futures with Promises

=head1 VERSION

version 0.95

=head1 DESCRIPTION

Here is the example Scala code, it assumes a function called C<fetch> 
which when given a URL will return a Future.

    def getThumbnail(url: String): Future[Webpage] = {
        val promise = new Promise[Webpage]
        fetch(url) onSuccess { page =>
            fetch(page.imageLinks(0)) onSuccess { p =>
                promise.setValue(p)
            } onFailure { exc =>
                promise.setException(exc)
            }
        } onFailure { exc =>
            promise.setException(exc)
        }
        promise
    }

If we take this and translate this into Perl code using the 
L<Mojo::UserAgent> library, the C<fetch> function would look 
like this:

    sub fetch {
        state $ua = Mojo::UserAgent->new;
        my $url   = shift;
        my $d     = deferred;
        $ua->get($url => sub {
            my ($ua, $tx) = @_;
            $d->resolve( $tx );
        });
        $d->promise;
    }

And if we were to take the C<get_thumbnail> function and 
translate it exactly, we would end up with this:

    sub get_thumbnail {
        my $url = shift;
        my $d   = deferred;
        fetch( $url )->then(
            sub {
                my $tx = shift;
                fetch( $tx->res->dom->find('img')->[0]->{'src'} )->then(
                    sub { $d->resolve( $_[0] ) },
                    sub { $d->reject( $_[0] ) },                
                )
            },
            sub { $d->reject( $_[0] ) }
        );
        $d->promise;
    }

Scala Futures have a method called C<flatMap>, which takes a 
function that given value will return another Future. Here is 
an example of how the C<getThumbnail> method can be simplified
by using it.

    def getThumbnail(url: String): Future[Webpage] =
        fetch(url) flatMap { page =>
             fetch(page.imageLinks(0))
        }

But since our C<then> method actually creates a new promise 
and wraps the callbacks to chain to that promise, we don't 
need this C<flatMap> combinator and so this, Just Works.

    sub get_thumbnail {
        my $url = shift;
        fetch( $url )->then(
            sub {
                my $tx = shift;
                fetch( $tx->res->dom->find('img')->[0]->{'src'} );
            }        
        );
    }

Scala Futures also have a C<rescue> method which can serve as 
a kind of catch block that potentially will return another 
Future.

    val f = fetch(url) rescue {
        case ConnectionFailed =>
          fetch(url)
    }

Just as with C<flatMap>, since our callbacks are wrapped and 
chained with a new Promise, we can do a rescue just by using 
the error callback The Promise returned by C<fetch> will get 
chained and so this will depend on it.

    sub get_thumbnail {
        my $url = shift;
        fetch( $url )->then(
            sub {
                my $page = shift;
                fetch( $page->image_links->[0] );
            },
            sub {
                given ( $_[0] ) {
                    when ('connection_failed') {
                        return fetch( $url );
                    }
                    default {
                        return "failed";
                    }
                }
            }
        );
    }

TODO ... figure out how retry can be generic ...

=head1 SEE ALSO

Systems Programming at Twitter - L<http://monkey.org/~marius/talks/twittersystems/>

=head1 AUTHOR

Stevan Little <stevan.little@iinteractive.com>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2014 by Infinity Interactive, Inc..

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/then.R
\name{then}
\alias{then}
\alias{catch}
\alias{finally}
\title{Access the results of a promise}
\usage{
then(promise, onFulfilled = NULL, onRejected = NULL)

catch(promise, onRejected, tee = FALSE)

finally(promise, onFinally)
}
\arguments{
\item{promise}{A promise object. The object can be in any state.}

\item{onFulfilled}{A function (or a formula--see Details) that will be
invoked if the promise value successfully resolves. When invoked, the
function will be called with a single argument: the resolved value.
Optionally, the function can take a second parameter \code{.visible} if you care
whether the promise was resolved with a visible or invisible value. The
function can return a value or a promise object, or can throw an error;
these will affect the resolution of the promise object that is returned
by \code{then()}.}

\item{onRejected}{A function taking the argument \code{error} (or a formula--see
Details). The function can return a value or a promise object, or can throw
an error. If \code{onRejected} is provided and doesn't throw an error (or return
a promise that fails) then this is the async equivalent of catching an
error.}

\item{tee}{If \code{TRUE}, ignore the return value of the callback, and use the
original value instead. This is useful for performing operations with
side-effects, particularly logging to the console or a file. If the
callback itself throws an error, and \code{tee} is \code{TRUE}, that error will still
be used to fulfill the the returned promise (in other words, \code{tee} only has
an effect if the callback does not throw).}

\item{onFinally}{A function with no arguments, to be called when the async
operation either succeeds or fails. Usually used for freeing resources that
were used during async operations.}
}
\description{
Use the \code{then} function to access the eventual result of a promise (or, if the operation fails, the reason for that failure). Regardless of the state of the promise, the call to \code{then} is non-blocking, that is, it returns immediately; so what it does \emph{not} do is immediately return the result value of the promise. Instead, you pass logic you want to execute to \code{then}, in the form of function callbacks (or formulas, see Details). If you provide an \code{onFulfilled} callback, it will be called upon the promise's successful resolution, with a single argument \code{value}: the result value. If you provide an \code{onRejected} callback, it will be called if the operation fails, with a single argument \code{reason}: the error that caused the failure.
}
\section{Formulas}{


For convenience, the \code{then()}, \code{catch()}, and \code{finally()} functions use
\code{\link[rlang:as_function]{rlang::as_function()}} to convert \code{onFulfilled}, \code{onRejected}, and
\code{onFinally} arguments to functions. This means that you can use formulas to
create very compact anonymous functions, using \code{.} to access the value (in
the case of \code{onFulfilled}) or error (in the case of \code{onRejected}).
}

\section{Chaining promises}{


The first parameter of \code{then} is a promise; given the stated purpose of the
function, this should be no surprise. However, what may be surprising is that
the return value of \code{then} is also a (newly created) promise. This new
promise waits for the original promise to be fulfilled or rejected, and for
\code{onFulfilled} or \code{onRejected} to be called. The result of (or error raised
by) calling \code{onFulfilled}/\code{onRejected} will be used to fulfill (reject) the
new promise.\preformatted{promise_a <- get_data_frame_async()
promise_b <- then(promise_a, onFulfilled = head)
}

In this example, assuming \code{get_data_frame_async} returns a promise that
eventually resolves to a data frame, \code{promise_b} will eventually resolve to
the first 10 or fewer rows of that data frame.

Note that the new promise is considered fulfilled or rejected based on
whether \code{onFulfilled}/\code{onRejected} returns a value or throws an error, not on
whether the original promise was fulfilled or rejected. In other words, it's
possible to turn failure to success and success to failure. Consider this
example, where we expect \code{some_async_operation} to fail, and want to consider
it an error if it doesn't:\preformatted{promise_c <- some_async_operation()
promise_d <- then(promise_c,
  onFulfilled = function(value) \{
    stop("That's strange, the operation didn't fail!")
  \},
  onRejected = function(reason) \{
    # Great, the operation failed as expected
    NULL
  \}
)
}

Now, \code{promise_d} will be rejected if \code{promise_c} is fulfilled, and vice
versa.

\strong{Warning:} Be very careful not to accidentally turn failure into success,
if your error handling code is not the last item in a chain!\preformatted{some_async_operation() \%>\%
  catch(function(reason) \{
    warning("An error occurred: ", reason)
  \}) \%>\%
  then(function() \{
    message("I guess we succeeded...?")  # No!
  \})
}

In this example, the \code{catch} callback does not itself throw an error, so the
subsequent \code{then} call will consider its promise fulfilled!
}

\section{Convenience functions}{


For readability and convenience, we provide \code{catch} and \code{finally} functions.

The \code{catch} function is equivalent to \code{then}, but without the \code{onFulfilled}
argument. It is typically used at the end of a promise chain to perform error
handling/logging.

The \code{finally} function is similar to \code{then}, but takes a single no-argument
function (or formula) that will be executed upon completion of the promise,
regardless of whether the result is success or failure. It is typically used
at the end of a promise chain to perform cleanup tasks, like closing file
handles or database connections. Unlike \code{then} and \code{catch}, the return value
of \code{finally} is ignored; however, if an error is thrown in \code{finally}, that
error will be propagated forward into the returned promise.
}

\section{Visibility}{


\code{onFulfilled} functions can optionally have a second parameter \code{visible},
which will be \code{FALSE} if the result value is \link[base:invisible]{invisible}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/promise.R
\name{promise}
\alias{promise}
\title{Create a new promise object}
\usage{
promise(action)
}
\arguments{
\item{action}{A function with signature \verb{function(resolve, reject)}.}
}
\value{
A promise object (see \code{\link{then}}).
}
\description{
\code{promise()} creates a new promise. A promise is a placeholder object for the
eventual result (or error) of an asynchronous operation. This function is not
generally needed to carry out asynchronous programming tasks; instead, it is
intended to be used mostly by package authors who want to write asynchronous
functions that return promises.
}
\details{
The \code{action} function should be a piece of code that returns quickly, but
initiates a potentially long-running, asynchronous task. If/when the task
successfully completes, call \code{resolve(value)} where \code{value} is the result of
the computation (like the return value). If the task fails, call
\code{reject(reason)}, where \code{reason} is either an error object, or a character
string.

It's important that asynchronous tasks kicked off from \code{action} be coded very
carefully--in particular, all errors must be caught and passed to \code{reject()}.
Failure to do so will cause those errors to be lost, at best; and the caller
of the asynchronous task will never receive a response (the asynchronous
equivalent of a function call that never returns, i.e. hangs).

The return value of \code{action} will be ignored.
}
\section{\verb{action=} formulas}{


\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

With \code{{promises}} depending on R >= 4.1, the shorthand of a formula, \code{~ fn(.)} for \code{action} is no longer recommended by the \code{{promises}} package or
tidyverse (for example,
\href{https://github.com/tidyverse/purrr/commit/670c3ed9920f15da0d4175068ecddc41f0f1f335#diff-c4dcc43795da5c7f6bf5a94d957b5507ce795fedd6d3eb092ccad03678c4f76dR15}{\code{{purrr}}})
as we now have access to the function shorthand, \verb{\\(x) fn(x)}. Please update
your \code{action} code to use the new function shorthand syntax \verb{\\(resolve, reject) resolve(arg1, arg2)} instead of \code{~ { resolve(arg1, arg2) }}. The
magically created \code{resolve}/\code{reject} functions can be confusing when chained
with other methods.
}

\examples{
# Create a promise that resolves to a random value after 2 secs
p1 <- promise(\(resolve, reject) {
  later::later(\() resolve(runif(1)), delay = 2)
})

p1 |> then(print)

# Create a promise that errors immediately
p2 <- promise(\(resolve, reject) {
  reject("An error has occurred")
})
then(p2,
  onFulfilled = \(value) message("Success"),
  onRejected = \(err) message("Failure")
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipe.R
\name{pipes}
\alias{pipes}
\alias{\%...>\%}
\alias{\%...T>\%}
\alias{\%...!\%}
\alias{\%...T!\%}
\title{Promise pipe operators}
\usage{
lhs \%...>\% rhs

lhs \%...T>\% rhs

lhs \%...!\% rhs

lhs \%...T!\% rhs
}
\arguments{
\item{lhs}{A promise object.}

\item{rhs}{A function call using the magrittr semantics. It can return either
a promise or non-promise value, or throw an error.}
}
\value{
A new promise.
}
\description{
Promise-aware pipe operators, in the style of \href{https://CRAN.R-project.org/package=magrittr/vignettes/magrittr.html}{magrittr}.
Like magrittr pipes, these operators can be used to chain together pipelines
of promise-transforming operations. Unlike magrittr pipes, these pipes wait
for promise resolution and pass the unwrapped value (or error) to the \code{rhs}
function call.
}
\details{
The \code{>} variants are for handling successful resolution, the \code{!} variants are
for handling errors. The \code{T} variants of each return the lhs instead of the
rhs, which is useful for pipeline steps that are used for side effects
(printing, plotting, saving).
\enumerate{
\item \code{promise \%...>\% func()} is equivalent to \code{promise \%>\% then(func)}.
\item \code{promise \%...!\% func()} is equivalent to \code{promise \%>\% catch(func)}.
\item \code{promise \%...T>\% func()} is equivalent to \code{promise \%T>\% then(func)}.
\item \code{promise \%...T!\% func()} is equivalent to \code{promise \%T>\%
catch(func)} or \code{promise \%>\% catch(func, tee = TRUE)}.
}

One situation where 3. and 4. above break down is when \code{func()} throws an
error, or returns a promise that ultimately fails. In that case, the failure
will be propagated by our pipe operators but not by the
magrittr-plus-function "equivalents".

For simplicity of implementation, we do not support the magrittr feature of
using a \code{.} at the head of a pipeline to turn the entire pipeline into a
function instead of an expression.
}
\examples{
\dontrun{
library(future)
plan(multisession)

future_promise(cars) \%...>\%
  head(5) \%...T>\%
  print()

# If the read.csv fails, resolve to NULL instead
future_promise(read.csv("http://example.com/data.csv")) \%...!\%
  { NULL }
}

}
\seealso{
https://rstudio.github.io/promises/articles/overview.html#using-pipes
}

package Promises::Cookbook::SynopsisBreakdown;

# ABSTRACT: A breakdown of the SYNOPSIS section of Promises

__END__

=pod

=head1 NAME

Promises::Cookbook::SynopsisBreakdown - A breakdown of the SYNOPSIS section of Promises

=head1 VERSION

version 1.00

=head1 SYNOPSIS

  use AnyEvent::HTTP;
  use JSON::XS qw[ decode_json ];
  use Promises qw[ collect deferred ];

  sub fetch_it {
      my ($uri) = @_;
      my $d = deferred;
      http_get $uri => sub {
          my ($body, $headers) = @_;
          $headers->{Status} == 200
              ? $d->resolve( decode_json( $body ) )
              : $d->reject( $body )
      };
      $d->promise;
  }

  my $cv = AnyEvent->condvar;

  collect(
      fetch_it('http://rest.api.example.com/-/product/12345'),
      fetch_it('http://rest.api.example.com/-/product/suggestions?for_sku=12345'),
      fetch_it('http://rest.api.example.com/-/product/reviews?for_sku=12345'),
  )->then(
      sub {
          my ($product, $suggestions, $reviews) = @_;
          $cv->send({
              product     => $product,
              suggestions => $suggestions,
              reviews     => $reviews,
          })
      },
      sub { $cv->croak( 'ERROR' ) }
  );

  my $all_product_info = $cv->recv;

=head1 DESCRIPTION

The example in the synopsis actually demonstrates a number of the
features of this module, this section will break down each part
and explain them in order.

  sub fetch_it {
      my ($uri) = @_;
      my $d = deferred;
      http_get $uri => sub {
          my ($body, $headers) = @_;
          $headers->{Status} == 200
              ? $d->resolve( decode_json( $body ) )
              : $d->reject( $body )
      };
      $d->promise;
  }

First is the C<fetch_it> function, the pattern within this function
is the typical way in which you might wrap an async function call
of some kind. The first thing we do it to create an instance of
L<Promises::Deferred> using the C<deferred> function, this is the
class which does the majority of the work or managing callbacks
and the like. Then within the callback for our async function,
we will call methods on the L<Promises::Deferred> instance. In the
case we first check the response headers to see if the request was
a success, if so, then we call the C<resolve> method and pass the
decoded JSON to it. If the request failed, we then call the C<reject>
method and send back the data from the body. Finally we call the
C<promise> method and return the promise 'handle' for this deferred
instance.

At this point out asynchronous operation will typically be in
progress, but control has been returned to the rest of our
program. Now, before we dive into the rest of the example, lets
take a quick detour to look at what promises do. Take the following
code for example:

  my $p = fetch_it('http://rest.api.example.com/-/user/bob@example.com');

At this point, our async operation is running, but we have not yet
given it anything to do when the callback is fired. We will get to
that shortly, but first lets look at what information we can get
from the promise.

  $p->status;

Calling the C<status> method will return a string representing the
status of the promise. This will be either I<in progress>, I<resolved>,
I<resolving> (meaning it is in the process of resolving), I<rejected>
or I<rejecting> (meaning it is in the process of rejecting).
(NOTE: these are also constants on the L<Promises::Deferred> class,
C<IN_PROGRESS>, C<RESOLVED>, C<REJECTED>, etc., but they are also
available as predicate methods in both the L<Promises::Deferred> class
and proxied in the L<Promises::Promise> class). At this point, this
method call is likely to return I<in progress>. Next is the C<result>
method:

  $p->result;

which will give us back the values that are passed to either C<resolve>
or C<reject> on the associated L<Promises::Deferred> instance.

Now, one thing to keep in mind before we go any further is that our
promise is really just a thin proxy over the associated L<Promises::Deferred>
instance, it stores no state itself, and when these methods are called on
it, it simply forwards the call to the associated L<Promises::Deferred>
instance (which, as I said before, is where all the work is done).

So, now, lets actually do something with this promise. So as I said above
the goal of the Promise pattern is to reduce the callback spaghetti that
is often created with writing async code. This does not mean that we have
no callbacks at all, we still need to have some kind of callback, the
difference is all in how those callbacks are managed and how we can more
easily go about providing some level of sequencing and control.

That all said, lets register a callback with our promise.

  $p->then(
      sub {
          my ($user) = @_;
          do_something_with_a_user( $user );
      },
      sub {
          my ($err) = @_;
          warn "An error was received : $err";
      }
  );

As you can see, we use the C<then> method (again, keep in mind this is
just proxying to the associated L<Promises::Deferred> instance) and
passed it two callbacks, the first is for the success case (if C<resolve>
has been called on our associated L<Promises::Deferred> instance) and
the second is the error case (if C<reject> has been called on our
associated L<Promises::Deferred> instance). Both of these callbacks will
receive the arguments that were passed to C<resolve> or C<reject> as
their only arguments, as you might have guessed, these values are the
same values you would get if you called C<result> on the promise
(assuming the async operation was completed).

It should be noted that the error callback is optional. If it is not
specified then errors will be silently eaten (similar to a C<try> block
that has not C<catch>). If there is a chain of promises however, the
error will continue to bubble to the last promise in the chain and
if there is an error callback there, it will be called. This allows
you to concentrate error handling in the places where it makes the most
sense, and ignore it where it doesn't make sense. As I alluded to above,
this is very similar to nested C<try/catch> blocks.

And really, that's all there is to it. You can continue to call C<then>
on a promise and it will continue to accumulate callbacks, which will
be executed in FIFO order once a call is made to either C<resolve>
or C<reject> on the associated L<Promises::Deferred> instance. And in
fact, it will even work after the async operation is complete. Meaning
that if you call C<then> and the async operation is already completed,
your callback will be executed immediately.

So, now lets get back to our original example. I will briefly explain
my usage of the L<AnyEvent> C<condvar>, but I encourage you to review
the docs for L<AnyEvent> yourself if my explanation is not enough.

So, the idea behind my usage of the C<condvar> is to provide a
merge-point in my code at which point I want all the asynchronous
operations to converge, after which I can resume normal synchronous
programming (if I so choose). It provides a kind of a transaction
wrapper if you will, around my async operations. So, first step is
to actually create that C<condvar>.

  my $cv = AnyEvent->condvar;

Next, we jump back into the land of Promises. Now I am breaking apart
the calling of C<collect> and the subsequent chained C<then> call here
to help keep things in digestible chunks, but also to illustrate that
C<collect> just returns a promise (as you might have guessed anyway).

  my $p = collect(
      fetch_it('http://rest.api.example.com/-/product/12345'),
      fetch_it('http://rest.api.example.com/-/product/suggestions?for_sku=12345'),
      fetch_it('http://rest.api.example.com/-/product/reviews?for_sku=12345'),
  );

So, what is going on here is that we want to be able to run multiple
async operations in parallel, but we need to wait for all of them to
complete before we can move on, and C<collect> gives us that ability.
As we know from above, C<fetch_it> is returning a promise, so obviously
C<collect> takes an array of promises as its parameters. As we said before
C<collect> also returns a promise, which is just a handle on a
C<Promises::Deferred> instance it created to watch and handle the
multiple promises you passed it. Okay, so now lets move onto adding
callbacks to our promise that C<collect> returned to us.

  $p->then(
      sub {
          my ($product, $suggestions, $reviews) = @_;
          $cv->send({
              product     => $product,
              suggestions => $suggestions,
              reviews     => $reviews,
          })
      },
      sub { $cv->croak( 'ERROR' ) }
  );

So, you will notice that, as before, we provide a success and an error
callback, but you might notice one slight difference in the success
callback. It is actually being passed multiple arguments, these are
the results of the three C<fetch_it> calls passed into C<collect>, and yes,
they are passed to the callback in the same order you passed them into
C<collect>. So from here we jump back into the world of C<condvars>, and
we call the C<send> method and pass it our newly assembled set of
collected product info. As I said above, C<condvars> are a way of
wrapping your async operations into a transaction like block, when
code execution encounters a C<recv>, such as in our next line of code:

  my $all_product_info = $cv->recv;

the event loop will block until a corresponding C<send> is called on
the C<condvar>. While you are not required to pass arguments to C<send>
it will accept them and the will in turn be the return values of
the corresponding C<recv>, which makes for an incredibly convenient
means of passing data around your asynchronous program.

It is also worth noting the usage of the C<croak> method on the
C<condvar> in the error callback. This is the preferred way of
dealing with exceptions in L<AnyEvent> because it will actually
cause the exception to be thrown from C<recv> and not somewhere
deep within a callback.

And that is all of it, once C<recv> returns, our program will go
back to normal synchronous operation and we can do whatever it is
we like with C<$all_product_info>.

=head1 AUTHOR

Stevan Little <stevan.little@iinteractive.com>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2019, 2017, 2014, 2012 by Infinity Interactive, Inc..

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/promises-package.R
\docType{package}
\name{promises-package}
\alias{promises}
\alias{promises-package}
\title{promises: Abstractions for Promise-Based Asynchronous Programming}
\description{
Provides fundamental abstractions for doing asynchronous programming in R using promises. Asynchronous programming is useful for allowing a single R process to orchestrate multiple tasks in the background while also attending to something else. Semantics are similar to 'JavaScript' promises, but with a syntax that is idiomatic R.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://rstudio.github.io/promises/}
  \item \url{https://github.com/rstudio/promises}
  \item Report bugs at \url{https://github.com/rstudio/promises/issues}
}

}
\author{
\strong{Maintainer}: Barret Schloerke \email{barret@posit.co} (\href{https://orcid.org/0000-0001-9986-114X}{ORCID})

Authors:
\itemize{
  \item Joe Cheng \email{joe@posit.co}
  \item Winston Chang \email{winston@posit.co} (\href{https://orcid.org/0000-0002-1576-2126}{ORCID})
  \item Charlie Gao \email{charlie.gao@posit.co} (\href{https://orcid.org/0000-0002-0750-061X}{ORCID})
}

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/domains.R
\name{with_promise_domain}
\alias{with_promise_domain}
\alias{new_promise_domain}
\title{Promise domains}
\usage{
with_promise_domain(domain, expr, replace = FALSE)

new_promise_domain(
  wrapOnFulfilled = identity,
  wrapOnRejected = identity,
  wrapSync = force,
  onError = force,
  ...,
  wrapOnFinally = NULL
)
}
\arguments{
\item{domain}{A promise domain object to install while \code{expr} is evaluated.}

\item{expr}{Any R expression, to be evaluated under the influence of
\code{domain}.}

\item{replace}{If \code{FALSE}, then the effect of the \code{domain} will be added
to the effect of any currently active promise domain(s). If \code{TRUE}, then
the current promise domain(s) will be ignored for the duration of the
\code{with_promise_domain} call.}

\item{wrapOnFulfilled}{A function that takes a single argument: a function
that was passed as an \code{onFulfilled} argument to \code{\link[=then]{then()}}. The
\code{wrapOnFulfilled} function should return a function that is suitable for
\code{onFulfilled} duty.}

\item{wrapOnRejected}{A function that takes a single argument: a function
that was passed as an \code{onRejected} argument to \code{\link[=then]{then()}}. The
\code{wrapOnRejected} function should return a function that is suitable for
\code{onRejected} duty.}

\item{wrapSync}{A function that takes a single argument: a (lazily evaluated)
expression that the function should \code{\link[=force]{force()}}. This expression represents
the \code{expr} argument passed to \code{\link[=with_promise_domain]{with_promise_domain()}}; \code{wrapSync} allows
the domain to manipulate the environment before/after \code{expr} is evaluated.}

\item{onError}{A function that takes a single argument: an error. \code{onError}
will be called whenever an exception occurs in a domain (that isn't caught
by a \code{tryCatch}). Providing an \code{onError} callback doesn't cause errors to
be caught, necessarily; instead, \code{onError} callbacks behave like calling
handlers.}

\item{...}{Arbitrary named values that will become elements of the promise
domain object, and can be accessed as items in an environment (i.e. using
\code{[[} or \code{$}).}

\item{wrapOnFinally}{A function that takes a single argument: a function
that was passed as an \code{onFinally} argument to \code{\link[=then]{then()}}. The
\code{wrapOnFinally} function should return a function that is suitable for
\code{onFinally} duty. If \code{wrapOnFinally} is \code{NULL} (the default), then the
domain will use both \code{wrapOnFulfilled} and \code{wrapOnRejected} to wrap the
\code{onFinally}. If it's important to distinguish between normal
fulfillment/rejection handlers and finally handlers, then be sure to
provide \code{wrapOnFinally}, even if it's just \code{\link[base:identity]{base::identity()}}.}
}
\description{
Promise domains are used to temporarily set up custom environments that
intercept and influence the registration of callbacks. Create new promise
domain objects using \code{new_promise_domain}, and temporarily activate a promise
domain object (for the duration of evaluating a given expression) using
\code{with_promise_domain}.
}
\details{
While \code{with_promise_domain} is on the call stack, any calls to \code{\link[=then]{then()}} (or
higher level functions or operators, like \code{\link[=catch]{catch()}} or the various \link{pipes})
will belong to the promise domain. In addition, when a \code{then} callback that
belongs to a promise domain is invoked, then any new calls to \code{then} will
also belong to that promise domain. In other words, a promise domain
"infects" not only the immediate calls to \code{then}, but also to "nested" calls
to \code{then}.

For more background, read the
\href{https://gist.github.com/jcheng5/b1c87bb416f6153643cd0470ac756231}{original design doc}.

For examples, see the source code of the Shiny package, which uses promise
domains extensively to manage graphics devices and reactivity.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/otel.R
\name{with_otel_span}
\alias{with_otel_span}
\alias{with_otel_promise_domain}
\alias{local_otel_promise_domain}
\title{OpenTelemetry integration}
\usage{
with_otel_span(name, expr, ..., tracer, attributes = NULL)

with_otel_promise_domain(expr)

local_otel_promise_domain(envir = parent.frame())
}
\arguments{
\item{name}{Character string. The name of the otel span.}

\item{expr}{An expression to evaluate within the otel span context.}

\item{...}{Additional arguments passed to \code{\link[otel:start_span]{otel::start_span()}}.}

\item{tracer}{(Required) An \code{{otel}} tracer. It is required to provide your
own tracer from your own package. See \code{\link[otel:get_tracer]{otel::get_tracer()}} for more
details.}

\item{attributes}{Attributes passed through \code{\link[otel:as_attributes]{otel::as_attributes()}} (when
not \code{NULL})}

\item{envir}{The "local" environment in which to add the promise domain. When
the environment is exited, the promise domain is removed.}
}
\description{
\pkg{otel} provides tools for integrating with OpenTelemetry, a framework for
observability and tracing in distributed systems.

These methods are intended to enhance the framework to be used with the
\pkg{promises} package, not as a generic replacement.

Developer note - Barret 2025/09: This otel span handoff promise domain topic is
complex and has been discussed over many hours. Many advanced Shiny/R
developers are not even aware of promise domains (very reasonable!),
therefore this topic requires more in-depth documentation and examples.
}
\section{Functions}{
\itemize{
\item \code{with_otel_span()}: Creates an OpenTelemetry span, executes the given expression within it, and
ends the span.

This method \strong{requires} the use of \code{with_otel_promise_domain()}
to be within the execution stack.

This function is designed to handle both synchronous and asynchronous
(promise-based) operations. For promises, the span is automatically ended
when the promise resolves or rejects.

Returns the result of evaluating \code{expr}. If \code{expr} returns a promise, the
span will be automatically ended when the promise completes.

This function differs from synchronous otel span operations in that it
installs a promise domain and properly handles asynchronous operations. In
addition, the internal span will be ended either when the function exits (for
synchronous operations) or when a returned promise completes (for
asynchronous operations).

If OpenTelemetry is not enabled, the expression will be evaluated without any
tracing context.

\item \code{with_otel_promise_domain()}: Adds an idempotent handoff Active OpenTelemetry span promise domain.

Package authors are required to use this function to have otel span context
persist across asynchronous boundaries. This method is only needed once per
promise domain stack. If you are unsure, feel free to call
\code{with_otel_promise_domain()} as the underlying promise domain will only be
added if not found within the current promise domain stack. If your package
\strong{only} works within Shiny apps, Shiny will have already added the domain so
no need to add it yourself. If your package works outside of Shiny and you
use \code{{promises}} (i.e. \code{{chromote}}), then you'll need to use this wrapper
method.

This method adds a \emph{handoff} Active OpenTelemetry span promise domain to the
expression evaluation. This \emph{handoff} promise domain will only run once on
reactivation. This is critical if there are many layered
\code{with_otel_span()} calls, such as within Shiny reactivity. For example, if
we nested many \code{with_otel_span()} calls of which each call added a promise
domain that reactivated each otel span on restore, we'd reactivate \code{k} otel
span objects (\code{O(k)}) when we only need to activate the \strong{last} span
(\code{O(1)}).

Returns the result of evaluating \code{expr} within the otel span promise domain.

\item \code{local_otel_promise_domain()}: Local OpenTelemetry span promise domain

Adds an OpenTelemetry span promise domain to the local scope. This is useful
for \code{{coro}} operations where encapsulating the coro operations inside a
\verb{with_*()} methods is not allowed.

When not using \code{{coro}}, please prefer to use \code{with_otel_span()} or
\code{with_otel_promise_domain()}.

}}
\section{Definitions}{

\itemize{
\item Promise domain: An environment in which has setup/teardown methods. These
environments can be composed together to facilitate execution context for
promises. In normal R execution, this can be achieved with scope / stack.
But for complex situations, such as the currently open graphics device,
async operations require promise domains to setup/teardown these contexts
to function properly. Otherwise a multi-stage promise that adds to the
graphics device at each stage will only ever print to the most recently
created graphics device, not the associated graphics device. These promise
domains are not automatically created, they must be manually added to the
execution stack, for example \code{with_otel_promise_domain()} does this for
OpenTelemetry spans.
\item Promise domain restoration: When switching from one promise chain to
another, the execution context is torn down and then re-established. This
re-establishment is called "promise domain restoration". During this
process, the promise domains are restored in their previously established
combination order.
\item Promise chain: A set of promise objects to execute over multiple async
ticks.
\item Async tick: the number of times an event loop must run to move computation
forward. (Similar to a JavaScript event loop tick.)
\item \code{then()} promise domain capture: When \code{then()} is called, it will capture
the current promise domain. This promise domain is restored (only if
needed) when evaluating the given \code{onFulfilled} and \code{onRejected} callbacks.
This captured promise domain does not go into any downstream promise chain
objects. The only way the promise domain is captured is exactly when the
\code{then()} method is called.
}

\code{with_otel_promise_domain()} creates a promise domain that restores the
currently active OpenTelemetry span from when a call to \code{promises::then()} is
executed. Given the special circumstance where only the current otel span is
needed to continue recording (not a full ancestry tree of otel spans), we can
capture \emph{just} the current otel span and reactivate that otel span during promise
domain restoration.
}

\section{When promise domains are captured}{


Asynchronous operation
\itemize{
\item Creates \code{async_op} otel span
\item Automatically ends the otel span (\code{async_op}) when the promise (p)
resolves or rejects
}

The code below illustrates an example of when the promise domain are
created/captured/restored and when otel span objects are
created/activated/reactivated/ended.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# t0.0
p2 <- with_otel_promise_domain(\{
  # t0.1
  p <- with_otel_span("async_op", \{
    # ... return a promise ...
    init_async_work() |> # t0.2
      then( # t0.3
        some_async_work # t1.0
      )
  \}) # t0.4, t1.0, t2.0
  p |>
    then( # t0.5
      more_async_work # t3.0
    )
\}) # t0.6

p_final <-
  p2 |> then( # t0.7
    final_work # t4.0
  )
}\if{html}{\out{</div>}}

An in-depth explanation of the execution timeline is below.
\itemize{
\item At the first initial tick, \verb{t0.*}:
\itemize{
\item \code{t0.0}: The code is wrapped in \code{with_otel_promise_domain()}
\item \code{t0.1}: The \code{async_op} otel span is created and activated
\item \code{t0.2}: Some async work is initiated
\item \code{t0.3}: \code{then()} is called, capturing the active \code{async_op} otel span (as it
is called within \code{with_otel_promise_domain()})
\item \code{t0.4}: The \code{with_otel_span()} call exits, but the \code{async_op} otel span is
not ended as the promise is still pending. The returned promise has a
\code{finally()} step added to it that will end the otel span \code{async_op} when \code{p}
is resolved.
\item \code{t0.5}: Another \code{then()} is called, but there is no active otel span to
capture
\item \code{t0.6}: The otel span promise domain call exits
\item \code{t0.7}: Another \code{then()} is called. No otel span will be captured as there is
no active otel span / promise domain
}
\item At the first followup tick, \code{t1.0}:
\itemize{
\item The active \code{async_op} otel span is reactivated during promise domain
restoration for the duration of the \code{then} callback
\item The \code{some_async_work} function is called
}
\item At tick, \code{t2.0}:
\itemize{
\item \code{some_async_work} has resolved
\item A hidden \code{finally()} step closes the otel span, \code{async_op}
\item \code{p} is now resolved
}
\item At tick, \code{t3.0}:
\itemize{
\item There is no active otel span at \code{t0.5}, so no otel span is reactivated during
promise domain restoration
\item The \code{more_async_work} function is executed
}
\item At tick, \code{t4.0}:
\itemize{
\item \code{more_async_work} has resolved, therefore \code{p2} is now resolved
\item There was no otel span promise domain at \code{t0.7}, so no attempt is made to
reactivate any otel span
\item The \code{final_work} function is executed
}
\item At tick, \code{t5.0}:
\itemize{
\item \code{p_final} has resolved
}
}
}

\section{Complexity}{


When reactivating the \code{k}th step in a promise chain, the currently active
otel span (during the call to \code{then()}) will be reactivated during promise domain
restoration (\code{O(1)}). To restore a chain of promises, the active otel span will
be restored at each step (\code{O(n)}) due to the \strong{\code{n}} calls to wrapping each
\code{onFulfilled} and \code{onRejected} callbacks inside \code{then()}.

If we did NOT have a handoff promise domain for otel span restoration, a regular
promise domain approach would be needed at each step to restore the active
otel span. Each step would call \code{with_active_span()} \code{k} times (\code{O(k)}, where as
handoff domain computes in \code{O(1)}). Taking a step back, to restore each otel span
at for every step in a promise chain would then take \code{O(n^2)} time, not
\code{O(n)}. The standard, naive promise domain approach does not scale for
multiple similar promise domain restorations.
}

\section{Execution model for \code{with_otel_promise_domain()}}{

\enumerate{
\item \code{with_otel_promise_domain(expr)} is called.
\itemize{
\item The following steps all occur within \code{expr}.
}
\item Create an otel span object using \code{otel::start_span()}.
\itemize{
\item We need the otel span to be active during the a followup async operation.
Therefore, \code{otel::start_local_active_span()} is not appropriate as the
otel span would be ended when the function exits, not when the promise chain
resolves.
}
\item Be sure your otel span is activated before calling \code{promises::then()}.
\itemize{
\item Activate it using \code{with_otel_span(name, expr)} (which also
creates/ends the otel span) or \code{otel::with_active_span(span, expr)}.
}
\item Call \code{promises::then()}
}
\itemize{
\item When \code{promises::then()} is called, the two methods (\code{onFulfilled} and
\code{onRejected}) capture the currently active spans. (Performed by the
initial \code{with_otel_promise_domain()})
}
\enumerate{
\item During reactivation of the promise chain step, the previously captured
otel span is reactivated via \code{with_active_span()}. (Performed by the initial
\code{with_otel_promise_domain()})
}
}

\section{OpenTelemetry span compatibility}{


For otel span objects to exist over may async ticks, the otel span must be created
using \code{otel::start_span()} and later ended using \code{otel::end_span()}. Ending
the otel span must occur \strong{after} any promise chain work has completed.

If we were to instead use \code{otel::start_local_active_span()}, the otel span would
be ended when the function exits, not when the promise chain completes. Even
though the local otel span is created, activated, and eventually ended, the otel span
will not exist during reactivation of the otel span promise domain.

\code{with_otel_span()} is a convenience method that creates, activates, and
ends the otel span only after the returned promise (if any) resolves. It also
properly handles both synchronous (ending the otel span within \code{on.exit()}) and
asynchronous operations (ending the otel span within \code{promises::finally()}).
}

\examples{
\dontrun{
# Common usage:
with_otel_promise_domain({
  # ... deep inside some code execution ...

  # Many calls to `with_otel_span()` within `with_otel_promise_domain()`
  with_otel_span("my_operation", {
    # ... do some work ...
  })
})
}
\dontrun{
with_otel_promise_domain({
  # ... deep inside some code execution ...

  # Synchronous operation
  # * Creates `my_operation` span
  result <- with_otel_span("my_operation", {
    # ... do some work ...
    print(otel::get_active_span()$name) # "my_operation"

    # Nest (many) more spans
    prom_nested <- with_otel_span("my_nested_operation", {
      # ... do some more work ...
      promise_resolve(42) |>
        then(\(value) {
          print(otel::get_active_span()$name) # "my_nested_operation"
          print(value) # 42
        })
    })

    # Since `then()` is called during the active `my_operation` span,
    # the `my_operation` span will be reactivated in the `then()` callback.
    prom_nested |> then(\(value) {
      print(otel::get_active_span()$name) # "my_operation"
      value
    })
  })

  # Since `then()` is called where there is no active span,
  # there is no _active_ span in the `then()` callback.
  result |> then(\(value) {
    stopifnot(inherits(otel::get_active_span(), "otel_span_noop"))
    print(value) # 42
  })
})
}
}
\seealso{
\code{\link[otel:start_span]{otel::start_span()}}, \code{\link[otel:with_active_span]{otel::with_active_span()}},
\code{\link[otel:end_span]{otel::end_span()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{promises-deprecated}
\alias{promises-deprecated}
\alias{with_ospan_async}
\alias{with_ospan_promise_domain}
\alias{local_ospan_promise_domain}
\title{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}: Deprecated functions in promises}
\usage{
with_ospan_async(name, expr, ..., tracer, attributes = NULL)

with_ospan_promise_domain(expr)

local_ospan_promise_domain(envir = parent.frame())
}
\description{
These functions have been deprecated and will be removed in a future version.
Please use the suggested alternatives instead.
}
\section{Functions}{
\itemize{
\item \code{with_ospan_async()}: Use \code{with_otel_span()} instead

\item \code{with_ospan_promise_domain()}: Use \code{with_otel_promise_domain()} instead

\item \code{local_ospan_promise_domain()}: Use \code{local_otel_promise_domain()} instead

}}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/promise.R
\name{is.promise}
\alias{is.promise}
\alias{is.promising}
\alias{as.promise}
\title{Coerce to a promise}
\usage{
is.promise(x)

is.promising(x)

as.promise(x)
}
\arguments{
\item{x}{An R object to test or coerce.}
}
\value{
\code{as.promise} returns a promise object, or throws an error if the
object cannot be converted.

\code{is.promise} returns \code{TRUE} if the given value is a promise object, and
\code{FALSE} otherwise.

\code{is.promising} returns \code{TRUE} if the given value is a promise object or
if it can be converted to a promise object using \code{as.promise}, and \code{FALSE}
otherwise.
}
\description{
Use \code{is.promise} to determine whether an R object is a promise. Use
\code{as.promise} (an S3 generic method) to attempt to coerce an R object to a
promise. This package includes support for converting \link[future:Future-class]{future::Future}
objects into promises.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future_promise.R
\name{future_promise_queue}
\alias{future_promise_queue}
\alias{future_promise}
\title{\pkg{future} promise}
\usage{
future_promise_queue()

future_promise(
  expr = NULL,
  envir = parent.frame(),
  substitute = TRUE,
  globals = TRUE,
  packages = NULL,
  ...,
  queue = future_promise_queue()
)
}
\arguments{
\item{expr}{An R expression. While the \code{expr} is eventually sent to \code{\link[future:future]{future::future()}}, please use the same precautions that you would use with regular \code{promises::promise()} expressions. \code{future_promise()} may have to hold the \code{expr} in a \code{\link[=promise]{promise()}} while waiting for a \pkg{future} worker to become available.}

\item{envir}{The \link{environment} from where global objects should be
identified.}

\item{substitute}{If TRUE, argument \code{expr} is
\code{\link[base]{substitute}()}:ed, otherwise not.}

\item{globals}{(optional) a logical, a character vector, or a named list
to control how globals are handled.
For details, see section 'Globals used by future expressions'
in the help for \code{\link[future:future]{future()}}.}

\item{packages}{(optional) a character vector specifying packages
to be attached in the \R environment evaluating the future.}

\item{...}{extra parameters provided to \code{\link[future:future]{future::future()}}}

\item{queue}{A queue that is used to schedule work to be done using \code{\link[future:future]{future::future()}}.  This queue defaults to \code{\link[=future_promise_queue]{future_promise_queue()}} and requires that method \code{queue$schedule_work(fn)} exist.  This method should take in a function that will execute the promised \pkg{future} work.}
}
\value{
Unlike [future::future()\verb{], }future_promise()\verb{returns a [promise()] object that will eventually resolve the \pkg{future}}expr`.
}
\description{
\ifelse{html}{\out{<a href='https://www.tidyverse.org/lifecycle/#experimental'><img src='figures/lifecycle-experimental.svg' alt='Experimental lifecycle'></a>}}{\strong{Experimental}}
}
\details{
When submitting \pkg{future} work, \pkg{future} (by design) will block the main R session until a worker becomes available.
This occurs when there is more submitted \pkg{future} work than there are available \pkg{future} workers.
To counter this situation, we can create a promise to execute work using future (using \code{future_promise()}) and only begin the work if a \pkg{future} worker is available.

Using \code{future_promise()} is recommended when ever a continuous runtime is used, such as with \pkg{plumber} or \pkg{shiny}.

For more details and examples, please see the \href{https://rstudio.github.io/promises/articles/future_promise.html}{\code{vignette("future_promise", "promises")}} vignette().
}
\section{Functions}{
\itemize{
\item \code{future_promise_queue}: Default \code{future_promise()} work queue to use. This function returns a \link{WorkQueue} that is cached per R session.

\item \code{future_promise}: Creates a \code{\link[=promise]{promise()}} that will execute the \code{expr} using \code{\link[future:future]{future::future()}}.
}}

\examples{
\donttest{# Relative start time
start <- Sys.time()
# Helper to force two `future` workers
with_two_workers <- function(expr) {
  if (!require("future")) {
    message("`future` not installed")
    return()
  }
  old_plan <- future::plan(future::multisession(workers = 2))
  on.exit({future::plan(old_plan)}, add = TRUE)
  start <<- Sys.time()
  force(expr)
  while(!later::loop_empty()) {Sys.sleep(0.1); later::run_now()}
  invisible()
}
# Print a status message. Ex: `"PID: XXX; 2.5s promise done"`
print_msg <- function(pid, msg) {
  message(
    "PID: ", pid, "; ",
    round(difftime(Sys.time(), start, units = "secs"), digits = 1), "s " ,
    msg
  )
}

# `"promise done"` will appear after four workers are done and the main R session is not blocked
# The important thing to note is the first four times will be roughly the same
with_two_workers({
  promise_resolve(Sys.getpid()) \%...>\% print_msg("promise done")
  for (i in 1:6) future::future({Sys.sleep(1); Sys.getpid()}) \%...>\% print_msg("future done")
})
{
#> PID: XXX; 2.5s promise done
#> PID: YYY; 2.6s future done
#> PID: ZZZ; 2.6s future done
#> PID: YYY; 2.6s future done
#> PID: ZZZ; 2.6s future done
#> PID: YYY; 3.4s future done
#> PID: ZZZ; 3.6s future done
}

# `"promise done"` will almost immediately, before any workers have completed
# The first two `"future done"` comments appear earlier the example above
with_two_workers({
  promise_resolve(Sys.getpid()) \%...>\% print_msg("promise")
  for (i in 1:6) future_promise({Sys.sleep(1); Sys.getpid()}) \%...>\% print_msg("future done")
})
{
#> PID: XXX; 0.2s promise done
#> PID: YYY; 1.3s future done
#> PID: ZZZ; 1.4s future done
#> PID: YYY; 2.5s future done
#> PID: ZZZ; 2.6s future done
#> PID: YYY; 3.4s future done
#> PID: ZZZ; 3.6s future done
}}
}
\seealso{
\code{\link{WorkQueue}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future_promise.R
\name{WorkQueue}
\alias{WorkQueue}
\title{Future promise work queue}
\description{
Future promise work queue

Future promise work queue
}
\details{
#' \ifelse{html}{\out{<a href='https://www.tidyverse.org/lifecycle/#experimental'><img src='figures/lifecycle-experimental.svg' alt='Experimental lifecycle'></a>}}{\strong{Experimental}}

An \pkg{R6} class to help with scheduling work to be completed. \code{WorkQueue} will only execute work if the \code{can_proceed()} returns \code{TRUE}. For the use case of \code{future}, \code{can_proceed()} defaults to \code{future::nbrOfFreeWorkers() > 0} which will not allow for work to be executed if a \pkg{future} worker is not available.

\code{WorkQueue} will constantly try to start new work once prior work item finishes.  However, if \code{can_proceed()} returns \code{FALSE} (no future workers are available) and there is more work to be done, then work is attempted later a random amount of time later using exponential backoff.  The exponential backoff will cap out at 10 seconds to prevent unnecessarily large wait times.

Each time \code{WorkQueue} tries to start more work, it will repeat until \code{can_proceed()} returns \code{FALSE} or there is no more work in the \code{queue}.
}
\section{Global event loop}{


The global loop is used by default as the internal \code{WorkQueue} "delayed check" uses a single delay check for the whole queue, rather than having each item in the queue attempt to process.
This behavior might change in the future, but we are not exactly sure how at this point.

If a private \code{later} loop wants to become synchronous by running until all jobs are completed but is waiting on a \code{future_promise()}, the private loop will not complete unless the global loop is allowed to move forward.

However, it is possible to use a private loop inside a user-defined \code{WorkQueue} may work which can be provided directly to \code{future_promise(queue=custom_queue)}. Having a concrete example (or need) will help us understand the problem better. If you have an example, please reach out .
}

\seealso{
\code{\link[=future_promise_queue]{future_promise_queue()}} which returns a \code{WorkQueue} which is cached per R session.
}
\keyword{internal}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-new}{\code{WorkQueue$new()}}
\item \href{#method-schedule_work}{\code{WorkQueue$schedule_work()}}
\item \href{#method-clone}{\code{WorkQueue$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-new"></a>}}
\if{latex}{\out{\hypertarget{method-new}{}}}
\subsection{Method \code{new()}}{
Create a new \code{WorkQueue}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WorkQueue$new(
  can_proceed = future_worker_is_free,
  queue = fastmap::fastqueue(),
  loop = later::global_loop()
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{can_proceed}}{Function that should return a logical value. If \code{TRUE} is returned, then the next scheduled work will be executed. By default, this function checks if \code{\link[future:nbrOfWorkers]{future::nbrOfFreeWorkers()} > 0}}

\item{\code{queue}}{Queue object to use to store the scheduled work. By default, this is a "First In, First Out" queue using \code{\link[fastmap:fastqueue]{fastmap::fastqueue()}}. If using your own queue, it should have the methods \verb{$add(x)}, \verb{$remove()}, \verb{$size()}.}

\item{\code{loop}}{\pkg{later} loop to use for calculating the next delayed check. Defaults to \code{\link[later:create_loop]{later::global_loop()}}.
Schedule work}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-schedule_work"></a>}}
\if{latex}{\out{\hypertarget{method-schedule_work}{}}}
\subsection{Method \code{schedule_work()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WorkQueue$schedule_work(fn)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{fn}}{function to execute when \code{can_proceed()} returns \code{TRUE}.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-clone"></a>}}
\if{latex}{\out{\hypertarget{method-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{WorkQueue$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{promise_all}
\alias{promise_all}
\alias{promise_race}
\title{Combine multiple promise objects}
\usage{
promise_all(..., .list = NULL)

promise_race(..., .list = NULL)
}
\arguments{
\item{...}{Promise objects. Either all arguments must be named, or all
arguments must be unnamed. If \code{.list} is provided, then these arguments are
ignored.}

\item{.list}{A list of promise objects--an alternative to \code{...}.}
}
\value{
A promise.

For \code{promise_all}, if all of the promises were successful, the returned
promise will resolve to a list of the promises' values; if any promise
fails, the first error to be encountered will be used to reject the
returned promise.

For \code{promise_race}, the first of the promises to either fulfill or reject
will be passed through to the returned promise.
}
\description{
Use \code{promise_all} to wait for multiple promise objects to all be successfully
fulfilled. Use \code{promise_race} to wait for the first of multiple promise
objects to be either fulfilled or rejected.
}
\examples{
p1 <- promise(~later::later(~resolve(1), delay = 1))
p2 <- promise(~later::later(~resolve(2), delay = 2))

# Resolves after 1 second, to the value: 1
promise_race(p1, p2) \%...>\% {
  cat("promise_race:\n")
  str(.)
}

# Resolves after 2 seconds, to the value: list(1, 2)
promise_all(p1, p2) \%...>\% {
  cat("promise_all:\n")
  str(.)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipe.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{\%>\%}
\alias{\%T>\%}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{magrittr}{\code{\link[magrittr:pipe]{\%>\%}}, \code{\link[magrittr:tee]{\%T>\%}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hybrid-then.R
\name{hybrid_then}
\alias{hybrid_then}
\title{Asynchronous or synchronous \code{then()}}
\usage{
hybrid_then(expr, on_success = NULL, on_failure = NULL, ..., tee = FALSE)
}
\arguments{
\item{expr}{An expression that evaluates to either a promise or a non-promise
value.}

\item{on_success}{A function to be called when no error occurs synchronously
or asynchronously. When invoked, the function will be called with a single
argument: the resolved value. Optionally, the function can take a second
parameter \code{.visible} if you care whether the promise was resolved with a
visible or invisible value. Can return a value or a promise.}

\item{on_failure}{A function to be called if an error occurs synchronously or
asynchronously. Takes one argument: the error object. Can return a value or
a promise to recover from the error, or throw a new error. If \code{on_failure}
is provided and doesn't throw an error (or return a promise that fails)
then this is the async equivalent of catching an error.}

\item{...}{Reserved for future use. Currently must be empty.}

\item{tee}{If \code{TRUE}, ignore the return value of the callback, and use the
original value of \code{expr} as the result. For \code{on_failure} with \code{tee = TRUE},
the callback executes but the original error is re-thrown afterward.}
}
\value{
\itemize{
\item If \code{expr} evaluates to a promise, a promise with a single followup
promise to handle the \code{on_success} or \code{on_failure} callbacks.
\item If \code{expr} evaluates to a non-promise value, the result of the synchronous
operation after being processed by \code{on_success} or \code{on_failure}.
\item If a callback returns a promise, the result is always a promise.
}
}
\description{
This is a helper function that behaves like \code{then}, however if
\code{is.promising()} returns \code{FALSE} then the handlers will be executed
immediately.
}
\details{
Execution paths:
\itemize{
\item If \code{expr} evaluates to a promise (\code{p}), it will call \code{p |> then(on_success, on_failure)}.
\item If \code{expr} evaluates to a non-promise value (\code{x}), it will call
\code{on_success(x)}.
\item If \code{expr} throws an error (\code{e}) during calculation, it will call
\code{on_failure(e)}.
}

In all cases, the \code{on_success} and \code{on_failure} callbacks are executed (when
provided).
}
\section{Utility}{


This function is useful for writing functions that need to execute followup
behavior \emph{now} or within a promise. This is different behavior than \code{then()}
where \emph{everything} is made into a promise.

\code{hybrid_then()} allows authors to keep synchronous execution on the same
\emph{tick} without requiring the use of a followup promise. This is particularly
appealing for situations where the author does not control the execution flow
for items that may be either synchronous or asynchronous, such as within
\code{{plumber2}}.
}

\section{Error Handling}{


If no \code{on_failure} callback is provided and an error occurs, the error is
re-thrown immediately (for synchronous errors) or propagated through the
returned promise (for asynchronous errors).

If an \code{on_failure} callback is provided but it throws an error, that new
error replaces the original error. With \code{tee = TRUE}, even if \code{on_failure}
executes successfully, the original error is still re-thrown.
}

\section{Callback Return Values}{


Callbacks can return any value, including promises. If a callback returns a
promise, the entire \code{hybrid_then()} call will return a promise, even if the
input was synchronous. This allows seamless transitions between synchronous
and asynchronous execution.
}

\examples{
# Basic usage - works with both sync and async values
add_to <- function(x, k) {
  hybrid_then(
    x,
    on_success = function(value) {
      value + k
    },
    on_failure = function(err) {
      message("Error: ", err$message)
      NA_real_
    }
  )
}

# Synchronous
42 |> add_to(100)
#> [1] 142

# Synchronous error
add_to({stop("Bad input!")}, 8)
#> Error: Bad input!
#> [1] NA

\dontrun{
# Asynchronous
promise_resolve(42) |>
  add_to(8) |>
  then(print)
# When resolved...
#> [1] 50

# Error handling - asynchronous
promise_resolve(stop("Bad async input!")) |>
  add_to(8) |>
  then(print)
# When resolved...
#> Error: Bad async input!
#> [1] NA

# Chaining multiple operations
# (Move the `promise_resolve()` around to see sync vs async behavior)
1 |>
  hybrid_then(on_success = \(x) x + 1) |>
  hybrid_then(on_success = \(x) promise_resolve(x * 2)) |>
  hybrid_then(on_success = \(x) x - 1) |>
  hybrid_then(print)
# When resolved...
#> [1] 3
}
}
\seealso{
\code{\link[=then]{then()}}, \code{\link[=is.promising]{is.promising()}}, \code{\link[=promise_resolve]{promise_resolve()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{promise_reduce}
\alias{promise_reduce}
\title{Promise-aware version of Reduce}
\usage{
promise_reduce(.x, .f, ..., .init)
}
\arguments{
\item{.x}{A vector or list to reduce. (Not a promise.)}

\item{.f}{A function that takes two parameters. The first parameter will be
the "result" (initially \code{.init}, and then set to the result of the most
recent call to \code{func}), and the second parameter will be an element of \code{.x}.}

\item{...}{Other arguments to pass to \code{.f}}

\item{.init}{The initial result value of the fold, passed into \code{.f} when it
is first executed.}
}
\value{
A promise that will resolve to the result of calling \code{.f} on the last
element (or \code{.init} if \code{.x} had no elements). If any invocation of \code{.f}
results in an error or a rejected promise, then the overall
\code{promise_reduce} promise will immediately reject with that error.
}
\description{
Similar to \code{\link[purrr:reduce]{purrr::reduce}} (left fold), but the function \code{.f} is permitted
to return a promise. \code{promise_reduce} will wait for any returned promise to
resolve before invoking \code{.f} with the next element; in other words, execution
is serial. \code{.f} can return a promise as output but should never encounter a
promise as input (unless \code{.x} itself is a list of promises to begin with, in
which case the second parameter would be a promise).
}
\examples{
# Returns a promise for the sum of e1 + e2, with a 0.5 sec delay
slowly_add <- function(e1, e2) {
  promise(~later::later(~resolve(e1 + e2), delay = 0.5))
}

# Prints 55 after a little over 5 seconds
promise_reduce(1:10, slowly_add, .init = 0) \%...>\% print()

}
